//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "console.h"
#include "debugger_private.h"
#include "debugger_m680x0.h"
#include "debugger_m88xx0.h"
#include "mainapp.h"
#include "mystring.h"
#include "mythread.h"
#include <algorithm>
#include <csignal>

static Debugger *gDebugger;
BreakpointMonitor *gBreakpointMonitor;
CVPrompt *gCVPrompt;

static void *debugger_run(void *);

// 初期化。
// この時点で VM が初期化されていること。
void
debugger_init()
{
	gDebugger = new Debugger();
	gDebugger->Init();

	gBreakpointMonitor = new BreakpointMonitor();
	gCVPrompt = new CVPrompt();

	// デバッガスレッド起動
	pthread_t th;
	pthread_create(&th, NULL, debugger_run, NULL);
}

// デバッガスレッドのエントリポイント
static void *
debugger_run(void *dummy)
{
	PTHREAD_SETNAME("Debugger");
	pthread_detach(pthread_self());

	gDebugger->ThreadRun();
	return NULL;
}

// コンストラクタ
Debugger::Debugger()
{
	if (gMPU680x0) {
		md = new DebuggerMD_m680x0(this, gMPU680x0->GetCPU());
	} else if (gMPU88xx0) {
		md = new DebuggerMD_m88xx0(this, gMPU88xx0->GetCPU());
	} else {
		throw "unknown mpu";
	}
}

// コマンドライン引数によって動作を決めるところ。
// 名前がこれでいいのかはあるけど。
void
Debugger::Init()
{
	// -d なら CPU 起動時点で停止してプロンプトを待つ
	if (gMainApp.debug_on_start) {
		md->ReqSet(CPU_REQ_PROMPT);
	}

	// -b <addr> ならブレークポイント設定
	for (auto& str : gMainApp.debug_breakaddr) {
		breakpoint_t bp;

		md->ReqSet(CPU_REQ_TRACE);

		// <addr>,<skip> で分離できたら <skip> を取り出す
		int pos = str.find(',');
		if (pos != std::string::npos) {
			std::string skipstr = str.substr(pos + 1);
			bp.skip = atoi(skipstr.c_str());

			// XXX これ出来るんだっけ?
			str[pos] = '\0';
		}
		// そしてどちらにしても <addr> を取り出す
		if (!ParseAddr(str.c_str(), &bp.addr)) {
			warnx("\"%s\": Invalid breakpoint address", str.c_str());
			return;
		}
		// 登録
		bp.type = BreakpointType::Address;
		AddBreakpoint(bp);
	}

	// どこでやるべか
	signal(SIGPIPE, SIG_IGN);
}

// デバッガスレッド
void
Debugger::ThreadRun()
{
	// XXX どこかでコンソールの選択とパラメータの取得
	if (gMainApp.debug_on_console) {
		cons = new ConsoleStdio();
	} else {
		cons = new ConsoleTCP();
	}
	if (cons->Init() == false) {
		delete cons;
		return;
	}

	for (;;) {
		// 着信待ち
		if (Accept() == false) {
			break;
		}

		cons->InitEditLine("> ");

		// デバッガプロンプトにいる間はトレースオン
		md->ReqSet(CPU_REQ_TRACE);

		bool first = true;
		for (;;) {
			// 接続後の1回目だけ実行するもの。
			if (first) {
				first = false;

				// greeting はプロンプトが取れる前にもう表示したい。
				// 何らかの事故でプロンプトが取れなくても、ここまでは接続
				// できてることが分かるように。
				cons->Print("This is debugger console\n");

				// 接続ごとに初期化する値
				// XXX もうちょっときれいにしたい
				d_last_addr = 0xffffffff;
				m_last_addr = 0xffffffff;
				n_enable = false;
				s_enable = false;
				t_enable = false;
			}

			// プロンプトが取れるのを待つ
			if (AcquirePrompt() == false) {
				break;
			}

			// プロンプトに来た時に表示するいつものやつ
			pc = md->GetPC();
			d_last_addr = pc;
			cmd_minus();

			// コマンド処理のメインループ
			auto action = MainLoop();

			// 次回との差分のため、今のレジスタセットをバックアップ
			md->BackupRegs();

			// メインループから戻ったということは Leave か Quit なので
			// どちらにしても、ここでプロンプトを手放す。
			gCVPrompt->NotifyRelease();

			// quit ならここでループを1段抜ける
			if (action == CommandAction::Quit) {
				break;
			}
		}

		// デバッガを抜けるのでトレースオフ
		md->ReqClr(CPU_REQ_TRACE);

		cons->Close();
		if (gMainApp.debug_on_console) {
			break;
		}
	}

	delete cons;
}

// コンソールに接続されるのを待つ。
// うーん、なんだこれ。
//
// +--- debug_on_console (-D オプション)
// | +- debug_on_start   (-d オプション)
// | |	動作
// 0 0	TCP、Accept で待機。
// 0 1	TCP、Accpet で待機。
// 1 0	stdio、起動時にプロンプトで止まらない -> キー入力待ち。
// 1 1	stdio、起動時にプロンプトで止まる。
bool
Debugger::Accept()
{
	if (gMainApp.debug_on_console == false) {
		// TCP なら accept が成功するのを待つ
		cons->Accept();
	} else {
		if (gMainApp.debug_on_start == false) {
			// stdio、起動時にプロンプトで止まらない。
			// この場合キー入力かブレークポイントなどで止まった時点で
			// プロンプトを出したいのでちょっと面倒なことに。
			for (;;) {
				bool is_prompt = gCVPrompt->WaitAcquire(200);
				if (is_prompt) {
					break;
				}

				int r = cons->Poll(200);
				if (r < 0) {
					cmd_q();
					return false;
				}
				if (r > 0) {
					break;
				}
			}
		} else {
			// stdio、起動時にプロンプトで止まる。
			// この場合は無条件に成功でよい。
		}
	}

	return true;
}

// プロンプトが取れるのを待つ
// 失敗すれば false を返す
bool
Debugger::AcquirePrompt()
{
	md->ReqSet(CPU_REQ_PROMPT);

	for (;;) {
		bool is_prompt = gCVPrompt->WaitAcquire(200);
		if (is_prompt)
			break;

		// コンソールを定期観測して、入力があればブレーク要求
		if (cons->Poll()) {
			// この入力は drop してみる
			if (cons->Gets(cmdbuf) == false) {
				// EOF or Error
				cmd_q();
				return false;
			}
			md->ReqSet(CPU_REQ_PROMPT);
		}
	}
	return true;
}

// コマンド処理のメインループ。
// コマンドが Stay ならこの中で処理を継続。
// コマンドが Leave か Quit ならその値を持って戻る (呼び出し側で処理する)。
Debugger::CommandAction
Debugger::MainLoop()
{
	for (;;) {
		cons->Prompt();
		if (cons->Gets(cmdbuf) == false) {
			// EOF or Error
			return CommandAction::Quit;
		}

		string_rtrim(cmdbuf);

		if (!cmdbuf.empty()) {
			// コマンドが入力されれば次回のために保存
			last_cmdbuf = cmdbuf;
		} else {
			// 空行が入力されれば直前のコマンドをもう一度
			// 前行がなければ何もせずもう一度プロンプトを表示するかね
			if (last_cmdbuf.empty())
				continue;
			cmdbuf = last_cmdbuf;
		}

		// 行を args に分解
		ParseCmdbuf();

		// 前回の値が有効なのはコマンドが連続した時だけ

		// コマンドをテーブルから探す
		auto it = std::find_if(cmdtable.begin(), cmdtable.end(),
			[=](auto x) { return strcmp(args[0].c_str(), x.name) == 0; });
		if (it != cmdtable.end()) {
			auto cmd = *it;

			// 見付かれば実行
			(this->*(cmd.func))();

			// Stay なら引き続きコマンド処理
			if (cmd.action == CommandAction::Stay) {
				continue;
			}
			// それ以外ならメインループ終了
			return cmd.action;
		}

		// (コマンドテーブルになくて) "r" から始まっていればレジスタ表示系
		if (args[0][0] == 'r') {
			// ShowRegister() は処理したら true を返す。
			// "r" 系コマンドはすべて Stay。
			if (md->ShowRegister(cons, args)) {
				continue;
			}
		}

		// 知らないコマンドも Stay 相当。
		cons->Print("%s: unknown command\n", args[0].c_str());
		continue;
	}
	__unreachable();
}

// ブレークポイントとかを調べる。
// CPU_REQ_TRACE フラグが立っている間1命令ごとに呼び出される。
bool
debugger_check()
{
	return gDebugger->Check();
}

// デバッガ実行中なら命令開始前にメインルーチンから呼ばれる。
// プロンプトに降りるなら true を返す。
bool
Debugger::Check()
{
	if (t_enable) {
		t_count--;
		if (t_count == 0) {
			t_enable = false;
			return true;
		}
	}

	if (s_enable) {
		s_count--;
		if (s_count == 0) {
			s_enable = false;
			return true;
		}
	}

	// n による1命令実行のブレークポイントか
	if (n_enable && n_breakenable && n_breakaddr == md->GetPC()) {
		n_breakenable = false;
		return true;
	}

	// ブレークポイント
	if (CheckAllBreakpoints()) {
		return true;
	}

	// ステップアウト
	if (so_enable) {
		if (md->IsStepOut()) {
			so_enable = false;
			return true;
		}
	}

	// アドレス指定付き continue
	if (bc_enable) {
		if (bc_addr == md->GetPC()) {
			bc_enable = false;
			return true;
		}
	}

	return false;
}

// ブレークポイントがどれかでも成立するかを調べる。
// 1つ以上成立してブレークするなら true を返す。
// 1つも成立しておらずブレークしないなら false を返す。
bool
Debugger::CheckAllBreakpoints()
{
	bool is_break = false;

	// 命令ごとにクリアする
	bi_inst = 0;
	bi_inst_bytes = 0;

	// 1つの条件でマッチしても(そこでブレークすること自体は確定するのだが)
	// 残りの他の条件も成立すればカウントを進める必要があるため、
	// 全部処理した上でどれか一つでもブレークしたかで判断する必要がある。
	for (auto& bp : bpoint) {
		switch (bp.type) {
		 case BreakpointType::Address:
			if (bp.addr != md->GetPC()) {
				continue;
			}
			break;

		 case BreakpointType::Memory:
			if (!md->CheckLEA(bp.addr)) {
				continue;
			}
			break;

		 case BreakpointType::Exception:
			if (bv_vector >= 0) {
				if (bp.vec1 <= bv_vector && bv_vector <= bp.vec2) {
					break;
				}
			}
			continue;

		 case BreakpointType::Instruction:
			if (CheckBreakpointInst(bp) == false) {
				continue;
			}
			break;

		 default:
			continue;
		}

		// 条件は成立したのでカウントとスキップ
		bp.matched++;

		// skip == -1 は常にブレークしない (カウントするだけ)
		if (bp.skip < 0) {
			continue;
		}

		// skip が 1 以上なら、remain を減算
		if (bp.skip > 0 && --bp.skipremain > 0) {
			continue;
		}

		// ブレーク成立
		is_break = true;
	}

	// 例外通知は通過ごとに常に下ろしておく
	bv_vector = -1;

	// ブレークポイントが1つでも成立したかどうかを返す
	return is_break;
}

// 命令ブレークポイントが成立するか調べる。
bool
Debugger::CheckBreakpointInst(breakpoint_t& bp)
{
	saddr_t laddr;
	uint64 paddr;

	// uint32 bi_inst が現在の PC 位置の命令データ (左詰め)
	// bi_inst_bytes が読み込んだバイト数(bi_inst の左からの有効バイト数)。
	// bi_need_bytes が現在のブレークポイントで読み込む必要のあるバイト数。

	laddr.addr = md->GetPC();
	laddr.super = md->IsSuper();
	laddr.logical = true;
	laddr.search = true;
	// 必要なバイト数に達するまで読み足す
	while (bi_inst_bytes < bi_need_bytes) {
		// アドレスを求める。
		// 論理アクセスで ATC になければテーブルサーチも試みる。
		if (md->MMUEnabled()) {
			paddr = md->TranslateAddr(laddr);
			if ((int64)paddr < 0) {
				// MMU 時点でバスエラー
				return false;
			}
		} else {
			paddr = laddr.addr;
		}

		// 命令長ずつ読み込む。
		for (int i = 0; i < md->inst_bits / 8; i++) {
			uint64 data = vm_phys_peek_8(paddr);
			if ((int64)data < 0) {
				return false;
			}
			bi_inst = (bi_inst << 8) | data;
			bi_inst_bytes++;
			laddr.addr++;
			paddr++;
		}
	}
	// 左詰めにする
	bi_inst <<= (4 - bi_inst_bytes) * 8;

	// 必要なバイト数取得できたので比較
	if ((bi_inst & bp.mask) == bp.inst) {
		return true;
	}
	return false;
}

// 例外通知 (MPU からの連絡用)
void
debugger_notifyexception(int vector)
{
	gDebugger->NotifyException(vector);
}

// 例外通知 (本体)
void
Debugger::NotifyException(int vector)
{
	bv_vector = vector;
}

// コマンド一覧。
// "r" から始まるレジスタ表示系コマンドはコード中で別処理してある。
/*static*/ std::vector<Debugger::cmddef_t> Debugger::cmdtable = {
	{ "bi",		&Debugger::cmd_bi,		Debugger::CommandAction::Stay },
	{ "bm",		&Debugger::cmd_bm,		Debugger::CommandAction::Stay },
	{ "bv",		&Debugger::cmd_bv,		Debugger::CommandAction::Stay },
	{ "bx",		&Debugger::cmd_bx,		Debugger::CommandAction::Stay },
	{ "b",		&Debugger::cmd_b,		Debugger::CommandAction::Stay },
	{ "brhist",	&Debugger::cmd_brhist,	Debugger::CommandAction::Stay },
	{ "c",		&Debugger::cmd_c,		Debugger::CommandAction::Leave },
	{ "d",		&Debugger::cmd_d,		Debugger::CommandAction::Stay },
	{ "dt",		&Debugger::cmd_dt,		Debugger::CommandAction::Stay },
	{ "D",		&Debugger::cmd_D,		Debugger::CommandAction::Stay },
	{ "exhist",	&Debugger::cmd_exhist,	Debugger::CommandAction::Stay },
	{ "h",		&Debugger::cmd_h,		Debugger::CommandAction::Stay },
	{ "help",	&Debugger::cmd_h,		Debugger::CommandAction::Stay },
	{ "L",		&Debugger::cmd_L,		Debugger::CommandAction::Stay },
	{ "m",		&Debugger::cmd_m,		Debugger::CommandAction::Stay },
	{ "mt",		&Debugger::cmd_mt,		Debugger::CommandAction::Stay },
	{ "M",		&Debugger::cmd_M,		Debugger::CommandAction::Stay },
	{ "n",		&Debugger::cmd_n,		Debugger::CommandAction::Leave },
	{ "q",		&Debugger::cmd_q,		Debugger::CommandAction::Quit },
	{ "quit",	&Debugger::cmd_q,		Debugger::CommandAction::Quit },
	{ "s",		&Debugger::cmd_s,		Debugger::CommandAction::Leave },
	{ "so",		&Debugger::cmd_so,		Debugger::CommandAction::Leave },
	{ "show",	&Debugger::cmd_show,	Debugger::CommandAction::Stay },
	{ "t",		&Debugger::cmd_t,		Debugger::CommandAction::Leave },
	{ "-",		&Debugger::cmd_minus,	Debugger::CommandAction::Stay },
};

// ヘルプ
void
Debugger::cmd_h()
{
	if (args.size() > 1) {
		if (args[1] == "r") {
			Help(md->GetRegisterHelp());
			return;
		}
	}
	// 引数なしか、あっても対応してないやつなら通常ヘルプを表示
	Help(HelpMsgMain);
	return;
}

/*static*/ const HelpMessages
Debugger::HelpMsgMain = {
	{ "b",						"ブレークポイント一覧表示" },
	{ "b #n",					"ブレークポイント n を削除" },
	{ "b <addr> [<skip>]",		"ブレークポイント設定" },
	{ "bm <addr> [<skip>]",		"メモリブレークポイント設定" },
	{ "bi <inst>[:<mask>] [<skip>]",	"命令ブレークポイント設定" },
	{ "bv <vec>[-<vec>] [<skip>]",	"例外ブレークポイント設定" },
	{ "bx",						"ブレークポイント全削除" },
	{ "brhist [<lines>]",		"ブランチ履歴表示" },
	{ "c [<addr>]",				"実行再開(continue)" },
	{ "d  [<addr> [<lines>]",	"逆アセンブル(論理、テーブルサーチなし)" },
	{ "dt [<addr> [<lines>]",	"逆アセンブル(論理、テーブルサーチあり)" },
	{ "D  [<addr> [<lines>]",	"逆アセンブル(物理)" },
	{ "exhist [<lines>]",		"例外履歴表示" },
	{ "h",						"ヘルプ(help)" },
	{ "L <name>=<lv>",			"ログレベル設定" },
	{ "m  [<addr> [<lines>]",	"メモリダンプ(論理、テーブルサーチなし)" },
	{ "mt [<addr> [<lines>]",	"メモリダンプ(論理、テーブルサーチあり)" },
	{ "M  [<addr> [<lines>]",	"メモリダンプ(物理)" },
	{ "n <cnt?:1>",				"ステップ実行 (サブルーチンを飛ばす)" },
	{ "q",						"quit" },
	{ "r*",						"各種レジスタを表示 (see \"help r\")" },
	{ "s [<cnt?:1>]",			"ステップ実行 (途中レジスタを表示しない)" },
	{ "so",						"ステップアウト" },
	{ "show <monitor>",			"モニター表示" },
	{ "t [<cnt?:1>]",			"トレース実行 (命令ごとにレジスタを表示)" },
};

void
Debugger::Help(HelpMessages msgs)
{
	for (const auto& pair : msgs) {
		cons->Print(" %-23s %s\n", pair.first.c_str(), pair.second.c_str());
	}
}

// cmdbuf を args... に分解する。
void
Debugger::ParseCmdbuf()
{
	int pos;
	int start;

	pos = 0;
	args.clear();

	// 引数を空白で分解
	for (;;) {
		// 先頭の空白はスキップ
		for (; cmdbuf[pos] != '\0'; pos++) {
			if (!isspace((unsigned int)cmdbuf[pos]))
				break;
		}
		if (cmdbuf[pos] == '\0')
			break;

		// 終わりを探す
		start = pos;
		for (; cmdbuf[pos] != '\0'; pos++) {
			if (isspace((unsigned int)cmdbuf[pos]))
				break;
		}
		args.push_back(cmdbuf.substr(start, pos - start));
	}
	// デバッグ表示
	if (0) {
		for (int i = 0 ; i < args.size(); i++) {
			printf("args[%d]=|%s|\n", i, args[i].c_str());
		}
	}
}

//
// デバッガコマンド
//

// ブレークポイント
// b                 ... 一覧表示
// b #n              ... #n を削除
// b <addr> [<skip>] ... 設定
void
Debugger::cmd_b()
{
	// 引数なしなら一覧表示
	if (args.size() < 2) {
		return cmd_b_list();
	}

	// 引数取得
	if (args[1][0] == '#') {
		// #<n> 形式なら、指定番号のブレークポイントを削除
		int i = atoi(&args[1][1]);
		if (i < 0 || i >= bpoint.size()) {
			cons->Print("invaild breakpoint number: #%d\n", i);
			return;
		}
		auto& bp = bpoint[i];
		if (bp.type == BreakpointType::Unused) {
			cons->Print("invalid breakpoint number: #%d\n", i);
			return;
		}
		cons->Print("breakpoint #%d (%08x) removed\n",
			i, bp.addr);
		bp.type = BreakpointType::Unused;
		// 今登録されている命令ブレークの必要命令長を再計算
		RecalcInstMask();
		return;
	}

	return cmd_b_set(BreakpointType::Address);
}

// メモリブレークポイントの設定
// bm <addr> [<skip>]
void
Debugger::cmd_bm()
{
	// XXX m68k では未サポート
	if (dynamic_cast<DebuggerMD_m680x0 *>(md)) {
		cons->Print("bm not supported yet on m68k\n");
		return;
	}
	cmd_b_set(BreakpointType::Memory);
}

// type が違うだけの各種ブレークポイント設定の共通部分。
void
Debugger::cmd_b_set(BreakpointType type)
{
	breakpoint_t bp;

	if (args.size() < 2) {
		cons->Print("usage: %s <addr> [<skipcount>]\n", args[0].c_str());
		return;
	}

	// アドレス
	if (!ParseAddr(args[1].c_str(), &bp.addr)) {
		return;
	}
	// あればスキップカウント
	if (args.size() > 2) {
		bp.skip = atoi(args[2].c_str());
	}

	// 空いてるところにセット
	// (よく似たエントリがあっても干渉しない)
	bp.type = type;
	int bi = AddBreakpoint(bp);
	if (bi == -1) {
		cons->Print("no free breakpoints\n");
	} else {
		cons->Print("breakpoint #%d added\n", bi);
	}
}

// 命令ブレークポイントの設定
// bi <insn>[:<mask>] [<skip>]
void
Debugger::cmd_bi()
{
	breakpoint_t bp;
	std::string inststr;
	std::string maskstr;
	int instlen;
	int masklen;

	if (args.size() < 2) {
		cons->Print("usage: bi <inst>[:<mask>] [<skipcount>]\n");
		return;
	}

	// 引数をまず分離
	auto pos = args[1].find(':');
	if (pos == std::string::npos) {
		// マスク指定なし
		inststr = args[1];
		instlen = inststr.size();
		masklen = -1;
	} else {
		// マスク指定あり
		inststr = args[1].substr(0, pos);
		instlen = inststr.size();
		maskstr = args[1].substr(pos + 1);
		masklen = maskstr.size();
	}

	// 命令部チェック
	if (ParseVerbHex(inststr.c_str(), &bp.inst) == false) {
		cons->Print("%s: invalid instruction value\n", args[1].c_str());
		return;
	}
	if (instlen % (md->inst_bits / 4) != 0) {
		cons->Print("%s: invalid instruction length\n", args[1].c_str());
		return;
	}

	// マスク部チェック
	bp.mask = 0xffffffff;
	if (masklen != -1) {
		if (ParseVerbHex(maskstr.c_str(), &bp.mask) == false) {
			cons->Print("%s: invalid mask value\n", args[1].c_str());
			return;
		}
		if (masklen != instlen) {
			cons->Print("%s: inst:mask must be the same length\n",
				args[1].c_str());
			return;
		}
	}
	// 8バイト未満なら左詰め。
	if (md->inst_bits < 32 && instlen < 8) {
		bp.inst <<= 32 - instlen * 4;
		bp.mask <<= 32 - instlen * 4;
	}

	// あればスキップカウント
	bp.skip = 0;
	if (args.size() > 2) {
		bp.skip = atoi(args[2].c_str());
	}

	// 空いてるところにセット
	bp.type = BreakpointType::Instruction;
	int bi = AddBreakpoint(bp);
	if (bi == -1) {
		cons->Print("no free breakpoints\n");
	} else {
		cons->Print("breakpoint #%d added\n", bi);
	}

	// 今登録されている命令ブレークの必要命令長を再計算
	RecalcInstMask();
}

// 例外ブレークポイントの設定
// bv <vector_number>[-<end_number>] [<skip>]
void
Debugger::cmd_bv()
{
	breakpoint_t bp;

	if (args.size() < 2) {
		cons->Print("usage: be <vec#>[-<end#>] [<skipcount>]\n");
		return;
	}

	auto pos = args[1].find('-');
	if (pos == std::string::npos) {
		// ベクタ番号が1つなら vec1, vec2 を同値にしておく。
		if (ParseVerbHex(args[1].c_str(), (uint32 *)&bp.vec1) == false) {
			cons->Print("%s: invalid vector number\n", args[1].c_str());
			return;
		}
		bp.vec2 = bp.vec1;
	} else {
		// ベクタ番号(範囲指定 [vec1, vec2])
		std::string str1 = args[1].substr(0, pos);
		std::string str2 = args[1].substr(pos + 1);

		if (ParseVerbHex(str1.c_str(), (uint32 *)&bp.vec1) == false) {
			cons->Print("%s: invalid first vector number\n", args[1].c_str());
			return;
		}
		if (ParseVerbHex(str2.c_str(), (uint32 *)&bp.vec2) == false) {
			cons->Print("%s: invalid last vector number\n", args[1].c_str());
			return;
		}
	}

	// 範囲チェック
	if (bp.vec1 < 0 || bp.vec1 >= md->vector_max) {
		cons->Print("$%x: invalid vector number\n", bp.vec1);
		return;
	}
	if (bp.vec2 < 0 || bp.vec2 >= md->vector_max) {
		cons->Print("$%x: invalid last vector number\n", bp.vec2);
		return;
	}

	// 大小が逆なら入れ替える?
	if (bp.vec1 > bp.vec2) {
		int tmp;
		tmp = bp.vec1;
		bp.vec1 = bp.vec2;
		bp.vec2 = tmp;
	}

	// あればスキップカウント
	bp.skip = 0;
	if (args.size() > 2) {
		bp.skip = atoi(args[2].c_str());
	}

	// 空いてるところにセット
	bp.type = BreakpointType::Exception;
	int bi = AddBreakpoint(bp);
	if (bi == -1) {
		cons->Print("no free breakpoints\n");
	} else {
		cons->Print("breakpoint #%d added\n", bi);
	}
}

// ブレークポイント一覧表示
void
Debugger::cmd_b_list()
{
	ShowMonitor(*gBreakpointMonitor);
}

// ブレークポイント一覧 (モニタ)
void
Debugger::MonitorBreakpoint(TextScreen& monitor)
{
	// 0         1         2         3         4         5         6
	// 0123456789012345678901234567890123456789012345678901234567890
	// No Type Parameter         Matched    Skip
	// #0 addr $01234567         123456789  123456789/123456789
	// #1 inst 00000000/00000000
	// #2 excp $00-$00

	monitor.Clear();
	monitor.Print(0, 0, "No Type Parameter");
	monitor.Print(26, 0, "Matched");
	monitor.Print(37, 0, "Skip");

	for (int i = 0; i < bpoint.size(); i++) {
		const auto& bp = bpoint[i];
		int y = i + 1;

		monitor.Print(0, y, "#%d", i);

		// 種別ごとの表示
		switch (bp.type) {
		 case BreakpointType::Unused:
			continue;

		 case BreakpointType::Address:
			monitor.Print(3, y, "addr $%08x", bp.addr);
			break;
		 case BreakpointType::Memory:
			monitor.Print(3, y, "mem  $%08x", bp.addr);
			break;

		 case BreakpointType::Exception:
			monitor.Print(3, y, "excp $%02x", bp.vec1);
			if (bp.vec2 != bp.vec1) {
				monitor.Print(11, y, "-$%02x", bp.vec2);
			}
			break;

		 case BreakpointType::Instruction:
			monitor.Print(3, y, "inst");
			if (md->inst_bits == 32) {
				if (bp.mask == 0xffffffff) {
					monitor.Print(8, y, "%08x", bp.inst);
				} else {
					monitor.Print(8, y, "%08x:%08x", bp.inst, bp.mask);
				}
			} else {
				if (bp.mask == 0xffffffff) {
					monitor.Print(8, y, "%08x", bp.inst);
				} else if (((bp.inst | bp.mask) & 0x0000ffff) != 0) {
					monitor.Print(8, y, "%08x:%08x", bp.inst, bp.mask);
				} else if (bp.mask == 0xffff0000) {
					monitor.Print(8, y, "%04x", bp.inst >> 16);
				} else {
					monitor.Print(8, y, "%04x:%04x",
						bp.inst >> 16, bp.mask >> 16);
				}
			}
			break;

		 default:
			monitor.Print(3, y, "type=%d", (int)bp.type);
			continue;
		}

		// マッチ回数
		monitor.Print(26, y, "%d", bp.matched);

		// スキップ
		if (bp.skip < 0) {
			monitor.Print(37, y, "forever");
		} else if (bp.skip > 0) {
			monitor.Print(37, y, "%d / %d", (bp.skip - bp.skipremain), bp.skip);
		}
	}
}

// ブレークポイント全削除
void
Debugger::cmd_bx()
{
	for (auto& bp : bpoint) {
		bp.type = BreakpointType::Unused;
	}
	cons->Print(" All breakpoints disabled\n");

	// 今登録されている命令ブレークの必要命令長を再計算
	RecalcInstMask();
}

// ブレークポイントを設定。
// new_bp のうち matched, skipremain はこちらで初期化する。
// それ以外を埋めてから呼ぶこと。
// 設定できればその番号、できなければ -1 を返す。
int
Debugger::AddBreakpoint(const breakpoint_t& new_bp)
{
	for (int i = 0; i < bpoint.size(); i++) {
		auto& bp = bpoint[i];
		if (bp.type == BreakpointType::Unused) {
			bp = new_bp;
			bp.matched = 0;
			if (bp.skip > 0) {
				bp.skipremain = bp.skip;
			} else {
				bp.skipremain = 0;
			}

			return i;
		}
	}
	return -1;
}

// すべての命令ブレークのマスクのうち最長のものを再計算する。
// 命令ブレークポイントの追加/削除のたびに呼び出すこと。
void
Debugger::RecalcInstMask()
{
	uint32 mask;
	bi_need_bytes = 0;

	// 登録されている命令ブレークの最長マスクを求める
	mask = 0;
	for (const auto& bp : bpoint) {
		if (bp.type == BreakpointType::Instruction) {
			mask |= bp.mask;
		}
	}

	// mask の Number of Trailing Zero を求める。
	// (x & -x) で x の最も下の立ってるビットだけを立てる、
	// (x & -x) -1 でそれより下の全ビットを立てる、
	// それを popcount で数えるので、$fffffff0 なら ntz = 4 になる。
	int ntz = __builtin_popcount((mask & -(int32)mask) - 1);
	// 上位側から数えたマスクに必要なビット数
	int mlen = 32 - ntz;
	// 命令語単位に切り上げる
	mlen = roundup(mlen, md->inst_bits);
	// バイト数に変換
	mlen /= 8;

	if (mlen > bi_need_bytes) {
		bi_need_bytes = mlen;
	}
}

// ブランチ履歴、例外履歴表示
// brhist [<maxlines>]
// exhist [<maxlines>]
// <maxlines> は表示する最大エントリ数。
// コンソールではデフォルトで下を新しいの順とする。<maxlines> を負数にすると
// (行数は絶対値して) 並び順を逆にしてモニタウィンドウと同じ上を新しいの順に
// する。
void
Debugger::cmd_brhist()
{
	cmd_hist_common(md->GetBrHist(), 0);
}
void
Debugger::cmd_exhist()
{
	cmd_hist_common(md->GetExHist(), BranchHistory::ExHist);
}

// ブランチ履歴、例外履歴表示の共通部分。
void
Debugger::cmd_hist_common(BranchHistory& hist, uint64 flag)
{
	// 表示最大行数(と向き)
	// 向きは bottom_to_top = true が新しいほうを下とする方向。
	int maxlines = 20;
	bool bottom_to_top = true;
	if (args.size() > 1) {
		maxlines = atoi(args[1].c_str());
		if (maxlines < 0) {
			bottom_to_top = false;
			maxlines = -maxlines;
		}

		if (maxlines < 1)
			maxlines = 1;
		if (maxlines > 256)
			maxlines = 256;
	}

	// コンソールでは空エントリの行は表示したくないので、
	// 先にエントリ数を調べる。
	int used = hist.GetUsed();

	// 表示行数 + 1行はヘッダ分
	int lines = std::min(used, maxlines) + 1;

	// MonitorUpdate() は TextScreen 高さに合わせて出力してくれる。
	auto size = hist.GetMonitorSize();
	TextScreen tscr;
	tscr.Init(size.width, lines);

	// コンソールでは下が新しいの順のほうがいい
	if (bottom_to_top) {
		flag |= BranchHistory::BottomToTop;
	}
	tscr.userdata = flag;

	// 表示
	hist.MonitorUpdate(tscr);
	ShowTextScreen(tscr);
}

// 実行再開(continue): c [<addr>]
// <addr> 指定があれば <addr> まで実行。
void
Debugger::cmd_c()
{
	if (args.size() > 1) {
		// 引数があれば
		if (!ParseAddr(args[1].c_str(), &bc_addr)) {
			return;
		}
		// 偶数番地に丸める
		bc_addr &= 0xfffffffe;

		bc_enable = true;
	}
}

// 引数などからアドレスを取得する共通ルーチン。
// 引数が1つ(以上)あれば args[1] をアドレスとしてパースする。
// なければ、lastaddr が指定されていればそれを使う。
// それもなければ現在の PC を使う。
bool
Debugger::GetAddr(uint32& addr, uint32& lastaddr)
{
	if (args.size() > 1) {
		// 1つ目の引数があれば、開始アドレス
		if (ParseAddr(args[1].c_str(), &addr) == false) {
			return false;
		}
		// 偶数番地からに丸める
		addr &= 0xfffffffe;
	} else if (lastaddr != 0xffffffff) {
		// 引数なしで前回値があれば前回の続き
		addr = lastaddr;
	} else {
		// そうでなければ、今の PC
		addr = md->GetPC();
	}
	return true;
}

// 逆アセンブル: d [<addr> [<cnt>]] (論理、テーブルサーチなし)
void
Debugger::cmd_d()
{
	cmd_d_common(MemdumpMode::Logical_atc);
}

// 逆アセンブル: dt [<addr>] [<cnt>]] (論理、テーブルサーチあり)
void
Debugger::cmd_dt()
{
	cmd_d_common(MemdumpMode::Logical_search);
}

// 逆アセンブル: D [<addr> [<cnt>]] (物理)
void
Debugger::cmd_D()
{
	cmd_d_common(MemdumpMode::Physical);
}

// 逆アセンブル共通
void
Debugger::cmd_d_common(MemdumpMode mode)
{
	uint32 addr;
	saddr_t saddr;
	int cnt;

	cnt = 10;

	// 開始アドレス
	if (GetAddr(addr, d_last_addr) == false) {
		return;
	}
	// 2つ目の引数があれば行数
	if (args.size() > 2) {
		cnt = strtol(args[2].c_str(), NULL, 10);
	}

	saddr.addr  = addr;
	saddr.super = md->IsSuper();

	// XXX そのうちなんとかする
	if (mode == MemdumpMode::Logical_search) {
		saddr.logical = true;
		saddr.search = true;
	} else if (mode == MemdumpMode::Logical_atc) {
		saddr.logical = true;
	}

	for (int i = 0; i < cnt; i++) {
		std::string addrbuf;
		std::string mnemonic;
		std::string mnembuf;
		std::vector<uint8> local_ir;

		if (FormatDumpAddr(saddr, addrbuf)) {
			md->Disassemble(saddr, mnemonic, local_ir);
			// オフライン用に整形
			mnembuf = md->FormatDisasm(mnemonic, local_ir);
		}
		// アドレス変換でバスエラーが起きてもここは表示する
		cons->Print("%-18s %s\n", addrbuf.c_str(), mnembuf.c_str());

		// 変換できなければここで終了
		if (local_ir.size() == 0)
			break;
		saddr.addr += local_ir.size();
	}

	d_last_addr = saddr.addr;
}

// ログレベル設定: L <name> <level>
void
Debugger::cmd_L()
{
	const char *name;
	int level;

	if (args.size() < 3) {
		cons->Print("L <name> <level>\n");
		return;
	}

	name = args[1].c_str();
	level = atoi(args[2].c_str());

	// 短縮形とかエイリアスとか
	// lib/mainapp.cpp に同じものがあるのでどうにかしたほうがいい
	if (strcmp(name, "sch") == 0)
		name = "scheduler";
	if (strcmp(name, "scc") == 0)
		name = "sio";

	// 複数のオブジェクトが同じ logname を持ってもよい (CMMUとか)。
	// all はどうするか
	std::string sname = name;
	bool found = false;
	for (auto& obj : gObjects) {
		if (obj->logname == sname) {
			obj->loglevel = level;
			cons->Print("%s=%d\n", name, obj->loglevel);
			found = true;
		}
	}
	// 見付からない場合はエラー
	if (!found) {
		cons->Print("%s: invalid logname\n", name);
	}
}

// メモリダンプ: m [<addr> [<cnt>]] (論理、テーブルサーチなし)
void
Debugger::cmd_m()
{
	cmd_m_common(MemdumpMode::Logical_atc);
}

// メモリダンプ: mt [<addr> [<cnt>]] (論理、テーブルサーチあり)
void
Debugger::cmd_mt()
{
	cmd_m_common(MemdumpMode::Logical_search);
}

// メモリダンプ: M [<addr> [<cnt>]] (物理)
void
Debugger::cmd_M()
{
	cmd_m_common(MemdumpMode::Physical);
}

// メモリダンプ共通
void
Debugger::cmd_m_common(MemdumpMode mode)
{
	uint32 addr;
	saddr_t saddr;
	int cnt;

	cnt = 10;

	// 開始アドレス
	if (GetAddr(addr, m_last_addr) == false) {
		return;
	}
	// 2つ目の引数があれば行数
	if (args.size() > 2) {
		cnt = strtol(args[2].c_str(), NULL, 10);
	}

	saddr.addr  = addr;
	saddr.super = md->IsSuper();
	// XXX
	if (mode == MemdumpMode::Logical_search) {
		saddr.logical = true;
		saddr.search = true;
	} else if (mode == MemdumpMode::Logical_atc) {
		saddr.logical = true;
	}

	for (int i = 0; i < cnt; i++) {
		std::string addrbuf;
		std::string dumpbuf;
		std::string charbuf;

		if (FormatDumpAddr(saddr, addrbuf) == false) {
			// アドレス変換でバスエラーならダンプもくそもない
			cons->Print("%s\n", addrbuf.c_str());
			// ここで終わる?
			// XXX 16バイト先を試してみるとか?
			break;
		}

		for (int j = 0; j < 8; j++) {
			uint64 paddr;
			// もう一回アドレス変換するのやや冗長だが仕方ない
			if (saddr.IsLogical() && md->MMUEnabled()) {
				// アドレス変換
				paddr = md->TranslateAddr(saddr);
				if ((int64)paddr < 0) {
					// バスエラーが起きたら改行して仕切り直し
					break;
				}
			} else {
				paddr = saddr.addr;
			}

			for (int k = 0; k < 2; k++) {
				uint64 b = vm_phys_peek_8(paddr + k);

				if ((int64)b < 0) {
					dumpbuf += "--";
					charbuf += " ";
				} else {
					uint32 c = (uint32)b;
					dumpbuf += string_format("%02x", c);
					charbuf += string_format("%c",
						(0x20 <= c && c <= 0x7e) ? c : '.');
				}
			}
			dumpbuf += " ";
			saddr.addr += 2;
		}
		cons->Print("%-18s %-40s %s\n",
			addrbuf.c_str(), dumpbuf.c_str(), charbuf.c_str());
	}
	m_last_addr = saddr.addr;
}

// プロンプトに来た最初に表示するいつものやつを再表示する
void
Debugger::cmd_minus()
{
	// 通常レジスタを表示
	std::vector<std::string> tmpargs {{ "r" }};
	md->ShowRegister(cons, tmpargs);

	// 論理アクセス時、ATC ミスが分かったほうが便利だと思う
	saddr_t laddr;
	laddr.addr  = pc;
	laddr.super = md->IsSuper();
	laddr.logical = true;
	laddr.search = false;

	// 現在の PC の位置を逆アセンブル。ここで ir を更新。
	std::string addrbuf;
	std::string mnemonic;
	std::string mnembuf;
	ir.clear();
	// アドレスを表示用に
	if (FormatDumpAddr(laddr, addrbuf)) {
		md->Disassemble(laddr, mnemonic, ir);
		// ライブ表示用に整形
		mnembuf = md->FormatDisasmLive(mnemonic, ir);
	}
	cons->Print("%-18s %s\n", addrbuf.c_str(), mnembuf.c_str());

	// XXX m88k?
	nextpc = pc + ir.size();
}

// ステップ実行: n[<count>?=1] (サブルーチンを飛ばす)
void
Debugger::cmd_n()
{
	if (args.size() > 1) {
		int count;
		count = strtol(args[1].c_str(), NULL, 10);
		if (count < 1) {
			cons->Print(" invalid step count: %d\n", count);
			return;
		}

		n_count = count;
	} else {
		n_count = 1;
	}
	n_enable = true;
	SetNBreakpoint();
}

// n コマンド用のサブルーチンを飛ばすブレークポイントを設定。
// 戻り値は、プロンプトをスキップするなら true。
bool
Debugger::SetNBreakpoint()
{
	// n コマンドは、通常はブレークポイントを次の命令に仕掛けながら
	// 実行するが、いくつかの命令ではブレークポイントではなく、
	// 命令トレース (t コマンド) のように動作させる。
	// トレース動作する命令は以下の通り。
	//  Bcc, BRA
	//  JMP
	//  RTD, RTR, RTS
	//  RTE
	//  FBcc
	//  FDBcc (2ワード目無視)
	//  (RTM は 68020 専用なので対応しない)
#define IsBcc(x)	((((x) & 0xf000) == 0x6000) && (((x) & 0xff00) != 0x6100))
#define IsJMP(x)	(((x) & 0xffc0) == 0x4ec0)
#define IsRTD(x)	(((x) & 0xffff) == 0x4e74)
#define IsRTE(x)	(((x) & 0xffff) == 0x4e73)
#define IsRTR(x)	(((x) & 0xffff) == 0x4e77)
#define IsRTS(x)	(((x) & 0xffff) == 0x4e75)
#define IsFBcc(x)	(((x) & 0xf180) == 0xf080)
#define IsFDBcc(x)	(((x) & 0xf1f8) == 0xf048)

	uint16 op;

	// XXX この辺統一的にしたほうがいい
	// XXX logical peek のような気がする
	op = (((uint16)vm_phys_peek_8(pc)) << 8)
	   | vm_phys_peek_8(pc + 1);

	if (IsBcc(op)
	 || IsJMP(op)
	 || IsRTD(op)
	 || IsRTE(op)
	 || IsRTR(op)
	 || IsRTS(op)
	 || IsFBcc(op)
	 || IsFDBcc(op)
	) {
		n_breakenable = false;
		s_enable = true;
		s_count = 1;
		return false;
	} else {
		n_breakenable = true;
		n_breakaddr = nextpc;
		return true;
	}
}

// デバッガ終了コマンド
void
Debugger::cmd_q()
{
	cons->Print("quit debugger console.\n");
}

// モニター表示
void
Debugger::cmd_show()
{
	if (args.size() != 2) {
		cons->Print("usage: show <monitor>\n");
		for (auto& d : gObjects) {
			if (d->logname != "?") {
				cons->Print(" %s\n", d->logname.c_str());
			}
		}
		return;
	}

	std::string name = args[1];
	for (auto& d : gObjects) {
		if (d->logname == name) {
			ShowMonitor(*d);
			return;
		}
	}
	// 一致しない
	cons->Print("show: no such objects\n");
}

// モニターを更新して表示。
void
Debugger::ShowMonitor(IMonitor& monitor)
{
	// モニタを更新
	TextScreen ts;

	auto size = monitor.GetMonitorSize();
	ts.Init(size.width, size.height);

	monitor.MonitorUpdate(ts);
	ShowTextScreen(ts);
}

// テキストスクリーンを表示。
void
Debugger::ShowTextScreen(TextScreen& monitor)
{
	char sbuf[1024];	// 適当

	// monitor 内部バッファから ShiftJIS バッファを作成。
	// その際属性もエスケープシーケンスで再現する。
	int col = monitor.GetCol();
	int row = monitor.GetRow();
	const uint16 *src = monitor.GetBuf();

	for (int y = 0; y < row; y++) {
		int sy = y * col;
		int sx;
		uint attr;
		char *d;

		memset(sbuf, 0, sizeof(sbuf));
		d = sbuf;
		attr = TA::Normal;
		for (sx = 0; sx < col; sx++) {
			uint a  = src[sy + sx] & 0xff00;
			uint ch = src[sy + sx] & 0x00ff;

			// 属性の変わり目
			if (attr != a) {
				switch (a) {
				 case TA::Normal:
				 case TA::Off:
					*d++ = 0x1b;
					*d++ = '[';
					*d++ = 'm';
					break;
				 case TA::On:
					*d++ = 0x1b;
					*d++ = '[';
					*d++ = '7';
					*d++ = 'm';
					break;
				 case TA::Disable:
					*d++ = 0x1b;
					*d++ = '[';
					*d++ = '2';
					*d++ = 'm';
					break;
				 case TA::Em:
					*d++ = 0x1b;
					*d++ = '[';
					*d++ = '1';
					*d++ = 'm';
					break;
				}
				attr = a;
			}
			*d++ = ch;
		}
		// 行の終わりで一旦属性を戻しておく
		if (attr != TA::Normal) {
			*d++ = 0x1b;
			*d++ = '[';
			*d++ = 'm';
		}
		*d = '\0';

		// XXX 日本語が使われてれば UTF-8 にしないといけないはず

		cons->Print("%s\n", sbuf);
	}
}

// ステップ実行: s[<count>?:1]
// 途中レジスタを表示せずに <count> 命令実行する
void
Debugger::cmd_s()
{
	if (args.size() > 1) {
		int count;
		count = strtol(args[1].c_str(), NULL, 10);
		if (count < 1) {
			cons->Print(" invalid step count: %d\n", count);
			return;
		}

		s_count = count;
	} else {
		s_count = 1;
	}
	s_enable = true;
}

// ステップアウト
void
Debugger::cmd_so()
{
	so_enable = true;
	md->SetStepOut();
}

// トレース実行: t[<count>?:1]
// レジスタを表示しながら <count> 命令実行する
void
Debugger::cmd_t()
{
	if (args.size() > 1) {
		int count;
		count = strtol(args[1].c_str(), NULL, 10);
		if (count < 1) {
			cons->Print(" invalid trace count: %d\n", count);
			return;
		}
		t_count = count;
	} else {
		t_count = 1;
	}
	t_enable = true;
}

// 引数で示されるプレフィックスなしの16進数値を返す。
// 正しく変換できれば値を valp に格納して true を返す。
// そうでなければ false を返す。
bool
Debugger::ParseVerbHex(const char *arg, uint32 *valp)
{
	uint32 val;
	char *end;

	errno = 0;
	val = strtoul(arg, &end, 16);
	if (arg[0] == '\0' || end[0] != '\0') {
		return false;
	}
	if (errno == ERANGE) {
		return false;
	}

	*valp = val;
	return true;
}

// 引数で示されるアドレスを返す。
// '%' で始まればレジスタ。
// そうでなければ16進数値。
// アドレスが正しく取得できればアドレスを addr に格納し真を返す。
// そうでなければエラーメッセージを表示して偽を返す。
bool
Debugger::ParseAddr(const char *arg, uint32 *addrp)
{
	uint32 addr;
	char *end;

	if (arg[0] == '%') {
		// '%' から始まればレジスタ
		uint64 r;
		r = md->GetRegAddr(arg + 1);
		if ((int64)r < 0) {
#if 0 // not yet
			cons->Print("valid register name are: %s%s\n",
				"%d0-%d7,%a0-%a7,%sp,%usp,%isp,%pc",
				",%msp,%vbr,%srp,%crp");
#endif
			cons->Print("invalid register name\n");
			return false;
		}
		addr = (uint32)r;
	} else {
		/* そうでなければ番地指定 */
		errno = 0;
		addr = strtoul(arg, &end, 16);
		if (arg[0] == '\0' || end[0] != '\0') {
			cons->Print("invalid address: '%s'\n", arg);
			return false;
		}
		if (errno == ERANGE) {
			cons->Print("out of range: %s\n", arg);
			return false;
		}
	}

	*addrp = addr;
	return true;
}

// (論理、物理)アドレスを表示用に整形。
// 逆アセンブル表示、メモリダンプ表示等で共通。
// アドレス変換でバスエラーが起きたら false を返す (この場合も str は有効)。
// 末尾を空白パディングはしていないので必要なら "%-18s" 等して使うこと。
//  123456789012345678
// "00112233:44556677:"
bool
Debugger::FormatDumpAddr(saddr_t laddr, std::string& str)
{
	bool rv = true;

	// laddr は常に表示
	str = string_format("%08x", laddr.addr);

	if (laddr.IsLogical() && md->MMUEnabled()) {
		// アドレス変換してみる
		uint64 paddr = md->TranslateAddr(laddr);
		if ((int64)paddr < 0) {
			// MMU 時点でバスエラー
			str += ":BusErr";
			rv = false;
		} else if (paddr == laddr.addr) {
			// PA==VA
			str += "=PA";
		} else {
			// PA!=VA
			str += string_format(":%08x:", (uint32)paddr);
		}
	} else {
		// 物理アクセス
		str += ":";
	}

	return rv;
}

//
// ブレークポイントモニター
//

// コンストラクタ
BreakpointMonitor::BreakpointMonitor()
{
	monitor_size = nnSize(55, 9);
}

void
BreakpointMonitor::MonitorUpdate(TextScreen& monitor)
{
	gDebugger->MonitorBreakpoint(monitor);
}
