//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "console.h"
#include "debugger_private.h"
#include "debugger_m88xx0.h"
#include "m88100disasm.h"
#include "m88100acc.h"
#include "mystring.h"

// newval が oldval から変化していれば太字属性にするマクロ
#define TSBOLD(newval, oldval)	(((newval) != (oldval)) ? TA::Em : TA::Off)

// cpu->name が変化していれば太字属性にするマクロ
#define TSBOLDC(name) TSBOLD(cpu->name, prev.name)

// name の mask で示される部分が変化していれば太字属性にする
#define TSBOLDM(name, mask) TSBOLD((cpu->name & (mask)), (prev.name & (mask)))

// コンストラクタ
DebuggerMD_m88xx0::DebuggerMD_m88xx0(Debugger *parent_, m88kcpu *cpu_)
	: inherited(parent_)
{
	cpu = cpu_;

	// 機種依存変数
	vector_max = 512;
	inst_bits = 32;

	// 奇数にしておくことで初回に必ずアドレス変換を起こさせる。
	last_lpage[0].addr = 0xffffffff;
	last_lpage[1].addr = 0xffffffff;

	cmp_rd = -1;
}

// デストラクタ
DebuggerMD_m88xx0::~DebuggerMD_m88xx0()
{
}

bool
DebuggerMD_m88xx0::MMUEnabled() const
{
	// XXX 命令/データの区別がまだないので、とりあえず命令のほうだけ見る
	uint32 xapr = cpu->cmmu[0].GetAPR(IsSuper() ? 1 : 0);
	return (xapr & m88200::APR_TE);
}

// アドレス変換
uint64
DebuggerMD_m88xx0::TranslateAddr(saddr_t laddr)
{
	uint64 paddr;
	int isdata;
	m88200 *cmmu;

	// XXX とりあえずね
	isdata = (laddr.data) ? 1 : 0;
	if (isdata == 0) {
		cmmu = &cpu->cmmu[0];
	} else {
		cmmu = &cpu->cmmu[1];
	}

	// ページ境界(と初回)でアドレス変換
	saddr_t lpage;
	lpage.addr  = laddr.addr & 0xfffff000;
	lpage.super = laddr.super;
	if (lpage.GetSaddr() != last_lpage[isdata].GetSaddr()) {
		last_lpage[isdata] = lpage;

		bool do_search = laddr.IsSearch();
		paddr = cmmu->TranslatePeek(laddr.addr, laddr.super, do_search);
		if ((int64)paddr < 0) {
			last_lpage[isdata].addr = 0xffffffff;
		}
		last_ppage[isdata] = paddr & 0xfffff000;
	} else {
		paddr = last_ppage[isdata] | (laddr.addr & 0x00000fff);
	}
	return paddr;
}

// op が bb0/tb0 命令なら true を返す
/*static*/ bool
DebuggerMD_m88xx0::IsBBTB0(uint32 op)
{
	return ((op & 0xf8000000) == 0xd0000000 ||		// bb0
	        (op & 0xfc00fe00) == 0xf000d000   );	// tb0
}

// op が bb1 命令なら true を返す
/*static*/ bool
DebuggerMD_m88xx0::IsBB1(uint32 op)
{
	return ((op & 0xf8000000) == 0xd8000000);		// bb1
}

// op が tb1 命令なら true を返す
/*static*/ bool
DebuggerMD_m88xx0::IsTB1(uint32 op)
{
	return ((op & 0xfc00fe00) == 0xf000d800);		// tb1
}

// op が bb1/tb1 命令なら true を返す
/*static*/ bool
DebuggerMD_m88xx0::IsBBTB1(uint32 op)
{
	return IsBB1(op) || IsTB1(op);
}

void
DebuggerMD_m88xx0::BackupRegs()
{
	prev = (m88100reg)*cpu;
}

void
DebuggerMD_m88xx0::SetStepOut()
{
	so_r1 = cpu->r[1];
}

bool
DebuggerMD_m88xx0::IsStepOut() const
{
	return (cpu->xip == so_r1);
}

// レジスタ名からそのレジスタ値を返す。
// メモリダンプのような用途なのでアドレスとして使えるアドレスのみ。
uint64
DebuggerMD_m88xx0::GetRegAddr(const char *name) const
{
	char buf[8];

	if (strcmp(name, "xip") == 0) return cpu->xip;
	if (strcmp(name, "nip") == 0) return cpu->nip;
	if (strcmp(name, "fip") == 0) return cpu->fip;

	// 制御レジスタ
	//if (strcmp(name, "pid") == 0)	return cpu->pid;
	//if (strcmp(name, "psr") == 0)	return cpu->psr;
	//if (strcmp(name, "epsr") == 0)	return cpu->epsr;
	//if (strcmp(name, "ssbr") == 0)	return cpu->ssbr;
	if (strcmp(name, "sxip") == 0)	return cpu->sxip & m88100reg::SIP_MASK;
	if (strcmp(name, "snip") == 0)	return cpu->snip & m88100reg::SIP_MASK;
	if (strcmp(name, "sfip") == 0)	return cpu->sfip & m88100reg::SIP_MASK;
	if (strcmp(name, "vbr") == 0)	return cpu->vbr;
	//if (strcmp(name, "dmt0") == 0)	return cpu->dmt0;
	if (strcmp(name, "dmd0") == 0)	return cpu->dmd0;
	if (strcmp(name, "dma0") == 0)	return cpu->dma0;
	//if (strcmp(name, "dmt1") == 0)	return cpu->dmt1;
	if (strcmp(name, "dmd1") == 0)	return cpu->dmd1;
	if (strcmp(name, "dma1") == 0)	return cpu->dma1;
	//if (strcmp(name, "dmt2") == 0)	return cpu->dmt2;
	if (strcmp(name, "dmd2") == 0)	return cpu->dmd2;
	if (strcmp(name, "dma2") == 0)	return cpu->dma2;
	if (strcmp(name, "sr0") == 0)	return cpu->sr[0];
	if (strcmp(name, "sr1") == 0)	return cpu->sr[1];
	if (strcmp(name, "sr2") == 0)	return cpu->sr[2];
	if (strcmp(name, "sr3") == 0)	return cpu->sr[3];

	for (int i = 0; i < countof(cpu->cr); i++) {
		snprintf(buf, sizeof(buf), "cr%d", i);
		if (strcmp(name, buf) == 0) {
			return cpu->cr[i];
		}
	}

	// FPU レジスタ
	// XXX not implement

	// 通常レジスタ
	for (int i = 0; i < countof(cpu->r); i++) {
		snprintf(buf, sizeof(buf), "r%d", i);
		if (strcmp(name, buf) == 0) {
			return cpu->r[i];
		}
	}

	return (uint64)-1;
}

// レジスタ表示系コマンドのヘルプ
/*static*/ const HelpMessages
DebuggerMD_m88xx0::HelpMsgRegister = {
	{ "r",				"通常レジスタを表示" },
	{ "rc",				"制御レジスタを表示" },
	{ "rf",				"FPU レジスタを表示" },
	{ "ro",				"その他のレジスタを表示" },
	{ "ra<id>",			"ATC を表示" },
	{ "rd<id> [<set>]",	"データキャッシュを表示" },
	{ "rm<id>",			"CMMU レジスタを表示" },
};

const HelpMessages&
DebuggerMD_m88xx0::GetRegisterHelp() const
{
	return HelpMsgRegister;
}

// レジスタ表示系コマンド
bool
DebuggerMD_m88xx0::ShowRegister(Console *cons,
	const std::vector<std::string>& args)
{
	int cmmu_id;

	// 余分な引数は無視する?

	if (args[0] == "r") {
		ShowRegMain(cons);
		return true;
	}
	if (args[0] == "rc") {
		ShowRegCtrl(cons);
		return true;
	}
	if (args[0] == "rf") {
		ShowRegFPU(cons);
		return true;
	}
	if (args[0] == "ro") {
		ShowRegOther(cons);
		return true;
	}

	if (MatchCMMUCmd(args[0], "ra", &cmmu_id)) {
		parent->ShowMonitor(*gMPU88200ATC[7 - cmmu_id]);
		return true;
	}
	if (MatchCMMUCmd(args[0], "rd", &cmmu_id)) {
		ShowRegCache(cons, args, cmmu_id);
		return true;
	}
	if (MatchCMMUCmd(args[0], "rm", &cmmu_id)) {
		ShowRegMMU(cons, cmmu_id);
		return true;
	}

	// 該当なし
	return false;
}

// CMMU ID 入りコマンド名を照合する。
// cmdname が basename + <id> の時、*idp に id を格納して true を返す。
// 一致しないか id が不正なら false を返す。
bool
DebuggerMD_m88xx0::MatchCMMUCmd(const std::string& cmdname,
	const char *basename, int *idp)
{
	int baselen = strlen(basename);

	if (strncmp(cmdname.c_str(), basename, baselen) != 0) {
		return false;
	}

	// 先頭が一致したら id を調べる
	int id;
	const char *start = cmdname.c_str() + baselen;
	char *end;
	errno = 0;
	id = strtoul(start, &end, 10);
	if (*start == '\0' || *end != '\0' || errno == ERANGE) {
		return false;
	}

	*idp = id;
	return true;
}

void
DebuggerMD_m88xx0::ShowRegMain(Console *cons)
{
	TextScreen s(80, 8);

/*
r0 :00000000  r8 :00000000  r16:00000000  r24:00000000   b:bs=1 a:lo=1 8:ls=0
r1 :00700000  r9 :00000000  r17:00000000  r25:00000000
r2 :0070e1a0  r10:00000000  r18:00000000  r26:00000000
*/

	for (int i = 0; i < countof(cpu->r); i++) {
		s.Print((i / 8) * 14, i % 8, TSBOLDC(r[i]), "r%-2d:%08x", i, cpu->r[i]);
	}

	// cmp 以降最初の分岐命令で結果レジスタの条件ビットマップを表示する。
	// 分岐命令は bb1 #3 みたいな形式なのでそれだけでは条件演算子が分からない
	// のと、bb1 #n 自体は cmp 以外に対しても使えるので、cmp 命令以降最初の
	// 分岐命令でのみ表示してみる。
	if (cmp_rd >= 0 && (IsBBTB0(cpu->opX) || IsBBTB1(cpu->opX))) {
		static const char * const cmpstr[] = {
			NULL, NULL, "eq", "ne", "gt", "le", "lt", "ge",
			"hi", "ls", "lo", "hs",
		};
		int x = 0;
		int y = 0;
		uint32 res = cpu->r[cmp_rd];
		for (int b = 11; b >= 2; b--) {
			s.Print(57 + x * 7, y, "%x:%s=%d", b, cmpstr[b],
				(res & (1 << b)) ? 1 : 0);
			x++;
			if (x > 2) {
				x = 0;
				y++;
			}
		}
	}

	parent->ShowTextScreen(s);
}

void
DebuggerMD_m88xx0::ShowRegCtrl(Console *cons)
{
	TextScreen s(80, 6);
	int x;
	int y;

/*
0         1         2         3         4         5         6         7
0123456789012345678901234567890123456789012345678901234567890123456789012345678
pid (cr0):12345678(Arch=$00 Ver=$00 M/C=Checker)       ssbr(cr3):00000000
psr (cr1):12345678(S,LE,SER,Cy SFD1,MXM,IND,SFRZ)      vbr (cr7):00000000
epsr(cr2):12345678(S,LE,SER,Cy SFD1,MXM,IND,SFRZ)      sr0(cr17):00000000
xip:00000000  opX:00000000(--)  sxip(cr4):00000000:VE  sr1(cr18):00000000
nip:00000000  opF:00000000(--)  snip(cr5):00000000:--  sr2(cr19):00000000
fip:00000000                    sfip(cr6):00000000:VE  sr3(cr20):00000000
*/

#define PRINT_B(x, y, reg, mask, fmt, arg)	do { \
	uint32 _new = cpu->reg & (mask);	\
	uint32 _old = prev.reg & (mask);	\
	s.Print((x), (y), TSBOLD(_new, _old), fmt, (arg));	\
} while (0)

	// pid(cr0)
	x = 0;
	y = 0;
	s.Print(x,      y, TSBOLDC(pid), "pid (cr0):%08x(", cpu->pid);
	PRINT_B(x + 19, y, pid, m88100reg::PID_REV_MASK, "Arch=$%02x",
		(_new >> 8) & 0xff);
	PRINT_B(x + 28, y, pid, m88100reg::PID_VER_MASK, "Ver=$%02x",
		(_new >> 1) & 0x7f);
	PRINT_B(x + 36, y, pid, m88100reg::PID_MASTER, "M/C=%s)",
		(_new & m88100reg::PID_MASTER) ? "Master" : "Checker");

	// psr(cr1)
	y++;
	s.Print(x,      y, TSBOLDC(psr), "psr (cr1):%08x", cpu->psr);
	s.Print(x + 18, y, "(s,le,ser,cy sfd1,mxm,ind,sfrz)");
	PRINT_B(x + 19, y, psr, m88100reg::PSR_SUPER, "%c", _new ? 'S' : '-');
	PRINT_B(x + 21, y, psr, m88100reg::PSR_BO_LE, "%s", _new ? "LE" : "BE");
	PRINT_B(x + 24, y, psr, m88100reg::PSR_SER,   "%s", _new ? "CON" : "SER");
	PRINT_B(x + 28, y, psr, m88100reg::PSR_C,     "%s", _new ? "Cy" : "--");
	PRINT_B(x + 31, y, psr, m88100reg::PSR_SFD1,  "%s", _new ? "SFD1" : "----");
	PRINT_B(x + 36, y, psr, m88100reg::PSR_MXM,   "%s", _new ? "MXM" : "---");
	PRINT_B(x + 40, y, psr, m88100reg::PSR_IND,   "%s", _new ? "IND" : "---");
	PRINT_B(x + 44, y, psr, m88100reg::PSR_SFRZ,  "%s", _new ? "SFRZ" : "----");

	// epsr(cr2)
	y++;
	s.Print(x,      y, TSBOLDC(psr), "epsr(cr2):%08x", cpu->epsr);
	s.Print(x + 18, y, "(s,le,ser,cy sfd1,mxm,ind,sfrz)");
	PRINT_B(x + 19, y, epsr, m88100reg::PSR_SUPER, "%c", _new ? 'S' : '-');
	PRINT_B(x + 21, y, epsr, m88100reg::PSR_BO_LE, "%s", _new ? "LE" : "BE");
	PRINT_B(x + 24, y, epsr, m88100reg::PSR_SER,   "%s", _new ? "CON" : "SER");
	PRINT_B(x + 28, y, epsr, m88100reg::PSR_C,     "%s", _new ? "Cy" : "--");
	PRINT_B(x + 31, y, epsr, m88100reg::PSR_SFD1,  "%s", _new ? "SFD1":"----");
	PRINT_B(x + 36, y, epsr, m88100reg::PSR_MXM,   "%s", _new ? "MXM" : "---");
	PRINT_B(x + 40, y, epsr, m88100reg::PSR_IND,   "%s", _new ? "IND" : "---");
	PRINT_B(x + 44, y, epsr, m88100reg::PSR_SFRZ,  "%s", _new ? "SFRZ":"----");

	// *IP
	x = 0;
	y = 3;
	s.Print(x,      y, "xip:%08x  opX:%08x(%c%c)",
		cpu->xip, (uint32)cpu->opX,
		cpu->OpIsBusErr(cpu->opX) ? 'B' : '-',
		cpu->OpIsDelay(cpu->opX)  ? 'D' : '-');
	s.Print(x,  y + 1, "nip:%08x  opF:%08x(%c%c)",
		cpu->nip, (uint32)cpu->opF,
		cpu->OpIsBusErr(cpu->opF) ? 'B' : '-',
		cpu->OpIsDelay(cpu->opX)  ? 'D' : '-');
	s.Print(x,  y + 2, "fip:%08x", cpu->fip);

	// S*IP
	x = 32;
	for (int i = 0; i < 3; i++) {
		s.Print(x, y + i, TSBOLDC(cr[4 + i]), "%s(cr%d):%08x:%c%c",
			m88100reg::sipname[i], 4 + i,
			(cpu->cr[4 + i] & m88100reg::SIP_MASK),
			(cpu->cr[4 + i] & m88100reg::SIP_V) ? 'V' : '-',
			(cpu->cr[4 + i] & m88100reg::SIP_E) ? 'E' : '-');
	}

	x = 55;
	y = 0;
	// ssbr(cr3)
	s.Print(x, y++, TSBOLDC(ssbr), "ssbr(cr3):%08x", cpu->ssbr);
	// vbr(cr7)
	s.Print(x, y++, TSBOLDC(vbr),  "vbr (cr7):%08x", cpu->vbr);

	// SR*
	for (int i = 0; i < 4; i++) {
		int rn = 17 + i;
		s.Print(x, y + i, TSBOLDC(cr[rn]), "sr%d(cr%d):%08x",
			i, rn, cpu->cr[rn]);
	}

	parent->ShowTextScreen(s);
}

// CMMU のデータキャッシュを表示
// rd<N>       なら概要表示。
// rd<N> <set> なら個別セットの詳細表示。
void
DebuggerMD_m88xx0::ShowRegCache(Console *cons,
	const std::vector<std::string>& args, int id)
{
	if (args.size() < 2) {
		ShowRegCacheOverview(cons, id);
	} else {
		int set;
		char *end;
		errno = 0;
		set = strtol(args[1].c_str(), &end, 16);
		if (end == &args[1][0] || *end != '\0' || errno == ERANGE) {
			cons->Print("<set> must be a number\n");
			return;
		}
		if (set < 0 || set > 256) {
			cons->Print("<set> must be in 00..ff\n");
			return;
		}
		ShowRegCacheSet(cons, id, set);
	}
}

// CMMU データキャッシュの概要表示
void
DebuggerMD_m88xx0::ShowRegCacheOverview(Console *cons, int id)
{
	TextScreen s(70, 17);
	m88200 *cmmu;

	// まだ今の所決め打ち
	if (id == 7) {
		cmmu = &cpu->cmmu[0];
	} else {
		cmmu = &cpu->cmmu[1];
	}
	cmmu->MonitorCacheOverview(s, 0, false);
	parent->ShowTextScreen(s);
}

// CMMU データキャッシュの指定セットの詳細表示
void
DebuggerMD_m88xx0::ShowRegCacheSet(Console *cons, int id, int setidx)
{
	TextScreen s(70, 5);
	m88200 *cmmu;

	// まだ今の所決め打ち
	if (id == 7) {
		cmmu = &cpu->cmmu[0];
	} else {
		cmmu = &cpu->cmmu[1];
	}
	cmmu->MonitorCacheSet(s, 0, setidx);
	parent->ShowTextScreen(s);
}

void
DebuggerMD_m88xx0::ShowRegFPU(Console *cons)
{
	static const char * const fcrname[] = {
		"fpecr",
		"fphs1",
		"fpls1",
		"fphs2",
		"fpls2",
		"fppt",
		"fprh",
		"fprl",
		"fpit",
		"fpsr",	// [9]  62
		"fpcr",	// [10] 63
	};
	TextScreen s(80, 4);

	for (int i = 0; i < 9; i++) {
		s.Print((i / 4) * 22, i % 4, TSBOLDC(fcr[i]),
			"%-5s(fcr%d):%08x", fcrname[i], i, cpu->fcr[i]);
	}

	s.Print(44, 2, TSBOLDC(fpsr), "fpsr(fcr62):%08x", cpu->fpsr);
	s.Print(44, 3, TSBOLDC(fpcr), "fpcr(fcr63):%08x", cpu->fpcr);

	parent->ShowTextScreen(s);
}

void
DebuggerMD_m88xx0::ShowRegMMU(Console *cons, int id)
{
	m88200 *cmmu;

	// まだ今の所決め打ち
	if (id == 7) {
		cmmu = &cpu->cmmu[0];
	} else {
		cmmu = &cpu->cmmu[1];
	}
	parent->ShowMonitor(*cmmu);
}

void
DebuggerMD_m88xx0::ShowRegOther(Console *cons)
{
	TextScreen s(80, 3);
/*
0         1         2         3         4         5         6         7
0123456789012345678901234567890123456789012345678901234567890123456789012345678
dmt0(cr8) :0000(B,S,D,L,Rxx,S,---B,W,V) dmd0(cr9) :00000000 dma0(cr10):00000000
dmt0(cr11):0000(B,U,D,L,Rxx,S,HH--,W,V) dmd0(cr9) :00000000 dma0(cr10):00000000
dmt0(cr14):0000(B, ,D,L,Rxx,S,LLLL,W,V) dmd0(cr9) :00000000 dma0(cr10):00000000
*/
	for (int i = 0; i <= 2; i++) {
		int dt =  8 + i * 3;
		int dd =  9 + i * 3;
		int da = 10 + i * 3;

		s.Print(0, i, TSBOLDC(cr[dt]), "dmt%d(cr%-2d):%04x",
			i, dt, (cpu->cr[dt] & 0xffff));
		s.Print(15, i, "(b,s,d,l,rx, s,en  ,w,v)");
		PRINT_B(16, i, cr[dt], m88100reg::DM_BO,     "%c", _new ? 'B' : '-');
		PRINT_B(18, i, cr[dt], m88100reg::DM_DAS,    "%c", _new ? 'S' : 'U');
		PRINT_B(20, i, cr[dt], m88100reg::DM_DOUB1,  "%c", _new ? 'D' : '-');
		PRINT_B(22, i, cr[dt], m88100reg::DM_LOCK,   "%c", _new ? 'L' : '-');
		// カンマを前詰め。かつ DREG がボールドでもカンマはボールドにしない
		PRINT_B(24, i, cr[dt], m88100reg::DM_DREG_MASK, "r%d", _new >> 7);
		s.Puts(",");
		PRINT_B(28, i, cr[dt], m88100reg::DM_SIGNED, "%c", _new ? 'S' : '-');
		PRINT_B(30, i, cr[dt], m88100reg::DM_EN_MASK,"%s",
			m88100reg::dmt_en_str[(_new >> 2) & 0xf]);
		PRINT_B(35, i, cr[dt], m88100reg::DM_WRITE,  "%c", _new ? 'W' : '-');
		PRINT_B(37, i, cr[dt], m88100reg::DM_VALID,  "%c", _new ? 'V' : '-');

		s.Print(40, i, TSBOLDC(cr[dd]), "dmd%d(cr%-2d):%08x",
			i, dd, cpu->cr[dd]);
		s.Print(60, i, TSBOLDC(cr[dd]), "dma%d(cr%2d):%08x",
			i, da, cpu->cr[da]);
	}

	parent->ShowTextScreen(s);
}

// 逆アセンブル
bool
DebuggerMD_m88xx0::Disassemble(saddr_t laddr,
	std::string& mnemonic, std::vector<uint8>& bin)
{
	m88100disasm dis(this);
	if (dis.Exec(laddr) == false) {
		return false;
	}
	mnemonic = dis.text;
	bin = dis.bin;
	return true;
}

// 逆アセンブルをフォーマット (オフライン版)
std::string
DebuggerMD_m88xx0::FormatDisasm(const std::string& mnemonic,
	const std::vector<uint8>& bin)
{
	// 命令はすべて32ビット固定長
	uint32 op = (bin[0] << 24) | (bin[1] << 16) | (bin[2] << 8) | bin[3];
	return string_format("%08x %s", op, mnemonic.c_str());
}

// 逆アセンブルをフォーマット (オンライン版)
std::string
DebuggerMD_m88xx0::FormatDisasmLive(const std::string& mnemonic,
	const std::vector<uint8>& bin)
{
	// 命令はすべて32ビット固定長
	uint32 op = (bin[0] << 24) | (bin[1] << 16) | (bin[2] << 8) | bin[3];

	// 遅延スロットは分かりやすく表示したい
	std::string delay;
	if (cpu->OpIsDelay(cpu->opX)) {
		delay = "(delayed) ";
	}

	const char *cond = CondStr(op);
	return string_format("%08x %s%s%s",
		op, delay.c_str(), mnemonic.c_str(), cond);
}

#define TAKE_IF(expr) do { \
	if ((expr)) \
		return " (will take)";	\
	else \
		return " (will not take)";	\
} while (0)

// 条件命令なら、成立可否などの文字列を返す。
const char *
DebuggerMD_m88xx0::CondStr(uint32 op)
{
	uint32 s = m88100opf_S1(op);

	if (IsBBTB0(op)) {
		// 11010N BBBBB SSSSS dddddd dd dddddddd	bb0
		// 111100 BBBBB SSSSS 110100 0V VVVVVVVV	tb0
		// rS の bit B が %0 ならブランチ/トラップ
		cmp_rd = -1;
		uint32 b = m88100opf_B5(op);
		TAKE_IF((cpu->r[s] & (1 << b)) == 0);
	}
	if (IsBB1(op)) {
		// 11011N BBBBB SSSSS dddddd dd dddddddd	bb1
		// rS の bit B が %1 ならブランチ
		cmp_rd = -1;
		// 対 r0 なら絶対成立しないので nop
		if (s == 0) {
			return " (nop)";
		}
		uint32 b = m88100opf_B5(op);
		TAKE_IF((cpu->r[s] & (1 << b)) != 0);
	}
	if (IsTB1(op)) {
		// 111100 BBBBB SSSSS 110110 0V VVVVVVVV	tb1
		// rS の bit B が %1 ならトラップ
		cmp_rd = -1;
		// 対 r0 なら sync として使っており逆アセンブラ側で表示を加工してある
		// のでこちら側では対処不要。
		if (s == 0) {
			return "";
		}
		uint32 b = m88100opf_B5(op);
		TAKE_IF((cpu->r[s] & (1 << b)) != 0);
	}
	if ((op & 0xf8000000) == 0xe8000000 ||	// bcnd
	    (op & 0xfc00fe00) == 0xf000e800)	// tcnd
	{
		// 11101N MMMMM SSSSS dddddd dd dddddddd	bcnd
		// rS が M5 で示される条件にマッチすればブランチ/トラップ
		cmp_rd = -1;
		uint32 m5 = m88100opf_M5(op);
		TAKE_IF(
		    (m5 == 0x2 && cpu->r[s] == 0) ||		// eq0
		    (m5 == 0xd && cpu->r[s] != 0) ||		// ne0
		    (m5 == 0x1 && (int32)cpu->r[s] > 0)  ||	// gt0
		    (m5 == 0xc && (int32)cpu->r[s] < 0)  ||	// lt0
		    (m5 == 0x3 && (int32)cpu->r[s] >= 0) ||	// ge0
		    (m5 == 0xe && (int32)cpu->r[s] <= 0)	// le0
		);
	}

	if ((op & 0xfc00fce0) == 0xf4007c00 ||	// cmp rD,rS1,rS2
	    (op & 0xfc000000) == 0x7c000000   )	// cmp rD,rS1,imm16
	{
		// 次の条件分岐命令で結果レジスタを表示するため、rD だけ覚えておく
		cmp_rd = m88100opf_D(op);
	}

	return "";
}
