//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "branchhistory.h"
#include "bus.h"
#include "cvprompt.h"
#include "textscreen.h"
#include <array>

using ConstStringPair = std::pair<const std::string, const std::string>;
using HelpMessages = std::vector<ConstStringPair>;

// 諸情報付きのアドレス
struct saddr_t
{
	uint32 addr = 0;		// アドレス(32bit)

	// addr が論理アドレスなら true。
	// MMU の有効/無効状態はここにセットするのではなく、メモリアクセスに
	// 際してアドレス変換を行う側がその時点で改めてチェックするので、
	// ここはどちらかというと物理アドレスなら false くらいの意味合い。
	bool logical = false;

	// addr が論理アドレスの場合に、テーブルサーチも行うなら true。
	// false なら TT、ATC のみ検索する。
	bool search = false;

	// addr がスーパバイザアクセスなら true。
	bool super = false;

	// addr がデータアクセスなら true。
	bool data = false;

	bool IsLogical() const	{ return logical; }
	bool IsSearch() const	{ return search; }
	bool IsSuper() const	{ return super; }
	bool IsData() const		{ return data; }
	bool IsInst() const		{ return !data; }

	// (論理)アドレスと Super/User ビットを合わせた 33bit アドレスを返す
	uint64 GetSaddr() const {
		uint64 sa = addr;
		if (IsSuper()) {
			sa |= 0x100000000ULL;
		}
		return sa;
	}
};

enum MemdumpMode {
	Physical = 0,		// 物理アドレス
	Logical_atc,		// 論理アドレス (ただし TT、ATC のみ検索)
	Logical_search,		// 論理アドレス (TT、ATC、テーブルサーチすべて行う)
};

class Console;
class DebuggerMD;
class Disasm;

class Debugger final
{
	friend class DebuggerMD_m680x0;
	friend class DebuggerMD_m88xx0;
 private:
	// 型
	enum CommandAction {
		Stay,		// デバッガプロンプトに留まる
		Leave,		// デバッガプロンプトを抜けて VM を実行する
		Quit,		// デバッガとの接続を終了する
	};
	using cmdfunc_t = void (Debugger::*)();
	typedef struct {
		const char *name;
		cmdfunc_t func;
		CommandAction action;
	} cmddef_t;

	// ブレークポイント種別
	enum BreakpointType {
		Unused = 0,
		Address,		// このアドレスに来る直前
		Memory,			// このアドレスをアクセスした直後
		Exception,		// この例外を検出した直後
		Instruction,	// この命令を実行する直前
	};

	// ブレークポイント
	typedef struct {
		BreakpointType type {};	// 種別
		// 種別ごとのパラメータ
		union {
			uint32 addr {};		// ターゲットアドレス (Address/Memory)
			struct {
				int vec1;		// ベクタ番号 (開始)
				int vec2;		// ベクタ番号 (終了)
			};
			struct {
				uint32 inst;	// 命令ワード
				uint32 mask;	// マスク
			};
		};

		uint32 matched {};		// 成立回数 (積算)

		// skip はユーザ指定値。ただし
		// -1 なら常にスキップ、つまりブレークせずカウントのみ行う、
		// 0 ならスキップなし、つまり成立ごとにブレーク(これがデフォルト)、
		// n(>0) なら n 回の成立をスキップし、(n+1) 回目でブレークの意。
		int32 skip {};			// スキップ回数 (ユーザ指定値)
		int32 skipremain {};	// 残りスキップ回数 (0 で成立)
	} breakpoint_t;
	static const int MAX_BREAKPOINTS = 8;

 public:
	Debugger();
	~Debugger() { }

	void Init();
	void ThreadRun();
	bool Check();
	void NotifyException(int vector);

	// ブレークポイントモニタを更新
	void MonitorBreakpoint(TextScreen& monitor);

 private:
	bool Accept();
	bool AcquirePrompt();
	CommandAction MainLoop();
	bool CheckAllBreakpoints();
	bool CheckBreakpointInst(breakpoint_t&);
	void ParseCmdbuf();
	// コマンド名は大文字小文字を区別する関係でスネークスタイル。
	void cmd_b();
	void cmd_bi();
	void cmd_bm();
	void cmd_bv();
	void cmd_b_list();
	void cmd_b_set(BreakpointType);
	void cmd_brhist();
	void cmd_bx();
	void cmd_c();
	void cmd_d();
	void cmd_dt();
	void cmd_D();
	void cmd_exhist();
	void cmd_h();
	void cmd_L();
	void cmd_m();
	void cmd_mt();
	void cmd_M();
	void cmd_minus();
	void cmd_n();
	void cmd_q();
	void cmd_s();
	void cmd_so();
	void cmd_show();
	void cmd_t();

	int AddBreakpoint(const breakpoint_t&);
	void RecalcInstMask();
	bool GetAddr(uint32& addr, uint32& lastaddr);
	void cmd_d_common(MemdumpMode mode);
	void cmd_hist_common(BranchHistory& hist, uint64 flag);
	void cmd_m_common(MemdumpMode mode);
	bool ParseVerbHex(const char *arg, uint32 *valp);
	bool ParseAddr(const char *arg, uint32_t *addrp);
	bool SetNBreakpoint();

	static const HelpMessages HelpMsgMain;
	void Help(const HelpMessages msgs);

	// モニターを更新して表示
	void ShowMonitor(IMonitor& monitor);
	// テキストスクリーンを表示
	void ShowTextScreen(TextScreen& screen);

	// アドレスをダンプ表示用に整形して返す
	bool FormatDumpAddr(saddr_t addr, std::string& addrstr);

	DebuggerMD *md = NULL;		// 機種依存部分

	Console *cons = NULL;		// 入出力
	const char *prompt;			// プロンプト文字列
	std::string cmdbuf {};		// 現在のコマンドライン
	std::string last_cmdbuf {};	// 直前のコマンドライン
	std::vector<std::string> args {};

	uint32 pc = 0;
	uint32 nextpc = 0;
	std::vector<uint8> ir {};	// この後実行する命令のバイナリ列
	std::array<breakpoint_t, MAX_BREAKPOINTS> bpoint {};
	bool bc_enable = false;
	uint32 bc_addr = 0;
	uint32 bi_inst = 0;
	int bi_inst_bytes = 0;
	int bi_need_bytes = 0;
	int bv_vector = -1;
	uint32 d_last_addr = 0;
	uint32 m_last_addr = 0;
	bool n_enable = false;
	bool n_breakenable = false;
	uint32 n_breakaddr = 0;
	uint32 n_count = 0;
	bool s_enable = false;
	uint32 s_count = 0;
	bool so_enable = false;
	bool t_enable = false;
	uint32 t_count = 0;

	// 一度VMを実行して再びプロンプトに来たら true
	bool is_continued = false;

	static std::vector<cmddef_t> cmdtable;
};

// デバッガの CPU 依存部分
class DebuggerMD
{
 public:
	DebuggerMD(Debugger *arg) {
		parent = arg;
	}
	virtual ~DebuggerMD() { }

	// CPU のリクエストフラグに flag を立てる
	virtual void ReqSet(uint32 flag) = 0;
	// CPU のリクエストフラグから flag を落とす
	virtual void ReqClr(uint32 flag) = 0;

	// 特権モードなら true を返す
	virtual bool IsSuper() const = 0;
	// MMU が有効なら true を返す
	virtual bool MMUEnabled() const = 0;

	// アドレス変換をする。
	// MMUEnabled() が true であることは呼び出し側が確認すること。
	// 変換できなければ (uint64)-1 を返す。
	virtual uint64 TranslateAddr(saddr_t laddr) = 0;

	// 現在の命令先頭アドレスを返す
	virtual uint32 GetPC() const = 0;

	// ステップアウトを設定する
	virtual void SetStepOut() = 0;
	// ステップアウトなら true を返す
	virtual bool IsStepOut() const = 0;

	// name で示されるのレジスタの値を返す。
	// メモリダンプで指定するケースを想定しているので、明らかにアドレスでは
	// ないレジスタは含まなくてよい (m680x0 の SR など)。ただし Dn は含めて
	// よい。また m680x0 の SRP/CRP は下位32ビットでよい。
	// name が正しくない場合は (uint64)-1 を返す。
	virtual uint64 GetRegAddr(const char *name) const = 0;

	// 現在のレジスタセットを内部にバックアップする
	virtual void BackupRegs() = 0;

	// レジスタ表示系のコマンドを実行。処理すれば true を返す。
	virtual bool ShowRegister(Console *cons,
		const std::vector<std::string>& args) = 0;
	// レジスタ表示系コマンドのヘルプメッセージを取得
	virtual const HelpMessages& GetRegisterHelp() const = 0;

	// ブランチ履歴・例外履歴(オブジェクト)を取得
	virtual BranchHistory& GetBrHist() const = 0;
	virtual BranchHistory& GetExHist() const = 0;

	// 直近の命令でアクセスした laddr が一致したら true を返す
	// 1 命令で複数にアクセスする場合を考慮して、md 側で判断する
	virtual bool CheckLEA(uint32 laddr) = 0;

	// 逆アセンブル。
	// laddr の位置の命令を一つ逆アセンブルする。
	// 成功すれば true を返し、失敗すれば(?) false を返す。
	// 成功した場合 mnemonic に逆アセンブルした文字列が、
	// bin にこの命令のバイナリ列が格納されている。
	// mnemonic は単純に命令部分のみ、16進ダンプや付随情報は含まない。
	// bin はメモリイメージのバイト順で格納すること。
	virtual bool Disassemble(saddr_t laddr,
		std::string& mnemonic, std::vector<uint8>& bin) = 0;

	// 逆アセンブルをフォーマット。
	// オフライン版は逆アセンブルウィンドウなど、
	// オンライン版(Live) はデバッガプロンプトで表示するやつ。
	virtual std::string FormatDisasm(const std::string& mnemonic,
		const std::vector<uint8>& bin) = 0;
	virtual std::string FormatDisasmLive(const std::string& mnemonic,
		const std::vector<uint8>& bin) = 0;

	// ベクタ番号の最大値 (+1)
	int vector_max = 0;
	// 1命令の最小ビット数
	uint inst_bits = 0;

 protected:
	Debugger *parent;		// 親クラス
};

#define NORM		"\x1b[0m"
#define BOLD		"\x1b[1m"
#define BOLDIF(n)	((n) ? BOLD : "")
