//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "netdriver_bpf.h"
#include "cvprompt.h"
#include "mystring.h"
#include <fcntl.h>
#include <ifaddrs.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#if defined(HAVE_SYS_EVENT_H)
#include <sys/event.h>
#endif
#if defined(HAVE_KQUEUE_SYS_EVENT_H)
#include <kqueue/sys/event.h>
#endif
#include <net/bpf.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_types.h>

// コンストラクタ
NetDriverBPF::NetDriverBPF(EthernetDevice *parent_,
	std::mutex *mtx_, std::condition_variable *cv_)
	: inherited("BPF", parent_, mtx_, cv_)
{
	fd = -1;
}

// デストラクタ
NetDriverBPF::~NetDriverBPF()
{
	Close();
}

// 初期化 & オープン
bool
NetDriverBPF::Init()
{
	struct ifreq ifr;
	struct bpf_program prog;
	u_int val;
	struct bpf_insn insn[] = {
		// すべてのパケットを全長受信
		BPF_STMT(BPF_RET + BPF_K, (u_int)-1),
	};

	// XXX tap はデバイスを開くとインタフェースが出来る、という方向だが、
	//     bpf はインタフェースを指定してデバイスを開く、という方向で、逆…。

	if (ifname.empty()) {
		// インタフェース指定がなければそれっぽいのを探すか
		ifname = FindInterface();
		if (ifname.empty()) {
			warn("NetDriverBPF: no interface specified");
			return false;
		}
	}

	// 最大値は適当
	fd = -1;
	for (int i = -1; i < 10; i++) {
		// i == -1 なら /dev/bpf、0..10 は /dev/bpfN。
		// まず /dev/bpf を cloning デバイスだと思って開いてみる。
		// そうでなければ従来どおり 0 から順番に試してみる。
		if (i == -1) {
			devpath = "/dev/bpf";
		} else {
			devpath = string_format("/dev/bpf%d", i);
		}

		fd = open(devpath.c_str(), O_RDWR);
		if (fd == -1) {
			warn("NetDriverBPF.Open: skip %s", devpath.c_str());
			continue;
		}
		break;
	}
	if (fd == -1) {
		warnx("NetDriverBPF.Open: no bpf device available");
		return false;
	}

	// インタフェースにバインド
	memset(&ifr, 0, sizeof(ifr));
	strlcpy(ifr.ifr_name, ifname.c_str(), sizeof(ifr.ifr_name));
	if (ioctl(fd, BIOCSETIF, &ifr) == -1) {
		warn("NetDriverBPF.Open: %s(%s) BIOCSETIF failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}
	// バインドまで出来たところでログを出しておきたい
	// XXX 本当は warn じゃないよね
	warnx("NetDriverBPF.Open: %s opened for %s",
		devpath.c_str(), ifname.c_str());

	if (ioctl(fd, BIOCPROMISC, (void *)0) == -1) {
		warn("NetDriverBPF.Open: %s(%s): BIOCPROMISC failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}

	val = 1;
	if (ioctl(fd, BIOCIMMEDIATE, &val) == -1) {
		warn("NetDriverBPF.Open: %s(%s): BIOCIMMEDIATE failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}

	if (ioctl(fd, BIOCSHDRCMPLT, &val) == -1) {
		warn("NetDriverBPF.Open: %s(%s): BIOCSHDRCMPLT failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}

	// XXX BIOCFEEDBACK...
	warn("NetDriverBPF.Open: %s(%s): ローカルホストとのみ通信はできません",
		devpath.c_str(), ifname.c_str());

	// フィルタをセット
	prog.bf_len = countof(insn);
	prog.bf_insns = insn;
	if (ioctl(fd, BIOCSETF, &prog) == -1) {
		warn("NetDriverBPF.Open: %s(%s): BIOCSETF failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}

	// バッファサイズを取得
	if (ioctl(fd, BIOCGBLEN, &bpfbuflen) == -1) {
		warn("NetDriverBPF.Open: %s(%s): BIOCGBLEN failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}

	// 受信用バッファを確保
	try {
		bpfbuf.reset(new char [bpfbuflen]);
	} catch (...) {
		warnx("NetDriverBPF.Open: %s(%s): memory allocation failed",
			devpath.c_str(), ifname.c_str());
		goto abort;
	}

	// オープンできたので受信スレッド起動。
	if (InitRecvThread() == false) {
		goto abort;
	}

	// bpf では ifup スクリプトは実行しない

	return true;

 abort:
	close(fd);
	fd = -1;
	return false;
}

// クローズ
// (private)
void
NetDriverBPF::Close()
{
	if (IsOpen()) {
		close(fd);
		fd = -1;
	}
}

// パケットを送信する
bool
NetDriverBPF::SendPacket(const std::vector<uint8>& data)
{
	int n;

	assert(IsOpen());

	putlog(2, "SendPacket %zu bytes", data.size());
	tx_pkts++;
	tx_bytes += data.size();

	n = write(fd, data.data(), data.size());
	if (n == -1) {
		warn("NetDriverBPF.SendPacket: write failed: fd=%d datalen=%zu",
			fd, data.size());
		return false;
	}

	if (n < (int)data.size()) {
		return false;
	}
	return true;
}

// 最初に見付かった使えそうなインタフェース名を返す。
// 物理インタフェースが明らかに1つしかないところで、
// インタフェース名の指定を省略できるようにするため。
std::string
NetDriverBPF::FindInterface()
{
	char name[IFNAMSIZ];
	struct ifaddrs *ifa_list, *ifa;
	struct sockaddr_dl *dl;

	if (getifaddrs(&ifa_list) == -1) {
		return "";
	}

	name[0] = '\0';
	for (ifa = ifa_list; ifa; ifa = ifa->ifa_next) {
		dl = (struct sockaddr_dl *)ifa->ifa_addr;
		if ((ifa->ifa_flags & IFF_UP) != 0 && dl->sdl_type == IFT_ETHER) {
			strlcpy(name, ifa->ifa_name, sizeof(name));
			break;
		}
	}
	freeifaddrs(ifa_list);

	return std::string(name);
}

// 受信スレッド
void
NetDriverBPF::ThreadRun()
{
	struct kevent kev;
	int kq;
	int r;

	kq = kqueue();
	if (kq == -1) {
		warn("kqueue failed");
		return;
	}

	EV_SET(&kev, fd, EVFILT_READ, EV_ADD, 0, 0, (kevent_udata_t)0);
	r = kevent_set(kq, &kev, 1);
	if (r == -1) {
		warn("kevent_set(fd=%d) failed", fd);
		close(kq);
		return;
	}

	for (;;) {
		int n;
		char *bpfptr;
		int bpfcaplen;

		// 何か起きるまで待つ
		r = kevent_poll(kq, &kev, 1, NULL);
		if (r == -1) {
			if (errno == EINTR) {
				continue;
			}
			// XXX ?
			warn("NetDriverBPF.ThreadRun: kevent_poll failed");
			break;
		}

		// 自動的にデバイスに着信があったことになるはず
		assert(r > 0);

		n = read(fd, bpfbuf.get(), bpfbuflen);
		if (n == -1) {
			if (errno == EINTR) {
				continue;
			}
			warn("NetDriverBPF.ThreadRun: read(fd=%d) failed", fd);
			break;
		}
		if (n == 0) {	// EOF
			return;
		}
		bpfptr = bpfbuf.get();
		bpfcaplen = n;

		while (bpfptr < bpfbuf.get() + bpfcaplen) {
			uint8 buf[1600];

			// BPF ヘッダを除いた部分
			struct bpf_hdr *bh = (struct bpf_hdr *)bpfptr;
			n = bh->bh_caplen;
			memcpy(buf, bpfptr + bh->bh_hdrlen, n);

			// 次のフレームの先頭を指す
			int framelen = BPF_WORDALIGN(bh->bh_hdrlen + bh->bh_caplen);
			bpfptr += framelen;

			// 60バイトパディングされないまま読み出せたらここでパディングする。
			if (n < 60) {
				memset(buf + n, 0, 60 - n);
				n = 60;
			}

			// CRC XXX 計算すること
			memset(buf + n, 0, 4);
			n += 4;

			// 受信処理に引き渡す
			// ただしデバッガプロンプトで停止してる状態なら VM に干渉しない。
			if (parent->RecvEnabled && gCVPrompt->IsPrompt() == false) {
				putlog(2, "RecvPacket %d bytes", n);
				rx_pkts++;
				rx_bytes += n;

				qvector<uint8> recvbuf(buf, buf + n);
				parent->RecvPacket(recvbuf);
			}
		}
	}
}
