//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "netdriver_tap.h"
#include "cvprompt.h"
#include <fcntl.h>
#include <poll.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#if defined(HAVE_SYS_EVENT_H)
#include <sys/event.h>
#endif
#if defined(HAVE_KQUEUE_SYS_EVENT_H)
#include <kqueue/sys/event.h>
#endif
#if defined(__linux__)
#include <linux/if.h>
#include <linux/if_tun.h>
#else
#include <ifaddrs.h>
#include <net/if.h>
#include <net/if_dl.h>
#include <net/if_tap.h>
#endif

// コンストラクタ
NetDriverTap::NetDriverTap(EthernetDevice *parent_,
	std::mutex *mtx_, std::condition_variable *cv_)
	: inherited("BSD tap", parent_, mtx_, cv_)
{
#if defined(__linux__)
	devpath = "/dev/net/tun";
#else
	devpath = "/dev/tap";
#endif
}

// デストラクタ
NetDriverTap::~NetDriverTap()
{
	Close();
}

// 初期化 & オープン
bool
NetDriverTap::Init()
{
	struct ifreq ifr;
	int r;

	// tap デバイスオープン
	fd = open(devpath.c_str(), O_RDWR);
	if (fd == -1) {
		warn("open \"%s\" failed", devpath.c_str());
		return false;
	}

	memset(&ifr, 0, sizeof(ifr));
#if defined(__linux__)
	// Linux なら、open() したデバイスに対して tap デバイス作成を指示、
	// その応答にインタフェース名が入っている。
	ifr.ifr_flags = IFF_TAP | IFF_NO_PI;	// tap として使う
	strlcpy(ifr.ifr_name, "tap%d", sizeof(ifr.ifr_name));
	// インタフェース作成。これには CAP_NET_ADMIN 権限が必要
	r = ioctl(fd, TUNSETIFF, (void *)&ifr);
	if (r == -1) {
		warn("Open: TUNSETIFF failed");
		if (errno == EPERM) {
			fprintf(stderr,
				" ** Set CAP_NET_ADMIN capability to the execute file.\n"
				" **  ex) sudo setcap cap_net_admin=ep path/to/%s\n",
				getprogname());
		}
		goto abort;
	}
#else
	// NetBSD なら、open() でインタフェースは作成されているので、
	// オープンされたこのディスクリプタのインタフェース名を取得。
	r = ioctl(fd, TAPGIFNAME, &ifr);
	if (r == -1) {
		warn("Open: TAPGIFNAME failed");
		goto abort;
	}
#endif

	// 作成できたインタフェース名を取得
	ifname = std::string(ifr.ifr_name);

	// オープンできたので受信スレッド起動。
	// ifup 実行前なので、ここで失敗しても ifdown 不要。
	if (InitRecvThread() == false) {
		goto abort;
	}

	// 最後に ifup スクリプトがあれば実行
	if (!Run_ifup()) {
		// どこまで実行できたか分からないので念のため ifdown も呼ぶ
		Close();
		return false;
	}

	return true;

	// ifdown を実行させないために fd = -1 にして帰る。
	// (まだ ifup を実行していないので)
 abort:
	close(fd);
	fd = -1;
	return false;
}

// クローズ
// (private)
void
NetDriverTap::Close()
{
	if (IsOpen()) {
		Run_ifdown();

		close(fd);
		fd = -1;
	}
}

// パケットを送信する
bool
NetDriverTap::SendPacket(const std::vector<uint8>& data)
{
	int n;

	assert(IsOpen());

	putlog(2, "SendPacket %zd bytes", data.size());
	tx_pkts++;
	tx_bytes += data.size();

	n = write(fd, data.data(), data.size());
	if (n == -1) {
		warn("SendPacket: write failed: fd=%d datalen=%zu",
			fd, data.size());
		return false;
	}

	if (n < (int)data.size()) {
		return false;
	}
	return true;
}

// 受信スレッド
void
NetDriverTap::ThreadRun()
{
	struct kevent kev;
	int kq;
	int r;

	kq = kqueue();
	if (kq == -1) {
		warn("kqueue failed");
		return;
	}

	EV_SET(&kev, fd, EVFILT_READ, EV_ADD, 0, 0, (kevent_udata_t)0);
	r = kevent_set(kq, &kev, 1);
	if (r == -1) {
		warn("kevent_set(fd=%d) failed", fd);
		close(kq);
		return;
	}

	for (;;) {
		uint8 buf[1600];
		int n;

		// 何か起きるまで待つ
		r = kevent_poll(kq, &kev, 1, NULL);
		if (r == -1) {
			if (errno == EINTR) {
				continue;
			}
			// XXX ?
			warn("kevent_poll failed");
			break;
		}

		// 自動的にデバイスに着信があったことになるはず
		assert(r > 0);

		n = read(fd, buf, sizeof(buf));
		if (n == -1) {
			if (errno == EINTR) {
				continue;
			}
			warn("read(fd=%d) failed", fd);
			break;
		}
		if (n == 0) {	// EOF
			return;
		}

		// カーネルが tap(4) に直接書き込んだパケット(というかフレームか)は
		// 60バイトパディングされないまま読み出せるので、ここでパディングする。
		if (n < 60) {
			memset(buf + n, 0, 60 - n);
			n = 60;
		}

		// CRC XXX 計算すること
		memset(buf + n, 0, 4);
		n += 4;

		// 受信処理に引き渡す
		// ただしデバッガプロンプトで停止してる状態なら VM に干渉しない。
		if (parent->RecvEnabled && gCVPrompt->IsPrompt() == false) {
			putlog(2, "RecvPacket %d bytes", n);
			rx_pkts++;
			rx_bytes += n;

			qvector<uint8> recvbuf(buf, buf + n);
			parent->RecvPacket(recvbuf);
		}
	}
}
