//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "mainapp.h"
#include "mystring.h"
#include <getopt.h>
#include <sys/stat.h>

// グローバルインスタンス
MainApp gMainApp;

// ヘルプメッセージ
void
MainApp::ShowHelp(bool all) const
{
	ShowVersion();

#define p(msg)	printf(" " msg "\n")

	printf("usage: %s [<options>]\n", getprogname());
	p("-A <file>           load and execute host binary (a.out or ELF)");
	p("-c <dir>            vm directory");
	p("-f                  fast mode");
	if (IsGUI())
		p("--fontsize <height> fontsize in monitors {12,16} (default:12)");
	p("-h                  show brief help message");
	p("--help              show all help message including for developers");
	if (IsGUI())
		p("-s,--scale <scale>  screen scale, default 1.0");
	p("--show-config       show configuration variables");
	p("-v                  show version");
	p("-V <key>=<val>      overwrite config option");

	if (all) {
		printf("\n(options for developers)\n");
		p("-b <addr>[,<skip>]  set breakpoint");
		p("-B <rom#>           benchmark mode");
		p("-C                  output log to console");
		p("-d                  debugger prompt on startup");
		p("-D                  debugger on console");
		p("-L <name>=<lv>[,..] set loglevel (-Lhelp displays names)");
		p("-M <name>[,..]      monitors to display at startup (-Mhelp displays names)");
		p("-X <human68k executable> human68k console emulation");
	}
}

// long option は enum と struct option をアルファベット順に並べる。
// enum は getopt() の1文字のオプションと衝突しなければいいので適当に
// 0x80 から始めておく。
enum {
	OPT_fontsize = 0x80,
	OPT_help,
	OPT_show_config,
	OPT_show_config_all,
};
static struct option longopts[] = {
	{ "fontsize",		required_argument,	NULL,	OPT_fontsize },
	{ "help",			no_argument,		NULL,	OPT_help },
	{ "show-config",	no_argument,		NULL,	OPT_show_config },
	// --show-config-all は開発用なのでヘルプには載せない
	{ "show-config-all",no_argument,		NULL,	OPT_show_config_all },
	{ NULL,				0,					NULL,	0 },
};

// 起動時の処理、VM の実行開始前まで。
// 所々 CLI と GUI で処理が違う。
bool
MainApp::Init(bool is_cli_, int ac, char *av[])
{
	is_cli = is_cli_;

	if (!ParseOpt(ac, av)) {
		return false;
	}

	// ログ機構は引数処理後なるはや
	// CLI ならログは常に標準出力へ(も)出力。
	// GUI なら -C で指定。
	gLogger.UseStdout(IsCLI() ? true : log_to_console);

	// 設定を作成。コンストラクタで初期値を用意
	gConfig.reset(new Config());

	// 引数で指定されたディレクトリの設定ファイルを読み込んで設定を更新
	ConfigFile file(vmdir + "nono.cfg");
	if (file.Load() == false) {
		return false;
	}
	if (gConfig->Update(file) == false) {
		return false;
	}
	// 最後にコマンドライン引数で更新
	if (gConfig->Update(config_options) == false) {
		return false;
	}

	// VM 種別を決定
	if (IsCLI() && human68k_file) {
		vmtype = VMTYPE_RXZ;
	} else {
		// VM 種別文字列から vmtype を決定
		const ConfigItem& item = gConfig->Find("vmtype");
		std::string vmstr = string_tolower(item.AsString());
		if (vmstr == "x68030") {
			vmtype = VMTYPE_X68030;
		} else if (vmstr == "luna") {
			vmtype = VMTYPE_LUNA1;
		} else if (vmstr == "luna88k") {
			vmtype = VMTYPE_LUNA88K;
		} else {
			if (item.GetFrom() == ConfigItem::FromInitial) {
				// 未指定の時
				warnx("vmtype must be specified");
			} else {
				// ユーザ由来の時
				item.Err("invalid vmtype");
			}
			return false;
		}
	}

	// VM 作成
	switch (vmtype) {
	 case VMTYPE_X68030:
		gVM.reset(new VM_X68030());
		break;

	 case VMTYPE_LUNA1:
		gVM.reset(new VM_LUNA());
		break;

	 case VMTYPE_RXZ:
		gVM.reset(new VM_RXZ());
		break;

	 case VMTYPE_LUNA88K:
		gVM.reset(new VM_LUNA88K());
		break;

	 default:
		__unreachable();
	}

	// 設定内容表示
	// (MPU コンストラクタでクロック数の初期値を決めるため、その後で表示)
	gConfig->Fix();
	if (show_config) {
		gConfig->Show(show_config - 1);
	}

	// 動的なコンストラクション
	if (!gVM->Create()) {
		return false;
	}

	// ログレベルを設定。コンストラクト後すぐに行う。
	// -L help もここで処理。
	if (!ParseLogopt()) {
		return false;
	}

	return true;
}

// VM の実行部分。
// Init() で引数を受け付けてから Start() でデバッガスレッドなどを起動するが、
// -Mhelp とかが指定された場合はここでスレッド開始する前にプロセスを終了する
// 必要があるので、分けてある。
bool
MainApp::Start()
{
	// VM 初期化。
	// これ以降はスレッドを開始したかもしれないので false を返す際には
	// VM をデストラクトすること。
	if (!gVM->Init()) {
		gVM.reset();
		return false;
	}
	// デバッガは VM オブジェクトリストに入っていない
	debugger_init();

	// 起動時設定の適用
	if (!gVM->Apply()) {
		gVM.reset();
		return false;
	}

	return true;
}

// コマンドライン引数を処理する。
// 知らない引数とかがあればこちらで usage を表示して false を返す。
bool
MainApp::ParseOpt(int ac, char *av[])
{
	int b;
	int c;

	fontsize = 12;
	screen_scale = 1.0;

	while ((c = getopt_long(ac, av, "A:b:B:c:CdDfhL:M:s:vV:X:",
	                        longopts, NULL)) != -1) {
		switch (c) {
		 case 'A':
			host_file = optarg;
			break;

		 case 'b':
			debug_breakaddr.push_back(optarg);
			break;

		 case 'B':
			b = atoi(optarg);
			if (b < 2 || b > 6) {
				fprintf(stderr, "-B <benchmark_mode>: 2..6\n");
				return false;
			}
			benchmark_mode = b;
			break;

		 case 'c':
			vmdir = std::string(optarg);
			break;

		 case 'C':
			log_to_console = true;
			break;

		 case 'd':
			debug_on_start = true;
			break;

		 case 'D':
			debug_on_console = true;
			break;

		 case 'f':
			fast_mode = true;
			break;

		 case 'L':
			if (logopt[0] != '\0') {
				strlcat(logopt, ",", sizeof(logopt));
			}
			strlcat(logopt, optarg, sizeof(logopt));
			break;

		 case 'M':
			if (!monitor_opt.empty()) {
				monitor_opt += ",";
			}
			monitor_opt += optarg;
			break;

		 case 's':
			if (IsGUI()) {
				char *end;
				errno = 0;
				screen_scale = strtod(optarg, &end);
				if (end == optarg || end[0] != '\0' || errno == ERANGE) {
					warnx("-s: invalid argument");
					return false;
				}
				// 上限は適当
				if (screen_scale <= 0.0 || screen_scale >= 10.0) {
					warnx("-s: invalid scale");
					return false;
				}
			}
			break;

		 case 'X':
			human68k_file = optarg;
			for (int i = optind; i < ac; i++) {
				if (i != optind) {
					strlcat(human68k_arg, " ", sizeof(human68k_arg));
				}
				strlcat(human68k_arg, av[i], sizeof(human68k_arg));
			}
			optind = ac;
			break;

		 case 'v':
			ShowVersion();
			exit(0);

		 case 'V':
			config_options.push_back(optarg);
			break;

		 case OPT_fontsize:
			fontsize = atoi(optarg);
			break;

		 case OPT_help:
			ShowHelp(true);
			return false;

		 case OPT_show_config:
			show_config = 1;
			break;

		 case OPT_show_config_all:
			show_config = 2;
			break;

		 case 'h':
		 default:
			ShowHelp(false);
			return false;
		}
	}

	// vmdir に '/' を付けておく
	if (vmdir.empty()) {
		vmdir = ".";
	}
	if (vmdir.back() != '/') {
		vmdir += '/';
	}

	// CLI 版で -Mhelp つけても黙って起動するのはさすがにどうかと思う。
	// とは言え GUI 版のコマンドラインを CLI 版用に書き換えた時に -M を
	// 消して回らないと起動できないのも面倒なので -M オプション自体は
	// 無意味だけど許容する。うーん。
	if (IsCLI()) {
		if (monitor_opt == "help") {
			warnx("-Mhelp is not available on CLI");
			return false;
		}
		if (!monitor_opt.empty()) {
			warnx("-M option is ignored on CLI");
			return true;
		}
	}

	return true;
}

// バージョンを表示
void
MainApp::ShowVersion() const
{
	// ここは実行ファイル名によらず nono にする
	fprintf(stderr, "nono version %d.%d.%d (%s)\n",
		NONO_MAJOR_VER, NONO_MINOR_VER, NONO_PATCH_VER, NONO_DATE);
}


// 引数 logopt のログ指定文字列をパースする。
// arg は "foo=1,bar=2" 形式の文字列で、これを分解して
// それぞれ担当するオブジェクトのログレベルにセットする。
bool
MainApp::ParseLogopt()
{
	char *buf;
	char *last;
	char *p;
	bool rv;

	// "help" (完全一致) なら識別子一覧を表示。
	if (strcmp(logopt, "help") == 0) {
		for (auto& obj : gObjects) {
			if (obj->logname != "?") {
				printf("%s\n", obj->logname.c_str());
			}
		}
		// エイリアスとか
		printf("sch -> scheduler\n");
		if (gMainApp.HasVMFeature(VMF_LUNA)) {
			printf("scc -> sio\n");
		}
		printf("all\n");
		return false;
	}

	rv = false;
	buf = strdup(logopt);
	for (p = strtok_r(buf, ",", &last);
		p;
		p = strtok_r(NULL, ",", &last))
	{
		const char *name;
		char *v;
		int val;

		name = p;

		v = strchr(p, '=');
		if (v) {
			*v++ = '\0';
			val = atoi(v);
		} else {
			val = 1;
		}
		// ここで name は変数名、val は値(省略されたら1)

		if (strlen(name) < 1) {
			printf("invalid logname '%s'\n", name);
			goto abort;
		}

		// 短縮形とかエイリアスとか
		if (strcmp(name, "sch") == 0)
			name = "scheduler";
		if (gMainApp.HasVMFeature(VMF_LUNA)) {
			if (strcmp(name, "scc") == 0)
				name = "sio";
		}

		// 比較
		if (strcmp(name, "all") == 0) {
			// "all" なら none 以外の全部にセット
			for (auto& obj : gObjects) {
				obj->loglevel = val;
			}
		} else {
			// それ以外は一致するキーを探してセット。
			// 複数のオブジェクトが同じ logname を持ってもよい (CMMUとか)。
			std::string sname = name;
			bool found = false;
			for (auto& obj : gObjects) {
				if (obj->logname == sname) {
					obj->loglevel = val;
					found = true;
				}
			}
			// 見つからない場合はエラー
			if (!found) {
				printf("invalid logname '%s'\n", name);
				goto abort;
			}
		}	
	}

	rv = true;
 abort:
	free(buf);
	return rv;
}

// 関連するファイルのパスを取得。
// 1. name がファイル名のみなら、VM ディレクトリとその親ディレクトリを検索。
//    ファイルが存在すればその時点で戻り値とする。ファイルが存在したけど
//    何らか不都合があったとしても (例えばファイルサイズが 0 だとか
//    パーミッションが足りないとか) 次を試すとかはしない。
// 2. name が '~' から始まっていればホームディレクトリに展開。
// 3. name が相対パスなら VM ディレクトリからの相対パスとする。
// 4. name は絶対パスのはずなので、そのまま使用する。
std::string
MainApp::SearchFile(const std::string& name) const
{
	std::string path;
	struct stat st;

	// 1. パス区切りを含んでなければ、ファイル名のみ
	if (name.find('/') == std::string::npos) {
		// 1a. VM ディレクトリ
		path = GetVMDir() + name;
		if (stat(path.c_str(), &st) == 0) {
			return path;
		}

		// 2. 親ディレクトリ
		path = GetVMDir() + "../" + name;
		if (stat(path.c_str(), &st) == 0) {
			return path;
		}

		// どちらもなければエラー
		return "";
	}

	// 2. 先頭の '~' を $HOME に展開
	path = name;
	if (path[0] == '~' && path[1] == '/') {
		const char *home = getenv("HOME");
		if (home == NULL) {
			home = "";
		}
		path = string_format("%s%s", home, path.c_str() + 1);
	}

	// 3. 相対パスなら、VM ディレクトリからの相対
	if (path[0] != '/') {
		path = GetVMDir() + path;
	}

	return path;
}

// 現在の VM が指定の feature を持っているか?
bool
MainApp::HasVMFeature(vmf_t feature) const
{
	uint32 vmf = 1 << GetVMType();
	return (vmf & feature) != 0;
}
