//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

//
// 外部公開用
//

#include "branchhistory.h"
#include <atomic>
#include <mutex>
#include <sys/cdefs.h>

__BEGIN_DECLS
#include "fpu_emulate.h"
__END_DECLS

class m68kcpu;

// 外部ループフラグ
// これらのフラグがいずれも立っていない間、CPU 実行は内側のディスパッチ
// ループを周り、どれかでも立てばディスパッチループを抜けてフラグを処理
// してから再びディスパッチループに戻る。
// それぞれ独立しているためビットを立てる際は代入ではなく OR すること。
#define CPU_REQ_STOP		(0x00000001)	// ストップ状態
#define CPU_REQ_HALT		(0x00000002)	// ホールト状態
#define CPU_REQ_TRACE		(0x00000004)	// デバッガ有効
#define CPU_REQ_PROMPT		(0x00000008)	// デバッガプロンプト
#define CPU_REQ_RELEASE		(0x00000010)	// 命令後に CPU 実行中断
#define CPU_REQ_INTR		(0x00000020)	// 命令後に割り込みチェックが必要

#define M68K_SR_T	(0xc000)
#define M68K_SR_S	(0x2000)
#define M68K_SR_M	(0x1000)
#define M68K_CCR_X	(0x0010)
#define M68K_CCR_N	(0x0008)
#define M68K_CCR_Z	(0x0004)
#define M68K_CCR_V	(0x0002)
#define M68K_CCR_C	(0x0001)

// SSW
#define SSW_SIZE(x)		(((x) & 0x03) << 4)
#define SSW_SIZE_BYTE	SSW_SIZE(1)
#define SSW_SIZE_WORD	SSW_SIZE(2)
#define SSW_SIZE_3BYTES	SSW_SIZE(3)
#define SSW_SIZE_LONG	SSW_SIZE(0)
#define SSW_SIZE_MASK	SSW_SIZE(0x03)

#define SSW_FC_MASK		(0x0007)	// FC_*
#define SSW_RW			(0x0040)	// SSW_BUS_{R,W}
#define SSW_RM			(0x0080)
#define SSW_DF			(0x0100)	// Data Fault
#define SSW_RB			(0x1000)
#define SSW_RC			(0x2000)
#define SSW_FB			(0x4000)
#define SSW_FC			(0x8000)

#define SSW_BUS_R		(0x0040)
#define SSW_BUS_W		(0x0000)

#define FC_DATA			(1)
#define FC_PROG			(2)
#define FC_USER			(0)
#define FC_SUPER		(4)
#define FC_UD			(FC_USER  | FC_DATA)
#define FC_UP			(FC_USER  | FC_PROG)
#define FC_SD			(FC_SUPER | FC_DATA)
#define FC_SP			(FC_SUPER | FC_PROG)

#define M68K_FPCR_MASK	(0x0000fff0)
#define M68K_FPSR_MASK	(0x0ffffff8)

#include "m68030ccr.h"
#include "m68030mmu.h"

// レジスタイメージを保持する構造体。
// この構造体はコピーとか比較をするのでポインタとかは置かないこと。
struct m68kreg {
 public:
	uint32 pc;
	uint32 da[16];

	m68030ccr ccr;
	bool s;				// SR の S ビット
	bool m;				// SR の M ビット
	int  intr_level;	// SR の割り込みレベル (0-7)

	// 現在の SP は常に da[15]。現在有効でない残り2つの SP をここに保存。
	// 例えばユーザ状態なら da[15] が USP、isp/msp は ISP/MSP を示し、
	// usp の内容は無効。
	uint32 usp;
	uint32 isp;
	uint32 msp;

	uint32 vbr;
	uint32 sfc;
	uint32 dfc;

	uint32 cacr;
	uint32 caar;

	union64 srp;
	union64 crp;
	uint32 tt[2];
	uint32 tc;
	uint16 mmusr;

	struct fpframe fpframe;

	// CPU の実行状態
	//  0            .. 通常状態
	//  CPU_REQ_STOP .. STOP 状態
	//  CPU_REQ_HALT .. HALT 状態
	uint32 state;

 public:
	// SR の上位8ビット部分のみを返す
	uint16 sr_h() {
		return (s ? M68K_SR_S : 0) | (m ? M68K_SR_M : 0) | (intr_level << 8);
	}
};

// バス状態
struct m68kbus {
	// 論理アドレスと SSW。
	// SSW はアクセス前に SSW_BUS_{R,W} と FC[012] をセットすること。
	DEF_UNION64(, ssw_laddr, ssw, laddr);

	uint32 paddr;			// 対応する物理アドレス
	// データバス。
	// 今の所書き込みだけで使う。読み込みデータは戻り値のみ。
	uint32 data;

	// SSW のうち現在の FC2 だけ別途覚えておく。
	// FC2 が変化するのは SR:S による特権モードが変わった時だけなので。
	uint16 fc2;

	// バスが Write なら true を返す
	bool IsWrite() const { return (ssw & SSW_BUS_R) == 0; }

	// このへんどうするか

	// FC2 をキャッシュしておく。
	void SetFC2(bool newfc2) {
		fc2 = newfc2 ? 4 : 0;
	}
	// FC を取得する。
	// これは MMU ルーチンから呼ばれる。そのためこの時点ではすでに
	// FC0/1 と FC2 は ssw に合成済みなので、これでよい。
	uint GetFC() const { return ssw & SSW_FC_MASK; }
};

class m68kcpu final {
 public:
	m68kcpu() { }
	~m68kcpu() { }

 public:
	m68kreg reg {};
	uint32 ppc = 0;			// 現在実行中の命令の先頭アドレス
	uint16 ir = 0;
	uint16 ir2 = 0;

	uint32 save_pd[8] {};	// -(An) を保存
	uint32 save_pi[8] {};	// (An)+ を保存
	uint8  flag_pd = 0;		// -(An) のどのレジスタを保存したか
	uint8  flag_pi = 0;		// (An)+ のどのレジスタを保存したか

	// MMU 状態
	bool mmu_enable = false;
	// SRP
	void SetSRP(uint32, uint32);
	uint64 GetSRP()  const { return reg.srp.q; }
	uint32 GetSRPh() const { return reg.srp.h; }
	uint32 GetSRPl() const { return reg.srp.l; }
	// CRP
	void SetCRP(uint32, uint32);
	uint64 GetCRP()  const { return reg.crp.q; }
	uint32 GetCRPh() const { return reg.crp.h; }
	uint32 GetCRPl() const { return reg.crp.l; }
	// TT
	void SetTT(int idx, uint32);
	uint64 GetTT(int idx) const { return reg.tt[idx]; }
	// TC
	bool SetTC(uint32);
	uint32 GetTC() const { return reg.tc; }
	// MMUSR
	void SetMMUSR(uint16);
	uint16 GetMMUSR() const { return reg.mmusr; }
	// MMU work
	m68030TT mmu_tt[2];
	uint8 mmu_tt_quick[256] {};
	m68030TC mmu_tc;
	m68030ATC atc {this};

 private:
	void mmu_enable_changed();

 public:
	int has_fpu = 0;		// 0:FPUなし、1:68881、2:68882(?)
	bool fpu_dirty = false;
	struct fpemu fe {};

	std::atomic<uint32> atomic_reqflag {}; // 外部ループ & リクエストフラグ

	// 割り込みペンディング
	// intr_vector[1] がレベル1割り込みのペンディング中ベクタ
	//  :
	// intr_vector[7] がレベル7割り込みのペンディング中ベクタ
	//
	// [0] を intr_pending (割り込みペンディング) として使う。
	// ペンディングになっているレベルのビットを立てる。
	//
	//   bit  7  6  5  4  3  2  1  0
	// Level  0  1  2  3  4  5  6  7
	//
	// intrmtx 取ってからアクセスすること。
	uint8 intr_vector[8] {};
#define intr_pending intr_vector[0]

	// 割り込みマスク
	// 現在の割り込みレベルで割り込みを受け付けるレベルのビットを立てる
	// I=0 だと割り込みレベル 1..7 を受け付けるので %01111111
	// I=5 だと割り込みレベル  6,7 を受け付けるので %00000011
	// I=6 だと割り込みレベル    7 を受け付けるので %00000001
	// ただし I=7 はこれを使わず別処理する
	//   bit  7  6  5  4  3  2  1  0
	// Level  0  1  2  3  4  5  6  7
	//
	// intrmtx 取ってからアクセスすること。
	uint8 intr_mask = 0;

	// 割り込み要求は別スレッドから呼ばれる場合もあるので
	// (自スレッドから呼び出した関数の向こうから呼ばれる場合もあることに注意)
	// それらの変数はこれで保護する。
	// 最終的には atomic_reqflag に CPU_REQ_INTR を立てることでメインループ
	// に通知する (この通知そのものは atomic なので特に保護しなくてもよい)。
	std::mutex intrmtx {};

	uint64 total_cycle = 0;		// 積算実行サイクル数
	int32 cycle_cache = 0;		// キャッシュケースの所用サイクル数
	int32 cycle_nocache = 0;	// ノーキャッシュケースの時の *追加分*
	uint64 inst_count = 0;		// 命令実行数

	// ブランチ履歴 (例外履歴を含む)
	BranchHistory_m680x0 brhist {};
	// 例外履歴のみ
	BranchHistory_m680x0 exhist {};

	// バス
	m68kbus bus {};

	// 手抜きのためリセットベクタ取得アドレスを知っている。
	// リセット時のためだけにデバイス側に細工を入れるのは実行時の
	// オーバーヘッドが大きそうなので。
	uint32 reset_vector = 0;

	// A-Line, F-Line 命令エミュレーションのコールバックとユーザ引数。
	// ユーザ引数には関与しないので呼び出し側が自由に指定してよい。
	// コールバック側で処理完了したら true を返す。
	// デフォルトの処理(すなわちm68k例外)にする場合は false を返す。
	bool (*aline_callback)(m68kcpu *, void *) = NULL;
	void *aline_arg = NULL;
	bool (*fline_callback)(m68kcpu *, void *) = NULL;
	void *fline_arg = NULL;

	// ダブルバスフォールトのコールバックとユーザ引数。
	void (*halt_callback)(void *) = NULL;
	void *halt_arg = NULL;
};

#define RegD(n)		(cpu->reg.da[(n)])
#define RegA(n)		(cpu->reg.da[(n) + 8])
#define RegR(n)		(cpu->reg.da[(n)])
#define RegPC			(cpu->reg.pc)
#define RegPPC			(cpu->ppc)
#define RegIR			(cpu->ir)
#define RegIR2			(cpu->ir2)
#define RegSR			(cpu->reg.sr_h() | cpu->reg.ccr.Get())
#define RegVBR			(cpu->reg.vbr)
#define RegUSP			(cpu->reg.usp)
#define RegISP			(cpu->reg.isp)
#define RegMSP			(cpu->reg.msp)
#define RegSFC			(cpu->reg.sfc)
#define RegDFC			(cpu->reg.dfc)
#define RegCACR		(cpu->reg.cacr)
#define RegCAAR		(cpu->reg.caar)

#define RegFP(n)	(cpu->reg.fpframe.fpf_regs[n])
#define RegFPCR		(cpu->reg.fpframe.fpf_fpcr)
#define RegFPSR		(cpu->reg.fpframe.fpf_fpsr)
#define RegFPIAR	(cpu->reg.fpframe.fpf_fpiar)

#define RegIRX			((RegIR >> 9) & 7)
#define RegIRY			(RegIR & 7)
#define RegDX			RegD(RegIRX)
#define RegAX			RegA(RegIRX)
#define RegDY			RegD(RegIRY)
#define RegAY			RegA(RegIRY)

// 外部公開ルーチン
extern m68kcpu *m68030_init(uint32 reset_vector);
extern uint32 m68030_run(m68kcpu *cpu, uint64 request);
extern void m68030_exception_reset(m68kcpu *cpu);
extern void m68030_interrupt(m68kcpu *cpu, int level, int vector);

// 外部で定義されるコールバック
extern void m68030_reset_inst(m68kcpu *cpu);
