//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m68030core.h"
#include "m68030bus.h"
#include "m68030fpu.h"
#include "debugger.h"

static void m68030_change_intr(m68kcpu *, int);
static void m68030_set_intr(m68kcpu *, int);
static void m68030_double_bus_fault(m68kcpu *, const char *);

// CPU コアの領域を確保して初期化
// reset_vector はリセット例外時に 0 番地に見せる(通常ROMの)アドレス。
m68kcpu *
m68030_init(uint32 reset_vector)
{
	m68kcpu *cpu;

	cpu = new m68kcpu();
	fpu_init(cpu);

	for (int i = 0; i < 16; i++) {
		RegR(i) = 0xcccccccc;
	}
	// XXX ただし X68030 IPLROM には A6 を初期化せずに書き込んでいるバグが
	// あるため、A6 を概ね RAM のあるあたりに指定しておく…。
	RegA(6) = 0x000ccccc;

	// 手抜きのためリセットベクタを知っている
	cpu->reset_vector = reset_vector;

	return cpu;
}

// 二重バスフォールト
static void
m68030_double_bus_fault(m68kcpu *cpu, const char *where)
{
	cpu->atomic_reqflag |= CPU_REQ_RELEASE | CPU_REQ_HALT;

	// コールバックがあれば呼び出す
	if (cpu->halt_callback) {
		cpu->halt_callback(cpu->halt_arg);
	}
}

// リセット例外を発生させる。
// 各レジスタを所定の状態にして、リセットベクタを読み込む。
// リセットベクタが読み込めなければ HALT 状態に移行。
void
m68030_exception_reset(m68kcpu *cpu)
{
	// 例外履歴ちょっとどうするかは置いておく

	// 1. T0, T1 クリア
	// 2. S をセット、M をクリア
	// 3. 割り込みレベルを 7

	// ここは set_sr を使わずに設定する。
	cpu->reg.s = true;
	cpu->reg.m = false;
	cpu->bus.SetFC2(cpu->reg.s);
	// ここは割り込みレベル変化のチェックは不要でセットだけでいい。
	m68030_set_intr(cpu, 7);

	// 4. VBR をクリア
	RegVBR = 0;
	// 5. キャッシュ
	RegCACR = 0;	// XXX
	// 6. キャッシュエントリクリア
	// 7. TC:E、TTn:E をクリア
	cpu->SetTC(cpu->GetTC() & ~m68030TC::TC_E);
	cpu->SetTT(0, cpu->GetTT(0) & ~m68030TT::E);
	cpu->SetTT(1, cpu->GetTT(1) & ~m68030TT::E);
	// 8. リセットベクタ生成

	// 9. SP をロード
	// 10. PC をロード

	// 手抜き。
	// XXX 本当は ROM を 0番地に見せかけてから 0番地からロードするのだが
	//     ぶっちゃけそれを外部から観測する手段もないのでなー。
	// XXX FC を気にしないといけない
	// XXX 読めなければ二重バスフォールト
	// 本当はアドレスエラーもチェックするのだが、起きないので省略。
	try {
		RegA(7) = m68030_read_32(cpu, cpu->reset_vector);
		RegPC = m68030_read_32(cpu, cpu->reset_vector + 4);
	} catch (int) {
		m68030_double_bus_fault(cpu, "reset exception");
	}
}

// MPU 外部からの割り込み受付。
// オートベクタの場合は vector にオートベクタの例外ベクタを指定のこと。
void
m68030_interrupt(m68kcpu *cpu, int level, int vector)
{
	std::unique_lock<std::mutex> lock(cpu->intrmtx);

	// この割り込みレベルがすでにペンディングなら何もしない
	if ((int8)(cpu->intr_pending << level) < 0) {
		return;
	}

	// 割り込みを受理
	cpu->intr_pending |= (0x80U >> level);
	cpu->intr_vector[level] = vector;
	cpu->atomic_reqflag |= CPU_REQ_INTR;
}

// 割り込みレベルの設定のみを行う。
// 割り込みチェックは行わない。
static void
m68030_set_intr(m68kcpu *cpu, int level)
{
	cpu->reg.intr_level = level;
	cpu->intr_mask = 0x7f >> level;
}

// 割り込みのチェックとレベルの変更。
// ステータスレジスタの割り込みレベルを変更し、必要な割り込みチェックを行う。
static void
m68030_change_intr(m68kcpu *cpu, int level)
{
	int oldlevel = cpu->reg.intr_level;

	if (oldlevel < 7 && level == 7) {
		// レベル 7 に引き上げられたら割り込み確認
		cpu->atomic_reqflag |= CPU_REQ_INTR;
	} else if (oldlevel == 7 && level < 7) {
		// レベル 7 から引き下げられたら割り込み確認
		cpu->atomic_reqflag |= CPU_REQ_INTR;
	} else if (oldlevel != level) {
		// それ以外でレベルが変わったら割り込み確認?
		cpu->atomic_reqflag |= CPU_REQ_INTR;
	}
	m68030_set_intr(cpu, level);
}

// SR (と CCR) に data を代入する。
// リセット例外時には呼ばれないのであっちで必要に応じて似た処理をする必要が
// ある。XXX なんとかしたほうがいい
void
m68030_set_sr(m68kcpu *cpu, uint16 data)
{
	bool oldS = cpu->reg.s;
	bool oldM = cpu->reg.m;

	// 現在の SP を保存
	if (oldS == false) {
		RegUSP = RegA(7);
	} else if (oldM == false) {
		RegISP = RegA(7);
	} else {
		RegMSP = RegA(7);
	}

	CCR.Set(data);
	cpu->reg.s = ((data & M68K_SR_S) != 0);
	cpu->reg.m = ((data & M68K_SR_M) != 0);
	m68030_change_intr(cpu, (data >> 8) & 7);

	// 新しい SP を用意
	if (cpu->reg.s == false) {
		RegA(7) = RegUSP;
	} else if (cpu->reg.m == false) {
		RegA(7) = RegISP;
	} else {
		RegA(7) = RegMSP;
	}

	// バスの FC2 を変更
	cpu->bus.SetFC2(cpu->reg.s);

	// XXX 他に何か状態変えないといけないことあったっけか。
}

// 命令中に (An)+ で変化したアドレスレジスタを命令実行前の状態に巻き戻す。
// flag_pi は bit7..0 が save_pi[7..0] の A7..A0 に対応している。
void
m68030_restore_reg_pi(m68kcpu *cpu)
{
	for (int i = 7; cpu->flag_pi; i--) {
		if ((int8)cpu->flag_pi < 0) {
			RegA(i) = cpu->save_pi[i];
		}
		cpu->flag_pi <<= 1;
	}
}

// 命令中に -(An) で変化したアドレスレジスタを命令実行前の状態に巻き戻す。
// flag_pd は bit7..0 が save_pd[7..0] の A7..A0 に対応している。
void
m68030_restore_reg_pd(m68kcpu *cpu)
{
	for (int i = 7; cpu->flag_pd; i--) {
		if ((int8)cpu->flag_pd < 0) {
			RegA(i) = cpu->save_pd[i];
		}
		cpu->flag_pd <<= 1;
	}
}

// 例外処理について。
//
// 0. リセット例外
//	リセット例外は外部から m68030_exception_reset() を呼び出す。
//
// 1. バスエラー例外
//	MMU による論理アドレスでのバスエラーは (m68030_mmu_translate_*() が
//	false を返した後) m68030bus.cpp ルーチンが (C++の) 例外をスローする。
//	物理アドレス空間アクセス後のバスエラーは (VM側の read*()/write*() が
//	(int64)-1 を返した後) m68030bus.cpp ルーチンが (C++) の例外をスローする。
//	どちらのケースも m68030core.cpp のメインループで catch し、その catch
//	ブロック内で m68030_exception() を呼び出す。
//
// 2. アドレスエラー例外
//	アドレスエラーは m68030_jump() 内でのみ発生するはずで、その場で
//	m68030_exception() を呼んで処理する。
//
// 3. 命令中で起きる例外
//	命令途中で起きる(確定的な)例外は m68030_exception() を呼び出す。
//	これに該当するのは
//	  o 特権違反
//	  o A系列/F系列命令
//	  o TRAP* 系命令、CHK/CHK2 命令
//	  o トレース例外 (未実装)
//	  o ゼロ除算?
//	  o MMU 構成例外
//	  o 不当命令 (*)
//	である。
//	ただし、不当命令についてはどこで不当命令と判定されるかによって
//	フローが複数存在することに注意。
//	  o 1ワード目が不当命令
//		-> 命令がそれぞれ m68030_exception() で処理。
//	  o 1ワード目の EA が不当命令
//		-> (C++の)例外をスローして catch 内で m68030_exception() を実行。
//	  o 2ワード目が不当命令
//		-> 命令がそれぞれ m68030_exception() で処理。
//
// 4. 割り込み例外
//	外部デバイスは m68030_interrpt() を呼び出すことで MPU に割り込みを要求
//	する。その結果 MPU は直後の命令境界で割り込みをチェックし、
//	m68030_exception_interrupt() を呼び出す。
//

// vector にジャンプする。vector はベクタ番号(アドレスではない)。
static inline void
m68030_jump_vector(m68kcpu *cpu, int vector)
{
	uint32 vecaddr;
	uint32 newpc;

	vecaddr = RegVBR + vector * 4;
	RegPC = vecaddr;
	newpc = m68030_fetch_32(cpu);
	// VBR 込みのジャンプ先が確定したところで今度はブランチ履歴に記録。
	// この時のブランチ元はベクタアドレス
	// (RegPC は fetch によって進んでしまうので先に保存)
	cpu->brhist.AddEntry(vecaddr, newpc, 0x8000);

	CHECK_ADDRESS_ERROR(newpc);
	RegPC = newpc;

	// 実際はここでパイプライン充填のために3ワードフェッチするので、
	// ベクタテーブルはフェッチ出来たけどそのジャンプ先がバスエラーに
	// なるケースはここでバスエラーが出なければいけないが、まだ
	// キャッシュを実装していないので、ここで1ワードフェッチしてみて
	// PC を戻すあほっぽい方法を取る。
	m68030_fetch_16(cpu);
	RegPC -= 2;
}

// 例外スタックフレーム $0 を生成する。
// vector はベクタ番号。
// pc はスタックに積むプログラムカウンタの値 (例外によって、この命令位置か
// 次の命令位置かが異なる)。
// sr は例外発生時の SR。
static inline void
m68030_generate_exframe_0(m68kcpu *cpu, int vector, uint32 pc, uint16 sr)
{
	// 例外スタックフレーム $0
	// SP+$00.W STATUS REGISTER
	//   +$02.L PROGRAM COUNTER		; 例外ごとに異なる
	//   +$06.W %0000 | VECTOR

	m68030_push_16(cpu, (0 << 12) | (vector * 4));
	m68030_push_32(cpu, pc);
	m68030_push_16(cpu, sr);
}

// 例外スタックフレーム $2 を生成する。
// vector はベクタ番号。
// スタックに積む PC は常に次の命令。
static inline void
m68030_generate_exframe_2(m68kcpu *cpu, int vector, uint16 sr)
{
	// 例外スタックフレーム $2
	// SP+$00.W STATUS REGISTER
	//   +$02.L PROGRAM COUNTER		; 次の命令位置
	//   +$06.W %0010 | VECTOR
	//   +$08.L INSTRUCTION ADDRESS	; 例外を起こした命令のアドレス

	m68030_push_32(cpu, RegPPC);	// 例外を起こした命令のアドレス
	m68030_push_16(cpu, (2 << 12) | (vector * 4));
	m68030_push_32(cpu, RegPC);		// 次の命令位置
	m68030_push_16(cpu, sr);
}

// 例外スタックフレーム $B を生成する。
static inline void
m68030_generate_exframe_b(m68kcpu *cpu, int vector, uint16 sr)
{
	uint16 ssw;
	uint32 fault_addr;
	uint32 stageBaddr;

	stageBaddr = RegPC + 2;
	// この後のスタック操作で bus が変化する前に保存
	fault_addr = cpu->bus.laddr;
	// DF ビットを立てておくと RTE した後この命令を再実行。
	ssw = cpu->bus.ssw | SSW_DF;

	// 例外スタックフレーム $B
	// SP+$00.W   STATUS REGISTER
	//   +$02.L   PROGRAM COUNTER
	//   +$06.W   %1011 | VECTOR
	//   +$08.W   (internal register)
	//   +$0a.W   SPECIAL STATUS REGISTER		; SSW
	//   +$0c.W   INSTRUCTION PIPE STAGE C
	//   +$0e.W   INSTRUCTION PIPE STAGE B
	//   +$10.L   DATA CYCLE FAULT ADDRESS
	//   +$14.W*2 (internal register)
	//   +$18.L   DATA OUTPUT BUFFER
	//   +$1c.W*4 (internal register, 4 words)
	//   +$24.L   STAGE B ADDRESS
	//   +$28.W*2 (internal register, 2 words)
	//   +$2c.L   DATA INPUT BUFFER
	//   +$30.W*3 (internal register, 3 words)
	//   +$36.W   VERSION# | internal information
	//   +$38.W*18 (internal register, 18 words)
	//   +$5a
	//	こっちから上に向かって PUSH

	// internal registers, 18 words
	// 内部レジスタ18ワードのうち、ここでは17ワード分を save_pd[8] (16ワード)、
	// flag_pd (1ワード) として使うことにする。
	// -(An) でバスエラーが起きると An は減算された状態になるが、DF ビットが
	// 立っていて再実行する時には現行構造では命令先頭から再実行せざるをえず、
	// そうするともう一度プリデクリメントが動いてしまう。そのため、RTE の時点
	// で DF=1 ならプリデクリメント発生前の An に戻すという処理をする。
	// RTE 時点で DF=0 なら命令を再実行しないため An も巻き戻してはならず、
	// そのためプリデクリメント分の巻き戻しは RTE 時点まで行えない。
	// 一方、ポストインクリメントのケースはこの少し下の
	// m68030_exception_buserr() 内のコメント参照。
	for (int i = 7; i >= 0; i--) {
		m68030_push_32(cpu, cpu->save_pd[i]);
	}
	m68030_push_16(cpu, cpu->flag_pd);
	RegA(7) -= 2;

	RegA(7) -= 2;						// VERSION#
	RegA(7) -= 6;
	m68030_push_32(cpu, 0);				// DATA INPUT BUFFER
	RegA(7) -= 4;
	m68030_push_32(cpu, stageBaddr);	// STAGE B ADDRESS
	RegA(7) -= 8;
	m68030_push_32(cpu, 0);				// DATA OUTPUT BUFFER
	RegA(7) -= 4;
	m68030_push_32(cpu, fault_addr);	// DATA CYCLE FAULT ADDRESS
	m68030_push_16(cpu, 0);				// INSTRUCTION PIPE STAGE B
	m68030_push_16(cpu, 0);				// INSTRUCTION PIPE STAGE C
	m68030_push_16(cpu, ssw);
	RegA(7) -= 2;
	m68030_push_16(cpu, 0xb000 | (vector << 2));
	// 本当は現在の PC だが、再実行のため PPC を積む。
	m68030_push_32(cpu, RegPPC);		// PC
	m68030_push_16(cpu, sr);			// 例外入り口で保存した SR
}

// 例外処理 (リセット・割り込み以外) を行う。
// vector はベクタ番号。
void
m68030_exception(m68kcpu *cpu, int vector)
{
	// バスエラー・アドレスエラーは別で処理。
	if (vector <= M68K_EXCEP_ADDRERR) {
		m68030_exception_buserr(cpu, vector);
		return;
	}

	// 例外履歴に記録 (例外発生はブランチ履歴にも記録)
	cpu->exhist.AddEntry(RegPPC, 0, 0x8000 | vector);
	cpu->brhist.AddEntry(RegPPC, 0, 0x8000 | vector);
	// デバッガに通知 (例外ブレーク用)
	debugger_notifyexception(vector);

	// 状態を保存
	uint16 sr = RegSR;

	// S を立てて T1,T0 を落とす
	m68030_set_sr(cpu, (RegSR | M68K_SR_S) & ~M68K_SR_T);

	try {
		switch (vector) {
		 case M68K_EXCEP_ILLEGAL:		// 不当命令
		 case M68K_EXCEP_PRIV:			// 特権違反例外
		 case M68K_EXCEP_ALINE:			// Aライン命令例外
		 case M68K_EXCEP_FLINE:			// Fライン命令例外
			// スタックに積む PC はこの命令のアドレス
			m68030_generate_exframe_0(cpu, vector, RegPPC, sr);
			break;

		 case M68K_EXCEP_TRAP0			// TRAP#0
		  ... M68K_EXCEP_TRAP15:		// TRAP#15
			// スタックに積む PC は次の命令
			m68030_generate_exframe_0(cpu, vector, RegPC, sr);
			break;

		 case M68K_EXCEP_ZERODIV:		// ゼロ除算
		 case M68K_EXCEP_CHK:			// CHK
		 case M68K_EXCEP_TRAPV:			// TRAPV
		 case M68K_EXCEP_MMU_CONFIG:	// MMU 構成例外
			m68030_generate_exframe_2(cpu, vector, sr);
			break;

		 default:
			PANIC("$%08x exception vector %d ?", RegPPC, vector);
			break;
		}

		// ベクタにジャンプ
		m68030_jump_vector(cpu, vector);
	} catch (int vector2) {
		assertmsg(vector2 == M68K_EXCEP_BUSERR, "vector=%d", vector);

		// バスエラー処理へ
		m68030_exception_buserr(cpu, vector2);
	}
}

// バスエラー・アドレスエラー例外処理を行う。
// 他の例外とは違いが多いので分けた。
void
m68030_exception_buserr(m68kcpu *cpu, int vector)
{
	assertmsg((vector == M68K_EXCEP_BUSERR || vector == M68K_EXCEP_ADDRERR),
		"vector=%d", vector);

	// 例外履歴に記録 (例外発生はブランチ履歴にも記録)
	cpu->exhist.AddEntry(RegPPC, 0, 0x8000 | vector);
	cpu->brhist.AddEntry(RegPPC, 0, 0x8000 | vector);
	// デバッガに通知 (例外ブレーク用)
	debugger_notifyexception(vector);

	// 状態を保存
	uint16 sr = RegSR;

	// ここで (An)+ で変化したレジスタを命令実行前の状態に巻き戻す。
	// EA 処理の時点で先にインクリメントしているのだが、バスエラーが
	// 起きたということは実際にはポストインクリメントは発生しないので、
	// ここで無条件に巻き戻してよい。SR を書き換えると A7 が切り替わって
	// しまうので、この巻き戻しは SR を書き換える前のここで行うこと。
	// 一方、プリデクリメントのケースはこの少し上の exframe_b の中の
	// コメント参照。
	m68030_restore_reg_pi(cpu);

	// S を立てて T1,T0 を落とす
	m68030_set_sr(cpu, (RegSR | M68K_SR_S) & ~M68K_SR_T);

	try {
		// フレームを作成。
		// 実際はショート($A)、ロング($B)の2種類あって、本来ならショートが
		// 使えればショートを使うのだが、ここでは全部ロングを使う。
		m68030_generate_exframe_b(cpu, vector, sr);
		// ベクタにジャンプ
		m68030_jump_vector(cpu, vector);
	} catch (int vector2) {
		// ここに来るのはバスエラーだけのはず
		assertmsg(vector2 == M68K_EXCEP_BUSERR, "vector=%d", vector);

		// バスエラー・アドレスエラー処理中のバスエラーは
		// ダブルバスフォールト。
		m68030_double_bus_fault(cpu, "bus/address error handling");
	}
}

// 割り込み例外処理を行う。
// intrmtx なしで呼び出すこと。
// 割り込みチェックは m68030core.cpp::m68030_flush_interrupt() で判定している。
void
m68030_exception_interrupt(m68kcpu *cpu, int level, int vector)
{
	// 例外履歴に記録 (例外発生はブランチ履歴にも記録)
	cpu->exhist.AddEntry(RegPPC, 0, 0x8000 | (level << 8) | vector);
	cpu->brhist.AddEntry(RegPPC, 0, 0x8000 | (level << 8) | vector);
	// デバッガに通知 (例外ブレーク用)
	debugger_notifyexception(vector);

	// 状態を保存
	uint16 sr = RegSR;

	// S を立てて T1,T0 を落とす
	m68030_set_sr(cpu, (RegSR | M68K_SR_S) & ~M68K_SR_T);
	// 割り込みレベルを変更
	m68030_change_intr(cpu, level);

	try {
		// 例外スタックフレーム $0 を生成。
		// スタックに積む PC は次の命令位置。
		// M ビットは未実装なのでスローアウェイとか無視
		m68030_generate_exframe_0(cpu, vector, RegPC, sr);

		// ベクタにジャンプ
		m68030_jump_vector(cpu, vector);
	} catch (int vector2) {
		assertmsg(vector2 == M68K_EXCEP_BUSERR, "vector=%d", vector);

		// バスエラー処理へ
		m68030_exception_buserr(cpu, vector2);
	}
}

// RTE 命令。
// 例外処理の近くに置いておくほうが便利なので。
void
m68030_rte(m68kcpu *cpu)
{
	uint32 pc;
	uint16 sr;
	uint16 frame;

	try {
		sr = m68030_pop_16(cpu);
		pc = m68030_pop_32(cpu);
		frame = m68030_pop_16(cpu);
		switch (frame >> 12) {
		 case 0:
			CYCLE2(18, 20);
			m68030_set_sr(cpu, sr);
			m68030_jump(cpu, pc);
			break;

		 case 2:
			CYCLE2(18, 20);
			// +08.L INSTRUCTION ADDRESS を読み捨てる
			RegA(7) += 4;
			m68030_set_sr(cpu, sr);
			m68030_jump(cpu, pc);
			break;

		 case 0x0b:
		 {
			uint32 stage_b;
			uint16 ssw;

			CYCLE(76);
			RegA(7) += 2;						// $08.W
			ssw = m68030_pop_16(cpu);			// $0a.W SSW
			RegA(7) += 4 * 6;					// $0c.L*6
			stage_b = m68030_pop_32(cpu);		// $24.L Stage B
			RegA(7) += 2 * 9;					// $28.W*9
			// $3a..$58 ここが save_pd の取り出し
			cpu->flag_pd = m68030_pop_16(cpu);
			for (int i = 0; i < 8; i++) {
				cpu->save_pd[i] = m68030_pop_32(cpu);
			}
			// A7 レジスタをリストアするかもしれない前に Super/User を
			// 復帰しておかないといけない
			m68030_set_sr(cpu, sr);

			// DFビットセットなら命令を再実行するため、-(An) で変化
			// したレジスタを巻き戻す。
			// DFビットクリアなら例外要因の命令はハンドラで処理された
			// とみなして、次の命令に進む。
			if ((ssw & SSW_DF)) {
				m68030_restore_reg_pd(cpu);
			} else {
				pc = stage_b - 2;
			}
			m68030_jump(cpu, pc);
			break;
		 }

		 default:
			PANIC("RTE frame=$%x not supported", frame >> 12);
			break;
		}
	} catch (int vector2) {
		assertmsg(vector2 == M68K_EXCEP_BUSERR, "vector2=%d", vector2);

		// RTE 処理中のバスエラーはダブルバスフォールト。
		m68030_double_bus_fault(cpu, "bus/address error handling");
	}
}

static const char * const exception_name_table[] = {
			//   0123456789012345678 <- 例外履歴欄の横幅
	/*  0 */	"Reset",
	/*  1 */	NULL,
	/*  2 */	"Bus Error",
	/*  3 */	"Address Error",
	/*  4 */	"Illegal Instruction",
	/*  5 */	"Zero Divide",
	/*  6 */	"CHK/CHK2 Insn",
	/*  7 */	"TRAPV, *TRAPcc Insn",
	/*  8 */	"PriviledgeViolation",
	/*  9 */	"Trace",
	/* 10 */	"Line 1010 emulator",
	/* 11 */	"Line 1111 emulator",
	/* 12 */	NULL,
	/* 13 */	"CoproProtoViolation",
	/* 14 */	"Format Error",
	/* 15 */	"Uninitialized Intr",
	/* 16 */	NULL, NULL, NULL, NULL,
	/* 20 */	NULL, NULL, NULL, NULL,
	/* 24 */	"Spurious Interrupt",
	/* 25 */	"Level 1 Interrupt",
	/* 26 */	"Level 2 Interrupt",
	/* 27 */	"Level 3 Interrupt",
	/* 28 */	"Level 4 Interrupt",
	/* 29 */	"Level 5 Interrupt",
	/* 30 */	"Level 6 Interrupt",
	/* 31 */	"Level 7 Interrupt",
	/* 32 */	"Trap #0",
	/* 33 */	"Trap #1",
	/* 34 */	"Trap #2",
	/* 35 */	"Trap #3",
	/* 36 */	"Trap #4",
	/* 37 */	"Trap #5",
	/* 38 */	"Trap #6",
	/* 39 */	"Trap #7",
	/* 40 */	"Trap #8",
	/* 41 */	"Trap #9",
	/* 42 */	"Trap #10",
	/* 43 */	"Trap #11",
	/* 44 */	"Trap #12",
	/* 45 */	"Trap #13",
	/* 46 */	"Trap #14",
	/* 47 */	"Trap #15",
			//   0123456789012345678
	/* 48 */	"FPCP Branch",
	/* 49 */	"FPCP InexcactResult",
	/* 50 */	"FPCP Divide by Zero",
	/* 51 */	"FPCP UnderFlow",
	/* 52 */	"FPCP Operand Error",
	/* 53 */	"FPCP OverFlow",
	/* 54 */	"FPCP Signaling NAN",
	/* 55 */	NULL,
	/* 56 */	"MMU Config Error",
	/* 57 */	NULL,
	/* 58 */	NULL,
	/* 59 */	NULL,
	/* 60 */	NULL,
	/* 61 */	NULL,
	/* 62 */	NULL,
	/* 63 */	NULL,
			//   0123456789012345678
	/* $40 */	"MFP Alarm",
	/* $41 */	"MFP EXPON",
	/* $42 */	"MFP POWSW",
	/* $43 */	"MFP FM IRQ",
	/* $44 */	"MFP Timer-D",
	/* $45 */	"MFP Timer-C",
	/* $46 */	"MFP V-Disp",
	/* $47 */	"MFP GPIP5",
	/* $48 */	"MFP Timer-B",
	/* $49 */	"MFP TX Error",
	/* $4a */	"MFP TX Empty",
	/* $4b */	"MFP RX Error",
	/* $4c */	"MFP RX Full",
	/* $4d */	"MFP Timer-A",
	/* $4e */	"MFP CRTC IRQ",
	/* $4f */	"MFP H-Sync",
	/* $50 */	"SCC#B TX Empty",
	/* $51 */	"SCC#B TX Empty",
	/* $52 */	"SCC#B",
	/* $53 */	"SCC#B",
	/* $54 */	"SCC#B",
	/* $55 */	"SCC#B",
	/* $56 */	"SCC#B",
	/* $57 */	"SCC#B",
	/* $58 */	"SCC#A TX Empty",
	/* $59 */	"SCC#A TX Empty",
	/* $5a */	"SCC#A",
	/* $5b */	"SCC#A",
	/* $5c */	"SCC#A",
	/* $5d */	"SCC#A",
	/* $5e */	"SCC#A",
	/* $5f */	"SCC#A",

			//   0123456789012345678
	/* $60 */	"I/O FDC Intr",
	/* $61 */	"I/O FDD Intr",
	/* $62 */	"I/O HDC Intr",
	/* $63 */	"I/O PRN Intr",
	/* $64 */	"DMAC#0 Complete",
	/* $65 */	"DMAC#0 Error",
	/* $66 */	"DMAC#1 Complete",
	/* $67 */	"DMAC#1 Error",
	/* $68 */	"DMAC#2 Complete",
	/* $69 */	"DMAC#2 Error",
	/* $6a */	"DMAC#3 Complete",
	/* $6b */	"DMAC#3 Error",
	/* $6c */	"SPC Interrupt",
	/* $6d */	NULL,
	/* $6e */	NULL,
	/* $6f */	NULL,

	/* $70 */	NULL, NULL, NULL, NULL,
	/* $74 */	NULL, NULL, NULL, NULL,
	/* $78 */	NULL, NULL, NULL, NULL,
	/* $7c */	NULL, NULL, NULL, NULL,
	/* $80 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $84 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $88 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $8c */	NULL, NULL, NULL, NULL,	// MIDI
	/* $90 */	NULL, NULL, NULL, NULL,
	/* $94 */	NULL, NULL, NULL, NULL,
	/* $98 */	NULL, NULL, NULL, NULL,
	/* $9c */	NULL, NULL, NULL, NULL,
	/* $a0 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $a4 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $a8 */	NULL, NULL, NULL, NULL,	// MIDI
	/* $ac */	NULL, NULL, NULL, NULL,	// MIDI
	/* $b0 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $b4 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $b8 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $bc */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $c0 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $c4 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $c8 */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $cc */	NULL, NULL, NULL, NULL,	// RS-232C
	/* $d0 */	NULL, NULL, NULL, NULL,
	/* $d4 */	NULL, NULL, NULL, NULL,
	/* $d8 */	NULL, NULL, NULL, NULL,
	/* $dc */	NULL, NULL, NULL, NULL,
	/* $e0 */	NULL, NULL, NULL, NULL,
	/* $e4 */	NULL, NULL, NULL, NULL,
	/* $e8 */	NULL, NULL, NULL, NULL,
	/* $ec */	NULL, NULL, NULL, NULL,

			//   0123456789012345678
	/* $f0 */	"PSX16x50 Interrupt",
	/* $f1 */	NULL,
	/* $f2 */	NULL,
	/* $f3 */	NULL,
	/* $f4 */	NULL,
	/* $f5 */	NULL,
	/* $f6 */	"ExSPC Interrupt",
	/* $f7 */	NULL,	// TS-6BSI
	/* $f8 */	"Nereid#1 Interrupt",
	/* $f9 */	"Neptune-X Interrupt",
	/* $fa */	"Nereid#1 USB",
	/* $fb */	"Nereid#0 USB",
	/* $fc */	NULL,
	/* $fd */	NULL,
	/* $fe */	NULL,
	/* $ff */	NULL,
};

// ベクタ番号から表示名を返す
const char *
m68030_get_exception_name(int vector)
{
	const char *name = NULL;
	if (vector < countof(exception_name_table)) {
		name = exception_name_table[vector];
	}
	if (name == NULL) {
		name = "(reserved?)";
	}
	return name;
}
