//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "header.h"
#include "branchhistory.h"
#include "bus.h"
#include "m88200.h"
#include <atomic>

// 外部リクエストフラグ
// XXX m680x0 側にもよく似たものがあってどうしたもんか
#define CPU_REQ_STOP		(0x00000001)	// ストップ状態
#define CPU_REQ_TRACE		(0x00000004)	// デバッガ有効
#define CPU_REQ_PROMPT		(0x00000008)	// デバッガプロンプト
#define CPU_REQ_RELEASE		(0x00000010)	// 命令後に CPU 実行中断
#define CPU_REQ_INTR		(0x00000020)	// 命令後に割り込みチェック

// m88100 レジスタイメージを保持する構造体。
// この構造体はコピーや比較をするので、ポインタとかは置かないこと。
struct m88100reg
{
	// PID(cr0)
	static const uint32 PID_REV_MASK	= 0x0000ff00; // 0 = MC88100
	static const uint32 PID_VER_MASK	= 0x000000fe;
	static const uint32 PID_MASTER		= 0x00000001;
	// PSR(cr1)
	static const uint32 PSR_SUPER		= 0x80000000;
	static const uint32 PSR_BO_LE		= 0x40000000; // ByteOrder(Little)
	static const uint32 PSR_SER			= 0x20000000;
	static const uint32 PSR_C			= 0x10000000; // Carry
	static const uint32 PSR_SFD1		= 0x00000008; // SFU1 Disable
	static const uint32 PSR_MXM			= 0x00000004; // MisalignedAccessEnable
	static const uint32 PSR_IND			= 0x00000002; // Interrupt Disable
	static const uint32 PSR_SFRZ		= 0x00000001; // Shadow Freeze

	// 例外処理で使用される S*IP
	static const uint32 SIP_MASK		= 0xfffffffcU;
	static const uint32 SIP_V			= 0x00000002;
	static const uint32 SIP_E			= 0x00000001;

	// DMTn、データ例外
	static const uint32 DM_BO			= 0x00008000;
	static const uint32 DM_DAS			= 0x00004000;
	static const uint32 DM_DOUB1		= 0x00002000;
	static const uint32 DM_LOCK			= 0x00001000;
	static const uint32 DM_DREG_MASK	= 0x00000f80;
	static const uint32 DM_SIGNED		= 0x00000040;
	static const uint32 DM_EN3			= 0x00000020;
	static const uint32 DM_EN2			= 0x00000010;
	static const uint32 DM_EN1			= 0x00000008;
	static const uint32 DM_EN0			= 0x00000004;
	static const uint32 DM_EN_MASK		= (DM_EN0 | DM_EN1 | DM_EN2 | DM_EN3);
	static const uint32 DM_WRITE		= 0x00000002;
	static const uint32 DM_VALID		= 0x00000001;
	// 内部識別フラグ
	static const uint32 DM_SIZE_B		= (1U << 16) | DM_EN3;
	static const uint32 DM_SIZE_H		= (1U << 17) | DM_EN3 | DM_EN2;
	static const uint32 DM_SIZE_W		= (1U << 18) | DM_EN_MASK;
	static const uint32 DM_USR			= (1U << 20);
	static const uint32 DM_BU			= DM_SIZE_B;
	static const uint32 DM_B			= DM_SIZE_B | DM_SIGNED;
	static const uint32 DM_HU			= DM_SIZE_H;
	static const uint32 DM_H			= DM_SIZE_H | DM_SIGNED;
	static const uint32 DM_W			= DM_SIZE_W;
	static const uint32 DM_D1			= DM_SIZE_W | DM_DOUB1;
	static const uint32 DM_D2			= DM_SIZE_W;

	// 頻繁にアクセスしそうな変数はオフセットの若いほうに持ってきておく。

	// 実行中の命令
	uint64 opX = 0;
	// プリフェッチ命令
	uint64 opF = 0;

	uint32 xip = 0;
	uint32 nip = 0;
	uint32 fip = 0;

	uint32 r[32] {};
	union {
		uint32 cr[21] {};
		struct {
			uint32 pid;		// cr0
			uint32 psr;		// cr1
			uint32 epsr;	// cr2
			uint32 ssbr;	// cr3
			uint32 sxip;	// cr4
			uint32 snip;	// cr5
			uint32 sfip;	// cr6
			uint32 vbr;		// cr7
			uint32 dmt0;	// cr8
			uint32 dmd0;	// cr9
			uint32 dma0;	// cr10
			uint32 dmt1;	// cr11
			uint32 dmd1;	// cr12
			uint32 dma1;	// cr13
			uint32 dmt2;	// cr14
			uint32 dmd2;	// cr15
			uint32 dma2;	// cr16
			uint32 sr[4];	// cr17,18,19,20
		};
	};
	union {
		uint32 fcr[11] {};
		struct {
			uint32 fpecr;	// fcr0
			uint32 fphs1;	// fcr1
			uint32 fpls1;	// fcr2
			uint32 fphs2;	// fcr3
			uint32 fpls2;	// fcr4
			uint32 fppt;	// fcr5
			uint32 fprh;	// fcr6
			uint32 fprl;	// fcr7
			uint32 fpit;	// fcr8
			uint32 fpsr;	// fcr62 [9]
			uint32 fpcr;	// fcr63 [10]
		};
	};

	// 直近のデータアクセスアドレス
	uint32 lastaddr;

	// スーパーバイザ/ユーザモードなら true
	bool IsSuper()	const { return (psr & PSR_SUPER) != 0; }
	bool IsUser()	const { return (psr & PSR_SUPER) == 0; }

	// MXM (Misaligned Access Enable) なら true
	bool IsMXM()	const { return (psr & PSR_MXM) != 0; }

	// キャリーフラグが立っていれば 1 を返す
	uint32 GetCY() const { return (psr >> 28) & 1; }
	// キャリーフラグをセットする。cy は最下位ビットのみ参照する
	void SetCY(uint cy) {
		psr &= ~PSR_C;
		psr |= ((cy & 1) << 28);
	}

	static const char * const sipname[3];
	static const char * const dmt_en_str[16];
};

class m88kcpu : public m88100reg
{
 public:
	m88kcpu();
	virtual ~m88kcpu();

 public:
	void Reset(uint32 reset_vector);
	uint32 Run(uint64 request);
	void Release();

	uint64 total_cycle = 0;		// 積算実行サイクル数

	void Interrupt();

	// PID.VERSION を設定する (初期化時に呼ぶ)
	void SetVersion(uint32 version);

 public:
	// [0] 命令バス
	// [1] データバス
	m88200 cmmu[2];

	uint64 fetch() {
		nip = fip;
		opF = cmmu[0].load_32(fip);
		fip += 4;
		return opF;
	}

	// op が>=0ならバスエラーは起きていない
	//  そのときに DELAYSLOT bit が 0 ならば通常
	//  bit32 が 1 ならば遅延スロット命令
	// op が負数ならバスエラーが起きた
	//  そのときに DELAYSLOT bit が 1 ならば通常バスエラー
	//  bit32 が 0 ならば遅延スロットでのバスエラー
	// バスエラーかどうかのチェックのほうが回数が多いため、
	// バスエラー優先にしたい。


	bool OpIsBusErr(uint64 op) const {
		return (int64)op < 0;
	}
	bool OpIsDelay(uint64 op) const {
		return (OpIsBusErr(op) ? ~op : op) & DELAYSLOT;
	}

	// PSR を変更する。
	// CY ビットは特別にこの関数を経由しないで変更してよいことにする。
	void SetPSR(uint32 newpsr)
	{
		psr = newpsr;
		// CMMU に S/U 信号を出す
		cmmu[0].SetSuper(IsSuper());
		cmmu[1].SetSuper(IsSuper());
	}

	// クラス外部からの xip 取得
	uint32 Getxip() const {
		return xip;
	}

	enum ExceptionKind
	{
		NORMAL,
		DATA,
		ERROR,
		INTR,
		TRAP,
	};

	void Exception(int vec);
	void ExceptionCore(int vec, ExceptionKind cause);
	void ReadDataException(uint32 addr, uint32 flag);
	void WriteDataException(uint32 addr, uint32 data, uint32 flag);
	void XmemDataException(uint32 addr, uint32 data, uint32 flag);

	// 例外名を返す
	static const char *GetExceptionName(int vector);

	// ブランチ履歴 (例外履歴を含む)
	BranchHistory_m88xx0 brhist {};
	// 例外履歴のみ
	BranchHistory_m88xx0 exhist {};

	// FPU
	bool IsFPUEnable() { return !(psr & PSR_SFD1); }
	void fpu_unimpl();

	std::atomic<uint32> atomic_reqflag {}; // リクエストフラグ

 private:
	const uint64 DELAYSLOT = 1ULL << 62;

	// OpSetDelay を一つの op に対して複数回呼ばないでください
	uint64 OpSetDelay(uint64 op) const {
		return op ^ DELAYSLOT;
	}

	// ブランチ処理用
	void EnterBranch();
	void DoBranch(uint32 toaddr);
	void ExitBranch();
	void ExitDelayBranch();
	uint32 nop_counter {};	// STOP 検出用の nop 命令カウンタ

	// ロードストア命令用の下処理
	bool ldst_usr(uint32& usr);
	uint32 ldst_scale(uint32 size);
	bool ldst_align(uint32 size, uint32& addr);

#define OP_PROTO(name)	void __CONCAT(op_,name)()
#include "m88100ops.h"
OP_PROTO(illegal);
#undef OP_PROTO

	void op_unimpl() {
		printf("unimplemented op XIP=%08x opX=%08x\n", xip, (uint32)opX);
	}

	// 例外名
	static const char * const exception_names[];
};

// とりあえず
static inline uint32
op32_to_12(uint32 op)
{
	uint32 op12 = ((op >> 20) & 0x0fc0) | ((op >> 10) & 0x003f);
	return op12;
}
