//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "m88100.h"
#include "m88100excep.h"
#include "debugger.h"
#include "cvprompt.h"
#include "mpu.h"

#define OP_DEF(name)	void __CONCAT(m88kcpu::op_,name)()
#define OP_FUNC(name)	__CONCAT(op_,name)()
#include "m88100ops.cpp"

// リセット例外のようなもの?
void
m88kcpu::Reset(uint32 reset_vector)
{
	// TODO: 未完成
	psr = 0x800003ff;
	// SFD1 = 0;
	psr &= ~PSR_SFD1;

#if defined(TEST_INTERRUPT)
psr &= ~PSR_IND;
psr &= ~PSR_SFRZ;
#endif

	SetPSR(psr);

	vbr = 0;

	cmmu[0].Reset();
	cmmu[1].Reset();

	// 実際には m88k MPU はリセット直後は 0 番地から実行を開始するので、
	// 大抵の場合システムでリセット直後は 0 番地付近に ROM を見せるように
	// してあるのだと思うが、ここではその代わりに ROM のアドレスを指定して、
	// 直接そこから実行を開始することにする。
	fip = reset_vector;
	nip = 0;
	xip = 0;

	fetch();
}

// request サイクル分実行する。
uint32
m88kcpu::Run(uint64 request)
{
	uint64 goal_cycle = total_cycle + request;
	bool is_release = false;
	bool is_intr = false;

	if ((atomic_reqflag & CPU_REQ_STOP)) {
		if ((atomic_reqflag & CPU_REQ_INTR) == 0) {
			// STOP 状態のままで割り込みが無ければ、
			// STOP 継続として戻る。
			total_cycle = goal_cycle;
			return atomic_reqflag;
		}
		// 状態が変わったら STOP 解除。
		atomic_reqflag &= ~CPU_REQ_STOP;
	}

	while (!is_release) {
		// shift pipeline
		opX = opF;
		xip = nip;

		// prefetch next inst
		fetch();

		// この時点で実行しようとする命令が opX に入っている

		if (atomic_reqflag != 0) {
			// 何かが起きた

			// 割り込みの認識
			// 割り込みの実行は命令実行完了後、次のプリフェッチの前
			if ((atomic_reqflag & CPU_REQ_INTR)) {
				atomic_reqflag &= ~CPU_REQ_INTR;
				is_intr = true;
#if defined(TEST_INTERRUPT)
					atomic_reqflag |= CPU_REQ_PROMPT;
#endif
			}

			if ((atomic_reqflag & CPU_REQ_TRACE)) {
				// デバッガプロンプトを出すかどうかを判断する
				//inst_count++;
				if (debugger_check()) {
					atomic_reqflag |= CPU_REQ_PROMPT;
				}
			}

			if ((atomic_reqflag & CPU_REQ_PROMPT)) {
				// デバッガプロンプトのための実行停止

				// プロンプト獲得を通知
				gCVPrompt->NotifyAcquire();

				// プロンプト解放を待機
				gCVPrompt->WaitRelease();

				// プロンプトを抜けたのでここでフラグを落とす
				atomic_reqflag &= ~CPU_REQ_PROMPT;
			}

			if ((atomic_reqflag & CPU_REQ_RELEASE)) {
				// 中断リクエスト受信
				// この命令を実行し終わったら抜ける
				atomic_reqflag &= ~CPU_REQ_RELEASE;
				is_release = true;
			}
		}

		if (OpIsBusErr(opX)) {
			// プリフェッチしていた命令がバスエラー
			Exception(M88K_EXCEP_INST);
			continue;
		}

		uint32 op12 = op32_to_12(opX);

		switch (op12) {
#include "m88100switch.inc"
		 default:
			OP_FUNC(illegal);
			break;
		}

		if (is_intr) {
			if (!(psr & PSR_IND)) {
				// XXX: 割り込み信号線はレベル割り込みなのだが
				// まだ処理できないのでとりあえずラッチ割り込みにしておく。
				is_intr = false;
				gMPU->putlog(4, "INTR take");
				ExceptionCore(M88K_EXCEP_INTERRUPT, ExceptionKind::INTR);
				continue;
			} else {
				gMPU->putlog(4, "INTR DISABLE");
			}
		}

		total_cycle++;

		if ((atomic_reqflag & CPU_REQ_STOP)) {
			// STOP 状態は、残りサイクルを消費したことにして戻る
			gMPU->putlog(2, "STOP 状態検出");
			total_cycle = goal_cycle;
			is_release = true;
		}

		if (total_cycle >= goal_cycle)
			break;
	}

	return atomic_reqflag;
}

// この命令後に実行を中断して VM に戻る
void
m88kcpu::Release()
{
	atomic_reqflag |= CPU_REQ_RELEASE;
}

// 通常の例外。
// DataAccessException, エラー、割り込み、TRAP からは使わない。
// vec はベクタ番号。
void
m88kcpu::Exception(int vec)
{
	ExceptionCore(vec, ExceptionKind::NORMAL);
}

// 例外処理コア部分
// vec はベクタ番号。
void
m88kcpu::ExceptionCore(int vec, ExceptionKind cause)
{
	// 例外履歴に記録 (例外発生はブランチ履歴にも記録)
	exhist.AddEntry(xip, 0, 0xfc000000 | vec);
	brhist.AddEntry(xip, 0, 0xfc000000 | vec);
	// デバッガに通知 (例外ブレーク用)
	debugger_notifyexception(vec);

	if (psr & PSR_SFRZ) {
		if (cause != ExceptionKind::TRAP) {
			gMPU->putlog(2, "SFRZ Error Exception");
			vec = M88K_EXCEP_ERROR;
			cause = ExceptionKind::ERROR;
		}
	} else {
		gMPU->putlog(2, "Exception %x", vec);
		if (vec != 1) {
			gMPU->putlog(1, "Exception %x x=%x n=%x f=%x", vec, xip, nip, fip);
		}
		sxip = xip;
		snip = nip;
		sfip = fip; 

		sxip |= SIP_V;
		snip |= SIP_V;
		// XXX:実機と異なり、fip は常にフェッチ前なので V=0
		// ^^^^ コレ自体怪しい。INTERRUPT とか。

		if (OpIsBusErr(opX)) {
			sxip |= SIP_E;	// E bit
		}
		if (OpIsBusErr(opF)) {
			snip |= SIP_E;	// E bit
		}

		if (vec == M88K_EXCEP_UNIMPL_OP) {
			sxip &= ~SIP_V;
		}
	}

	if (cause != ExceptionKind::DATA) {
		// データアクセス例外以外は、
		// データパイプラインはフラッシュ状態とする。
		// ほんとうは違う。かも。
		dmt0 = 0;
	} else {
		// xip: ld/st/xmem
		// nip: next
		// fip: fetch

		// rom1.2 は DAE が起きると、

		// ROM挙動               CPU の RTE 挙動
		// sfip = sxip
		// sxip -> ~V (INVALID)  CPU はそもそも sxip を評価しない
		// snip = 0 (INVALID)    CPU はこれを nop 相当に処理する
		// rte

		// してくる。そのまま読むと命令を再実行しようとしているように
		// 読めるが、再実行すると無限ループになってしまう。
		// DAE の時点で xip が進んでいることを期待して利用している?

		// openbsd は
		// xmem の write バスエラーを検知すると命令を再実行するために
		// fip = nip, nip = xip して rte する。ということは xip = xmem に
		// なっていることを期待している。

		// というわけで
		// ld で DataAccessException のときは、実CPUはすでに次の命令を
		// 実行している?
		// もう少し考えると、スコアボードでレジスタの干渉がなければ、
		// 実行している?

		// 仕方がないのでハックでごまかす。
		// rte のほうも参照のこと。

		if (IsSuper() && xip >= 0x41000000 && xip < 0x42000000) {
			// 遅延例外機構がないので、ROM 合わせのハック。
			sxip = snip;
		}
	}

	epsr = psr;
	SetPSR(psr | PSR_SUPER | PSR_SFD1 | PSR_IND | PSR_SFRZ);

	fip = vbr + (vec << 3);
	fetch();
	// ここでベクタに飛ぶのだが、ベクタ内の2命令でおそらく必ずもう一度
	// 分岐してそっちでも履歴が残るので、ここでブランチ履歴を残すのは
	// ちょっと冗長という気もする。

	if (OpIsBusErr(opF)) {
		if (cause == ExceptionKind::ERROR) {
			// double bus fault
			// 実機は無限ループ状態
			PANIC("Infinite ERROR (double bus fault)");
		} else {
			ExceptionCore(M88K_EXCEP_ERROR, ExceptionKind::ERROR);
		}
	}
}

// データアクセス例外(Read)
void
m88kcpu::ReadDataException(uint32 addr, uint32 flag)
{
	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	dma0 = addr & ~0x3;
	dmt0 = ((psr & PSR_BO_LE) ? DM_BO : 0)
	     | (((psr & PSR_SUPER) && !(flag & DM_USR)) ? DM_DAS : 0)
	     | (flag & DM_DOUB1)	// ダブルワードの第1アクセスのときセット
	     | (flag & DM_LOCK)		// xmem でセット
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	     | 0 // DM_WRITE = 0
	     | DM_VALID;
	gMPU->putlog(1, "ReadDAE xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// データアクセス例外(Write)
void
m88kcpu::WriteDataException(uint32 addr, uint32 data, uint32 flag)
{
	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	dma0 = addr & ~0x3;
	dmt0 = ((psr & PSR_BO_LE) ? DM_BO : 0)
	     | (((psr & PSR_SUPER) && !(flag & DM_USR)) ? DM_DAS : 0)
	     | (flag & DM_DOUB1)	// ダブルワードの第1アクセスのときセット
	     | (flag & DM_LOCK)		// xmem でセット
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	     | DM_WRITE
	     | DM_VALID;

	dmd0 = data;
	gMPU->putlog(1, "WriteDAE xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

// データアクセス例外(xmem)
void
m88kcpu::XmemDataException(uint32 addr, uint32 data, uint32 flag)
{
	dma0 = addr & ~0x3;
	dmt0 = ((psr & PSR_BO_LE) ? DM_BO : 0)
	     | (((psr & PSR_SUPER) && !(flag & DM_USR)) ? DM_DAS : 0)
	     | (flag & DM_DOUB1)	// ダブルワードの第1アクセスのときセット
	     | (flag & DM_LOCK)		// xmem でセット
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	     | 0 // DM_WRITE = 0
	     | DM_VALID;

	dma1 = addr & ~0x3;
	dmt1 = ((psr & PSR_BO_LE) ? DM_BO : 0)
	     | (((psr & PSR_SUPER) && !(flag & DM_USR)) ? DM_DAS : 0)
	     | (flag & DM_DOUB1)	// ダブルワードの第1アクセスのときセット
	     | (flag & DM_LOCK)		// xmem でセット
	     | (FLD_D << 7)
	     | (flag & DM_SIGNED)
	     | ((flag & DM_EN_MASK) >> (addr & 0x03))
	     | DM_WRITE
	     | DM_VALID;

	dmd1 = data;

	// ミスアラインドはこの例外より前にチェックされていて、
	// ここにはアラインドしかこない。
	if (cmmu[1].fault_code != m88200::FAULT_CODE_BUSERR
	 || cmmu[1].acc_read) {
		// アドレス変換フォルトまたは read 時点
		// nop
	} else {
		// xmem の write 時点
		dma0 = dma1;
		dmt0 = dmt1;
		dmd0 = dmd1;
		dmt1 &= ~DM_VALID;
	}

	gMPU->putlog(1, "xmemDAE xip=%x addr=%x", xip, addr);
	ExceptionCore(M88K_EXCEP_DATA, ExceptionKind::DATA);
}

void
m88kcpu::Interrupt()
{
	atomic_reqflag |= CPU_REQ_INTR;
//	gMPU->putlog(1, "INTERRUPT IGNORED");
	gMPU->putlog(4, "INTERRUPT ACK (アドレスは不正確)");
}

void
m88kcpu::fpu_unimpl()
{
	printf("%s\n", __func__);
}

OP_DEF(illegal)
{
	// 本来は不当命令例外?
	gMPU->putlog(0, "Illegal instruction %08x", (uint32)opX);
}
