//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "bt454.h"
#include "bitmap.h"
#include "mpu.h"

// 仕様書には $c100_0000 と書いてあるが、PROM も NetBSD カーネルも
// $c110_0000 にアクセスしている。実際には 4バイトごとにミラーが見えて
// いるので問題はない。

// 下位ニブルは 0x0f が読めるようだ
#define LOW_NIBBLE	(0x0f)

std::unique_ptr<BT454Device> gBT454;

BT454Device::BT454Device()
{
	logname = "bt454";
	devname = "BT454";
	devaddr = 0xc1000000;
}

BT454Device::~BT454Device()
{
}

bool
BT454Device::Init()
{
	nplane = gBitmap->GetPlaneCount();
	return true;
}

void
BT454Device::ResetHard()
{
	reg.addr = 0;
}

/* static */ const char
BT454Device::rgbstr[3] = { 'R', 'G', 'B' };

uint64
BT454Device::Read(uint32 addr)
{
	uint32 data;

	switch (addr) {
	 case 0:	// アドレスレジスタ
		// 今の(今からアクセスする)パレット番号
		data = reg.addr / 3;
		data = (data << 4) | LOW_NIBBLE;
		putlog(2, "アドレスレジスタ -> $%02x", data);
		break;

	 case 1:	// カラーマップレジスタ
		// reg.color[] は上位4ビットが有効なのでそのまま読み出せる。
		data = reg.color[reg.addr];
		data |= LOW_NIBBLE;
		putlog(2, "カラーマップ $%x(%c) -> $%02x",
			reg.addr / 3, rgbstr[reg.addr % 3], data);
		// アクセスするたびにインクリメントする。
		reg.addr++;
		if (reg.addr >= countof(reg.color)) {
			reg.addr = 0;
		}
		break;

	 case 2:
	 case 3:
		putlog(0, "未実装バイト読み込み $%08x", gMPU->GetPaddr());
		data = 0xff;
		break;

	 default:
		__unreachable();
	}

	return data;
}

uint64
BT454Device::Write(uint32 addr, uint32 data)
{
	// 書き込みデータのうち有効なのは上位ニブルだけ

	switch (addr) {
	 case 0:	// アドレスレジスタ
		// 今からアクセスするパレット番号
		// data はパレット番号だが reg.addr は配列のインデックスなので 3倍。
		putlog(2, "アドレスレジスタ <- $%02x", data);
		reg.addr = (data >> 4) * 3;
		break;

	 case 1:	// カラーマップレジスタ
		// reg.color[] は上位4ビットが有効なのでそのまま書き込む。
		// 下位4ビットはどうするか。
		reg.color[reg.addr] = (data & 0xf0);
		if (loglevel == 1) {
			// ログレベル1なら RGB が書き込まれた時点で表示。
			// 大体はこれくらいの表示で構わないはず。
			if (reg.addr % 3 == 2) {
				putlog("カラーマップ $%x = (%x,%x,%x)",
					reg.addr / 3,
					reg.color[reg.addr - 2] >> 4,
					reg.color[reg.addr - 1] >> 4,
					reg.color[reg.addr    ] >> 4);
			}
		} else if (loglevel >= 2) {
			putlog("カラーマップ $%x(%c) <- $%02x",
				reg.addr / 3, rgbstr[reg.addr % 3], data);
		}
		// ホストカラーを計算。
		MakeHostColor(reg.addr);

		// ホストカラー計算後にインクリメントする。
		reg.addr++;
		if (reg.addr >= countof(reg.color)) {
			reg.addr = 0;
		}
		break;

	 case 2:
	 case 3:
		putlog(0, "未実装バイト書き込み $%08x <- %02x", gMPU->GetPaddr(), data);
		break;

	 default:
		__unreachable();
	}
	return 0;
}

uint64
BT454Device::Peek(uint32 addr)
{
	switch (addr) {
	 case 0:	// アドレスレジスタ
		// XXX 下位ニブルは?
		return ((reg.addr / 3) << 4) | LOW_NIBBLE;

	 case 1:	// カラーマップレジスタ
		// 今読み出そうとしたら読み出せるやつ
		return reg.color[reg.addr] | LOW_NIBBLE;

	 case 2:
	 case 3:
		// 未実装
		return 0xff;

	 default:
		__unreachable();
	}
}

// addr 番目のレジスタ値からホストパレットを作成。
void
BT454Device::MakeHostColor(int addr)
{
	int col = addr / 3;
	int idx = addr % 3;

	if (nplane == 1) {
		// 1bpp なら R,G,B のうち G の値だけを使用
		if (idx == 1) {
			reg.hostcolor[col * 4 + 0] = reg.color[col * 3 + 1];
			reg.hostcolor[col * 4 + 1] = reg.color[col * 3 + 1];
			reg.hostcolor[col * 4 + 2] = reg.color[col * 3 + 1];
		}
	} else {
		// 4bpp なら代入するだけ
		reg.hostcolor[col * 4 + idx] = reg.color[col * 3 + idx];
	}
}
