//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// M88200 CMMU の I/O デバイス部分

#include "cmmu.h"
#include "mpu88xx0.h"

// CMMU が確定したらそっちのデバイスログから出力する。
// CMMU が確定しない場合はこのデバイスのログを使う。

std::unique_ptr<CMMUDevice> gCMMU;

// コンストラクタ
CMMUDevice::CMMUDevice()
{
	// m88200 側の logname が "cmmu7" とかなので、こっちを "cmmu" で
	// 受け付けてしまうのは紛らわしそうなので。
	logname = "cmmudev";
	devname = "CMMU";
}

// デストラクタ
CMMUDevice::~CMMUDevice()
{
}

bool
CMMUDevice::Init()
{
	// CMMU をポイントしておく
	cmmus[7] = &gMPU88xx0->GetCPU()->cmmu[0];
	cmmus[6] = &gMPU88xx0->GetCPU()->cmmu[1];

	return true;
}

// アドレスデコーダ
/*static*/ inline uint32
CMMUDevice::Decoder(uint32 addr)
{
	uint32 offset = addr & 0xfff;

	// BWP* の A5 はデコードされてない
	if (0x420 <= offset && offset < 0x440) {
		offset &= ~0x20;
	}
	// CDP*,CTP* の A4,A5 はデコードされていない
	if (0x800 <= offset && offset < 0x880) {
		offset &= ~0x30;
	}

	return offset;
}

uint64
CMMUDevice::Read32(uint32 addr)
{
	uint32 id = (addr >> 12) & 0xff;

	m88200 *cmmu = GetCMMU(id);
	if (cmmu == NULL) {
		putlog(3, "未搭載 CMMU[%d] 読み込み $%08x", id, addr);
		return 0xffffffff;
	}

	uint32 offset = Decoder(addr);
	uint32 data;
	switch (offset) {
	 case IDR:
		data = cmmu->GetIDR();
		cmmu->putlog(2, "IDR  -> $%08x", data);
		return data;

	 case SCR:
		data = cmmu->GetSCR();
		cmmu->putlog(2, "SCR  -> $%08x", data);
		return data;

	 case SSR:
		data = cmmu->GetSSR();
		cmmu->putlog(2, "SSR  -> $%08x", data);
		return data;

	 case SAR:
		data = cmmu->GetSAR();
		cmmu->putlog(2, "SAR  -> $%08x", data);
		return data;

	 case SCTR:
		data = cmmu->GetSCTR();
		cmmu->putlog(2, "SCTR -> $%08x", data);
		return data;

	 case PFSR:
		data = cmmu->GetPFSR();
		cmmu->putlog(2, "PFSR -> $%08x", data);
		return data;

	 case PFAR:
		data = cmmu->GetPFAR();
		cmmu->putlog(2, "PFAR -> $%08x", data);
		return data;

	 case SAPR:
		data = cmmu->GetSAPR();
		cmmu->putlog(2, "SAPR -> $%08x", data);
		return data;

	 case UAPR:
		data = cmmu->GetUAPR();
		cmmu->putlog(2, "UAPR -> $%08x", data);
		return data;

	 case BWP0 ... BWP7:
		// BWP は書き込み専用 (何が読めるだろうか)
		return 0xffffffff;

	 case CDP0 ... CDP3:
	 case CTP0 ... CTP3:
		break;

	 case CSSP:
		data = cmmu->GetCSSP();
		cmmu->putlog(2, "CSSP -> $%08x", data);
		return data;

	 default:
		break;
	}
	putlog(0, "未実装読み込み $%08x", addr);
	return 0xffffffff;
}

uint64
CMMUDevice::Write32(uint32 addr, uint32 data)
{
	uint32 id = (addr >> 12) & 0xff;

	m88200 *cmmu = GetCMMU(id);
	if (cmmu == NULL) {
		putlog(3, "未搭載 CMMU[%d] 書き込み $%08x <- $%08x", id, addr, data);
		return 0;
	}

	uint32 offset = Decoder(addr);
	uint n;
	switch (offset) {
	 case IDR:
	 {
		uint32 newid = data >> 24;
		if (id != newid) {
			cmmu->putlog(1, "ID 更新 $%02x", newid);
			cmmu->SetID(newid);
		} else {
			cmmu->putlog(2, "IDR 書き込み(ID変更なし) <- $%08x", data);
		}
		return 0;
	 }

	 case SCR:
		cmmu->SetSCR(data);
		return 0;

	 case SSR:
		cmmu->putlog(1, "SSR  <- $%08x", data);
		cmmu->SetSSR(data);
		return 0;

	 case SAR:
		cmmu->putlog(2, "SAR  <- $%08x", data);
		cmmu->SetSAR(data);
		return 0;

	 case SCTR:
		cmmu->SetSCTR(data);
		return 0;

	 case PFSR:
		cmmu->SetPFSR(data);
		return 0;

	 case PFAR:
		cmmu->SetPFAR(data);
		return 0;

	 case SAPR:
		cmmu->SetSAPR(data);
		return 0;

	 case UAPR:
		cmmu->SetUAPR(data);
		return 0;

	 case BWP0 ... BWP7:
		n = (offset - BWP0) / 4;
		assert(n < 8);
		cmmu->SetBATC(n, data);
		return 0;

	 case CDP0 ... CDP3:
	 case CTP0 ... CTP3:
		break;

	 case CSSP:
		cmmu->SetCSSP(data);
		return 0;

	 default:
		break;
	}
	putlog(0, "未実装書き込み $%08x <- $%08x", addr, data);
	PANIC(" ");
	return 0;
}

// Peek8() の下請け。32ビットで処理して後でバイトに分割するほうが楽。
uint32
CMMUDevice::Peek32(uint32 addr) const
{
	uint32 id = (addr >> 12) & 0xff;

	m88200 *cmmu = GetCMMU(id);
	if (cmmu == NULL) {
		return 0xffffffff;
	}

	uint32 offset = Decoder(addr);
	switch (offset) {
	 case IDR:
		return cmmu->GetIDR();

	 case SCR:
		return cmmu->GetSCR();

	 case SSR:
		return cmmu->GetSSR();

	 case SAR:
		return cmmu->GetSAR();

	 case SCTR:
		return cmmu->GetSCTR();

	 case PFSR:
		return cmmu->GetPFSR();

	 case PFAR:
		return cmmu->GetPFAR();

	 case SAPR:
		return cmmu->GetSAPR();

	 case UAPR:
		return cmmu->GetUAPR();

	 case BWP0 ... BWP7:
		return 0xffffffff;

	 case CDP0 ... CDP3:
	 case CTP0 ... CTP3:
		break;

	 case CSSP:
		return cmmu->GetCSSP();

	 default:
		break;
	}
	return 0xffffffff;
}

uint64
CMMUDevice::Peek8(uint32 addr)
{
	uint32 data = Peek32(addr & ~3U);
	switch (addr & 3) {
	 case 0:
		return data >> 24;
	 case 1:
		return (data >> 16) & 0xff;
	 case 2:
		return (data >>  8) & 0xff;
	 case 3:
		return data & 0xff;
	}
	__unreachable();
}
