//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "keyboard.h"
#include "mfp.h"
#include "mpu680x0.h"
#include "mystring.h"
#include "scheduler.h"

std::unique_ptr<MFPDevice> gMFP;

// コンストラクタ
MFPDevice::MFPDevice()
{
	logname = "mfp";
	devname = "MFP";
	devaddr = baseaddr;
	devlen  = 0x2000;
	monitor_size = nnSize(65, 17);

	for (int ch = 0; ch < countof(event); ch++) {
		event[ch].dev = this;
		event[ch].func = (DeviceCallback_t)&MFPDevice::TimerCallback;
		event[ch].code = ch;
		event[ch].SetName(string_format("MFP Timer-%c", ch + 'A'));
	}

	// キーボード周りは x68kkbd.cpp のコメント参照
	key_event.dev = this;
	key_event.func = (DeviceCallback_t)&MFPDevice::KeyCallback;
	key_event.time = 3750_usec;
	key_event.SetName("MFP USART");
}

// デストラクタ
MFPDevice::~MFPDevice()
{
}

// MPU の RESET 命令によるリセット
void
MFPDevice::ResetSoft()
{
	// TxDR, UDR, TSR はクリアされない

	// ?
	mfp.tsr = 0x80;

	// それ以外のすべてのレジスタはクリア?
	mfp.vr = 0;

	// タイマーはすべてストップなのでイベントを削除
	for (int ch = 0; ch < countof(event); ch++) {
		event[ch].Stop();
	}
	key_event.Stop();

	// 信号線
	rr = false;

	// ペンディング割り込みクリア
}

uint64
MFPDevice::Read(uint32 addr)
{
	uint8 data;

	switch (addr) {
	 case MFP::GPIP:
		data = mfp.gpip;
		break;
	 case MFP::AER:
		data = mfp.aer;
		break;
	 case MFP::DDR:
		data = mfp.ddr;
		break;
	 case MFP::IERA:
		data = mfp.ier.a;
		break;
	 case MFP::IERB:
		data = mfp.ier.b;
		break;
	 case MFP::IPRA:
		data = mfp.ipr.a;
		break;
	 case MFP::IPRB:
		data = mfp.ipr.b;
		break;
	 case MFP::ISRA:
		data = mfp.isr.a;
		break;
	 case MFP::ISRB:
		data = mfp.isr.b;
		break;
	 case MFP::IMRA:
		data = mfp.imr.a;
		break;
	 case MFP::IMRB:
		data = mfp.imr.b;
		break;
	 case MFP::VR:
		data = mfp.vr;
		break;
	 case MFP::TACR:
		data = mfp.tcr[0];
		break;
	 case MFP::TBCR:
		data = mfp.tcr[1];
		break;
	 case MFP::TCDCR:
		data = (mfp.tcr[2] << 4) | mfp.tcr[3];
		break;
	 case MFP::TADR:
		data = GetTDR(0);
		break;
	 case MFP::TBDR:
		data = GetTDR(1);
		break;
	 case MFP::TCDR:
		data = GetTDR(2);
		break;
	 case MFP::TDDR:
		data = GetTDR(3);
		break;
	 case MFP::SCR:
		data = mfp.scr;
		break;
	 case MFP::UCR:
		data = mfp.ucr;
		break;
	 case MFP::RSR:
		data = mfp.rsr;
		break;
	 case MFP::TSR:
		data = mfp.tsr;
		break;
	 case MFP::UDR:
		data = mfp.udr;
		break;
	 default:
		data = 0xff;
		break;
	}
	return data;
}

uint64
MFPDevice::Write(uint32 addr, uint32 data)
{
	switch (addr) {
	 case MFP::GPIP:
		putlog(0, "GPIP <- $%02x 未実装書き込み", data);
		mfp.gpip = data;
		break;
	 case MFP::AER:
		putlog(0, "AER <- $%02x 未実装書き込み", data);
		mfp.aer = data;
		break;
	 case MFP::DDR:
		if (data != 0) {
			putlog(0, "DDR <- $%02x 未実装書き込み", data);
		}
		mfp.ddr = data;
		break;
	 case MFP::IERA:
		putlog(2, "IERA <- $%02x", data);
		mfp.ier.a = data;
		break;
	 case MFP::IERB:
		putlog(2, "IERB <- $%02x", data);
		mfp.ier.b = data;
		break;
	 case MFP::IPRA:
		putlog(2, "IPRA <- $%02x", data);
		mfp.ipr.a = data;
		break;
	 case MFP::IPRB:
		putlog(2, "IPRB <- $%02x", data);
		mfp.ipr.b = data;
		break;
	 case MFP::ISRA:
		putlog(2, "ISRA <- $%02x", data);
		mfp.isr.a = data;
		break;
	 case MFP::ISRB:
		putlog(2, "ISRB <- $%02x", data);
		mfp.isr.b = data;
		break;
	 case MFP::IMRA:
		putlog(2, "IMRA <- $%02x", data);
		mfp.imr.a = data;
		break;
	 case MFP::IMRB:
		putlog(2, "IMRB <- $%02x", data);
		mfp.imr.b = data;
		break;
	 case MFP::VR:
		mfp.vr = data & 0xf0;
		putlog(2, "VR <- $%02x", mfp.vr);
		break;
	 case MFP::TACR:
		putlog(2, "TACR <- $%02x", data);
		SetTCR(0, data);
		break;
	 case MFP::TBCR:
		putlog(2, "TBCR <- $%02x", data);
		SetTCR(1, data);
		break;
	 case MFP::TCDCR:
		putlog(2, "TCDCR <- $%02x", data);
		SetTCR(2, data >> 4);
		SetTCR(3, data & 7);
		break;
	 case MFP::TADR:
		putlog(2, "TADR <- $%02x", data);
		SetTDR(0, data);
		break;
	 case MFP::TBDR:
		putlog(2, "TBDR <- $%02x", data);
		SetTDR(1, data);
		break;
	 case MFP::TCDR:
		putlog(2, "TCDR <- $%02x", data);
		SetTDR(2, data);
		break;
	 case MFP::TDDR:
		putlog(2, "TDDR <- $%02x", data);
		SetTDR(3, data);
		break;
	 case MFP::SCR:
		putlog(0, "SCR <- $%02x 未実装書き込み", data);
		mfp.scr = data;
		break;
	 case MFP::UCR:
		putlog(2, "UCR <- $%02x", data);
		SetUCR(data);
		break;
	 case MFP::RSR:
		putlog(2, "RSR <- $%02x", data);
		SetRSR(data);
		break;
	 case MFP::TSR:
		putlog(0, "TSR <- $%02x 未実装書き込み", data);
		mfp.tsr = (mfp.tsr & 0xf0) | (data & 0x0f);
		break;
	 case MFP::UDR:
		if (data == 0x40 || data == 0x41) {
			// MSCTRL だけ頻度が高いのでログレベルを上げておく
			putlog(3, "UDR <- $%02x", data);
		} else {
			putlog(2, "UDR <- $%02x", data);
		}
		SetUDR(data);
		break;
	 default:
		break;
	}
	return 0;
}

uint64
MFPDevice::Peek(uint32 addr)
{
	uint8 data;

	switch (addr) {
	 case MFP::GPIP:
		data = mfp.gpip;
		break;
	 case MFP::AER:
		data = mfp.aer;
		break;
	 case MFP::DDR:
		data = mfp.ddr;
		break;
	 case MFP::IERA:
		data = mfp.ier.a;
		break;
	 case MFP::IERB:
		data = mfp.ier.b;
		break;
	 case MFP::IPRA:
		data = mfp.ipr.a;
		break;
	 case MFP::IPRB:
		data = mfp.ipr.b;
		break;
	 case MFP::ISRA:
		data = mfp.isr.a;
		break;
	 case MFP::ISRB:
		data = mfp.isr.b;
		break;
	 case MFP::IMRA:
		data = mfp.imr.a;
		break;
	 case MFP::IMRB:
		data = mfp.imr.b;
		break;
	 case MFP::VR:
		data = mfp.vr;
		break;
	 case MFP::TACR:
		data = mfp.tcr[0];
		break;
	 case MFP::TBCR:
		data = mfp.tcr[1];
		break;
	 case MFP::TCDCR:
		data = (mfp.tcr[2] << 4) | mfp.tcr[3];
		break;
	 case MFP::TADR:
		data = GetTDR(0);
		break;
	 case MFP::TBDR:
		data = GetTDR(1);
		break;
	 case MFP::TCDR:
		data = GetTDR(2);
		break;
	 case MFP::TDDR:
		data = GetTDR(3);
		break;
	 case MFP::SCR:
		data = mfp.scr;
		break;
	 case MFP::UCR:
		data = mfp.ucr;
		break;
	 case MFP::RSR:
		data = mfp.rsr;
		break;
	 case MFP::TSR:
		data = mfp.tsr;
		break;
	 case MFP::UDR:
		data = mfp.udr;
		break;
	 default:
		data = 0xff;
	}
	return data;
}

void
MFPDevice::MonitorUpdate(TextScreen& monitor)
{
	int x;
	int y;

	monitor.Clear();

	// 0         1         2         3         4         5         6
	// 012345678901234567890123456789012345678901234567890123456789012345
	//                 IER    IMR  IPR  ISR           Mode        TDR Cnt
	// An:MPSC TxEmpty Enable Mask Pend Serv  Timer-A Delay(50us) 255 255

	// 割り込み関連
	monitor.Print(0, 0, "   %-12s %-6s %-4s %-4s %-4s",
		"", "IER", "IMR", "IPR", "ISR");
	y = 1;
	for (int i = 15; i >= 0; i--, y++) {
		uint16 mask = (1 << i);
		monitor.Print(0, y, "%c%d:%-12s %-6s %-4s %-4s %-4s",
			i >= 8 ? 'A' : 'B',
			(i % 8),
			intrname[i],
			(mfp.ier.w & mask) ? "Enable" : "",
			(mfp.imr.w & mask) ? ""       : "Mask",	// %0 でマスクする
			(mfp.ipr.w & mask) ? "Pend"   : "",
			(mfp.isr.w & mask) ? "Serv"   : "");
	}

	// タイマー関連
	x = 39;
	y = 1;
	for (int ch = 0; ch < 4; ch++) {
		y = ch + 1;
		monitor.Print(x, y, "Timer-%c", ch + 'A');
		if (mfp.tcr[ch] == 0) {
			monitor.Print(x + 8, y, "Stopped");
		} else if (mfp.tcr[ch] == 8) {
			monitor.Print(x + 8, y, "Event");
		} else if (mfp.tcr[ch] > 8) {
			monitor.Print(x + 8, y, "Pulse(%s)", prescale_str[mfp.tcr[ch] - 8]);
		} else {
			monitor.Print(x + 8, y, "Delay(%s)", prescale_str[mfp.tcr[ch]]);
		}

		monitor.Print(x + 19, y, "%3d", GetTDR(ch));
		monitor.Print(x + 23, y, "%3d", mfp.tdr[ch]);
	}

	// GPIP 関連
	x = 39;
	y = 8;
	monitor.Puts(x, y++, "GPIP Name GPIP AER DDR");
	for (int i = 7; i >= 0; i--, y++) {
		uint mask = (1 << i);
		monitor.Print(x, y, "b%d %-6s  %d   %d   %s", i, gpipname[i],
			(mfp.gpip & mask) ? 1 : 0,
			(mfp.aer & mask) ? 1 : 0,
			(mfp.ddr & mask) ? "OUT" : "IN");
	}
}

/*static*/ const char *
MFPDevice::intrname[] = {
	"RTC Alarm",
	"EXPON",
	"POW SW",
	"FM IRQ",
	"Timer-D",
	"Timer-C",
	"V-Disp",
	"---",
	"Timer-B",
	"MPSC TxErr",
	"MPSC TxEmpty",
	"MPSC RxErr",
	"MPSC RxFull",
	"Timer-A",
	"CRTC IRQ",
	"H-Sync",
};

/*static*/ const char *
MFPDevice::gpipname[] = {
	"ALARM",
	"EXPON",
	"POW SW",
	"FMIRQ",
	"V-Disp",
	"------",
	"CIRQ",
	"H-Sync",
};

// TxCR の設定値から分周期間というかメインカウンタ1回分の時間を返す
/*static*/ const int
MFPDevice::prescale_ns[8] = {
	0,
	1000,	// /4
	2500,	// /10
	4000,	// /16
	12500,	// /50
	16000,	// /64
	25000,	// /100
	50000,	// /200
};

/*static*/ const char *
MFPDevice::prescale_str[8] = {
	"",
	"1us",
	"2.5us",
	"4us",
	"12.5us",
	"16us",
	"25us",
	"50us",
};

// タイマーのチャンネルからベクタ(ビット番号) に変換する
/*static*/ const int
MFPDevice::timer_vector[4] = {
	MFP::INTR_TIMER_A,
	MFP::INTR_TIMER_B,
	MFP::INTR_TIMER_C,
	MFP::INTR_TIMER_D,
};

// このタイマーで割り込みを上げる必要があるかどうかを返す
bool
MFPDevice::IsTimerIntrEnable(uint ch) const
{
	assertmsg(ch < 4, "ch=%u", ch);
	__assume(ch < 4);

	// 割り込みが無効なら false
	if ((mfp.ier.w & (1 << timer_vector[ch])) == 0) {
		return false;
	}

	// 割り込みがマスクされていれば false
	if ((mfp.imr.w & (1 << timer_vector[ch])) == 0) {
		return false;
	}

	// そうでなければ、割り込みを上げる必要がある
	return true;
}

// TxCR をセットする
void
MFPDevice::SetTCR(int ch, uint32 data)
{
	uint32 prev;

	prev = mfp.tcr[ch];
	data &= 0x0f;

	mfp.tcr[ch] = data;
	if (mfp.tcr[ch] > 0x08) {
		putlog(0, "T%cCR サポートしていないモード $%02x",
			ch + 'A', mfp.tcr[ch]);
		return;
	}
	if (mfp.tcr[ch] == 8) {
		// イベントカウントモード

		// TDR をロードする
		mfp.initial_tdr[ch] = mfp.tdr[ch];
		putlog(1, "Timer-%c イベントカウントモード", ch + 'A');
		return;

	} else if (mfp.tcr[ch] == 0) {
		// タイマー停止
		if (prev != 0) {
			putlog(1, "Timer-%c 停止", ch + 'A');
		}
		mfp.timer_start[ch] = 0;
		event[ch].Stop();
		return;

	} else {
		// タイマー開始
		int scale = prescale_ns[mfp.tcr[ch]];
		if (prev == 0) {
			// タイマー停止中に開始
			// ここで initial_tdr にロード
			mfp.initial_tdr[ch] = mfp.tdr[ch] ?: 256;
			mfp.timer_start[ch] = gScheduler->GetVirtTime();
		} else {
			// タイマー動作中に(再)開始

			// タイマー動作中に TxCR に書き込むと、TxDR は影響を受けずに、
			// プリスケーラのカウンタがリセットされる。
			// 例えば、TCR = 0x01 (プリスケール 10)、TDR = 0x05 で、
			// プリスケールカウントが12回進行した時点で TCR に再び 0x01
			// (プリスケール10) が書き込まれた場合、以下のようになるはず。
			//
			//             time ->
			// prescale cnt| 0 1 2 3 4 5 6 7 8 9 0 1 2
			// main count  | 5 ----------------> 4 -->
			//                                        * ここで TCR 書き込み
			// TCR書き込み後のprescale cnt |          0 1 2 3 4 5 6 7 8 9 0 1
			// TCR書き込み後のmain count   |          4 ----------------> 3 ->
			uint64 ellapsed;
			uint deccount;

			// タイマー開始時からの経過時間
			ellapsed = gScheduler->GetVirtTime() - mfp.timer_start[ch];

			// タイマー開始時から今までのメインカウンタ減算回数
			deccount = ellapsed / scale;

			// 再設定
			mfp.initial_tdr[ch] -= deccount;
			mfp.timer_start[ch] += gScheduler->GetVirtTime();
		}

		// 割り込みを上げる必要があれば、カウントダウン終了時にイベントを
		// 起こす。
		// 割り込みを上げる必要がない場合は、TxDR を読み出す時に正しい値が
		// 算出できるだけでよくて、それは timer_start 等により行えているので
		// 他には何もしなくてよい。Timer-B 対策。
		if (IsTimerIntrEnable(ch)) {
			event[ch].time = mfp.initial_tdr[ch] * scale;
			event[ch].Start();
		}

		putlog(1, "Timer-%c ディレイモード開始(%d.%d x %d usec)",
			ch + 'A',
			scale / 1000,
			(scale / 100) % 10,
			mfp.initial_tdr[ch]);
	}
}

// TxDR の読み出し
// タイマーデータレジスタの現在値は保持しておらず、
// タイマー開始時刻と減算量が分かっているので読み出しのたびに求める。
// この操作に副作用はないので Peek/Monitor でも使用可能。
uint8
MFPDevice::GetTDR(int ch) const
{
	uint64 ellapsed;
	uint64 count;
	uint32 data;

	// イベントカウントモードなら initial_tdr が読める?
	if (mfp.tcr[ch] == 8) {
		return mfp.initial_tdr[ch];
	}

	// タイマー停止中なら TxDR と同じでよい?
	if (mfp.tcr[ch] == 0) {
		return mfp.tdr[ch];
	}

	// タイマー開始時からの経過時間
	ellapsed = gScheduler->GetVirtTime() - mfp.timer_start[ch];

	// タイマー開始時から現在までのメインカウンタ減算回数
	count = ellapsed / prescale_ns[mfp.tcr[ch]];

	// カウント終了時に割り込みが上がる設定になっていれば
	// count は必ず 0 から initail_tdr-1 までの範囲だが、
	// カウント終了時に割り込みが上がる設定になっていない場合は
	// 経過時間とそれに伴い count はひたすら増える。
	// そのため initial_tdr で modulo する。
	count %= mfp.initial_tdr[ch];

	// count は減算回数だが、
	// TxDR レジスタの読み出し値は減算カウンタの現在値。
	// TxDR = 10 で開始して count = 0 のうちは TxDR 読み出しは 10。
	// TxDR = 10 で開始して count = 1 のうちは TxDR 読み出しは 9。
	//  :
	// TxDR = 10 で開始して count = 9 のうちは TxDR 読み出しは 1。
	data = mfp.initial_tdr[ch] - count;

	return data;
}

// TxDR への書き込み
void
MFPDevice::SetTDR(int ch, uint32 data)
{
	uint64 ellapsed;
	uint64 scale;
	uint64 period;
	uint64 timer_end;

	mfp.tdr[ch] = data;

	// イベントカウントモードなら即ロード
	if (mfp.tcr[ch] == 8) {
		mfp.initial_tdr[ch] = mfp.tdr[ch];
	}

	// カウント終了時に割り込みが上がる設定になっていれば、
	// そのイベントコールバックで tdr を initial_tdr にロードするが、
	// 割り込みが上がらない設定の場合、そのままでは tdr をリロードする
	// 機会がない。そのためこのケースでは書き込み時点の次のカウント終了
	// 時にイベントを起こす必要がある。
	//
	// TCR=1(プリスケール10)、TDR=5 で 38カウント目に TDR がセットされた
	// 場合、ここから 12カウント目がリロードタイミング。
	//
	// pre | 0 ... 0 ... 0 ... 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0
	// main| 5 --> 4 --> 3 --> 2 ------------->*
	//                                         |<--------------------->|
	//                                          イベントを起こすまでの時間
	if (IsTimerIntrEnable(ch) == false) {
		// タイマー開始時から現在までの経過時間
		ellapsed = gScheduler->GetVirtTime() - mfp.timer_start[ch];
		// プリスケールカウント時間 (上でいう1カウントの時間)
		scale = prescale_ns[mfp.tcr[ch]];
		if (__predict_true(scale > 0)) {
			// 1タイマー期間(上でいう50カウントに相当する時間)
			period = scale * mfp.initial_tdr[ch];
			// タイマー開始時から次のタイムアップ時刻までの時間
			timer_end = ((ellapsed + period - 1) / period) * period;
			// タイムアップ時刻
			timer_end = mfp.timer_start[ch] + timer_end;
		} else {
			// タイムアップ時刻
			timer_end = mfp.timer_start[ch];
		}
		// 現在から次のタイムアップ時刻まで
		event[ch].time = timer_end - gScheduler->GetVirtTime();
		event[ch].Start();

		putlog(3, "Timer-%c TDR 更新予約 %d.%09d", ch + 'A',
			(int)(event[ch].time / (1000 * 1000 * 1000)),
			(int)(event[ch].time % (1000 * 1000 * 1000)));
	}
}

// イベントコールバック
void
MFPDevice::TimerCallback(int arg)
{
	int ch = arg;
	int prevtdr;
	uint64 ellapsed;
	uint64 scale;
	uint64 period;
	uint64 timer_end;

	// タイムアップした時刻を計算。現在時刻ではない。
	// 現在時刻はタイムアップした直後の命令境界の時刻なので。

	// タイマー開始時から現在までの経過時間
	ellapsed = gScheduler->GetVirtTime() - mfp.timer_start[ch];
	// 1プリスケール期間
	scale = prescale_ns[mfp.tcr[ch]];
	// 1タイマー期間
	period = scale * mfp.initial_tdr[ch];
	// タイマー開始時から直近のタイムアップ時刻まで
	timer_end = (ellapsed / period) * period;

	// TxDR がゼロになったので TxDR をリロード
	prevtdr = mfp.initial_tdr[ch];
	mfp.initial_tdr[ch] = mfp.tdr[ch] ?: 256;
	// 今タイムアップ時刻を次のタイマー起点に更新
	mfp.timer_start[ch] += timer_end;

	// 割り込み有効なら
	if (IsTimerIntrEnable(ch)) {
		// 1タイマー期間を新しい TDR で再計算
		period = prescale_ns[mfp.tcr[ch]] * mfp.initial_tdr[ch];
		// イベント再設定
		event[ch].time = period;
		event[ch].Start();

		// 割り込みを上げる
		gMPU680x0->Interrupt(this, 6, mfp.vr + timer_vector[ch]);
	}

	if (mfp.initial_tdr[ch] != prevtdr) {
		putlog(1, "Timer-%c ディレイモード更新(%u.%u x %u usec)",
			ch + 'A',
			(uint)(scale / 1000),
			(uint)((scale / 100) % 10),
			(uint)mfp.initial_tdr[ch]);
	}
}

// VDisp 入力
void
MFPDevice::SetVDisp(bool vdisp)
{
	// TAI への入力はイベントカウントモードの時のみ
	if (mfp.tcr[0] == 8) {
		// AER が示す方向と同じエッジ方向の時
		bool aer = (mfp.aer & (1 << MFP::GPIP_VDISP));
		if (aer == vdisp) {
			// カウンタを減算
			mfp.initial_tdr[0]--;
			if (mfp.initial_tdr[0] == 0) {
				// 0 になったらリロード
				mfp.initial_tdr[0] = mfp.tdr[0];

				// XXX ここで割り込みを上げる
			}
		}
	}

	// GPIP への入力
	SetGPIP(MFP::GPIP_VDISP, vdisp);
}

// GPIP への入力。
// num はビット番号(0..7)。
void
MFPDevice::SetGPIP(int num, bool val)
{
	bool aer = (mfp.aer & (1 << num));
	bool old = (mfp.gpip & (1 << num));

	// 立ち上がりか立ち下がりで割り込みを上げる (AER による)
	// AER OLD VAL   Interrupt
	//   0   0   0   0
	//   0   0   1   0
	//   0   1   0   1
	//   0   1   1   0
	//   1   0   0   0
	//   1   0   1   1
	//   1   1   0   0
	//   1   1   1   0
	if ((old != val) && (aer == val)) {
		// XXX ここで割り込みを上げる
	}

	if (val) {
		mfp.gpip |= (1 << num);
	} else {
		mfp.gpip &= ~(1 << num);
	}
}

// UCR への書き込み
void
MFPDevice::SetUCR(uint32 data)
{
	mfp.ucr = data;

	if ((mfp.ucr & MFP::UCR_CLK) == 0) {
		putlog(0, "UCR CLK=0 未実装");
	}
	if (((mfp.ucr & MFP::UCR_WL) >> 5) != 0) {
		putlog(0, "UCR WL=%d 未実装", (mfp.ucr >> 5) & 3);
	}
	if (((mfp.ucr & MFP::UCR_ST) >> 3) != 1) {
		putlog(0, "UCR ST=%d 未実装", (mfp.ucr >> 3) & 3);
	}
}

// RSR への書き込み
void
MFPDevice::SetRSR(uint32 data)
{
	uint32 old = mfp.rsr;

	mfp.rsr = data;

	if ((old & MFP::RSR_RE) == 0 && (mfp.rsr & MFP::RSR_RE)) {
		// RE 0 -> 1
		rr = true;
	}
	if ((old & MFP::RSR_RE) && (mfp.rsr & MFP::RSR_RE) == 0) {
		// RE 1 -> 0
		rr = false;
		mfp.rsr &= 0x0f;	// 上位4ビットのステータスを %0 に
	}
}

// UDR への書き込み
void
MFPDevice::SetUDR(uint32 data)
{
	if ((mfp.tsr & MFP::TSR_TE)) {
		gKeyboard->Command(data);
	}
}

// システムポート#4 の KEY CTRL の状態を受け取る。
void
MFPDevice::SetKeyCtrl(bool val)
{
	key_ctrl = val;
	putlog(1, "KEYCTRL = %d", key_ctrl);
}

// RR (Receiver Ready) 線の状態を返す。true なら Ready。
// これはシステムポート#4 の KEY CTRL の状態も加味したもの。
bool
MFPDevice::IsRR() const
{
	return rr && key_ctrl;
}

// キーボードからデータを受信
void
MFPDevice::Rx(uint32 data)
{
	// Ready を下げる
	rr = false;

	key_event.code = data;
	key_event.Start();
}

// キーボードからのデータが受信し終わった頃に呼ばれるイベント。
void
MFPDevice::KeyCallback(int code)
{
	// Ready を元に戻す
	rr = true;

	mfp.udr = code;
	mfp.rsr |= MFP::RSR_BF;

	// XXX この辺要再考察
	mfp.isr.a |= 0x10;
	// 割り込み有効でマスクされてない時だけ
	if ((mfp.ier.a & 0x10) && (mfp.imr.a & 0x10)) {
		gMPU680x0->Interrupt(this, 6, mfp.vr + MFP::INTR_RXFULL);
	}
}
