//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// MPU (m680x0)

#include "mpu680x0.h"
#include "config.h"
#include "m68030excep.h"
#include "mainapp.h"
#include "vm.h"
#include <math.h>

std::unique_ptr<MPUATC> gMPUATC;

// コンストラクタ
//
// reset_vector にはリセットベクタの番地を指定する。
// 本来の MPU の動作は、リセット例外時にメモリの $0, $4 番地 (もっと言えば
// FC がプログラム空間の) から SP, PC を読み込むのだが、そのためにメモリ空間
// を切り替える処理はリセット時にしか必要ないわりに結構手間である。そしてこの
// m68k 恒例のブートストラップ動作はソフトウェア的には一切観測する手段がない。
// であれば、リセット例外時にこのアドレスから SP, PC の値を直接読み込んだと
// しても動作にはまったく違いは出ない。そしてバス・メモリ空間の実装も簡略化
// できる。すごい楽。
MPU680x0Device::MPU680x0Device(uint32 reset_vector)
{
	monitor_size = nnSize(79, 16);

	cpu = m68030_init(reset_vector);

	// モニタ用の別オブジェクト
	gMPUATC.reset(new MPUATC(cpu));
}

// デストラクタ
MPU680x0Device::~MPU680x0Device()
{
}

bool
MPU680x0Device::Init()
{
	// FPU (6888x)
	if (gMainApp.GetVMType() == VMTYPE_LUNA1) {
		// LUNA は 68881 固定 (設定は見ない)
		cpu->has_fpu = 1;
	} else {
		// X68k は設定による
		const ConfigItem& item = gConfig->Find("mpu-has-fpu");
		cpu->has_fpu = item.AsInt();
		if (cpu->has_fpu < 0 || cpu->has_fpu > 2) {
			item.Err();
			return false;
		}
	}

	return true;
}

// リセット
// (MPU が動くのはこの後 Scheduler::Run() を発行した後)
void
MPU680x0Device::ResetHard()
{
	m68030_exception_reset(cpu);
}

// MPU を request で指定されたクロックサイクルだけ実行する。
// 今回の実行サイクル数が request を越えた命令境界まで実行するので
// 実際には少しはみ出る。
// また実行中にデバイスアクセスなどによりスケジューラにイベントが
// 追加されるとその次の命令境界で処理を打ち切って戻ってくる。
// 戻り値は CPU の outer フラグ。
uint32
MPU680x0Device::Run(uint64 request)
{
	return m68030_run(cpu, request);
}

// 割り込み発生を MPU に通知 (オートベクタ)
void
MPU680x0Device::Interrupt(Object *src, int level)
{
	int vector = M68K_EXCEP_LEVEL_BASE + level;
	m68030_interrupt(cpu, level, vector);
}

// 割り込み発生を MPU に通知 (ベクタ)
void
MPU680x0Device::Interrupt(Object *src, int level, int vector)
{
	m68030_interrupt(cpu, level, vector);
}

// A-Line 命令エミュレーションのコールバックを指定。
void
MPU680x0Device::SetALineCallback(bool (*callback)(m68kcpu *, void*), void *arg)
{
	cpu->aline_callback = callback;
	cpu->aline_arg = arg;
}

// F-Line 命令エミュレーションのコールバックを指定。
void
MPU680x0Device::SetFLineCallback(bool (*callback)(m68kcpu *, void*), void *arg)
{
	cpu->fline_callback = callback;
	cpu->fline_arg = arg;
}

// ダブルバスフォールトのコールバックを指定。
void
MPU680x0Device::SetHaltCallback(void (*callback)(void *), void *arg)
{
	cpu->halt_callback = callback;
	cpu->halt_arg = arg;
}

// モニター更新 (レジスタウィンドウ)
void
MPU680x0Device::MonitorUpdate(TextScreen& monitor)
{
	m68kreg reg;
	uint32 ppc;
	int x;
	int y;

	monitor.Clear();

	// ローカルにコピー
	// ただし PPC だけ reg 構造体ではなく cpu クラス側にある
	reg = cpu->reg;
	ppc = cpu->ppc;

	// データレジスタ、アドレスレジスタ
	for (int i = 0; i < 4; i++) {
		monitor.Print(0, i, "D%d:%08x D%d:%08x  A%d:%08x A%d:%08x",
			i + 0, reg.da[i],
			i + 4, reg.da[i + 4],
			i + 0, reg.da[i + 8],
			i + 4, reg.da[i + 12]);
	}

	// 5列目
	x = 50;
	// SR
	monitor.Print(x, 0, "SR:%04x(%c%c%c%c%c)",
		(reg.sr_h() | reg.ccr.Get()),
		(reg.ccr.IsX()) ? 'X' : '-',
		(reg.ccr.IsN()) ? 'N' : '-',
		(reg.ccr.IsZ()) ? 'Z' : '-',
		(reg.ccr.IsV()) ? 'V' : '-',
		(reg.ccr.IsC()) ? 'C' : '-');

	// VBR
	monitor.Print(x, 1, "VBR:%08x", reg.vbr);

	// *SP
	uint ms = (reg.s ? 2 : 0) | (reg.m ? 1 : 0); // T1 T0 S M
	if (ms == 3) {
		// Supervisor (Master) mode (A7=MSP, USP/ISP)
		monitor.Print(x, 2, TA::Disable, "USP:%08x", reg.usp);
		monitor.Print(x, 3, TA::Disable, "ISP:%08x", reg.isp);
	} else if (ms == 2) {
		// Supervisor (Interrupt) mode (A7=ISP, USP/MSP)
		monitor.Print(x, 2, TA::Disable, "USP:%08x", reg.usp);
		monitor.Print(x, 3, TA::Disable, "MSP:%08x", reg.msp);
	} else {
		// User mode (A7=USP, ISP/MSP)
		monitor.Print(x, 2, TA::Disable, "ISP:%08x", reg.isp);
		monitor.Print(x, 3, TA::Disable, "MSP:%08x", reg.msp);
	}

	// 6列目
	// PC:  01234567
	// SFC:1   DFC:1
	// CACR:01234567
	// CAAR:01234567
	x = 66;
	monitor.Print(x, 0, "PC:  %08x", ppc);
	monitor.Print(x, 1, "SFC:%d  DFC:%d", reg.sfc, reg.dfc);
	monitor.Print(x, 2, "CACR:%08x", reg.cacr);
	monitor.Print(x, 3, "CAAR:%08x", reg.caar);

	// XXX どこに置くのがいいか
	if (reg.state == 0) {
		monitor.Puts(25, 4, "State: Running");
	} else if (reg.state == CPU_REQ_STOP) {
		monitor.Puts(25, 4, "State: STOP instruction");
	} else if (reg.state == CPU_REQ_HALT) {
		monitor.Puts(25, 4, TA::On, "State: Double Bus Fault");
	}

	// MMU
	x = 0;
	y = 4;
	monitor.Puts(x, y++, "<MMU>");
	monitor.Print(x, y,     "SRP:%08x_%08x", reg.srp.h, reg.srp.l);
	monitor.Print(x, y + 1, "CRP:%08x_%08x", reg.crp.h, reg.crp.l);
	x += 23;
	for (int i = 0; i < 2; i++) {
		uint32 tt = reg.tt[i];
		monitor.Print(x, y + i, "TT%d:%08x(%c%c%c)",
			i, tt,
			(tt & m68030TT::E)   ? 'E' : '-',
			(tt & m68030TT::CI)  ? 'C' : '-',
			(tt & m68030TT::RWM) ? '-' : ((tt & m68030TT::RW)  ? 'R' : 'W'));
	}
	x += 19;
	monitor.Print(x, y, "TC:%08x(%c%c%c)",
		reg.tc,
		(reg.tc & m68030TC::TC_E)   ? 'E' : '-',
		(reg.tc & m68030TC::TC_SRE) ? 'S' : '-',
		(reg.tc & m68030TC::TC_FCL) ? 'F' : '-');
	monitor.Print(x, y + 1, "MMUSR: %04x(%c%c%c%c%c%c%c N=%d)",
		reg.mmusr,
		(reg.mmusr & m68030MMUSR::B) ? 'B' : '-',
		(reg.mmusr & m68030MMUSR::L) ? 'L' : '-',
		(reg.mmusr & m68030MMUSR::S) ? 'S' : '-',
		(reg.mmusr & m68030MMUSR::W) ? 'W' : '-',
		(reg.mmusr & m68030MMUSR::I) ? 'I' : '-',
		(reg.mmusr & m68030MMUSR::M) ? 'M' : '-',
		(reg.mmusr & m68030MMUSR::T) ? 'T' : '-',
		(reg.mmusr & m68030MMUSR::N));

	// FPU
	x = 0;
	y = 7;
	monitor.Puts(x, y++, "<FPU>");
	for (int i = 0; i < 8; i++) {
		monitor.Print(x, y + i, "FP%d:%04x_%08x_%08x (%-20s)",
			i,
			reg.fpframe.fpf_regs[i * 3 + 0] >> 16,
			reg.fpframe.fpf_regs[i * 3 + 1],
			reg.fpframe.fpf_regs[i * 3 + 2],
			"not yet");
	}

	x = 51;

	// FPCR
	static const char * const rpstr[] = {
		".EXT",
		".SGL",
		".DBL",
		".???",
	};
	static const char * const rmstr[] = {
		"Near",
		"Zero",
		"Minus",
		"Plus",
	};
	uint32 fpcr = reg.fpframe.fpf_fpcr;
	monitor.Print(x, y++, "FPCR:%04x", fpcr);
	monitor.Print(x + 2, y++, "%s %s %s %s %s %s %s %s",
		(fpcr & 0x8000) ? "BS" : "--",
		(fpcr & 0x4000) ? "SN" : "--",
		(fpcr & 0x2000) ? "OP" : "--",
		(fpcr & 0x1000) ? "OV" : "--",
		(fpcr & 0x0800) ? "UF" : "--",
		(fpcr & 0x0400) ? "DZ" : "--",
		(fpcr & 0x0200) ? "I2" : "--",
		(fpcr & 0x0100) ? "I1" : "--");
	monitor.Print(x + 2, y++, "RP=%s RM=%s",
		rpstr[(fpcr >> 6) & 3],
		rmstr[(fpcr >> 4) & 3]);

	// FPSR
	uint32 fpsr = reg.fpframe.fpf_fpsr;
	uint32 cc = fpsr >> 24;
	monitor.Print(x, y++, "FPSR:%08x", fpsr);
	monitor.Print(x + 2, y++, "%c %c %s %s Q=$%02x",
		(cc & 0x08) ? 'N' : '-',
		(cc & 0x04) ? 'Z' : '-',
		(cc & 0x02) ? "Inf" : "---",
		(cc & 0x01) ? "NAN" : "---",
		(fpsr >> 16) & 0xff);
	monitor.Print(x + 2, y++, "%s %s %s %s %s %s %s %s",
		(fpsr & 0x8000) ? "BS" : "--",
		(fpsr & 0x4000) ? "SN" : "--",
		(fpsr & 0x2000) ? "OP" : "--",
		(fpsr & 0x1000) ? "OV" : "--",
		(fpsr & 0x0800) ? "UF" : "--",
		(fpsr & 0x0400) ? "DZ" : "--",
		(fpsr & 0x0200) ? "I2" : "--",
		(fpsr & 0x0100) ? "I1" : "--");
	monitor.Print(x + 2, y++, "%s %s %s %s %s",
		(fpsr & 0x80) ? "IOP"  : "---",
		(fpsr & 0x40) ? "OVFL" : "----",
		(fpsr & 0x20) ? "UNFL" : "----",
		(fpsr & 0x10) ? "DZ"   : "--",
		(fpsr & 0x08) ? "INEX" : "----");

	// FPIAR
	monitor.Print(x, y++, "FPIAR:%08x", reg.fpframe.fpf_fpiar);
}

// CPU 側からのコールバック。RESET 命令ハンドラ
void
m68030_reset_inst(m68kcpu *cpu)
{
	gVM->ResetSoft();
}

//
// ATC モニタ
//
MPUATC::MPUATC(m68kcpu *cpu_)
{
	cpu = cpu_;
#if !defined(ATC_SINGLE)
	monitor_size = nnSize(131, m68030ATCTable::LINES + 3);
#else
	monitor_size = nnSize(31, 24);
#endif
}

void
MPUATC::MonitorUpdate(TextScreen& monitor)
{
	m68030ATCTable *table;
#if !defined(ATC_SINGLE)
	struct {
		int fc;
		const char *name;
	} fclist[] = {
		{ 1,	"User/Data" },
		{ 2,	"User/Program" },
		{ 5,	"Super/Data" },
		{ 6,	"Super/Program" },
	};

	monitor.Clear();

	// 行数は左端にだけ
	for (int i = 0; i < m68030ATCTable::LINES; i++) {
		monitor.Print(0, i + 2, "%02d:", i);
	}

	for (int t = 0; t < countof(fclist); t++) {
		int x = t * 32 + 4;
		int fc = fclist[t].fc;
		const char *name = fclist[t].name;
		double r;
		int i;

		table = &cpu->atc.tables[fc];
		monitor.Print(x, 0, "FC=%d %s", fc, name);
		monitor.Puts(x, 1, "LAddr    PAddr   Flag Hit%  Age");
		// 先に統計用の母数を計算
		uint64 total = 0;
		for (i = 0; i < countof(table->hit); i++) {
			total += table->hit[i];
		}
		total += table->hit_head;

		// エントリ表示
		for (i = 0; i < countof(table->line); i++) {
			m68030ATCLine& a = table->line[i];
			TA attr;
			if (a.IsInvalid()) {
				attr = TA::Disable;
			} else if (table->head == &a) {
				attr = TA::Em;
			} else {
				attr = TA::Normal;
			}

			monitor.Print(x, i + 2, attr, "%08x %08x %c%c%c",
				a.GetLAddr(),
				a.GetPAddr(),
				a.IsBusError() ? 'B' : '-',
				a.IsWProtect() ? 'P' : '-',
				a.IsModified() ? 'M' : '-');
		}
		// ヘッドのヒット率
		monitor.Puts(x + 0, i + 2, "head");
		r = (double)table->hit_head * 100 / total;
		if (!isnan(r)) {
			monitor.Print(x + 5, i + 2, "%4.1f%%", r);
		}
		// 配列のヒット率とミス率
		monitor.Puts(x + 17, i + 2, "miss");
		for (i = 0; i < countof(table->hit); i++) {
			r = (double)(table->hit[i] * 100) / total;
			if (!isnan(r)) {
				monitor.Print(x + 22, i + 2, "%4.1f%%", r);
			}
		}
		// カウンタ
		// 分かりやすさのため最大値からの差(距離)を表示
		for (i = 0; i < countof(table->line); i++) {
			m68030ATCLine& a = table->line[i];
			if (!a.IsInvalid()) {
				monitor.Print(x + 27, i + 2, "%4d", table->head->age - a.age);
			}
		}
	}
#else
	m68030ATCLine *a;
	int i;

	table = &cpu->atc.tables[0];
	monitor.Clear();

	monitor.Puts(0, 0, "No. FC    LAddr    PAddr   Flag");
	for (a = table->head, i = 0; a; a = a->next, i++) {
		monitor.Print(0, i + 1, "%02d:", i);
		if (!a->IsInvalid()) {
			char dp;
			switch (a->fc & 3) {
			 case 1:	dp = 'D';	break;
			 case 2:	dp = 'P';	break;
			 default:	dp = '?';	break;
			}
			monitor.Print(4, i + 1, "%d(%c%c):%08x %08x %c%c%c",
				a->fc,
				(a->fc & 4) ? 'S' : 'U',
				dp,
				a->GetLAddr(), a->GetPAddr(),
				a->IsBusError() ? 'B' : '-',
				a->IsWProtect() ? 'W' : '-',
				a->IsModified() ? 'M' : '-');
		}
	}
#endif
}
