//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "device.h"

class LunaPROMEmuDevice : public IODevice
{
	using inherited = IODevice;

	// このデバイス先頭アドレス
	static const uint32 baseaddr		= 0x41000000;

	// 謎の I/O 空間
	static const uint32 ROMIO_BASE		= 0x41010000;
	static const uint32 ROMIO_INIT		= (ROMIO_BASE + 0);
	static const uint32 ROMIO_KEYIN		= (ROMIO_BASE + 4);
	static const uint32 ROMIO_CLOSE		= (ROMIO_BASE + 8);

 public:
	LunaPROMEmuDevice();
	~LunaPROMEmuDevice() override;

	bool Init() override;
	void ResetHard() override;

	uint64 Read8(uint32 addr) override;
	uint64 Read16(uint32 addr) override;
	uint64 Read32(uint32 addr) override;
	uint64 Write8(uint32 addr, uint32 data) override;
	uint64 Write16(uint32 addr, uint32 data) override;
	uint64 Write32(uint32 addr, uint32 data) override;
	uint64 Peek8(uint32 addr) override;

 private:
	uint64 Decoder(uint32 addr) const;

	uint32 ROM_Init();
	uint32 ROM_Close();
	uint32 ROM_Keyin();
	void ConsoleUpdate();
	uint32 Command();
	void CommandG(const std::vector<std::string>& arg);
	void CommandGH();
	uint32 LoadHostFile();
	void CommandK(const std::vector<std::string>& arg);
	bool CommandKInput(const std::string& val);
	uint32 CommandX();

	std::unique_ptr<uint8[]> romdata;

	// NVRAM
	uint8 CalcNVRAMCsum();
	void WriteNVRAMCsum();
	bool VerifyNVRAM();
	bool InitNVRAM();

	// NVRAM の設定に従って起動
	uint32 LoadByNVRAM(const std::string& fname);

	// 内蔵 ROM 使用後はアクセスできないようにするため
	bool is_closed = false;

	// 'g' コマンドで読み込んだプログラムのエントリポイント。
	// なければ 0。'x' コマンドで使う。
	uint32 entrypoint;
	// エラーメッセージの受け渡し用
	// 基本的にエラーが起きた時だけ更新するので、ここにエラーメッセージを
	// 返す可能性がある関数をコールする前に呼び出し側が clear() しておき、
	// 関数から戻ったら empty() かどうか確認する、errno みたいな使い方。
	std::string errmsg;

	// ROM モニタ用コンソール画面
	TextConsole text;

	// 入力バッファ
	std::string inputbuf {};
	int inputpos {};			// カーソル位置
	bool is_shift;				// SHIFT キー押し下げ中
	int mousecnt {};			// マウス入力をスキップするためのカウンタ

	// プロンプトと入力モード
	std::string prompt {};		// 現在のプロンプト文字列
	// プロンプトの状態。
	// 0 が通常プロンプト。
	// 1 以降は k コマンドのプロンプトを示す
	int prompt_idx {};
	const std::string default_prompt = ">";	// デフォルトプロンプト

	// NVRAM
	// 本来他デバイスはすべてグローバル参照できるのだが、LUNA の NVRAM は
	// RTC と兼用になっていて、RTCDevice *gRTC として見えるので
	// MK48T02Device にするには dynamic_cast する必要がある。
	MK48T02Device *nvram;

	// LUNA キーコード -> ASCII 変換テーブル
	static const uint8 lunakey2asciitable[128];
	static const uint8 lunakey2shifttable[128];
};
