//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "rtc.h"
#include <sys/time.h>

// XXX rtc の読み書きはロックしたほうがいいような気がする

std::unique_ptr<RTCDevice> gRTC;

RTCDevice::RTCDevice()
{
	// 秒への繰り上がりを許可。
	rtc.enable_sec = true;
}

RTCDevice::~RTCDevice()
{
}

// ホストの現在時刻を rtc に取得。
// (派生クラスの InitRTC() から呼ばれる)
// localtime を内部時刻にするときは use_localtime == true
// UTC を内部時刻にするときは use_localtime == false にすること。
void
RTCDevice::LoadHostRTC(bool use_localtime)
{
	struct timeval tv;
	struct tm tm;

	// 起動した時の現在時刻で設定する
	gettimeofday(&tv, NULL);
	if (use_localtime) {
		localtime_r(&tv.tv_sec, &tm);
	} else {
		gmtime_r(&tv.tv_sec, &tm);
	}
	// 秒未満の初期値もそれなりの精度にしておく (1000*1000 / 32 = 31250)
	rtc.cnt  = tv.tv_usec / 31250;
	rtc.sec  = tm.tm_sec;
	rtc.min  = tm.tm_min;
	rtc.hour = tm.tm_hour;
	// wday も tm_wday も日曜が 0
	rtc.wday = tm.tm_wday;
	rtc.mday = tm.tm_mday;
	rtc.mon  = tm.tm_mon + 1;
	rtc.year = tm.tm_year + 1900;
}

// 32Hz カウンタの処理。
// 戻り値は更新が起きた位。
// 派生クラスで用意した ClockIn() から呼び出すこと。
uint32
RTCDevice::CountUp()
{
	uint32 updates;

	updates = 0;

	// 内部カウンタアップ
	rtc.cnt++;
	rtc.cnt &= 63;

	// RP5C15 の Timer Enable は内部カウンタは動作していて、
	// 秒へのカウントアップをするかどうかのフラグ。
	if (rtc.enable_sec == false) {
		return updates;
	}

	// 秒の繰り上がり処理。
	// 32カウントで1秒だが、RP5C15 はタイマー停止中でも 1秒への繰り上がり
	// 1回は内部で覚えているので、ここだけ 32 以上だったら 32 を引く運用。
	if (rtc.cnt >= 32) {
		rtc.cnt -= 32;
		updates |= CarrySec(updates);
	}
	return updates;
}

// 秒への繰り上がり処理。
// 戻り値は更新が起きた位。
uint32
RTCDevice::CarrySec(uint32 updates)
{
	rtc.sec++;
	updates |= RTCDevice::UPDATE_SEC;

	if (rtc.sec >= 60) {
		rtc.sec = 0;
		updates |= CarryMin(updates);
	}
	return updates;
}

// 分への繰り上がり処理。
// 戻り値は更新が起きた位。
uint32
RTCDevice::CarryMin(uint32 updates)
{
	rtc.min++;
	updates |= RTCDevice::UPDATE_MIN;

	if (rtc.min >= 60) {
		rtc.min = 0;
		updates |= CarryHour(updates);
	}
	return updates;
}

// 時への繰り上がり処理。時間は24時間計。
// 戻り値は更新が起きた位。
uint32
RTCDevice::CarryHour(uint32 updates)
{
	rtc.hour++;
	updates |= RTCDevice::UPDATE_HOUR;

	if (rtc.hour >= 24) {
		rtc.hour = 0;
		updates |= CarryDay(updates);
	}
	return updates;
}

// 日への繰り上がり処理。
// 戻り値は更新が起きた位。
uint32
RTCDevice::CarryDay(uint32 updates)
{
	// まず曜日
	rtc.wday = (rtc.wday + 1) % 7;

	// 日にち
	rtc.mday++;
	updates |= RTCDevice::UPDATE_DAY;

	// 日が繰り上がった結果、月が繰り上がるかどうか
	static const int mdays[] = {
		31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
	};
	if (rtc.mday > (mdays[rtc.mon] + (IsLeapMon() ? 1 : 0))) {
		rtc.mday = 1;
		updates |= CarryMon(updates);
	}
	return updates;
}

// 月への繰り上がり処理。
// 戻り値は更新が起きた位。
uint32
RTCDevice::CarryMon(uint32 updates)
{
	rtc.mon++;
	updates |= RTCDevice::UPDATE_MON;

	if (rtc.mon >= 12) {
		rtc.mon = 1;
		updates |= CarryYear(updates);
	}
	return updates;
}

// 年への繰り上がり処理。
// 戻り値は更新が起きた位。
uint32
RTCDevice::CarryYear(uint32 updates)
{
	rtc.year++;
	updates |= RTCDevice::UPDATE_YEAR;

	// 特にすることはない

	return updates;
}

// 閏年の2月なら true を返す。
bool
RTCDevice::IsLeapMon() const
{
	// 1900年から2100年までの間は 4 で割るだけでよい。
	if ((rtc.year % 4 == 0) && (rtc.mon == 2)) {
		return true;
	}
	return false;
}

// 曜日文字列
/*static*/ const char * const
RTCDevice::wdays[7] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat",
};
