//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "scc.h"
#include "mpu680x0.h"

//          IODevice
//              |
//              v
//        upd7201Device  (uPD7201 と Z8530 の共通部分)
//              |
//    +---------+----------+
//    v                    v
// SIODevice           SCCDevice
// (LUNA, uPD7201)     (X68030, Z8530)

std::unique_ptr<SCCDevice> gSCC;

SCCDevice::SCCDevice()
{
	logname = "scc";
	devname = "SCC";
	devaddr = baseaddr;
	devlen  = 0x2000;

	cr.chan[0].desc = "Serial Port";
	cr.chan[1].desc = "Mouse";
}

SCCDevice::~SCCDevice()
{
}

// Z8530 は RD と WR を同時に Low にするとリセットがかかる。
// 回路図からは PEDEC から繋がってるという以上のことは分からないけど
// 普通に考えればハードリセットでリセットされるよな。
void
SCCDevice::ResetHard()
{
	Reset();
}

uint64
SCCDevice::Read(uint32 addr)
{
	switch (addr) {
	 case 0:	// $E98001
		return ReadCtrl(&cr.chan[1]);

	 case 1:	// $E98003
		// マウス非接続は 0x00 ?
		putlog(0, "ChB データレジスタ読み込み (未実装)");
		return 0x00;

	 case 2:	// $E98005
		return ReadCtrl(&cr.chan[0]);

	 case 3:	// $E98007
		putlog(0, "ChA データレジスタ読み込み (未実装)");
		return ReadData(&cr.chan[0]);
	}
	__unreachable();
}

uint64
SCCDevice::Write(uint32 addr, uint32 data)
{
	switch (addr) {
	 case 0:	// $E98001
		WriteCtrl(&cr.chan[1], data);
		return 0;

	 case 1:	// $E98003
		WriteData(&cr.chan[1], data);
		return 0;

	 case 2:	// $E98005
		WriteCtrl(&cr.chan[0], data);
		return 0;

	 case 3:	// $E98007
		WriteData(&cr.chan[0], data);
		return 0;
	}
	__unreachable();
}

uint64
SCCDevice::Peek(uint32 addr)
{
	switch (addr) {
	 case 0:	// $E98001
		return PeekCtrl(&cr.chan[1]);

	 case 1:	// $E98003
		// マウス非接続は 0x00 ?
		return 0x00;

	 case 2:	// $E98005
		return PeekCtrl(&cr.chan[0]);

	 case 3:	// $E98007
		return PeekData(&cr.chan[0]);
	}
	__unreachable();
}

// Z8530 制御ポート読み込み
uint8
SCCDevice::ReadCtrl(struct SIO::channel *chan)
{
	return inherited::ReadCtrl(chan);
}

// Z8530 制御ポート書き込み
void
SCCDevice::WriteCtrl(struct SIO::channel *chan, uint32 data)
{
	switch (chan->ptr) {
	 case 0:	// WR0
		chan->cr0 = data;

		switch ((chan->cr0 & SIO::WR0_CMD) >> 3) {
		 case 0:
		 case 1:
			// コマンド 0、1 がレジスタ切り替え
			chan->ptr = chan->cr0 & SIO::WR0_PTR;
			putlog(3, "WR0%s <- $%02x (ptr=%d)", chan->name, data, chan->ptr);
			break;

		 case 2:	// Reset Ext/Status Interrupts
			putlog(0, "WR0 Reset Ext/Status Interrupts 未実装");
			break;
		 case 3:	// Send Abort
			putlog(0, "WR0 Send Abort 未実装");
			break;
		 case 4:	// Enable Int on Next Rx Character
			putlog(0, "WR0 Enable Int on Next Rx Character 未実装");
			break;
		 case 5:	// Reset Tx Int Pending
			putlog(0, "WR0 Reset Tx Int Pending 未実装");
			break;
		 case 6:	// Error Reset
			putlog(0, "WR0 Error Reset 未実装");
			break;
		 case 7:	// Reset Highest IUS
			putlog(0, "WR0 Reset Highest IUS");
			break;
		}
		break;

	 case 2:	// WR2
		// Z8530 の WR2 は A/B 共通でベクタ設定
		putlog(2, "WR2 <- $%02x ベクタ設定", data);
		cr.vector = data;
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 1:	// WR1
	 case 3:
	 case 4:
	 case 5:
	 case 6:
	 case 7:
		inherited::WriteCtrl(chan, data);
		break;

	 case 8:	// WR8
		WriteData(chan, data);
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 9:	// WR9
		putlog(2, "WR%d%s <- $%02x", chan->ptr, chan->name, data);
		switch ((data & 0xc0) >> 6) {
		 case 0:	// No Reset
			break;
		 case 1:	// Channel Reset A
			ResetChannel(&cr.chan[0]);
			break;
		 case 2:	// Channel Reset B
			ResetChannel(&cr.chan[1]);
			break;
		 case 3:	// Force Hardware Reset
			Reset();
			break;
		 default:
			__unreachable();
		}
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 11:	// WR11
		putlog(2, "WR%d%s <- $%02x (未実装)", chan->ptr, chan->name, data);
		chan->wr11 = data;
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 12:	// WR12
		chan->tc = (chan->tc & 0xff00) | data;
		putlog(2, "WR%d%s TC(L)=%04x", chan->ptr, chan->name, chan->tc);
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 13:	// WR13
		chan->tc = (chan->tc & 0x00ff) | (data << 8);
		putlog(2, "WR%d%s TC(H)=%04x", chan->ptr, chan->name, chan->tc);
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 14:	// WR14
		putlog(2, "WR%d%s <- $%02x (未実装)", chan->ptr, chan->name, data);
		chan->wr14 = data;
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 case 15:	// WR15
		putlog(2, "WR%d%s <- $%02x (未実装)", chan->ptr, chan->name, data);
		chan->wr15 = data;
		// アクセス後は WR0 へ戻す
		chan->ptr = 0;
		break;

	 default:
		PANIC("未実装レジスタ書き込み WR%d%s <- $%02x",
			chan->ptr, chan->name, data);
		break;
	}
}

// 割り込みを上げる
void
SCCDevice::Interrupt()
{
	putlog(0, "割り込み(未実装)");
	gMPU680x0->Interrupt(this, 5, 0x50/*XXX*/);
}
