//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "scheduler.h"
#include "config.h"
#include "mainapp.h"
#include "mfp.h"
#include "mpu680x0.h"
#include "mpu88xx0.h"
#include "mythread.h"
#include "rtc.h"
#include "vm.h"

// VM スケジューラは、仮想時間とイベントを管理する。
//
// 仮想時間は MPU の起動時からのクロックサイクルの経過数で表すこととする。
// このエミュレータはシングルプロセッサシステムのみがターゲットであり、
// 実行中に動作クロックが変わることはないので、これで問題ない。
// VM 界からは gMPU->total_cycle() で取得する。

static void *scheduler_run(void *arg);

std::unique_ptr<Scheduler> gScheduler;

// コンストラクタ
Scheduler::Scheduler()
{
	logname = "scheduler";
	devname = "Scheduler";

	eventlist.clear();
}

// デストラクタ
Scheduler::~Scheduler()
{
	Terminate();
}

// 初期化
// ここで mpu_clock が(非ゼロに)初期化されるとこれ以降 putlog() が使える。
// そのため、Scheduler デバイスをデバイスリストの先頭のほうに置くことで、
// これ以降のオブジェクトの Init() は putlog() を使ってよいとできる。
// BusErrDevice だけは Scheduler より先に作るが BusErrDevice に Init() は
// ないので (= 問題は起きないので)、よいことにする。
bool
Scheduler::Init()
{
	const ConfigItem& item = gConfig->Find("mpu-clock");
	const std::string& val = item.AsString();
	if (val.empty()) {
		item.Err();
		return false;
	}
	// 文字列を double に変換
	char *end;
	errno = 0;
	double f = strtod(val.c_str(), &end);
	if (end == val.c_str() || end[0] != '\0' || errno == ERANGE) {
		item.Err();
		return false;
	}
	// 設定の "mpu-clock" は MHz 単位だが、変数 mpu_clock は kHz 単位。
	mpu_clock = (int)(f * 1000);
	// 1MHz 以下はエラー。実際 10MHz 以下でもいい気がするけど
	if (mpu_clock < 1000) {
		item.Err();
		return false;
	}

	// 起動時引数指定の高速モード
	SetFullSpeed(gMainApp.fast_mode);

	// スケジューラ(VM)スレッド起動
	pthread_create(&thread, NULL, scheduler_run, NULL);
	thread_created = true;

	return true;
}

// スケジューラを開始する
void
Scheduler::Run()
{
	std::unique_lock<std::mutex> lock(cvmtx);
	atomic_reqflag |= REQ_POWER_ON;
	cv.notify_one();
}

// スレッドエントリ関数
void *
scheduler_run(void *arg)
{
	PTHREAD_SETNAME("VM");

	gScheduler->ThreadRun();
	return NULL;
}

// スレッドエントリ関数(の実体)
void
Scheduler::ThreadRun()
{
	uint64 cycle;	// 今回実行するサイクル数
	uint32 req;

	// 最初の電源オンを待つ
	for (;;) {
		std::unique_lock<std::mutex> lock(cvmtx);
		cv.wait(lock, [&] { return (atomic_reqflag != 0); });
		req = atomic_reqflag.load();
		// 終了
		if ((req & REQ_EXIT)) {
			return;
		}
		// REQ_POWER_ON なら他のフラグを書き戻して抜ける。
		// REQ_{SYNC|FAST} は以降のメインループで処理するため。
		if ((req & REQ_POWER_ON)) {
			req &= ~REQ_POWER_ON;
			atomic_reqflag.store(req);
			break;
		}
	}
	// ここからは電源オン

	// RTC 時刻だけは最初から必要。
	rtimestart = GetRealTime();
	rtc_last_clock = rtimestart;

	cycle = INT64_MAX;
	mode = 0;
	for (;;) {
		req = atomic_reqflag.exchange(0);
		if (req) {
			if ((req & REQ_SYNC)) {
				// スケジューラを同期モードに
				mode |= SCHED_SYNC;
			}
			if ((req & REQ_FAST)) {
				// スケジューラを高速モードに
				mode &= ~SCHED_SYNC;
			}
			if ((req & REQ_RUN)) {
				// MPU が通常状態になった
				mode &= ~SCHED_STOP;
			}
			if ((req & REQ_STOP)) {
				// MPU が STOP 状態になった
				mode |= SCHED_STOP;
			}
			if ((req & REQ_MODEMASK)) {
				// mode がどれかにでも変わったら
				static const char *modestr[] = {
					"高速",
					"高速 & STOP状態",
					"通常",
					"通常 & STOP状態",
				};
				putlog(1, "モード変更 mode=%d (%s)",  mode, modestr[mode]);

				// 基準時刻をリセット
				rtimebase = ::GetRealTime();
				vtimebase = GetVirtTime();

				// 該当ビットを落としておく
				req &= ~REQ_MODEMASK;
			}

			if ((req & REQ_POWER_OFF)) {
				// 電源オフ要求の場合

				// 全デバイス電源オフ
				gVM->DevicePowerOff();

				req &= ~REQ_POWER_OFF;
			}

			if ((req & REQ_EXIT)) {
				// 終了要求
				// この for ブロックをいきなり抜けるだけなので
				// もうフラグの値も関係ないのだが一応。
				req &= ~REQ_EXIT;
				break;
			}

			// 電源オン要求は来ないはず
			assertmsg(req == 0, "req=$%x", req);
		}

		// RTC は常にホスト時刻で動いており
		// 32Hz = 31.25msec ごとにパルスを入れる。
		while (::GetRealTime() > rtc_last_clock + 31.25_msec) {
			gRTC->ClockIn();
			rtc_last_clock += 31.25_msec;
		}

		// まず CPU を駆動する。MPU::Run() はデバイスアクセスなどによって
		// スケジューラにイベントが登録されると、その命令の終了とともに
		// 処理を打ち切って戻ってくる。
		// ここでイベントリストを調べ、その時点から最も早く発生するイベント
		// までの必要サイクル数が、次のループでの駆動サイクル数になる。

		// CPU 駆動
		uint32 outer;
		outer = gMPU->Run(cycle);

		// イベント
		for (;;) {
			evcs.lock();
			if (eventlist.empty()) {
				cycle = INT64_MAX;
				break;
			}
			Event *e = eventlist.front();
			if (gMPU->total_cycle() < e->cycle) {
				// まだ時刻に到達していない
				cycle = e->cycle - gMPU->total_cycle();
				break;
			}

			// 到達したのでこのイベントをリストから削除
			e->active = false;
			eventlist.pop_front();
			evcs.unlock();

			// コールバック
			e->dev->putlog(3, "イベント '%s' 時刻到達", e->GetName().c_str());
			((e->dev)->*(e->func))(e->code);

			// 同時刻のイベントがあるかも知れないのでなくなるまで調べる
		}
		evcs.unlock();

		// 同期モードなら実時間調整。
		// あるいは高速モードであってもストップ状態なら実時間駆動する。
		if (mode != 0) {
			// 通常モード
			uint64 vtime;
			uint64 rtime;

			// 基準時からの経過時間
			vtime = GetVirtTime() - vtimebase;
			rtime = GetRealTime() - rtimebase;

			if (vtime > rtime) {
				// 仮想時間のほうが進んでいれば、スリープして待つ
				uint64 diff = vtime - rtime;
				struct timespec ts;
				ts.tv_sec  = diff / (1000 * 1000 * 1000);
				ts.tv_nsec = diff % (1000 * 1000 * 1000);
				nanosleep(&ts, NULL);
			} else {
				// 実時間のほうが進んでいれば、間に合ってない

				// XXX ここで間引き運転とか
			}
		}

		// CPU の STOP 状態をこっちのモードに反映。
		// このために SCHED_STOP と CPU_REQ_STOP は同じビット位置にしてある。
		// CPU の HALT はスケジューラ的には STOP と同じ扱いでいい。
		if (((mode ^ outer) & SCHED_STOP)) {
			// 状態変更要求
			if ((outer & (CPU_REQ_STOP | CPU_REQ_HALT)) != 0) {
				atomic_reqflag |= REQ_STOP;
			} else {
				atomic_reqflag |= REQ_RUN;
			}
		}
	}
}

// 必要ならスレッドの終了を指示して完了まで待つ。
void
Scheduler::Terminate()
{
	if (thread_created) {
		atomic_reqflag |= REQ_EXIT;
		pthread_join(thread, NULL);
		thread_created = false;
	}
}

// 動作モードを設定する。
// true なら高速モード、false なら同期モード。
void
Scheduler::SetFullSpeed(bool enable)
{
	if (enable) {
		atomic_reqflag |= REQ_FAST;
	} else {
		atomic_reqflag |= REQ_SYNC;
	}
}

// 指定のイベントを開始する。
// すでに同イベントが登録されている場合は古いイベントを削除してから
// 新しいイベントを再登録となる。
// イベントはワンショットのみ。
void
Scheduler::StartEvent(Event *newev)
{
	bool inserted = false;
	bool updated = false;

	// 相対 time から絶対 cycle を計算
	newev->cycle = gMPU->total_cycle() + Vtime2Cycle(newev->time);

	evcs.lock();
	// すでにあれば削除
	if (newev->active) {
		for (auto it = eventlist.begin(); it != eventlist.end(); ++it) {
			Event *e = *it;
			if (e == newev) {
				eventlist.erase(it);
				updated = true;
				break;
			}
		}
	}
	// ソートされているところに自身を挿入
	for (auto it = eventlist.begin(); it != eventlist.end(); ++it) {
		Event *e = *it;
		if (newev->cycle <= e->cycle) {
			// このイベントの前に入れる
			eventlist.insert(it, newev);
			inserted = true;
			break;
		}
	}
	if (inserted == false) {
		// 最後に追加
		eventlist.push_back(newev);
	}
	newev->active = true;
	evcs.unlock();

	gMPU->Release();

	newev->dev->putlog(3, "イベント '%s' %s %d.%03d usec 後",
		newev->GetName().c_str(),
		updated ? "更新" : "開始",
		(int)(newev->time / 1000),
		(int)(newev->time % 1000));
}

// 指定のイベントを停止する。
// 指定されたイベントが登録されていなければ何もしない。
void
Scheduler::StopEvent(Event *event)
{
	bool found = false;

	evcs.lock();
	for (auto it = eventlist.begin(); it != eventlist.end(); ++it) {
		Event *e = *it;
		if (e == event) {
			e->active = false;
			eventlist.erase(it);
			found = true;
			break;
		}
	}
	evcs.unlock();

	// イベントを停止した場合は MPU の実行中断はしなくてよい。
	// 1msec 後にイベントを追加した後、やっぱりそのイベントを取り消した
	// 場合 (LUNA の電源オフとか) はここで MPU 処理を中断するよりも
	// そのまま 1msec 走って問題ない。

	if (found) {
		event->dev->putlog(3, "イベント  '%s' 停止", event->GetName().c_str());
	}
}

// 経過時間 t を文字列にして返す。
static const std::string
TimeToStr(uint64 t)
{
	char buf[32];
	char *p;
	size_t len;
	int n;

	uint ns = t % 1000;
	t /= 1000;
	uint us = t % 1000;
	t /= 1000;
	uint ms = t % 1000;
	t /= 1000;
	uint s = t % 60;
	t /= 60;
	uint m = t % 60;
	t /= 60;
	uint h = t;

	p = buf;
	len = sizeof(buf);
	if (h) {
		n = snprintf(p, len, "%d:%02d:%02d", h, m, s);
		p += n;
		len -= n;
	} else if (m) {
		n = snprintf(p, len, "%d:%02d", m, s);
		p += n;
		len -= n;
	} else {
		n = snprintf(p, len, "%d", s);
		p += n;
		len -= n;
	}
	n = snprintf(p, len, ".%03d'%03d'%03d", ms, us, ns);
	p += n;
	len -= n;

	return std::string(buf, p - buf);
}

// モニターサイズ取得
nnSize
Scheduler::GetMonitorSize()
{
	// 本当は Init() 完了時点でイベント個数は確定しているはずだけど、
	// そのタイミングで一度だけ呼ばれるフックが今の所ない。
	return nnSize(68, 11 + gEvents.size());
}

void
Scheduler::MonitorUpdate(TextScreen& monitor)
{
	int x, y;

	monitor.Clear();
	x = 0;
	y = 0;

	monitor.Print(0, y++, "Mode: %s %s",
		(mode & SCHED_SYNC) ? "Sync" : "Full",
		(mode & SCHED_STOP) ? "STOP" : "Run ");
	monitor.Print(0, y++, "Req: %08x", (uint32)atomic_reqflag);
	monitor.Print(0, y++, "MPU Speed: %d.%03dMHz",
		(mpu_clock / 1000), (mpu_clock % 1000));
	uint64 cycle = gMPU->total_cycle();
	uint64 vt = GetVirtTime();
	uint64 rt = GetRealTime();
	uint64 relvt = vt - vtimebase;
	uint64 relrt = rt - rtimebase;
	// RealTime は実時間なのでスケジューラ開始からの時間に変換
	rt -= rtimestart;
	monitor.Print(0, y++, "Total  Real Time: %18s", TimeToStr(rt).c_str());
	monitor.Print(0, y++, "    Virtual Time: %18s", TimeToStr(vt).c_str());
	double ratio = (double)vt / rt * 100;
	monitor.Print(0, y++, "    Ratio       : %3d.%01d%%",
		(int)ratio, ((int)(ratio * 10) % 10));
	monitor.Print(0, y++, "Moment Real Time: %18s", TimeToStr(relrt).c_str());
	monitor.Print(0, y++, "    Virtual Time: %18s", TimeToStr(relvt).c_str());
	// XXX 移動平均をとる
	double relratio = (double)relvt / relrt * 100;
	monitor.Print(0, y++, "    Ratio       : %3d.%01d%%",
		(int)relratio, ((int)(relratio * 10) % 10));

	// 0         1         2         3         4         5         6
	// 0123456789012345678901234567890123456789012345678901234567890123456
	// Set Time        Remain Time     Description               Code
	//   3.123'456'789   3.123'456'789 0123456789012345678901234 $01234567
	x = 0;
	y++;
	monitor.Puts(x, y, "Set Time");
	monitor.Puts(x + 16, y, "Remain Time");
	monitor.Puts(x + 32, y, "Description");
	monitor.Puts(x + 58, y, "Code");
	y++;
	for (const auto& e : gEvents) {
		uint64 rem;
		TA attr;
		if (e->active) {
			attr = TA::Normal;
			rem = Cycle2Vtime(e->cycle - cycle);
		} else {
			attr = TA::Disable;
			rem = 0;
		}
		monitor.Print(x, y++, attr,
			"%3u.%03u'%03u'%03u %3u.%03u'%03u'%03u %-25s $%08x",
			(uint)(e->time / (1000 * 1000 * 1000)),
			(uint)((e->time / 1000 / 1000) % 1000),
			(uint)((e->time / 1000) % 1000),
			(uint)(e->time % 1000),
			(uint)(rem / (1000 * 1000 * 1000)),
			(uint)((rem / 1000 / 1000) % 1000),
			(uint)((rem / 1000) % 1000),
			(uint)(rem % 1000),
			e->GetName().c_str(),
			e->code);
	}
}
