//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "spc.h"
#include "mainapp.h"
#include "mpu680x0.h"
#include "mpu88xx0.h"
#include "mystring.h"
#include "scheduler.h"
#include "vm.h"

// time 時間後に呼び出すコールバックをセットする。
// func の書式が面倒なのを省略して書きたいため。
#define CallAfter(func_, name_, time_)	do { \
	event.func = (DeviceCallback_t)&SPCDevice::func_; \
	event.SetName(name_); \
	event.code = 0; \
	event.time = time_; \
	event.Start(); \
} while (0)

std::unique_ptr<SPCDevice> gSPC;

// コンストラクタ
SPCDevice::SPCDevice()
{
	logname = "spc";
	devname = "SPC";
	// これらは XIOSpace で必要なので、LUNA ではダミー。
	// もうちょっといい方法を考えたほうがいい。
	devaddr = 0xe96000;
	devlen = 0x2000;

	SetConfigName("spc0");
	myid = -1;

	monitor_size = nnSize(70, 15);

	// 汎用イベント
	// 名前とコールバック関数は都度セットする。
	event.dev = this;
	event.SetName("SPC phase");

	// TC カウンタイベント
	event_tc.dev = this;
	event_tc.code = 0;
	event_tc.SetName("SPC Timer");
}

// デストラクタ
SPCDevice::~SPCDevice()
{
}

bool
SPCDevice::Init()
{
	if (!inherited::Init()) {
		return false;
	}

	// 実行中に別インスタンスを頻繁に参照しないようにここで vmtype を覚えておく
	vmtype = gMainApp.GetVMType();

	// 入力クロックは機種によって異なる。
	// X68k は 5MHz が入っている。
	// LUNA-I はおそらく 6.144MHz だと思われる。LUNA88K は未確認。
	const ConfigItem& item = gConfig->Find("!spc-tCLF");
	t_CLF = item.AsInt();

	return true;
}


// 本体リセットスイッチによるリセット
void
SPCDevice::ResetHard()
{
	// LUNA は本体回路が分からないのでとりあえずこっち。
	if (vmtype != VMTYPE_X68030) {
		Reset();
	}
}

// MPU の RESET 命令によるリセット
void
SPCDevice::ResetSoft()
{
	// X680x0 はこっちだと分かっている。
	if (vmtype == VMTYPE_X68030) {
		Reset();
	}
}

// SPC リセット
void
SPCDevice::Reset()
{
	putlog(1, "リセット");

	// 初期値 (p.61)
	spc.sctl = SPC::SCTL_RESET;
	spc.ints = 0x00;
	spc.ssts &= 0x0f;
	spc.serr = 0x00;

	// XXX マニュアルには書いていないけど、
	// SDGC bit5 の転送ごとに割り込みを発生させるビットが
	// OFFになることを NetBSD が期待しているし、
	// このビットは落ちるのではないだろうか
	spc.sdgc = 0x00;

	// 内部状態
	prev_ints = (uint32)-1;
	prev_ssts = (uint32)-1;
	set_atn_in_selection = false;
	transfer_command_running = false;
	event.Stop();
	event_tc.Stop();
}

uint64
SPCDevice::Read(uint32 addr)
{
	uint8 data;

	switch (addr) {
	 case SPC::BDID:
		data = 1 << myid;
		putlog(3, "BDID -> $%02x", data);
		return data;

	 case SPC::SCTL:
		data = spc.sctl;
		putlog(3, "SCTL -> $%02x", data);
		return data;

	 case SPC::SCMD:
		data = spc.scmd;
		putlog(3, "SCMD -> $%02x", data);
		return data;

	 case SPC::TMOD:
		// MB89352 には存在しないレジスタ。
		// 下位2ビットは %0 らしいので意味はないだろうけど真似ておく。
		return 0xfc;

	 case SPC::INTS:
		data = spc.ints;
		if (loglevel >= 3) {
			if (event_tc.active) {
				// イベント中は値が変わった時だけ表示
				if (prev_ints != data) {
					putlog("INTS -> $%02x", data);
					prev_ints = data;
				}
			} else {
				putlog("INTS -> $%02x", data);
				prev_ints = -1U;
			}
		}
		return data;

	 case SPC::PSNS:
		data = GetPSNS();
		putlog(3, "PSNS -> $%02x", data);
		return data;

	 case SPC::SSTS:
		data = GetSSTS();
		if (loglevel >= 3) {
			if (event_tc.active) {
				// イベント中は値が変わった時だけ表示
				if (prev_ssts != data) {
					putlog("SSTS -> $%02x", data);
					prev_ssts = data;
				}
			} else {
				putlog("SSTS -> $%02x", data);
				prev_ssts = -1U;
			}
		}
		return data;

	 case SPC::SERR:
		data = spc.serr;
		putlog(3, "SERR -> $%02x", data);
		return data;

	 case SPC::PCTL:
		data = GetPCTL();
		putlog(3, "PCTL -> $%02x", data);
		return data;

	 case SPC::MBC:
		putlog(0, "MBC 未実装読み込み");
		return 0xff;

	 case SPC::DREG:
		return ReadDREG();

	 case SPC::TEMP:
		data = spc.temp_in;
		putlog(3, "TEMP -> $%02x", data);
		return data;

	 case SPC::TCH:
		data = spc.tc >> 16;
		putlog(3, "TCH  -> $%02x", data);
		return data;

	 case SPC::TCM:
		data = (spc.tc >> 8) & 0xff;
		putlog(3, "TCM  -> $%02x", data);
		return data;

	 case SPC::TCL:
		data = spc.tc & 0xff;
		putlog(3, "TCL  -> $%02x", data);
		return data;

	 case SPC::EXBF:
		// MB89352 には存在しないレジスタ
		return 0xff;
	}
	__unreachable();
}

uint64
SPCDevice::Write(uint32 addr, uint32 data)
{
	switch (addr) {
	 case SPC::BDID:
		putlog(3, "BDID <- $%02x", data);
		if (myid == data) {
			// 同じなら何もしない
			return 0;
		}
		if (myid == -1) {
			// 初回なのでバスに参加
			myid = data;
			scsibus->Attach(this, myid);
		} else {
			// そうでなければ移動?
			putlog(0, "BDID 変更 %d to %d not implemented", myid, data);
			return 0;
		}
		return 0;

	 case SPC::SCTL:
		WriteSCTL(data);
		return 0;

	 case SPC::SCMD:
		WriteSCMD(data);
		return 0;

	 case SPC::TMOD:
		// MB89352 には存在しないレジスタ
		return 0;

	 case SPC::INTS:
		WriteINTS(data);
		return 0;

	 case SPC::SDGC:
		spc.sdgc = data;
		putlog(3, "SDGC <- $%02x", spc.sdgc);
		return 0;

	 case SPC::SSTS:
		// XXX 書き込みはどうなる?
		return 0;

	 case SPC::SERR:
		// XXX 書き込みはどうなる?
		return 0;

	 case SPC::PCTL:
		data &= SPC::PCTL_MASK;
		putlog(3, "PCTL <- $%02x", data);
		spc.busfree_intr_enable = (data & SPC::PCTL_BFINT_EN);
		SetXfer(data);
		return 0;

	 case SPC::MBC:
		// XXX 書き込みはどうなる?
		return 0;

	 case SPC::DREG:
		putlog(3, "DREG <- $%02x", data);
		return WriteDREG(data);

	 case SPC::TEMP:
		spc.temp_out = data;
		putlog(3, "TEMP <- $%02x", spc.temp_out);

		// Inside (p479) によると、SPC はセレクションタイムアウト時 SEL を
		// 下げないので、代わりに TEMP に $00 を書き込むことで SEL を下げる、
		// そうなのだが、データシート PDF には記述が見当たらないし、
		// luna68k の /boot はここで SEL が下げてしまうと期待した動作に
		// ならないようだ。
		// SPC はセレクションタイムアウトしたからと言って、データバスに出した
		// ID を自動的に取り下げたりはしないので、手動でデータバスをクリア
		// してあげないといけない (TEMP に $00 を書き込む)、だったんでは。
		if (0 && GetPhase() == SCSI::Phase::Selection) {
			NegateSEL();
		}
		// Out 方向ならバスに出す
		if (GetIO() == false) {
			SetData(spc.temp_out);
		}
		return 0;

	 case SPC::TCH:
		spc.tc = (spc.tc & 0x00ffff) | (data << 16);
		putlog(3, "TCH  <- $%02x (TC=$%06x)", data, spc.tc);
		return 0;
	 case SPC::TCM:
		spc.tc = (spc.tc & 0xff00ff) | (data << 8);
		putlog(3, "TCM  <- $%02x (TC=$%06x)", data, spc.tc);
		return 0;
	 case SPC::TCL:
		spc.tc = (spc.tc & 0xffff00) | data;
		putlog(3, "TCL  <- $%02x (TC=$%06x)", data, spc.tc);
		return 0;

	 case SPC::EXBF:
		// MB89352 には存在しないレジスタ
		return 0;
	}
	__unreachable();
}

uint64
SPCDevice::Peek(uint32 addr)
{
	switch (addr) {
	 case SPC::BDID:
		return 1 << myid;

	 case SPC::SCTL:
		return spc.sctl;

	 case SPC::SCMD:
		return spc.scmd;

	 case SPC::TMOD:
		// MB89352 には存在しないレジスタ。
		// 下位2ビットは %0 らしいので意味はないだろうけど真似ておく。
		return 0xfc;

	 case SPC::INTS:
		return spc.ints;

	 case SPC::PSNS:
		return GetPSNS();

	 case SPC::SSTS:
		return GetSSTS();

	 case SPC::SERR:
		return spc.serr;

	 case SPC::PCTL:
		return GetPCTL();

	 case SPC::MBC:
		return 0xff;

	 case SPC::DREG:
		// X68k は SPC の DREQ 信号から DTACK 信号を作っているため、
		// DREQ 信号を出さないプログラム転送モードでは DREG へのアクセスが
		// すべてバスエラーになってしまう。
		if (vmtype == VMTYPE_X68030) {
			if ((spc.scmd & SPC::SCMD_PROGRAM)) {
				return (uint64)-1;
			}
		}
		return spc.dreg.Peek(0);

	 case SPC::TEMP:
		return spc.temp_in;

	 case SPC::TCH:
		return spc.tc >> 16;

	 case SPC::TCM:
		return (spc.tc >> 8) & 0xff;

	 case SPC::TCL:
		return spc.tc & 0xff;

	 case SPC::EXBF:
		// MB89352 には存在しないレジスタ
		return 0xff;
	}
	__unreachable();
}

void
SPCDevice::MonitorUpdate(TextScreen& monitor)
{
	int x;
	int y;
	uint32 sctl;
	uint32 scmd;
	uint32 ints;
	uint32 psns;
	uint32 sdgc;
	uint32 ssts;
	uint32 serr;
	uint32 pctl;
	uint32 tc;

	monitor.Clear();

	sctl = spc.sctl;
	scmd = spc.scmd;
	ints = spc.ints;
	psns = GetPSNS();
	sdgc = spc.sdgc;
	ssts = GetSSTS();
	serr = spc.serr;
	pctl = GetPCTL();
	tc   = spc.tc;

	y = 0;
	monitor.Print(0, y++, "BDID: $%02x (ID=%d)", (1 << myid), myid);
	monitor.Print(0, y++, "SCTL: $%02x", sctl);
	monitor.Print(0, y++, "SCMD: $%02x", scmd);
	monitor.Print(0, y++, "INTS: $%02x", ints);
	monitor.Print(0, y++, "PSNS: $%02x", psns);
	monitor.Print(0, y++, "SDGC: $%02x", sdgc);
	monitor.Print(0, y++, "SSTS: $%02x", ssts);
	monitor.Print(0, y++, "SERR: $%02x", serr);
	monitor.Print(0, y++, "PCTL: $%02x", pctl);
	monitor.Print(0, y++, "MBC : $XX");
	monitor.Print(0, y,   "DREG: ");
	{
		uint i;
		for (i = 0; i < spc.dreg.Length(); i++) {
			monitor.Print(6 + i * 4, y, "$%02x", spc.dreg.Peek(i));
		}
		for (; i < 8; i++) {
			monitor.Print(6 + i * 4, y, TA::Disable, "---");
		}
	}
	y++;
	monitor.Print(0, y++, "TEMP: $%02x(IN) $%02x(OUT)",
		spc.temp_in, spc.temp_out);
	monitor.Print(0, y++, "TC  : $%06x", tc);

	x = 10;
	y = 1;

	// SCTL
	static const char * const sctl_bits[] = {
		"RST", "CTRL", "DIAG", "ARB", "PARI", "SEL", "RESL", "INTR"
	};
	MonitorReg(monitor, x, y, sctl, sctl_bits);
	y++;

	// SCMD
	static const char * const scmd_bits[] = {
		"", "", "", "RST", "INTC", "PROG", "", "TERM"
	};
	MonitorReg(monitor, x, y, scmd, scmd_bits);
	static const char * const scmd_cmd[] = {
		//12345678901234
		"(Bus Release)",
		"(Select)",
		"(Reset ATN)",
		"(Set ATN)",
		"(Transfer)",
		"(XferPause)",
		"(ResetACK/REQ)",
		"(Set ACK/REQ)",
	};
	monitor.Puts(x, y, scmd_cmd[scmd >> 5]);
	y++;

	// INTS
	static const char * const ints_bits[] = {
		"Sel", "Resl", "Disc", "Cmpl", "SReq", "TOut", "HErr", "Rst"
	};
	MonitorReg(monitor, x, y, ints, ints_bits);
	y++;

	// PSNS
	static const char * const psns_bits[] = {
		"REQ", "ACK", "ATN", "SEL", "BSY", "MSG", "C/D", "I/O"
	};
	MonitorReg(monitor, x, y, psns, psns_bits);
	y++;

	// SDGC
	static const char * const sdgc_bits[] = {
		"",    "",    "XfEn", "",   "",    "",    "",    ""
	};
	MonitorReg(monitor, x, y, sdgc, sdgc_bits);
	y++;

	// SSTS
	static const char * const ssts_bits[] = {
		"", "", "SBSY", "XInP", "RSTI", "TCZ", "Full", "Empt"
	};
	MonitorReg(monitor, x, y, ssts, ssts_bits);
	static const char * const ssts_conn[] = {
		//12345678
		"(NotConn)",
		"(As Tgt)",
		"(As Init)",
		"(undef?)",
	};
	monitor.Puts(x, y, ssts_conn[ssts >> 6]);
	y++;

	// SERR
	static const char * const serr_bits[] = {
		"", "", "XOut", "", "PErr", "", "Shrt", ""
	};
	MonitorReg(monitor, x, y, serr, serr_bits);
	y++;

	// PCTL
	static const char * const pctl_bits[] = {
		"BIE", "", "", "", "", "MSG", "C/D", "I/O"
	};
	MonitorReg(monitor, x, y, pctl, pctl_bits);
	y++;

	// バス状態
	x = 52;
	y = 0;
	SCSI::Phase ph {};
	uint8 rst = 0;
	uint8 bsy = 0;
	uint8 sel = 0;
	uint8 msg = 0;
	uint8 cd = 0;
	uint8 io = 0;
	uint8 atn = 0;
	uint8 req = 0;
	uint8 ack = 0;
	uint8 data = 0;
	if (bus) {
		ph  = GetPhase();
		rst = GetRST();
		bsy = GetBSY();
		sel = GetSEL();
		msg = GetMSG();
		cd  = GetCD();
		io  = GetIO();
		atn = GetATN();
		req = GetREQ();
		ack = GetACK();
		data = GetData();
	}
	monitor.Print(x, y++, "Phase: %s", SCSI::GetPhaseName(ph));
	monitor.Puts(x, y++, TA::OnOff(rst), "RST");
	monitor.Puts(x, y++, TA::OnOff(bsy), "BSY");
	monitor.Puts(x, y++, TA::OnOff(sel), "SEL");
	monitor.Puts(x, y++, TA::OnOff(msg), "MSG");
	monitor.Puts(x, y++, TA::OnOff(cd),  "CD");
	monitor.Puts(x, y++, TA::OnOff(io),  "IO");
	monitor.Puts(x, y++, TA::OnOff(atn), "ATN");
	monitor.Puts(x, y++, TA::OnOff(req), "REQ");
	monitor.Puts(x, y++, TA::OnOff(ack), "ACK");
	monitor.Print(x, y++, "DATA:$%02x", data);

	y = 2;
	monitor.Puts(x + 4, y++, "(________)");
	monitor.Puts(x + 4, y++, "(________)");
	monitor.Puts(x + 4, y++, "+");
	monitor.Puts(x + 4, y++, "|");
	monitor.Puts(x + 4, y++, "+");

	// BSY, SEL の内訳。"(76543210)" の順。
	y = 2;
	for (int i = 0; i < 8; i++) {
		if ((bsy & (1 << i))) {
			monitor.Print(x + 4 + 8 - i, y, "%d", i);
		} else {
			monitor.Print(x + 4 + 8 - i, y, "_");
		}
	}
	y++;
	for (int i = 0; i < 8; i++) {
		if ((sel & (1 << i))) {
			monitor.Print(x + 4 + 8 - i, y, "%d", i);
		} else {
			monitor.Print(x + 4 + 8 - i, y, "_");
		}
	}

	// 転送フェーズ中なら現在の XferPhase
	y += 2;
	if (ph == SCSI::Phase::Transfer) {
		monitor.Puts(x + 5, y, SCSI::GetXferPhaseName(GetXfer()));
	}

	// ターゲット
	x = 0;
	y = 14;
	if (bus) {
		SCSIDevice *d = bus->GetConnectedTarget();
		SCSITarget *t = dynamic_cast<SCSITarget*>(d);
		if (t) {
			std::vector<uint8> cmds = t->GetCmdSeq();
			if (cmds.size() > 0) {
				const char *cmdname = SCSI::GetCommandName(cmds[0]);
				monitor.Print(x, y, "Command: $%02x", cmds[0]);
				if (cmdname) {
					monitor.Puts(x + 14, y, cmdname);
				}
			}
		}
	}
}

// レジスタをビットごとに表示する。MonitorUpdate の下請け。
void
SPCDevice::MonitorReg(TextScreen& monitor,
	int x, int y, uint32 reg, const char * const *names)
{
	for (int i = 0; i < 8; i++) {
		bool b = reg & (1 << (7 - i));
		monitor.Puts(x + i * 5, y, TA::OnOff(b), names[i]);
	}
}

// SCTL レジスタへの書き込み
void
SPCDevice::WriteSCTL(uint32 data)
{
	// リセットの立ち上がり検出が必要
	uint32 up = ~spc.sctl & data;

	spc.sctl = data;
	putlog(3, "SCTL <- $%02x", spc.sctl);
	if ((up & SPC::SCTL_RESET)) {
		// リセット
		Reset();
	} else {
		// 通常動作
	}
	if ((spc.sctl & SPC::SCTL_CTL_RESET)) {
		putlog(2, "Control Reset");
	}
	if ((spc.sctl & SPC::SCTL_DIAG_MODE)) {
		putlog(0, "SCTL 未実装 Diag Mode");
	}
}

// SCMD レジスタへの書き込み
void
SPCDevice::WriteSCMD(uint32 data)
{
	uint32 chg = spc.scmd ^ data;	// 変更点
	spc.scmd = data;
	putlog(3, "SCMD <- $%02x", spc.scmd);

	// Intercept, Program, Terminate Mode の未実装判定は
	// 転送開始時のみ表示する。

	// RST Out 操作は SCTL Reset がオフの時だけ有効
	// と書いてあるけど、SCTL Reset がオンならほぼ全機能無効なのでは?
	if ((spc.sctl & SPC::SCTL_RESET) == 0 && (chg & SPC::SCMD_RST_OUT)) {
		if ((spc.scmd & SPC::SCMD_RST_OUT)) {
			// 0 -> 1
			// すでに RST が立ってたらどうするか
			if (GetRST() == false) {
				putlog(1, "SCMD RST Out");
				BusRelease();
				AssertRST();
				SetINTS(SPC::INTS_RESET_COND);
			}
		} else {
			// 1 -> 0
			// 特にすることはない?
			NegateRST();
		}
	}

	switch (spc.scmd >> 5) {
	 case 0:
		putlog(2, "SCMD Bus Release コマンド");
		BusFree();
		break;
	 case 1:
		SelectCommand();
		break;
	 case 2:
		putlog(1, "SCMD Reset ATN コマンド (未実装)");
		break;
	 case 3:
		SetATNCommand();
		break;
	 case 4:
		TransferCommand();
		break;
	 case 5:
		putlog(2, "SCMD Transfer Pause コマンド (未実装)");
		break;
	 case 6:
		putlog(2, "SCMD Reset ACK/REQ コマンド");
		// イニシエータなら ACK を、ターゲットなら REQ をリセットする。
		// ターゲットモードはサポートしないので、操作するのは ACK のみ。
		NegateACK();
		break;
	 case 7:
		putlog(2, "SCMD Set ACK/REQ コマンド");
		// イニシエータなら ACK を、ターゲットなら REQ をセットする。
		// ターゲットモードはサポートしないので、操作するのは ACK のみ。
		AssertACK();
		break;
	}
}

// INTS レジスタへの書き込み
void
SPCDevice::WriteINTS(uint32 data)
{
	uint32 ack;

	// クリアすることになるビットを覚えておく
	ack = spc.ints & data;

	// '1' が書き込まれたビットをクリア
	spc.ints &= ~data;
	putlog(3, "INTS <- $%02x (INTS = $%02x)", data, spc.ints);

	// SPC では INTS でビットを落とす行為は、
	// ホストが割り込み要因に対して処理を完了したことを
	// SPC に通知する意味が含まれている。

	// DISCONNECTED ビットが立っているときは SPC は SCSI から
	// セレクション、リセレクション要求が来ても応答しない。
	// DISCONNECTED ビットが下ろされると、応答するようになる。

	// TIMEOUT ビットをクリアした時が
	if ((ack & SPC::INTS_TIMEOUT)) {
		// セレクションフェーズなら
		if (GetPhase() == SCSI::Phase::Selection) {
			// TC がゼロならバスフリー
			// そうでなければリスタート
			if (spc.tc == 0) {
				BusFree();
			} else {
				event_tc.time = spc.tc * t_CLF * 2;
				event_tc.func = (DeviceCallback_t)&SPCDevice::SelectionTimeout;
				event_tc.Start();
			}
		}
	}
}

// INTS 設定。
// INTS にビットを立てる要因が起きた時に data にそれを指定して呼び出す。
// この関数は INTS を書き換えて、必要なら割り込みを上げる。
// INTS レジスタへの書き込み(割り込み要因のクリア)は WriteINTS() のほう。
void
SPCDevice::SetINTS(uint32 data)
{
	// 立ち上がりでだけ割り込みがかかるわけではないが表示のため。
	uint32 up = ~spc.ints & data;

	// 指定された要因を立てる
	spc.ints |= data;

	// Disconnected 割り込みが起きたら ATN を下げる
	if ((spc.ints & SPC::INTS_DISCONNECTED) && GetATN()) {
		NegateATN();
	}

	// 割り込み許可なら割り込み。
	// RESET_COND はノンマスカブル割り込み。
	if ((spc.sctl & SPC::SCTL_INTR_EN) || (spc.ints & SPC::INTS_RESET_COND)) {
		putlog(3, "割り込み ints=$%02x up=$%02x", spc.ints, up);
		if (vmtype != VMTYPE_X68030) {
			gMPU->Interrupt(this, 2);
		} else {
			// とりあえず内蔵だけ
			gMPU680x0->Interrupt(this, 1, 0x6c);
		}
	}
}

// PSNS レジスタ値の取得。
// Read*() と Peek8() から呼ばれる。
// scsibus 接続前にもモニタの更新などで呼ばれることに注意。
uint32
SPCDevice::GetPSNS() const
{
	uint32 data = 0;

	if (__predict_false(bus == NULL)) {
		return 0;
	}

	if (GetREQ()) {
		data |= SPC::PSNS_REQ;
	}
	if (GetACK()) {
		data |= SPC::PSNS_ACK;
	}
	if (GetATN()) {
		data |= SPC::PSNS_ATN;
	}
	if (GetSEL()) {
		data |= SPC::PSNS_SEL;
	}
	if (GetBSY()) {
		data |= SPC::PSNS_BSY;
	}
	data |= (uint32)GetXfer();

	return data;
}

// SSTS レジスタ値の取得。
// Read*() と Peek8() から呼ばれる。
// scsibus 接続前にもモニタの更新などで呼ばれることに注意。
uint32
SPCDevice::GetSSTS() const
{
	uint32 data;

	data = spc.ssts & 0xf0;
	if (__predict_true(bus != NULL) && GetRST()) {		// RST 信号
		data |= SPC::SSTS_RST_IN;
	}
	if (spc.tc == 0) {	// TC がゼロか
		data |= SPC::SSTS_TC_ZERO;
	}
	if (spc.dreg.Length() == 8) {
		data |= SPC::SSTS_DREG_FULL;
	}
	if (spc.dreg.Length() == 0) {
		data |= SPC::SSTS_DREG_EMPTY;
	}

	return data;
}

// PCTL レジスタ値の取得。
// Read*() と Peek8() から呼ばれる。
// scsibus 接続前にもモニタの更新などで呼ばれることに注意。
uint32
SPCDevice::GetPCTL() const
{
	uint32 data;

	data = 0;
	if (__predict_true(bus != NULL)) {
		data |= (uint32)GetXfer();
	}
	if (spc.busfree_intr_enable) {
		data |= SPC::PCTL_BFINT_EN;
	}
	return data;
}

// DREG レジスタの読み込み
// (副作用があるので Peek 系からは呼ばないこと)
uint64
SPCDevice::ReadDREG()
{
	uint32 data;

	// X68k は SPC の DREQ 信号から DTACK 信号を作っているため、
	// DREQ 信号を出さないプログラム転送モードでは DREG へのアクセスが
	// すべてバスエラーになってしまう。
	if (vmtype == VMTYPE_X68030) {
		if ((spc.scmd & SPC::SCMD_PROGRAM)) {
			return (uint64)-1;
		}
	}

	// XFER_OUT は DREG アクセスでリセットされる
	spc.serr &= ~SPC::SERR_XFER_OUT;

	// 受信データがあれば取り出す
	if (__predict_true(spc.dreg.Length() > 0)) {
		bool q_was_full = spc.dreg.IsFull();
		data = spc.dreg.Dequeue();
		if (loglevel >= 4) {
			putlog("DREG -> $%02x (q=%d TC=0x%x)",
				data, spc.dreg.Length(), spc.tc);
		} else {
			putlog(3, "DREG -> $%02x", data);
		}
		// この読み出しが最後だったらこれで転送完了。
		if (spc.tc == 0 && spc.dreg.Length() == 0) {
			TransferComplete();
		}
		// この読み出しによってバッファに空きが出来たら転送再開。
		if (transfer_command_running && GetREQ() && q_was_full) {
			// XXX 時間は適当
			CallAfter(HardwareTransfer, "SPC ReadDREG to Transfer", t_CLF);
		}
	} else {
		// 受信データがない場合
		if ((spc.scmd & SPC::SCMD_PROGRAM)) {
			// XXX プログラム転送中ならどうなる?
			data = 0xff;
			putlog(3, "DREG -> $%02x (empty)", data);
		} else {
			// DMA 転送
			// XXX どうする?
			PANIC("DMA転送中の DREG empty");
		}
	}
	return data;
}

// DREG レジスタへの書き込み
uint64
SPCDevice::WriteDREG(uint32 data)
{
	// X68k は SPC の DREQ 信号から DTACK 信号を作っているため、
	// DREQ 信号を出さないプログラム転送モードでは DREG へのアクセスが
	// すべてバスエラーになってしまう。
	if (vmtype == VMTYPE_X68030) {
		if ((spc.scmd & SPC::SCMD_PROGRAM)) {
			return (uint64)-1;
		}
	}

	// XFER_OUT は DREG アクセスでリセットされる
	spc.serr &= ~SPC::SERR_XFER_OUT;

	// DREG キューに入れて..
	spc.dreg.Enqueue(data);
	// ハードウェア転送中で REQ が来てれば送信。
	if (transfer_command_running && GetREQ()) {
		// XXX 時間は適当
		CallAfter(HardwareTransfer, "SPC WriteDREG to Transfer", t_CLF);
	}
	return 0;
}

// バスフリーフェーズに移行する
void
SPCDevice::BusFree()
{
	putlog(3, "BusFree");

	// ここは BDID 書き込み前 (まだ scsibus に自身をつなげる前) にも呼ばれる。
	// その場合は何もしない。バスフリーのはずでもあるし。
	if (bus == NULL) {
		return;
	}

	BusRelease();
}

// バスをリリースする。バスフリーとバスリセットから呼ばれる
void
SPCDevice::BusRelease()
{
	// 自発的にバスをリリースする場合、今のフェーズから自分が変えていい
	// BSY、SEL だけを落とす。それ以外の信号線は正しくはないけどとりあえず
	// 一括して SCSIHostDevice::BusFreeCB() が落とすことにしてある。

	switch (GetPhase()) {
	 case SCSI::Phase::BusFree:
	 case SCSI::Phase::Transfer:
		break;
	 case SCSI::Phase::Arbitration:
	 case SCSI::Phase::Selection:
	 case SCSI::Phase::Reselection:
		NegateSEL();
		NegateBSY();
		break;
	}
}

// Select コマンド実行
void
SPCDevice::SelectCommand()
{
	std::string idstr;
	uint32 data;
	uint32 tgtid;
	uint64 wait;

	// ターゲット ID を表示用に解析
	data = spc.temp_out;
	if (data == 0) {
		// ID が一切立ってない?
		idstr = "TEMP==0; No IDs?";
	} else if ((data & (1 << myid)) == 0) {
		// 自 ID が立ってない?
		idstr = string_format("TEMP=$%02x; No BDID?", data);
	} else {
		// ここで少なくとも自 ID は立っている
		data &= ~(1 << myid);

		if (data == 0) {
			// ターゲット ID がない
			idstr = string_format("#%d", myid);
		} else {
			tgtid = DecodeID(data);
			idstr = string_format("#%d -> #%d", myid, tgtid);
		}
	}
	putlog(2, "Select コマンド (%s)", idstr.c_str());

	// アービトレーションありの場合
	//
	// BSY  X\______/~~~~~~~~~~~~~~~~~~~~~~~~~~\_ .. _/^^^^^^^^^^^^^^^
	//       |t_BFBL|                          |      |t_BIDH|
	// DB*  ________<XXXXXXXXXXXXXXXXXXXXXXXXXXXX .. XXXXXXXX>________
	//              |t_ARB|t_AWSL|             |      |t_BLSH|
	// SEL  _____________________/~~~~~~~~~~~~~~~ .. ~~~~~~~~\________
	//                           |t_SIDA|t_IDBH|             |
	// ATN  ____________________________/~~~~~~~~ .. ~~~~~~~~~~~~~~~~~
	//                                                       |t_SHIR|
	// INTR _____________________________________ .. _______________/~
	//
	// t_BFBL = <Min> (6 + TCL) * t_CLF, <Max> (7 + TCL) * t_CLF + 60
	// t_ARB  = <Min> 32 * t_CLF - 60
	// t_AWSL = <Min> 0, <Max> 80
	// t_SIDA = <Min> 11 * t_CLF - 30
	// t_IDBH = <Min> 2 * t_CLF - 80
	//
	// ここではバスフリー検出から少なくとも t_BFBL 以上 + t_ARB + t_AWSL 後に
	// SEL と DB* と ATN を立てて、そこから t_SIDA + t_IDBH 後に BSY を落とす。
	//
	// アービトレーションなしの場合
	//
	// BSY  X\_______________ .. _/~~~~~~~~~~~~~~~
	//       |t_FRID|             |t_BIDH|
	// DB*  ________<XXXXXXXX .. XXXXXXXX><XXXXXXX
	//              |t_IDSL|      |t_BLSH|
	// SEL  _______________/~ .. ~~~~~~~~\________
	//                                   |t_SHIR|
	// INTR _________________ .. _______________/~
	//
	// t_FRID = <Min> (6 + TCL) * t_CLF, <Max> (7 + TCL) * t_CLF + 140
	// t_IDSL = <Min> 11 * t_CLF - 80
	// t_BIDH = <Min> t_BLSH = 2 * t_CLF
	// t_SHIR = <Max> 60
	//

	// t_FRID/t_BFBL はたぶん今からではなくバスフリーになった時点から起算。
	// t_FRID と t_BFBL は同じ式なのでここでは t_BFBL で統一。
	uint64 t_BFBL = (6 + spc.GetTCL()) * t_CLF;	// Min
	uint64 now = gScheduler->GetVirtTime();
	if ((last_busfree_time + t_BFBL) > now) {
		wait = (last_busfree_time + t_BFBL) - now;
	} else {
		wait = 0;
	}

	spc.ssts &= ~(SPC::SSTS_CONN_INIT | SPC::SSTS_CONN_TARG);
	spc.ssts |= SPC::SSTS_SPC_BUSY;

	// TCH,TCM によるタイマーを始動。
	// N が 0 なら無限大なのでタイマーは動かさない。
	uint32 N = (spc.tc >> 8) << 8;
	if (N != 0) {
		event_tc.time = (N + 15) * t_CLF * 2;
		event_tc.func = (DeviceCallback_t)&SPCDevice::SelectionTimeout;
		event_tc.Start();
	}

	if ((spc.sctl & SPC::SCTL_ARBIT_EN)) {
		// アービトレーションありの場合

		// ここでは複数のデバイスが同時にアービトレーションフェーズに
		// 入ることはないため、アービトレーションを開始したデバイスが
		// 必ずバスの使用権を得る。

		// バスフリーから規定時間後に SEL (と DB* と ATN) を上げる。
		uint64 t_ARB  = 32 * t_CLF - 60;
		uint64 t_AWSL = 40;	// 適当に中間
		CallAfter(Arbitration1, "SPC Arbitration", wait + t_ARB + t_AWSL);
	} else {
		// アービトレーションなしの場合

		// バスフリーから規定時間後に SEL (と DB*) を上げる。
		uint64 t_IDSL = 11 * t_CLF - 80;
		CallAfter(Selection1, "SPC Selection", wait + t_IDSL);
	}
}

// アービトレーションフェーズ開始後、規定時間経過したので SEL を立てるところ。
void
SPCDevice::Arbitration1(int dummy)
{
	putlog(3, "Arbitration1");

	spc.ssts |= SPC::SSTS_SPC_BUSY;

	// 本当は DB* -> SEL -> ATN の順でタイミングも規定されてるけど省略。
	// ここでは SEL がトリガーになるので SEL のアサートは最後に行う。
	SetData(spc.temp_out);
	if (set_atn_in_selection) {
		AssertATN();
		set_atn_in_selection = false;
	}
	AssertSEL();

	uint64 t_SIDA = 11 * t_CLF - 30;
	uint64 t_IDBH = 2 * t_CLF - 80;
	CallAfter(Arbitration2, "SPC Arbitration2", t_SIDA + t_IDBH);
}

// アービトレーションフェーズ中、SEL を立ててから規定時間経過したので BSY を
// 下げるところ。
void
SPCDevice::Arbitration2(int dummy)
{
	putlog(3, "Arbitration2");

	NegateBSY();
}

// アービトレーションフェーズを使わない場合のセレクションフェーズで、
// 規定時間経過したので SEL を立てるところ。
void
SPCDevice::Selection1(int dummy)
{
	putlog(3, "Selection1");

	spc.ssts |= SPC::SSTS_SPC_BUSY;

	// 本当は DB* -> SEL の順でタイミングも規定されてるけど省略。
	// ここでは SEL がトリガーになるので SEL のアサートは最後に行う。
	SetData(spc.temp_out);
	AssertSEL();
}

// セレクションコマンドがタイムアウトした時に呼ばれるコールバック。
void
SPCDevice::SelectionTimeout(int dummy)
{
	if (loglevel >= 3) {
		int tgtid;
		uint8 data = GetData();
		data &= ~(1 << myid);
		tgtid = DecodeID(data);
		if (tgtid == -1) {
			putlog("Selection timeout");
		} else {
			putlog("Selection timeout (#%d)", tgtid);
		}
	}

	spc.ssts &= ~(SPC::SSTS_CONN_INIT | SPC::SSTS_SPC_BUSY);

	// TC は実際には減算カウンタのようだが、とりあえず辻褄合わせ
	spc.tc = 0;

	SetINTS(SPC::INTS_TIMEOUT);
}

// SetATN コマンド実行
void
SPCDevice::SetATNCommand()
{
	putlog(2, "SCMD Set ATN Command");

	// Select コマンド前なら、セレクションフェーズで ATN を立てる指示。
	// 接続中(転送フェーズ中?) なら即座に ATN を立てる。
	switch (GetPhase()) {
	 case SCSI::Phase::BusFree:
		set_atn_in_selection = true;
		break;
	 case SCSI::Phase::Arbitration:
	 case SCSI::Phase::Selection:
	 case SCSI::Phase::Reselection:
		// Select Command 実行中なので実質ここで呼ばれることはないはずだが
	 case SCSI::Phase::Transfer:
		AssertATN();
		break;
	}
}

// Transfer コマンド実行
void
SPCDevice::TransferCommand()
{
	putlog(2, "SCMD Transfer Command (xfer=%s len=0x%x)",
		SCSI::GetXferPhaseName(GetPCTL() & 7), spc.tc);

	// SCMD レジスタ書き込みのうち Transfer コマンドに関わる未実装機能の
	// ログ表示は SCMD レジスタ書き込み時ではなくここで表示。
	if ((spc.scmd & SPC::SCMD_TERM_MODE)) {
		putlog(0, "SCMD Padding mode not implemented");
	}
	if ((spc.scmd & SPC::SCMD_INTERCEPT)) {
		putlog(0, "SCMD Intercept Transfer not implemented");
	}

	transfer_command_running = true;
	spc.ssts |= SPC::SSTS_SPC_BUSY | SPC::SSTS_IN_PROGRESS;

	// 転送開始前に DREG キューを初期化する。
	// MPU 側が DREG を全部読まずにフェーズを中断とかはたぶん出来るので。
	// XXX 転送開始前なのかリセットなのか…
	spc.dreg.Clear();

	// Transfer コマンド発行時に REQ がアサートされていなければ
	// アサートされるまで待つ。この場合、ここでは何もすることはなく、
	// TransferReqCB() が呼ばれてそこから転送フェーズが動き始める。
	if (GetREQ()) {
		// コマンドから転送開始までちょっとだけかかる
		// XXX 4 * t_CLF だと NetBSD カーネルがレースコンディションにはまる。
		CallAfter(TransferCommandStart, "SPC SCMD to Start", 0 * t_CLF);
	}
}

// バスフリーコールバック (SCSIBus から呼ばれる)
// id はこのバスフリーを行ったデバイス (0..7)
void
SPCDevice::BusFreeCB(int id)
{
	putlog(4, "BusFree Callback");

	inherited::BusFreeCB(id);
	last_busfree_time = gScheduler->GetVirtTime();

	transfer_command_running = false;
	spc.ssts &= ~(SPC::SSTS_CONN_INIT | SPC::SSTS_CONN_TARG);
	spc.ssts &= ~(SPC::SSTS_SPC_BUSY | SPC::SSTS_IN_PROGRESS);

	// マニュアルが分かりにくいんだけど、
	// o イニシエータとして動作時に、SPC が自らバスフリーにした時
	// o BusFree INT enable ビットが立っていて、ターゲットがバスフリーに
	//   した時
	// のいずれかで INTS のバスフリー検出ビットを立てる
	if (id != myid || spc.busfree_intr_enable) {
		SetINTS(SPC::INTS_DISCONNECTED);
	}
}

// ターゲットがセレクションに応答したコールバック (SCSIBus から呼ばれる)
void
SPCDevice::SelectionAckCB()
{
	// ターゲットが BSY を立てて応答したので、
	// こちらはデータバスをクリアして SEL を下げる。
	// これによりセレクションフェーズは成功で完了する。
	putlog(4, "SelectionAck Calback");
	SetData(0);
	NegateSEL();

	// タイムアウトイベントを停止
	event_tc.Stop();

	spc.ssts &= ~SPC::SSTS_SPC_BUSY;
	spc.ssts |= SPC::SSTS_CONN_INIT;

	// Command Complete 割り込みを起こす
	SetINTS(SPC::INTS_COMPLETE);
}

// 転送フェーズでターゲットが REQ を立てたコールバック (SCSIBus から呼ばれる)
void
SPCDevice::TransferReqCB()
{
	// REQ が立てば Transfer Command 発行前でも Xfer In Progress は立てる
	spc.ssts |= SPC::SSTS_IN_PROGRESS;

	if (transfer_command_running) {
		// ハードウェア転送で SDGC_XFER_ENABLE がセットされていれば、
		// データ要求が来ている時に XFER_OUT をセット(?)
		if ((spc.sdgc & SPC::SDGC_XFER_ENABLE)) {
			spc.serr |= SPC::SERR_XFER_OUT;
		}

		// Transfer Command 実行中なら REQ 待ちだったので
		// ここでハードウェア転送を開始。
		TransferCommandStart(0);
	} else {
		// ハードウェア転送中でなくて IN 方向ならここで TEMP に取り込む。
		if (GetIO()) {
			spc.temp_in = GetData();
		}
	}
}

// Transfer Command の実行開始部分。
// TransferCommand() と TransferReqCB() から呼ばれる。
void
SPCDevice::TransferCommandStart(int dummy)
{
	// Transfer コマンドの発行と REQ がアサートの両方が成立した時点で、
	// SCSI バスの状態と PCTL レジスタの下位3ビットの状態を比較する。
	// 一致していれば転送開始、そうでなければ Service Required 割り込み。
	if ((GetPCTL() & 7) == GetXfer()) {
		HardwareTransfer(0);
	} else {
		transfer_command_running = false;
		SetINTS(SPC::INTS_SERV_REQ);
	}
}

// ハードウェア転送の1バイト分。
// イベントコールバックとして呼び出すこと。
// REQ がアサートされているのを確認してから呼び出すこと。
// (1バイト送信するとターゲットが再び REQ を上げるなどしてくるので、
// TransferReqCB() か DREG 書き込みから再びここが呼ばれる)
void
SPCDevice::HardwareTransfer(int dummy)
{
	// SDGC_XFER_ENABLE は転送要求ごとに割り込みを上げる。
	// (SPC から見て) 入力操作時は読み込めるデータがある場合、出力操作時は
	// キューにバイトを書き込む必要がある場合に、割り込みを上げる。

	// DREG が送受信不可能ならここでは何もせずに帰るだけでよい。
	// その後 DREG アクセスによって送受信可能になればまたこれが呼ばれる。
	if (GetIO()) {
		// Input (ターゲット → イニシエータ)
		bool queued = spc.dreg.Enqueue(GetData());
		// SDGC_XFER_ENABLE なら読み込めるデータがある場合に割り込み。
		if ((spc.sdgc & SPC::SDGC_XFER_ENABLE)) {
			putlog(3, "SDGC XFER_OUT Interrupt");
			SetINTS(0);
		}
		if (!queued) {
			putlog(4, "Hardware Transfer (DREG full)");
			return;
		}
	} else {
		// Output (イニシエータ → ターゲット)
		if (spc.dreg.Length() != 0) {
			SetData(spc.dreg.Dequeue());
		} else {
			putlog(4, "Hardware Transfer (DREG empty)");
			// SDGC_XFER_ENABLE ならキューに書き込む必要がある場合に割り込み。
			if ((spc.sdgc & SPC::SDGC_XFER_ENABLE)) {
				putlog(3, "SDGC XFER_OUT Interrupt");
				SetINTS(0);
			}
			return;
		}
	}
	// SDGC XFER_OUT のログと順序が微妙になるけど、
	// DREG full/empty で2行出るのを避けたいので。
	putlog(4, "Hardware Transfer");

	// XXX XFER_OUT はどこで落とすんやろうか。
	// データ要求が来たら上げるのだから用が済んだら下ろすのかしら。
	spc.serr &= ~SPC::SERR_XFER_OUT;

	AssertACK();

	// 転送完了
	spc.tc--;
	putlog(4, "Hardware Transfer: %s q=%d tc=%d",
		(GetIO() ? "IN" : "OUT"), spc.dreg.Length(), spc.tc);
	if (spc.tc == 0) {
		// どちらでも転送自体はここで完了(?)
		transfer_command_running = false;

		if (GetIO()) {
			// Input 時は TC がゼロになって DREG キューを全部読み出した
			// ところで完了なので、ここでは何もしない。
		} else {
			// Output ならここで転送完了。
			TransferComplete();
		}
	}

	// メッセージアウトフェーズの最終バイト送信中(?)に ATN を下げる
	if (GetXfer() == SCSI::XferPhase::MsgOut && spc.tc == 0) {
		NegateATN();
	}

	if (GetXfer() == SCSI::XferPhase::MsgIn && spc.tc == 0) {
		// ただしメッセージインフェーズの最終バイトなら ACK を下げない。
	} else {
		NegateACK();
	}
}

// 転送完了。
// HardwareTransfer() と Read() から呼ばれる。
void
SPCDevice::TransferComplete()
{
	spc.ssts &= ~(SPC::SSTS_SPC_BUSY | SPC::SSTS_IN_PROGRESS);
	SetINTS(SPC::INTS_COMPLETE);
}
