//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// LUNA88K のシステムコントローラのような何か

#include "sysctlr.h"

// この辺を全部一つのデバイスで受け付けておくか。BusIO はまた後で考える。
// $6500000n.b	HW 割り込みレベル (n*4 = CPU#n)
//   $65000000.b	CPU#0
//   $65000004.b	CPU#1
//   $65000008.b	CPU#2
//   $6500000c.b	CPU#3
// $6900000n	ソフトウェア割り込み(?)
// $6d00000n	CPU#n をリセット(?)
// $6d000010	全 CPU をリセット

// intmask[] について:
// I/O ポート 0x6500_00xx での設定値は、
// bit 31 30 29 28 27 26 25...
// lvl 6  5  4  3  2  1  (not used)
// の対応になっている。intmask では 25bit 右シフトして、
// ビット番号が割り込みレベルに対応するようにマップした。


// LUNA88K
// 割り込みコントロールレジスタ
// $6500_000n
//  n: CPU ID * 4 (0,4,8,c)
//  この16Mページ内はミラーされている模様。
//  下位2bitしかデコードしていない?
// 書き込み時:
//    3                   2                   1
//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
// +-+-+-+-+-+-+---------------------------------------------------+
// |6|5|4|3|2|1|                                                   |
// +-+-+-+-+-+-+---------------------------------------------------+
// bit31: 1=割り込み 6 イネーブル
// bit30: 1=割り込み 5 イネーブル
// bit29: 1=割り込み 4 イネーブル
// bit28: 1=割り込み 3 イネーブル
// bit27: 1=割り込み 2 イネーブル
// bit26: 1=割り込み 1 イネーブル
// m68k と異なり、個別の割り込みごとに許可/禁止できる。
// 該当ビットに 1 を書き込むと許可。
// 空欄部分は無視される模様。
//
// 読み込み時
//    3                   2                   1
//  1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0 9 8 7 6 5 4 3 2 1 0
// +-----+---------+-+-+-+-+-+-+-----------------------------------+
// |Level|         |6|5|4|3|2|1|                                   |
// +-----+---------+-+-+-+-+-+-+-----------------------------------+
// bit31-29: 発生している最上位レベルの割り込み。範囲は7-0
// bit23-18: 書き込んだ割り込みマスク。
// 空欄部分は「不定」の模様。
//
// bit31-29 のレベルは、割り込みコントローラに入力
// されている割り込み信号線6本を、マスク処理した後、
// プライオリティバイナリエンコーダの値がそのまま出力されている模様。
// ラッチ機構は無いようだ。
// したがってマスクされている割り込み(マスクビットが 0) の
// 値が出力されることはない。
// 同時に入力があった場合は、数字の大きいほうが出力される。
// レベル7 はマスクなしでプライオリティエンコーダに入力されているようだ。
//
// 割り込み入力がないときは 0 になる模様。
// 本体 ABRT スイッチを押すと、押している間、すべての CPU の
// level が 111(=7) になる。離すと戻る。
// 88100 には NMI が無いので、通常の割り込みと同じように処理している？
// (なので IND 状態だとABRT できない。いいのかそれで)
// (OpenBSD6.7 の割り込みハンドラはABRTには対応していないように読める)
//
// 想定される回路
//
// [割り込みソース]
//  1:SOFTINT レジスタ $6900000n
//  |   2:(XP 低レベル?)
//  |   |   3:SPC
//  |   |   |   4:(LANCE|拡張スロット?)
//  |   |   |   |   5:(拡張PC98?|シリアル?)
//  |   |   |   |   |   6:(XP 高レベル?|システムクロック?)
//  |   |   |   |   |   |   7:ABRTスイッチ
//  V   V   V   V   V   V   |
// +---------------------+  |
// |   割り込みマスク    |  |  (R:bit23-18, W:bit31-26)
// +---------------------+  |
//  |   |   |   |   |   |   |
//  V   V   V   V   V   V   V
// +-------------------------+
// |プライオリティエンコーダ |-> (R:bit31-29)
// +-------------------------+
//       (入力のWired-OR)
//           |
//       [M88K INT]
//
// LUNA-I とはレベル配置が違う。1 レベルのズレに加え、シリアルが
// 6 -> 5 に変更されているようだ？
//
// 割り込みソース調査状況
// 1:SOFTINT レジスタ $6900000n
//   ROM のメモリコマンドで確認。OpenBSD machdep.c の記述の通り。
// 2:XP 低レベル
//   LUNA-I からの類推及び、OpenBSD mainbus.c に記述あり。OpenBSD のほうは
//   低レベルかどうかは不明。
// 3:SPC
//   ROM1.2, OpenBSD ともに使用。
// 4:LANCE
//   ROM はまだ不明。OpenBSD は使用。
// 4:拡張スロット
//   LUNA-I からの類推のみ。不明。
// 5:拡張PC98
//   LUNA-I からの類推のみ。LUNA88K にはスロットがないようなので、
//   LUNA88K2 が必要。
//   OpenBSD mainbus.c に記述あり。
// 5:シリアル(7201)
//   OpenBSD mainbus.c に記述。
// 6:XP 高レベル
//   LUNA-I からの類推。存在するかどうかも不明。
//   ただ、これをなくすと XP ファームウェアまで作り直しなので、
//   なくさないんじゃないかなあ。OMRON ファームで使ってなければ
//   さっくり廃止されているかもしれない(PSGPAM が使用...レベル1 に
//   変えればいいだけだが)
// 6:システムクロック
//   ROM のメモリコマンドではレベル6割り込みが常時上がっているように観測
//   できるので、おそらくシステムクロック割り込みが上がっている。
//   OpenBSD maibus.c にも記述。
// 7:ABRTスイッチ
//   ROM のメモリコマンドで確認。
//
//
// SOFTINT レジスタ $6900000n
//  $69000000: CPU0
//  $69000004: CPU1
//  $69000008: CPU2
//  $6900000c: CPU3
// 挙動:
// この32ビットのどこかにライトアクセスすると、INT1 をアサートする。
// この32ビットのどこかをリードアクセスすると、INT1 をネゲートする。
// データバスは無視される。
// 0x00 を書いても、アサートされることに注意。
// リードアクセスで読み込まれる値も不定のようだ。
// 要はアドレスバスとR/W信号線しか見ていない。


std::unique_ptr<SysCtlrDevice> gSysCtlr;


// コンストラクタ
SysCtlrDevice::SysCtlrDevice()
{
	logname = "sysctlr";
	devname = "SysCtlr";

	monitor_size = nnSize(60, 5);
}

// デストラクタ
SysCtlrDevice::~SysCtlrDevice()
{
}

uint64
SysCtlrDevice::Read32(uint32 addr)
{
	int n = (addr & 15) >> 2;

	// 読み込みでは割り込みレベルを返すようだ
	// bit 31..29: アクティブな割り込みのレベル
	//     24 level7 割り込みが発生したら 1
	//     23 level6 ..
	//     22 level5
	//     21 level4
	//     20 level3
	//     19 level2
	//     18 level1
	// XXX: もう資料がないとわからない
	if (0x65000000 <= addr && addr < 0x65000010) {
		int maxlev = 0;
		// 7 to 1
		for (int i = 7; i > 0; i--) {
			if (intlevel[n] & (1 << i)) {
				maxlev = i;
				break;
			}
		}

		uint32 data = (maxlev << 29) | (intlevel[n] << 17);
		putlog(2, "割り込み保持読み出し(CPU#%d) -> $%08x", n, data);
		return data;
	}

	if (0x69000000 <= addr && addr < 0x69000010) {
		putlog(1, "ソフトウェア割り込み(CPU#%d)クリア(未実装)", n);
		return 0xffffffff;
	}

	if (0x6d000000 <= addr && addr < 0x6d000010) {
		putlog(1, "CPU%d リセット 未実装読み込み", n);
		return 0xffffffff;
	}
	if (addr == 0x6d000010) {
		putlog(1, "CPU ALL リセット 未実装読み込み");
		return 0xffffffff;
	}

	putlog(0, "未実装読み込み $%08x", addr);
	return 0xffffffff;
}

uint64
SysCtlrDevice::Write32(uint32 addr, uint32 data)
{
	int n = (addr & 15) >> 2;

	if (0x65000000 <= addr && addr < 0x65000010) {
		intmask[n] = data >> 25;

		// XXX: クリアしてみる
		intlevel[n] = 0;

		putlog(1, "割り込みマスク(CPU#%d) <- $%08x", n, intmask[n] << 25);
		return 0;
	}
	if (0x69000000 <= addr && addr < 0x69000010) {
		// 詳細不明。0xFFFFFFFF が書き込まれている、ようだ。
		intlevel[n] &= ~data;
		// bit 表示したい
		putlog(0, "割り込み(CPU#%d)クリア 0x%x 残り=0x%x", n, data, intlevel[n]);
		return 0;
	}
	if (0x6d000000 <= addr && addr < 0x6d000010) {
		putlog(0, "CPU%d リセット 未実装書き込み", n);
		return 0;
	}
	if (addr == 0x6d000010) {
		putlog(0, "CPU ALL リセット 未実装書き込み");
		return 0;
	}

	putlog(0, "未実装書き込み $%08x <- $%08x", addr, data);
	return 0;
}

uint64
SysCtlrDevice::Peek8(uint32 addr)
{
	uint32 data;
	int n = (addr & 15) >> 2;

	if (0x65000000 <= addr && addr < 0x65000010) {
		data = intmask[n] << 25;
	} else {
		data = 0xffffffff;
	}
	switch (addr & 3) {
	 case 0:
		return data >> 24;
	 case 1:
		return (data >> 16) & 0xff;
	 case 2:
		return (data >>  8) & 0xff;
	 case 3:
		return data & 0xff;
	 default:
		break;
	}
	__unreachable();
}

void
SysCtlrDevice::MonitorUpdate(TextScreen& monitor)
{
	monitor.Clear();
	monitor.Puts(6, 0, "IntrEnable     SWIntr");
	for (int i = 0; i < countof(intmask); i++) {
		monitor.Print(0, i + 1, "CPU#%d  $%02x", i, intmask[i]);
	}
}
