//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "upd7201.h"

// コンストラクタ
uPD7201Device::uPD7201Device()
{
	monitor_size = nnSize(48, 19);
	cr.chan[0].id = 0;
	cr.chan[0].name = "A";
	cr.chan[1].id = 1;
	cr.chan[1].name = "B";

	cr.chan[0].desc = NULL;
	cr.chan[1].desc = NULL;
}

// デストラクタ
uPD7201Device::~uPD7201Device()
{
}

// リセット
// どのタイミングで呼ばれるかは上位による
void
uPD7201Device::Reset()
{
	putlog(2, "リセット");

	// CR2A はチップリセットで bit6 以外を '0' に
	cr.cr2a &= ~0xbf;

	// チャンネルリセット
	ResetChannel(&cr.chan[0]);
	ResetChannel(&cr.chan[1]);
}

// チャンネルリセット
void
uPD7201Device::ResetChannel(struct SIO::channel *chan)
{
	putlog(2, "チャンネル %s リセット", chan->name);

	// CR0 b2-0
	chan->cr0 &= ~SIO::CR0_PTR;
	chan->ptr = 0;

	// CR1 b0,1,3,4,7
	chan->cr1 &= ~(SIO::CR1_WAIT_EN | SIO::CR1_RXINT |
	               SIO::CR1_TXINT_EN | SIO::CR1_ESINT_EN);

	// CR3 b0-7
	chan->cr3 = 0;

	// CR5 b1,2,3,4,7
	chan->cr5 &= ~(SIO::CR5_DTR | SIO::CR5_SENDBREAK | SIO::CR5_TX_EN |
	               SIO::CR5_CRC16 | SIO::CR5_RTS);

	// SR0 は b0,1 が '0'、b2,6 が '1'、残りは不定
	chan->sr0 &= ~(SIO::SR0_INTPEND | SIO::SR0_RXAVAIL);
	chan->sr0 |= (SIO::SR0_TXUNDER | SIO::SR0_TXEMPTY);

	// SR1 は b4-7 が '0'
	chan->sr1 &= ~(SIO::SR1_ENDOFFRAME | SIO::SR1_CRCERROR |
	               SIO::SR1_RXOVERRUN | SIO::SR1_PARITYERROR);

	// E/S ラッチ解除
	chan->es_ratched = false;

	// 受信バッファをクリア
	memset(chan->rxbuf, 0, sizeof(chan->rxbuf));
	chan->rxlen = 0;
}

// 制御ポートの読み込み
uint8
uPD7201Device::ReadCtrl(struct SIO::channel *chan)
{
	uint8 data;

	switch (chan->ptr) {
	 case 0:	// SR0
		data = ReadSR0(chan);
		putlog(2, "SR%d%s -> $%02x", chan->ptr, chan->name, data);
		// アクセス後は SR0 へ戻す
		chan->ptr = 0;
		break;

	 case 1:	// SR1
		data = ReadSR1(chan);
		putlog(2, "SR%d%s -> $%02x", chan->ptr, chan->name, data);
		// アクセス後は SR0 へ戻す
		chan->ptr = 0;
		break;

	 case 2:
		// SR2B はベクタの読み出し
		if (chan->id == 1) {
			data = cr.vector;
			putlog(2, "SR%d%s -> $%02x", chan->ptr, chan->name, data);
			// アクセス後は SR0 へ戻す
			chan->ptr = 0;
			break;
		}

		// SR2A はない
		FALLTHROUGH;

	 default:
		putlog(0, "未実装レジスタ読み込み SR%d%s",
			chan->ptr, chan->name);
		// SR0 へ戻る?
		chan->ptr = 0;
		return 0xff;
	}
	return data;
}

// 制御ポートの書き込み
void
uPD7201Device::WriteCtrl(struct SIO::channel *chan, uint32 data)
{
	uint cmd;

	switch (chan->ptr) {
	 case 0:	// CR0
		// XXX b7,b6 は未対応。
		// b5,b4,b3 がコマンド。
		// b2,b1,b0 が次にアクセスするレジスタ(chan->ptr)の切り替え。
		chan->cr0 = data;
		chan->ptr = (chan->cr0 & SIO::CR0_PTR);

		cmd = (chan->cr0 & SIO::CR0_CMD) >> 3;
		if (cmd == 0) {
			// コマンド0 ならレジスタ切り替え
			putlog(3, "CR0%s <- $%02x (ptr=%d)", chan->name, data, chan->ptr);
		} else {
			// 0 以外ならコマンド発行
			putlog(2, "CR0%s <- $%02x", chan->name, data);
		}
		switch (cmd) {
		 case 0:	// No operation
			break;
		 case 1:	// Send Abort
			if (chan->mode == SIO::Mode::HDLC) {
				putlog(0, "CR0: Send Abort 未実装");
			} else {
				// XXX Send Abort コマンドは「HDLC モードのみ有効」と書いて
				//     あるがそれ以外の時にどうなるか不明。
				putlog(0, "CR0: 未定義コマンド %d", cmd);
			}
			break;
		 case 2:	// Reset External/Status Interrupt
			putlog(3, "E/S ラッチ解除");
			chan->es_ratched = false;
			break;
		 case 3:	// Channel Reset
			putlog(1, "チャンネル%sリセット", chan->name);
			ResetChannel(chan);
			break;
		 case 4:	// Enable Interrupt On Next Receive Character
			putlog(0, "CR0: 未実装コマンド %d", cmd);
			break;
		 case 5:	// Reset Transmitter Interrupt/DMA Pending
			// 本来これは上がっている Tx 割り込みを落とすためのようだが、
			// SIO から MPU への割り込み通知はもう終わっていて下げる必要は
			// ないので、何もすることはない。
			break;
		 case 6:	// Error Reset
		 case 7:	// End of Interrupt
			putlog(0, "CR0: 未実装コマンド %d", cmd);
			break;
		 default:
			__unreachable();
		}
		break;

	 case 1:	// CR1
		chan->cr1 = data;
		putlog(2, "CR%d%s <- $%02x", chan->ptr, chan->name, data);
		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;

	 case 2:
		if (chan->id == 0) {	// CR2A
			putlog(2, "CR2A <- $%02x 書き込み(b7-2未実装)", data);
			cr.cr2a = data;
		} else {				// CR2B
			putlog(2, "CR2B <- $%02x ベクタ設定", data);
			cr.vector = data;
		}
		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;

	 case 3:	// CR3
		chan->cr3 = data;
		putlog(2, "CR%d%s <- $%02x", chan->ptr, chan->name, data);
		putlog(1, "チャンネル%s RX %s, %dbits",
			chan->name,
			(chan->cr3 & SIO::CR3_RX_EN) ? "Enable" : "Disable",
			bits_per_char[(chan->cr3 & SIO::CR3_RXBITS) >> 6]);
		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;

	 case 4:	// CR4
	 {
		chan->cr4 = data;
		putlog(2, "CR%d%s <- $%02x", chan->ptr, chan->name, data);
		uint8 stopbits = (chan->cr4 & SIO::CR4_STOPBITS);
		uint8 syncmode = (chan->cr4 & SIO::CR4_SYNCMODE);

		// モード分け、優先順位は分からない
		if (stopbits == SIO::SYNC_MODE) {
			if (syncmode == SIO::SYNC_HDLC) {
				chan->mode = SIO::Mode::HDLC;
				putlog(2, "チャンネル%s HDLC モード", chan->name);
			} else {
				chan->mode = SIO::Mode::SYNC;
				const char *sm[] = { "8bit", "16bit", "HDLC?", "Ext" };
				putlog(2, "チャンネル%s Sync (%s) モード",
					chan->name, sm[syncmode >> 4]);
			}
		} else {
			chan->mode = SIO::Mode::ASYNC;
			const char *cm[] = { "x1", "x16", "x32", "x64" };
			const char *sm[] = { "?", "1", "1.5", "2" };
			putlog(2, "チャンネル%s Async (%s clock, stopbit %s) モード",
				chan->name,
				cm[(chan->cr4 & SIO::CR4_CLKRATE) >> 6],
				sm[stopbits >> 2]);
		}

		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;
	 }

	 case 5:	// CR5
	 {
		chan->cr5 = data;
		putlog(2, "CR%d%s <- $%02x", chan->ptr, chan->name, data);
		putlog(1, "チャンネル%s TX %s, %dbits",
			chan->name,
			(chan->cr5 & SIO::CR5_TX_EN) ? "Enable" : "Disable",
			bits_per_char[(chan->cr5 & SIO::CR5_TXBITS) >> 5]);
		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;
	 }

	 case 6:	// CR6
		putlog(0, "CR6%s <- $%02x 書き込み(未実装)", chan->name, data);
		chan->cr6 = data;
		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;

	 case 7:	// CR7
		putlog(0, "CR7%s <- $%02x 書き込み(未実装)", chan->name, data);
		chan->cr7 = data;
		// アクセス後は CR0 へ戻す
		chan->ptr = 0;
		break;

	 default:
		putlog(0, "未実装レジスタ書き込み CR%d%s <- $%02x",
			chan->ptr, chan->name, data);
		break;
	}
}

// 制御ポートの Peek
uint8
uPD7201Device::PeekCtrl(const struct SIO::channel *chan) const
{
	switch (chan->ptr) {
	 case 0:	// SR0
		return ReadSR0(chan);

	 case 1:	// SR1
		return ReadSR1(chan);

	 case 2:	// SR2
		// SR2A はない。SR2B はベクタの読み出し
		if (chan->id == 1) {
			return cr.vector;
		}
		FALLTHROUGH;

	 default:
		// 未調査
		return 0xff;
	}
}

// SR0 の読み込み (Read, Peek 共通)
uint8
uPD7201Device::ReadSR0(const struct SIO::channel *chan) const
{
	uint32 data;

	// 送信バッファは常に空
	data = SIO::SR0_TXEMPTY;
	// 受信データがあるか
	if (chan->rxlen > 0) {
		data |= SIO::SR0_RXAVAIL;
	}
	return data;
}

// SR1 の読み込み (Read, Peek 共通)
uint8
uPD7201Device::ReadSR1(const struct SIO::channel *chan) const
{
	// エラーとか起きてないという状態だけ再現。
	return SIO::SR1_ALL_SENT;
}


// データポートの読み込み
uint8
uPD7201Device::ReadData(struct SIO::channel *chan)
{
	uint8 data;

	if (chan->rxlen > 0) {
		data = chan->rxbuf[0];
		chan->rxlen--;
		// 少ないので毎回前詰めしちゃえ
		if (chan->rxlen > 0) {
			memmove(chan->rxbuf + 1, chan->rxbuf, chan->rxlen);
		}
	} else {
		// XXX 何が読めるか
		data = 0xff;
	}

	return data;
}

// データポートの書き込み
void
uPD7201Device::WriteData(struct SIO::channel *chan, uint32 data)
{
	char charbuf[4];

	if (isprint((int)data)) {
		snprintf(charbuf, sizeof(charbuf), "'%c'", data);
	} else {
		charbuf[0] = '\0';
	}
	putlog(1, "Ch%s データレジスタ書き込み $%02x %s",
		chan->name, data, charbuf);

	// 1文字送信

	// 今のところ内部バッファがなく常に送信できるので
	// 必ず Tx Buffer Empty が発生する。
	chan->sr0 |= SIO::SR0_TXEMPTY;

	// 割り込みをあげる
	if ((chan->cr1 & SIO::CR1_TXINT_EN)) {
		putlog(2, "割り込み (TX Empty)");
		Interrupt();
	}
}

// データポートの Peek
uint8
uPD7201Device::PeekData(const struct SIO::channel *chan) const
{
	if (chan->rxlen > 0) {
		// 受信バッファにデータがあれば先頭バイトが見える
		return chan->rxbuf[0];
	} else {
		// 空の時は?
		return 0xff;
	}
}

// TX/RX の bits/char のレジスタ値を実際の bits/char の値に変換する。
/* static */
const int
uPD7201Device::bits_per_char[] = { 5, 7, 6, 8 };

// 1バイト受信する。
// 受理されれば true を返す。
bool
uPD7201Device::Rx(int ch, uint32 data)
{
	struct SIO::channel *chan = &cr.chan[ch];

	// Rx Enable でなければ受け取らない?
	if ((chan->cr3 & SIO::CR3_RX_EN) == 0) {
		return false;
	}

	// バッファが一杯なら Rx Overrun Error
	if (chan->rxlen == sizeof(chan->rxbuf)) {
		putlog(2, "Rx Overrun");
		chan->sr1 |= SIO::SR1_RXOVERRUN;
		return false;
	}

	// 受信
	chan->rxbuf[chan->rxlen++] = data;

	// 必要なら割り込みを上げる
	if ((chan->cr1 & SIO::CR1_RXINT) == SIO::RXINT_FIRSTCHAR) {
		// First character interrupt モードなら最初の1バイトでだけ割り込み
		if (chan->rxrecved == false) {
			Interrupt();
		}
		chan->rxrecved = true;
	} else if ((chan->cr1 & SIO::CR1_RXINT) >= SIO::RXINT_ALL_INC_SPECIAL) {
		// All character interrupt モード (Special のことはちょっと置いとく)
		chan->rxrecved = true;
		Interrupt();
	}

	return true;
}

// モニター
void
uPD7201Device::MonitorUpdate(TextScreen& monitor)
{
	monitor.Clear();
	for (int i = 0; i < 2; i++) {
		int x = 0;
		int y = i * 10;
		MonitorUpdateCh(monitor, i, x, y);
	}
}

// モニター (1チャンネル)
void
uPD7201Device::MonitorUpdateCh(TextScreen& monitor,
	int ch, int xbase, int ybase)
{
	struct SIO::channel *chan = &cr.chan[ch];
	uint cr0;
	uint cr1;
	uint cr3;
	uint cr4;
	uint cr5;
	uint sr0;
	uint sr1;
	uint rxlen;
	int x;
	int y;

	cr0 = chan->cr0;
	cr1 = chan->cr1;
	cr3 = chan->cr3;
	cr4 = chan->cr4;
	cr5 = chan->cr5;
	sr0 = chan->sr0;
	sr1 = chan->sr1;
	rxlen = chan->rxlen;

	// 0         1         2         3         4         5         6
	// 012345678901234567890123456789012345678901234567890123456789012345
	//          0123 0123 0123 0123 0123 0123 0123 0123
	// CR0:$xx  CRC=x     CMD=x          PTR=x
	// CR1:$xx  WE   0    WR   RXInt=x   0    TXEn ESEn

	x = xbase;
	y = ybase;
	monitor.Print(x, y++, "Channel %s: %s", chan->name, chan->desc);
	monitor.Print(x, y++, "CR0:$%02x", cr0);
	monitor.Print(x, y++, "CR1:$%02x", cr1);
	monitor.Print(x, y++, "CR3:$%02x", cr3);
	monitor.Print(x, y++, "CR4:$%02x", cr4);
	monitor.Print(x, y++, "CR5:$%02x", cr5);
	monitor.Print(x, y++, "SR0:$%02x", sr0);
	monitor.Print(x, y++, "SR1:$%02x", sr1);
	monitor.Print(x, y,   "RXbuf:%d", rxlen);
	int i;
	for (i = 0; i < rxlen; i++) {
		monitor.Print(x + 8 + i * 4, y, "$%02x", chan->rxbuf[i]);
	}
	for (; i < sizeof(chan->rxbuf); i++) {
		monitor.Print(x + 8 + i * 4, y, "---");
	}

	y = ybase + 1;
	x = xbase;
	// CR0
	monitor.Print(x + 9, y, "CRC=%d", (cr0 >> 6));
	static const char * const cmd_str[] = {
	//   0123456789012345
		"No operation",
		"Send Abort",
		"Reset E/S Int",
		"Channel Reset",
		"Enable IntNextCh",
		"Reset TxIntPend",
		"Error Reset",
		"End of Interrupt",
	};
	uint cmd = (cr0 >> 3) & 7;
	monitor.Print(x + 19, y, "CMD=%d(%s)", cmd, cmd_str[cmd]);
	monitor.Print(x + 43, y, "PTR=%d", (cr0 & 7));
	y++;

	// CR1
	static const char * const cr1_str[] = {
		"WaEn", "----", "WaRx", "", "", "----", "TxIE", "ESEn"
	};
	MonitorReg(monitor, x + 9, y, cr1, cr1_str);
	monitor.Print(x + 24, y, "RXInt=%d", (cr1 >> 3) & 3);
	y++;

	// CR3
	static const char * const cr3_str[] = {
		"", "", "AuEn", "----", "----", "----", "----", "RXEn"
	};
	MonitorReg(monitor, x + 9, y, cr3, cr3_str);
	monitor.Print(x + 9, y, "RXbits=%d", 5 + ((cr3 >> 6) & 3));
	y++;

	// CR4
	static const char * const cr4_str[] = {
		"", "", "", "", "", "", "PaEv", "PaEn"
	};
	MonitorReg(monitor, x + 9, y, cr4, cr4_str);

	static const char * const clkrate_str[] = {
		"1", "16", "32", "64"
	};
	monitor.Print(x + 9, y, "Rate=x%s", clkrate_str[(cr4 >> 6) & 3]);
	static const char * const syncmode_str[] = {
		"Mono", "Bi", "HDLC", "Ext"
	};
	monitor.Print(x + 19, y, "Sync=%s", syncmode_str[(cr4 >> 4) & 3]);
	static const char * const stopbit_str[] = {
		"Sync", "1", "1.5", "2"
	};
	monitor.Print(x + 29, y, "Stop=%s", stopbit_str[(cr4 >> 2) & 3]);
	y++;

	// CR5
	static const char * const cr5_str[] = {
		"DTR", "", "", "SBrk", "TXEn", "CRC", "!RTS", "TCEn"
	};
	MonitorReg(monitor, x + 9, y, cr5, cr5_str);
	monitor.Print(x + 14, y, "TXbits=%d", 5 + ((cr5 >> 5) & 3));
	y++;

	// SR0
	static const char * const sr0_str[] = {
		"BrAb", "TxUn", "CTS", "SYNC", "DCD", "TxEm", "IntP", "RxAv"
	};
	MonitorReg(monitor, x + 9, y, sr0, sr0_str);
	y++;

	// SR1
	static const char * const sr1_str[] = {
		"EOF", "CRCE", "RxOv", "PaEr", "", "", "", "Sent"
	};
	MonitorReg(monitor, x + 9, y, sr1, sr1_str);
	monitor.Print(x + 29, y, "ResidueCode=%d", (sr1 >> 1) & 7);
	y++;
}

// レジスタをビットごとに表示する。MonitorUpdate の下請け。
void
uPD7201Device::MonitorReg(TextScreen& monitor,
	int x, int y, uint32 reg, const char * const *names)
{
	for (int i = 0; i < 8; i++) {
		if (strcmp(names[i], "----") == 0) {
			monitor.Puts(x + i * 5, y, TA::Disable, names[i]);
		} else {
			bool b = reg & (1 << (7 - i));
			monitor.Puts(x + i * 5, y, TA::OnOff(b), names[i]);
		}
	}
}
