//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

// キーボードは MFP の USART に繋がっている。
// MFP USART の RR 線 (Receiver Ready) は受信可能なら 'H'、受信中は 'L' に
// なっている。キーボードは RR が 'H' の時だけデータを送信する。
// また、システムポート#4 の KEY CTRL = 0 (キーデータ送信不可) はこの RR を
// 'L' に落とす論理回路になっている。
// キーボードからの送信は 2400bps、ストップビット1、パリティなしなので、
// 1バイトの送信に 3.75 msec かかる。
// 実装では、キーボードは MFP の RR が 'H' なら MFP にキーデータを送信。
// MFP 側で RR を 'L' にして 3.75 msec をイベントで待った後、キーを処理して
// RR を 'H' に戻すことにする。

#include "x68kkbd.h"
#include "mfp.h"

// コンストラクタ
X68030Keyboard::X68030Keyboard()
{
	led.resize(7);
}

// デストラクタ
X68030Keyboard::~X68030Keyboard()
{
}

// 本体リセットスイッチによるリセット
void
X68030Keyboard::ResetHard()
{
	rept_delay = 500;
	rept_time  = 110;

	for (int i = 0; i < led.size(); i++) {
		led[i] = false;
	}
	led_darkness = 0;
}

// キーを押した
void
X68030Keyboard::MakeKey(uint keycode)
{
	uint x68kkey;

	// 共通キーコードから X680x0 キーコードに
	x68kkey = keycode2x68kkey(keycode);

	// X680x0 にないキーは無視
	if (x68kkey == NoKey) {
		return;
	}

	// LED は MPU 側から点灯/消灯指示が来るのでここでは何もしなくてよい

	putlog(1, "MakeKey  %d(%s)", x68kkey, keyname[x68kkey]);
	SendKey(x68kkey);
}

// キーを離した
void
X68030Keyboard::BreakKey(uint keycode)
{
	uint x68kkey;

	// 共通キーコードから X680x0 キーコードに
	x68kkey = keycode2x68kkey(keycode);

	// X680x0 にないキーは無視
	if (x68kkey == NoKey) {
		return;
	}

	putlog(2, "BreakKey %d(%s)", x68kkey, keyname[x68kkey]);
	SendKey(x68kkey | 0x80);
}

// マウス入力
void
X68030Keyboard::MouseInput(int x, int y, bool rb, bool lb, bool mb)
{
}

// キーデータを送信
void
X68030Keyboard::SendKey(uint data)
{
	if (gMFP->IsRR()) {
		gMFP->Rx(data);
	}
}

// ホストからの制御
void
X68030Keyboard::Command(uint32 data)
{
	if (data < 0x40) {
		// $00-$3f TV 制御
		// (実装不要)
		putlog(2, "Command TV 制御 $%02x", (data & 0x1f));

	} else if (data < 0x48) {
		// $40-$47 MSCTRL 信号制御
		// こいつは頻度が高いのでログレベルを上げておく
		putlog(3, "Command MSCTRL 制御 %d", (data & 1));

	} else if (data < 0x50) {
		// $48-$4f キーデータ送出許可
		putlog(1, "Command キー送出許可 %d (未実装)", (data & 1));

	} else if (data < 0x54) {
		// $50-$53 ディスプレイ制御モード選択
		// (実装不要)
		putlog(2, "Command ディスプレイ制御 X68000モード %d", (data & 1));

	} else if (data < 0x58) {
		// $54-$57 LED 明るさ選択
		// 0 が明るく 3 が暗いので、明るさというより暗さ。
		led_darkness = (data & 3);
		putlog(1, "Command LED 明るさ %d", led_darkness);

	} else if (data < 0x5c) {
		// $58-$5b 本体からのディスプレイ制御許可
		// (実装不要)
		putlog(2, "Command 本体からディスプレイ制御 %d", (data & 1));

	} else if (data < 0x60) {
		// $5c-$5f OPT.2 によるディスプレイ制御許可
		// (実装不要)
		putlog(2, "Command OPT2からディスプレイ制御 %d", (data & 1));

	} else if (data < 0x70) {
		// $60-$6f リピートまでの時間
		rept_delay = 200 + (data & 15) * 100;
		putlog(1, "Command Repeat Delay %d (%d msec)", (data & 15), rept_delay);

	} else if (data < 0x80) {
		// $70-$7f リピート間隔
		rept_time = 30 + (data & 15) * (data & 15) * 5;
		putlog(1, "Command Repeat time  %d (%d msec)", (data & 15), rept_time);

	} else {
		// $80-$ff LED 制御
		data &= 0x7f;
		putlog(1, "Command LED 制御 $%02x", data);
		uint32 b = 0x01;
		for (int i = 0; i < led.size(); i++) {
			// data は %0 で点灯、led[] は true で点灯
			led[i] = ((data & b) == 0);
			b <<= 1;
		}
	}
}

// 共通キーコードを X680x0 キーコードに変換して返す。
// X680x0 キーがなければ NoKey を返す。
uint
X68030Keyboard::keycode2x68kkey(uint keycode)
{
	if (keycode >= countof(keycode2x68kkey_table)) {
		return NoKey;
	}

	return keycode2x68kkey_table[keycode];
}

// キーコード変換表
// 共通キーコードから X680x0 キーコードに変換する。
/*static*/ const uint
X68030Keyboard::keycode2x68kkey_table[133] = {
	NoKey,					// [00]
	X68kKey_BREAK,			// [01]
	X68kKey_COPY,			// [02]
	X68kKey_romaji,			// [03]
	X68kKey_code,			// [04]
	X68kKey_kigou,			// [05]
	X68kKey_touroku,		// [06]
	X68kKey_HELP,			// [07]
	X68kKey_HOME,			// [08]
	X68kKey_TAB,			// [09]
	X68kKey_CTRL,			// [0a]
	X68kKey_kana,			// [0b]
	X68kKey_SHIFT,			// [0c]
	NoKey,					// [0d]
	X68kKey_CAPS,			// [0e]
	NoKey,					// [0f]

	X68kKey_ESC,			// [10]
	X68kKey_BS,				// [11]
	X68kKey_enter,			// [12]
	X68kKey_INS,			// [13]
	X68kKey_space,			// [14]
	X68kKey_DEL,			// [15]
	NoKey,					// [16]
	NoKey,					// [17]
	NoKey,					// [18]
	NoKey,					// [19]
	NoKey,					// [1a]
	NoKey,					// [1b]
	X68kKey_up,				// [1c]
	X68kKey_left,			// [1d]
	X68kKey_right,			// [1e]
	X68kKey_down,			// [1f]

	X68kKey_ROLLUP,			// [20]
	X68kKey_ROLLDOWN,		// [21]
	X68kKey_1,				// [22]
	X68kKey_2,				// [23]
	X68kKey_3,				// [24]
	X68kKey_4,				// [25]
	X68kKey_5,				// [26]
	X68kKey_6,				// [27]
	X68kKey_7,				// [28]
	X68kKey_8,				// [29]
	X68kKey_9,				// [2a]
	X68kKey_0,				// [2b]
	X68kKey_minus,			// [2c]
	X68kKey_circum,			// [2d]
	X68kKey_backslash,		// [2e]
	X68kKey_UNDO,			// [2f]

	X68kKey_OPT1,			// [30]
	X68kKey_OPT2,			// [31]
	X68kKey_Q,				// [32]
	X68kKey_W,				// [33]
	X68kKey_E,				// [34]
	X68kKey_R,				// [35]
	X68kKey_T,				// [36]
	X68kKey_Y,				// [37]
	X68kKey_U,				// [38]
	X68kKey_I,				// [39]
	X68kKey_O,				// [3a]
	X68kKey_P,				// [3b]
	X68kKey_at,				// [3c]
	X68kKey_bracketleft,	// [3d]
	X68kKey_hira,			// [3e]
	X68kKey_zenkaku,		// [3f]

	NoKey,					// [40]
	NoKey,					// [41]
	X68kKey_A,				// [42]
	X68kKey_S,				// [43]
	X68kKey_D,				// [44]
	X68kKey_F,				// [45]
	X68kKey_G,				// [46]
	X68kKey_H,				// [47]
	X68kKey_J,				// [48]
	X68kKey_K,				// [49]
	X68kKey_L,				// [4a]
	X68kKey_semicolon,		// [4b]
	X68kKey_colon,			// [4c]
	X68kKey_bracketright,	// [4d]
	NoKey,					// [4e]
	NoKey,					// [4f]

	NoKey,					// [50]
	NoKey,					// [51]
	X68kKey_Z,				// [52]
	X68kKey_X,				// [53]
	X68kKey_C,				// [54]
	X68kKey_V,				// [55]
	X68kKey_B,				// [56]
	X68kKey_M,				// [57]
	X68kKey_N,				// [58]
	X68kKey_comma,			// [59]
	X68kKey_period,			// [5a]
	X68kKey_slash,			// [5b]
	X68kKey_underscore,		// [5c]
	NoKey,					// [5d]
	NoKey,					// [5e]
	NoKey,					// [5f]

	NoKey,					// [60]
	X68kKey_PAD_plus,		// [61]
	X68kKey_PAD_minus,		// [62]
	X68kKey_PAD_7,			// [63]
	X68kKey_PAD_8,			// [64]
	X68kKey_PAD_9,			// [65]
	X68kKey_PAD_4,			// [66]
	X68kKey_PAD_5,			// [67]
	X68kKey_PAD_6,			// [68]
	X68kKey_PAD_1,			// [69]
	X68kKey_PAD_2,			// [6a]
	X68kKey_PAD_3,			// [6b]
	X68kKey_PAD_0,			// [6c]
	X68kKey_PAD_decimal,	// [6d]
	X68kKey_PAD_enter,		// [6e]
	X68kKey_PAD_CLR,		// [6f]

	NoKey,					// [70]
	NoKey,					// [71]
	X68kKey_F1,				// [72]
	X68kKey_F2,				// [73]
	X68kKey_F3,				// [74]
	X68kKey_F4,				// [75]
	X68kKey_F5,				// [76]
	X68kKey_F6,				// [77]
	X68kKey_F7,				// [78]
	X68kKey_F8,				// [79]
	X68kKey_F9,				// [7a]
	X68kKey_F10,			// [7b]
	X68kKey_PAD_multiply,	// [7c]
	X68kKey_PAD_divide,		// [7d]
	X68kKey_PAD_equal,		// [7e]
	X68kKey_PAD_comma,		// [7f]

	X68kKey_XF1,			// [80]
	X68kKey_XF2,			// [81]
	X68kKey_XF3,			// [82]
	X68kKey_XF4,			// [83]
	X68kKey_XF5,			// [84]
};

/*static*/ const char * const
X68030Keyboard::keyname[] =
{
	"",			// 00
	"ESC",		// 01
	"1!",		// 02
	"2\"",		// 03
	"3#",		// 04
	"4$",		// 05
	"5%",		// 06
	"6&",		// 07
	"7'",		// 08
	"8(",		// 09
	"9)",		// 0a
	"0",		// 0b
	"-=",		// 0c
	"^~",		// 0d
	"\\|",		// 0e
	"BS",		// 0f

	"TAB",		// 10
	"Q",		// 11
	"W",		// 12
	"E",		// 13
	"R",		// 14
	"T",		// 15
	"Y",		// 16
	"U",		// 17
	"I",		// 18
	"O",		// 19
	"P",		// 1a
	"@`",		// 1b
	"[{",		// 1c
	"Enter",	// 1d
	"A",		// 1e
	"S",		// 1f

	"D",		// 20
	"F",		// 21
	"G",		// 22
	"H",		// 23
	"J",		// 24
	"K",		// 25
	"L",		// 26
	";+",		// 27
	":*",		// 28
	"]}",		// 29
	"Z",		// 2a
	"X",		// 2b
	"C",		// 2c
	"V",		// 2d
	"B",		// 2e
	"N",		// 2f

	"M",		// 30
	",<",		// 31
	".>",		// 32
	"/?",		// 33
	" _",		// 34
	"Space",	// 35
	"HOME",		// 36
	"DEL",		// 37
	"ROLLUP",	// 38
	"ROLLDOWN",	// 39
	"UNDO",		// 3a
	"←",		// 3b
	"↑",		// 3c
	"→",		// 3d
	"↓",		// 3e
	"PADCLR",	// 3f

	"PAD/",		// 40
	"PAD*",		// 41
	"PAD-",		// 42
	"PAD7",		// 43
	"PAD8",		// 44
	"PAD9",		// 45
	"PAD+",		// 46
	"PAD4",		// 47
	"PAD5",		// 48
	"PAD6",		// 49
	"PAD=",		// 4a
	"PAD1",		// 4b
	"PAD2",		// 4c
	"PAD3",		// 4d
	"PADEnter",	// 4e
	"PAD0",		// 4f

	"PAD,",		// 50
	"PAD.",		// 51
	"記号",		// 52
	"登録",		// 53
	"HELP",		// 54
	"XF1",		// 55
	"XF2",		// 56
	"XF3",		// 57
	"XF4",		// 58
	"XF5",		// 59
	"かな",		// 5a
	"ローマ字",	// 5b
	"コード",	// 5c
	"CAPS",		// 5d
	"INS",		// 5e
	"ひらがな",	// 5f

	"全角",		// 60
	"BREAK",	// 61
	"COPY",		// 62
	"F1",		// 63
	"F2",		// 64
	"F3",		// 65
	"F4",		// 66
	"F5",		// 67
	"F6",		// 68
	"F7",		// 69
	"F8",		// 6a
	"F9",		// 6b
	"F10",		// 6c
	"",			// 6d
	"",			// 6e
	"",			// 6f

	"SHIFT",	// 70
	"CTRL",		// 71
	"OPT1",		// 72
	"OPT2",		// 73
};
