//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "xiospace.h"
#include "areaset.h"
#include "busio.h"
#include "crtc.h"
#include "dmac.h"
#include "fdc.h"
#include "gvram.h"
#include "ioctlr.h"
#include "mfp.h"
#include "pio.h"
#include "pluto.h"
#include "printer.h"
#include "ram.h"
#include "romimg_x68k.h"
#include "rp5c15.h"
#include "scc.h"
#include "spc.h"
#include "sprite.h"
#include "sram.h"
#include "sysport.h"
#include "tvram.h"
#include "videoctlr.h"
#include "vm.h"
#include <typeinfo>

//
// X68030 のメイン空間 ($000000-$FFFFFF) 担当。
//

// デバイス
static std::unique_ptr<AreaSetDevice> areaset;
static std::unique_ptr<CGROMDevice> cgrom;
static std::unique_ptr<FDCDevice> fdc;
static std::unique_ptr<IPLROM1Device> iplrom1;
static std::unique_ptr<IPLROM2Device> iplrom2;
static std::unique_ptr<PlutoDevice> pluto;
static std::unique_ptr<PPIDevice> ppi;
static std::unique_ptr<PrinterDevice> printer;
static std::unique_ptr<SpriteDevice> sprite;
static std::unique_ptr<SysportDevice> sysport;

// 1つの 8KB ブロックを複数のデバイスで共有する場合
class MuxDevice
	: public IODevice
{
	using inherited = IODevice;
 public:
	MuxDevice();
	virtual ~MuxDevice();

	virtual uint64 Read8(uint32 addr);
	virtual uint64 Read16(uint32 addr);
	virtual uint64 Write8(uint32 addr, uint32 data);
	virtual uint64 Write16(uint32 addr, uint32 data);
	virtual uint64 Peek8(uint32 addr);

	IODevice *SearchDevice(uint32 addr) const;

	void AddDevice(IODevice *dev);

	std::vector<IODevice*> devlist;
};

MuxDevice::MuxDevice()
{
	logname = "none";
}

MuxDevice::~MuxDevice()
{
}

// デバイスを追加する。
// ここに追加するデバイスは dev->devlen を指定すること。
void
MuxDevice::AddDevice(IODevice *dev)
{
	if (dev->devlen == 0) {
		printf("MuxDevice::AddDevice: '%s' devlen not set!\n",
			dev->devname.c_str());
		exit(1);
	}
	devlist.push_back(dev);
}

IODevice *
MuxDevice::SearchDevice(uint32 addr) const
{
	for (const auto& d : devlist) {
		if (d->devaddr <= addr && addr < d->devaddr + d->devlen) {
			return d;
		}
	}
	return nullptr;
}

uint64
MuxDevice::Read8(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Read8(addr);
}

uint64
MuxDevice::Read16(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Read16(addr);
}

uint64
MuxDevice::Write8(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Write8(addr, data);
}

uint64
MuxDevice::Write16(uint32 addr, uint32 data)
{
	IODevice *d = SearchDevice(addr);
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Write16(addr, data);
}

uint64
MuxDevice::Peek8(uint32 addr)
{
	IODevice *d = SearchDevice(addr);
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Peek8(addr);
}


// コンストラクタ
XIOSpaceDevice::XIOSpaceDevice()
{
	logname = "xiospace";
	devname = "IOSpace";

	// メモリ
	// 通常メイン RAM は bus から直接アクセスするが、ウェイトを入れたい
	// 場合や上位8ビットに値が入っていて 24bit にミラーされたアドレスへの
	// アクセスになる場合はここを通る。
	for (int i = 0; i < ram_size / 8192; i++) {
		table[i] = gRAM.get();
	}

#define ADD_DEVICE(varname, clsname)	do {	\
	varname.reset(new clsname);	\
	AddDevice(varname.get());	\
} while (0)

	// デバイスクラス作成 (順序に注意)
	ADD_DEVICE(areaset, AreaSetDevice());
	ADD_DEVICE(cgrom, CGROMDevice());
	ADD_DEVICE(gCRTC, BusIO_W<CRTCDevice>());
	ADD_DEVICE(gDMAC, BusIO_B<DMACDevice>());
	ADD_DEVICE(fdc, BusIO_EB<FDCDevice>());
	ADD_DEVICE(gGVRAM, GVRAMDevice());
	ADD_DEVICE(gIOCtlr, BusIO_EB<IOCtlrDevice>());
	ADD_DEVICE(iplrom1, IPLROM1Device());
	ADD_DEVICE(iplrom2, IPLROM2Device());
	ADD_DEVICE(gMFP, BusIO_EB<MFPDevice>());
	ADD_DEVICE(pluto, PlutoDevice());
	ADD_DEVICE(ppi, BusIO_EB<PPIDevice>());
	ADD_DEVICE(printer, PrinterDevice());
	ADD_DEVICE(gRTC, BusIO_EB<RP5C15Device>());
	ADD_DEVICE(gSCC, BusIO_EB<SCCDevice>());
	ADD_DEVICE(gSPC, BusIO_X68kSPC<SPCDevice>());
	ADD_DEVICE(sprite, SpriteDevice());
	ADD_DEVICE(gSRAM, SRAMDevice());
	ADD_DEVICE(sysport, BusIO_FB<SysportDevice>());
	ADD_DEVICE(gTVRAM, TVRAMDevice());
	ADD_DEVICE(gVideoCtlr, BusIO_W<VideoCtlrDevice>());

	// デバッグ表示
	if (0) {
		for (int i = 0; i < 512; i++) {
			printf("[%3d] $%06x ", i, 0xc00000 + i * 8192);
			if (table[i] == nullptr) {
				printf("null\n");
			} else if (dynamic_cast<MuxDevice*>(table[i])) {
				printf("MuxDevice\n");
			} else if (dynamic_cast<IODevice*>(table[i])) {
				printf("%s\n", table[i]->devname.c_str());
			}
		}
	}
}

// デストラクタ
XIOSpaceDevice::~XIOSpaceDevice()
{
}

// デバイステーブルに dev を追加する。
// ここに追加するデバイスは dev->devlen を指定すること。
void
XIOSpaceDevice::AddDevice(IODevice *dev)
{
	if (dev->devaddr == 0) {
		printf("XIOSpaceDevice:AddDevice: '%s' devaddr not set!\n",
			dev->devname.c_str());
		exit(1);
	}
	if (dev->devlen == 0) {
		printf("XIOSpaceDevice:AddDevice: '%s' devlen not set!\n",
			dev->devname.c_str());
		exit(1);
	}

	int idx = dev->devaddr / 8192;
	for (int i = idx; i < idx + (dev->devlen + 8191) /  8192; i++) {
		if (table[i] == nullptr) {
			table[i] = dev;
		} else if (typeid(*table[i]) == typeid(MuxDevice)) {
			(static_cast<MuxDevice*>(table[i]))->AddDevice(dev);
		} else {
			MuxDevice *muxdev = new MuxDevice();
			muxdev->AddDevice(table[i]);
			muxdev->AddDevice(dev);
			table[i] = muxdev;
		}
	}
}

uint64
XIOSpaceDevice::Read8(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Read8(addr);
}

uint64
XIOSpaceDevice::Read16(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Read16(addr);
}

uint64
XIOSpaceDevice::Read32(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Read32(addr);
}

uint64
XIOSpaceDevice::Write8(uint32 addr, uint32 data)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Write8(addr, data);
}

uint64
XIOSpaceDevice::Write16(uint32 addr, uint32 data)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Write16(addr, data);
}

uint64
XIOSpaceDevice::Write32(uint32 addr, uint32 data)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Write32(addr, data);
}

uint64
XIOSpaceDevice::Peek8(uint32 addr)
{
	addr &= 0x00ffffff;
	IODevice *d = SearchDevice(addr);
	// BusErrDevice 導入すればこの if は不要
	if (d == nullptr) {
		return (uint64)-1;
	}
	return d->Peek8(addr);
}
