//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

//
// LUNA LCD パネル
//

#include "wxlcdwindow.h"
#include "wxcolor.h"
#include "wxtextscreen.h"
#include "lcd.h"

// パネルの大きさ (定数)
// N はとりあえずの倍率。
#define N (2)
#define PANEL_WIDTH  ((N) * (5 + 1) * 16 + 1)
#define PANEL_HEIGHT ((N) * (8 + 1) *  2 + 1)

//
// LUNA LCD パネル
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXLCDPanel, inherited)
	EVT_TIMER(wxID_ANY, WXLCDPanel::OnTimer)
	EVT_PAINT(WXLCDPanel::OnPaint)
wxEND_EVENT_TABLE()

// コンストラクタ
WXLCDPanel::WXLCDPanel(wxWindow *parent)
	: inherited(parent, wxID_ANY, wxDefaultPosition,
		wxSize(PANEL_WIDTH, PANEL_HEIGHT))
{
	bmpbuf.reset(new uint8 [PANEL_WIDTH * PANEL_HEIGHT * 3]);
	memset(&bmpbuf[0], 0x70, PANEL_WIDTH * PANEL_HEIGHT * 3);
	image.reset(new wxImage(PANEL_WIDTH, PANEL_HEIGHT, bmpbuf.get(), true));
	bitmap.reset(new wxBitmap(*image));

	count = (uint32)-1;
	ddram = gLCD->GetDDRAM();

	// 更新タイマー。そんなに応答速度速くなくていい
	timer.SetOwner(this);
	timer.Start(50);
}

// デストラクタ
WXLCDPanel::~WXLCDPanel()
{
	bitmap.reset();
	image.reset();
	bmpbuf.reset();
}

// タイマーイベント
void
WXLCDPanel::OnTimer(wxTimerEvent& event)
{
	uint32 devcount;

	// 更新があれば..
	devcount = gLCD->GetCounter();
	if (count != devcount) {
		count = devcount;

		// 再描画を起こす
		Refresh();
	}
}

// 描画イベント (UI スレッドから呼ばれる)
void
WXLCDPanel::OnPaint(wxPaintEvent& event)
{
	// メモリ DC に描画
	wxMemoryDC memDC;
	memDC.SelectObject(*bitmap);
	Draw(memDC);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	dstDC.Blit(0, 0, PANEL_WIDTH, PANEL_HEIGHT, &memDC, 0, 0);
}

// バックグラウンドバッファに描画
void
WXLCDPanel::Draw(wxDC& dc)
{
	// 背景
	dc.SetPen(UD_CREAM);
	dc.SetBrush(UD_CREAM);
	dc.DrawRectangle(wxPoint(0, 0), wxSize(PANEL_WIDTH, PANEL_HEIGHT));

	for (int y = 0; y < 2; y++) {
		for (int x = 0; x < 16; x++) {
			uint8 ch = ddram[0x40 * y + x];
			DrawChar(dc, x, y, ch);
		}
	}
}

// 1文字描画
void
WXLCDPanel::DrawChar(wxDC& dc, int x, int y, uint8 ch)
{
	wxSize dotsize = wxSize(N, N);
	const uint8 *font;

	font = gLCD->GetFont(ch);

	for (int py = 0; py < 8; py++) {
		uint8 line = *font++;
		line <<= 3;
		for (int px = 0; px < 5; px++) {
			wxColour color;
			if ((int8)line < 0) {
				color = UD_BROWN;
			} else {
				color = UD_YELLOW_GREEN;
			}
			dc.SetPen(color);
			dc.SetBrush(color);
			dc.DrawRectangle(
				wxPoint((1 + x * 6 + px) * N, (1 + y * 9 + py) * N),
				dotsize);

			line <<= 1;
		}
	}
}

//
// LCD ウィンドウ
//

// コンストラクタ
WXLCDWindow::WXLCDWindow(wxWindow *parent, wxWindowID id)
	: inherited(parent, id, _("Front LCD"))
{
	panel = new WXLCDPanel(this);

	// 大きさをそれに固定
	Layout();
	Fit();
}

// デストラクタ
WXLCDWindow::~WXLCDWindow()
{
}

//
// LCD キャラクタパネル
//

#define CP_WIDTH  (N * (1 + 6 * 16))
#define CP_HEIGHT (N * (1 + 9 * 16))

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXLCDCharPanel, inherited)
	EVT_PAINT(WXLCDCharPanel::OnPaint)
wxEND_EVENT_TABLE()

// コンストラクタ
WXLCDCharPanel::WXLCDCharPanel(wxWindow *parent)
	: inherited(parent, wxID_ANY, wxDefaultPosition,
		wxSize(CP_WIDTH, CP_HEIGHT))
{
	bmpbuf.reset(new uint8 [CP_WIDTH * CP_HEIGHT * 3]);
	memset(&bmpbuf[0], 0x70, CP_WIDTH * CP_HEIGHT * 3);
	image.reset(new wxImage(CP_WIDTH, CP_HEIGHT, bmpbuf.get(), true));
	bitmap.reset(new wxBitmap(*image));
}

// デストラクタ
WXLCDCharPanel::~WXLCDCharPanel()
{
	bitmap.reset();
	image.reset();
	bmpbuf.reset();
}

// 描画イベント (UI スレッドから呼ばれる)
void
WXLCDCharPanel::OnPaint(wxPaintEvent& event)
{
	// メモリ DC に描画
	wxMemoryDC memDC;
	memDC.SelectObject(*bitmap);
	Draw(memDC);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	dstDC.Blit(0, 0, CP_WIDTH, CP_HEIGHT, &memDC, 0, 0);
}

// バックグラウンドバッファに描画
void
WXLCDCharPanel::Draw(wxDC& dc)
{
	wxSize dotsize = wxSize(N, N);

	// 背景
	// XXX LIGHT_GREY を textscreen と共用すること
	dc.SetPen(*wxLIGHT_GREY);
	dc.SetBrush(*wxLIGHT_GREY);
	dc.DrawRectangle(wxPoint(0, 0), wxSize(CP_WIDTH, CP_HEIGHT));

	// フォントは 5x8 (最下行はカーソル用に空けてあるので実際は 5x7)。
	// 実際には 0xe0 以降には 5x10 フォントが混在しているのだが、その
	// モードはサポートしていないので今の所全部 5x8 としてある。
	for (int ch = 0; ch < 256; ch++) {
		// この文字の左上座標 (スケール前)
		int cx = 1 + (ch % 16) * 6;
		int cy = 1 + (ch / 16) * 9;

		const uint8 *font = gLCD->GetFont(ch);
		for (int py = 0; py < 8; py++) {
			uint8 line = *font++;
			line <<= 3;
			for (int px = 0; px < 5; px++) {
				wxColour color;
				if ((int8)line < 0) {
					color = *wxBLACK;
				} else {
					color = *wxWHITE;
				}
				dc.SetPen(color);
				dc.SetBrush(color);
				dc.DrawRectangle(
					wxPoint((cx + px) * N, (cy + py) * N),
					dotsize);

				line <<= 1;
			}
		}
	}
}

//
// LCD モニタウィンドウ
//

// コンストラクタ
WXLCDMonitor::WXLCDMonitor(wxWindow *parent, wxWindowID id)
	: inherited(parent, id, _("LCD"))
{
	wxBoxSizer *topsizer = new wxBoxSizer(wxVERTICAL);

	monpanel = new WXTextScreen(this, *gLCD);
	topsizer->Add(monpanel, 0, wxEXPAND | wxALL, DefaultPadding);

	chrpanel = new WXLCDCharPanel(this);
	topsizer->Add(chrpanel, 0, wxEXPAND | wxRIGHT | wxLEFT | wxBOTTOM,
		DefaultPadding);

	SetSizer(topsizer);

	// 大きさをそれに固定
	Layout();
	Fit();
}

// デストラクタ
WXLCDMonitor::~WXLCDMonitor()
{
}
