//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxmainview.h"
#include "keyboard.h"
#include "mystring.h"
#include "renderer.h"
#include "scheduler.h"
#include <wx/region.h>
#include <wx/event.h>

//#define KEYDEBUG 1
//#define MOUSEDEBUG 1

#if defined(KEYDEBUG)
#define DKPRINTF(...)	printf(__VA_ARGS__)
#else
#define DKPRINTF(...) /**/
#endif

#if defined(MOUSEDEBUG)
#define DMPRINTF(...)	printf(__VA_ARGS__)
#else
#define DMPRINTF(...) /**/
#endif

// 再描画矩形の指定方法について。
// DoRefresh() にて、VM スレッドが自発的に画面を更新する際は、
// パフォーマンスの観点からも更新矩形だけを再描画すべき。
// これを wxWidgets に指示するのが wxWindow::Refresh{,Rect}() であるが
// 一部のプラットフォームではこれを呼び出すと、アプリケーションウィンドウ
// 全体がささってしまい、MainView だけでなくメニューバーも含めたすべてが
// 再描画できなくなってしまう。仕方がないので、これらのプラットフォームでは
// Refresh{,Rect}() を呼ばず、再描画矩形を画面全体として運用する。
// ちょっと無駄なだけなので (動かないことに比べれば) 気にしない。
//
// 今のところ NetBSD (5.0 以降) と OpenBSD(/amd64)(-6.7)
// (と Linux (CentOS 6.3/x86_64)) が該当する。
#if defined(__NetBSD__) || defined(__OpenBSD__)
#define DONT_USE_REFRESH 1
#endif

static void wxmainview_refresh();

WXMainView *gMainView;

wxBEGIN_EVENT_TABLE(WXMainView, inherited)
	EVT_PAINT(WXMainView::OnPaint)
	EVT_KEY_DOWN(WXMainView::OnKeyDown)
	EVT_KEY_UP(WXMainView::OnKeyUp)
	EVT_MOUSE_EVENTS(WXMainView::OnMouse)
	EVT_TIMER(wxID_ANY, WXMainView::OnTimer)
wxEND_EVENT_TABLE()

WXMainView::WXMainView(wxWindow *parent)
	: inherited(parent)
{
	// 仮想画面(VM が描画する画面)
	rendwidth  = gRenderer->GetWidth();
	rendheight = gRenderer->GetHeight();

	// レンダラバッファは常に仮想画面サイズ
	int buflen = rendwidth * rendheight * 3;
	ImageBuf.reset(new uint8_t[buflen]);
	memset(&ImageBuf[0], 0x70, buflen);
	// レンダラに通知
	gRenderer->SetBuf(ImageBuf.get(), wxmainview_refresh);

	// 実画面 (ウィンドウの大きさ)
	DoResize(screen_scale);

	// 描画を律速するためのタイマー
	timer.SetOwner(this);
	timer.Start(33);

	// 最初に一度描画
	Refresh();
}

WXMainView::~WXMainView()
{
}

// リフレッシュ要求コールバック。
// VM から呼ばれるので一旦グローバル関数を経由する。
void
wxmainview_refresh(void)
{
	gMainView->DoRefresh();
}

// リフレッシュ要求コールバック。こっちが実体。
void
WXMainView::DoRefresh()
{
#if defined(DONT_USE_REFRESH)
	// Refresh{,Rect}() が使えないので、再描画範囲は常に全体とする
#else
	// 再描画範囲を更新矩形に限定
	// XXX のはずだが、今の所全体
	RefreshRect(wxRect(0, 0, viewwidth, viewheight), false);
#endif

	// ここは VM Renderer が画面を更新した時に呼ばれるので、本来なら
	// ここで単に Refresh() を呼んで再描画を指示すればいいのだが、
	// VM 高速モードになると、VM Renderer が高速で回ってここも高速で
	// 呼ばれてしまい、その結果 Refresh() を発行しすぎて Mac では
	// システムイベントが拾えなくなってしまう。
	// それを避けるため、ここではフラグを立てるだけにしておいて、
	// こちらで管理している実時間タイマーで Refresh() を律速する。
	paint_request = true;
}

// タイマーイベント
void
WXMainView::OnTimer(wxTimerEvent& event)
{
	bool req = paint_request.exchange(false);
	if (req) {
		// 実際に再描画する
		Refresh();
	}
}

// 描画イベント
void
WXMainView::OnPaint(wxPaintEvent& event)
{
	if (viewwidth == 0 || viewheight == 0)
		return;

	wxImage image(rendwidth, rendheight, ImageBuf.get(), true);
	if (viewwidth != rendwidth || viewheight != rendheight) {
		// wxIMAGE_QUALITY_NORMAL だと速いが、品質優先。
		image.Rescale(viewwidth, viewheight, wxIMAGE_QUALITY_HIGH);
	}

	wxBitmap bitmap(image);
	wxMemoryDC memDC;
	memDC.SelectObject(bitmap);

	wxPaintDC dc(this);
	wxRegion region(0, 0, viewwidth, viewheight);
	wxRegionIterator ri(region);
	for (; ri; ri++) {
		int dx, dy, dw, dh;
		dx = ri.GetX();
		dy = ri.GetY();
		dw = ri.GetWidth();
		dh = ri.GetHeight();
		int sx = dx;
		int sy = dy;
		dc.Blit(dx, dy, dw, dh, &memDC, sx, sy, wxCOPY);
	}
}

// リサイズ (UI スレッドから呼ぶこと)
void
WXMainView::DoResize(double scale)
{
	screen_scale = scale;
	int width  = (double)rendwidth  * scale;
	int height = (double)rendheight * scale;
	if (viewwidth != width || viewheight != height) {
		// ウィンドウサイズ変更
		viewwidth  = width;
		viewheight = height;
		SetClientSize(viewwidth, viewheight);
		SetMinSize(wxSize(viewwidth, viewheight));
	}
}

// 現在のスケールを返す。
// screen_scale は public static なので見えるんだけど一応お行儀で。
double
WXMainView::GetScale() const
{
	return screen_scale;
}

// 現在のスケール
/*static*/ double WXMainView::screen_scale;


//
// マウス入力
//

// マウスモード設定
void
WXMainView::SetMouseMode(bool mode)
{
	// XXX マウスモード ON にした時の座標を記憶してそこに戻るのはどうか

	mousemode = mode;
	if (mousemode) {
		// マウスモードを ON にする
		DMPRINTF("SetMouseMode ON\n");

		// マウスカーソル OFF
		::wxSetCursor(wxCursor(wxCURSOR_BLANK));

		CaptureMouse();

		// このウィンドウから外れないように移動を検出し続けるため、
		// マウスカーソルをウィンドウ中央に移す。
		// ウィンドウとメニューバーが分離してる wxOSX のために必要。
		MyWarpPointer(GetClientSize() / 2);
	} else {
		// マウスモードを OFF にする
		DMPRINTF("SetMouseMode OFF\n");

		ReleaseMouse();

		// 分かりやすさのため、マウスカーソルを中央へ戻してから復帰
		wxSize pt = GetClientSize() / 2;
		WarpPointer(pt.x, pt.y);

		// マウスカーソル ON
		::wxSetCursor(wxNullCursor);
	}
}

// マウスカーソルをクライアント座標 (x, y) に移動
void
WXMainView::MyWarpPointer(wxSize pos)
{
	DMPRINTF("MyWarpPointer %d,%d\n", pos.x, pos.y);

	// この後の MDWarpPointer() でもう一度 EVT_MOTION が飛ぶので
	// それを判別するためのフラグ。
	motion_by_warp_pointer = true;

	WarpPointer(pos.x, pos.y);
}

// マウスイベント
void
WXMainView::OnMouse(wxMouseEvent& event)
{
	// マウスモードの時だけ処理。
	// XXX 本当はイベントごと Disconnect したい
	if (mousemode) {
		bool leaving = event.Leaving();
		bool moving  = event.Moving();
		bool dragging = event.Dragging();
		// LeftDown() 等は押された時、LeftIsDown() 等は押されてる間 true。
		bool rb = event.RightIsDown();
		bool lb = event.LeftIsDown();
		bool mb = event.MiddleIsDown();
		const wxPoint& pos = event.GetPosition();
		int dx = 0;
		int dy = 0;

#if defined(MOUSEDEBUG)
		std::string dstr = "OnMouse";
		if (moving || dragging || motion_by_warp_pointer) {
			dstr += string_format(" (%d,%d)", pos.x, pos.y);
		}
		if (moving)
			dstr += " Moving";
		if (dragging)
			dstr += " Dragging";
		if (motion_by_warp_pointer)
			dstr += " MotionByWarp";
		if (lb || mb || rb) {
			dstr += string_format(" Button(%c%c%c)",
				(lb ? 'L' : '-'),
				(mb ? 'M' : '-'),
				(rb ? 'R' : '-'));
		}
		if (leaving)
			dstr += " Leaving";
		printf("%s\n", dstr.c_str());
#endif

		if (leaving) {
			// タスクスイッチした時とかのための自動リリース
			SetMouseMode(false);
		}

		if (moving || dragging) {
			// (My)WarpPointer() で自発的にマウスカーソルを移動させたことで
			// 来る EVT_MOTION は無視する。
			if (motion_by_warp_pointer) {
				motion_by_warp_pointer = false;
				// ここを基準点にする
				prevpos = pos;
			} else {
				// 前回からの差を求める
				dx = pos.x - prevpos.x;
				dy = pos.y - prevpos.y;
				prevpos = pos;

				// マウスカーソルを中心に戻す
				if (dx != 0 || dy != 0) {
					MyWarpPointer(GetClientSize() / 2);
				}
			}
		}

		// VM に入力
		if (dx != 0 ||
		    dy != 0 ||
		    lb != prevlb ||
		    mb != prevmb ||
		    rb != prevrb   )
		{
			DMPRINTF("        Input dx=%d dy=%d %c%c%c\n",
				dx, dy,
				(lb ? 'L' : '-'),
				(mb ? 'M' : '-'),
				(rb ? 'R' : '-'));

			gKeyboard->MouseInput(dx, dy, rb, lb, mb);
		}

		prevlb = lb;
		prevmb = mb;
		prevrb = rb;
	}

	// EVT_LEFT_DOWN をデフォルト処理するため
	event.Skip();
}

// アクティベートイベント。
// この(トップレベル)ウィンドウにフォーカスが当たるか外れるかで飛んでくる。
// 実際には WXMainFrame 宛てに飛んできたイベントをこちらに転送してある。
// マウスモード中にこのウィンドウからフォーカスが外れたら (例えば ALT+TAB の
// ようなウィンドウを切り替える操作)、マウスモードを解除するため。
void
WXMainView::OnActivate(wxActivateEvent& event)
{
	// フォーカスが外れる際にマウスモード中ならマウスモードを解除
	if (event.GetActive() == false) {
		DMPRINTF("OnActivate FocusOut\n");
		if (mousemode) {
			SetMouseMode(false);
		}
	} else {
		DMPRINTF("OnActivate FocusIn\n");
	}
}


//
// キー入力
//

// wxKeyEvent でキーコードを得るためのメソッドは
// GetKeyCode()、GetRawKeyCode()、GetRawKeyFlags() の3つが用意されていて
// wxWigets の port (wxGTK、wxOSX とか) によってそれぞれ内容の定義が違う。
//
// wxGTK の場合
//  GetKeyCode() は wxWidgets の WXK_* (/usr/pkg/include/wx-3.0/wx/defs.h)。
//   これは左右の SHIFT が区別できないが、
//   それ以外はだいたい使えそうではある。
//
//  GetRawKeyCode() は GDK の keycode っぽい?
//   (/usr/pkg/include/gtk-2.0/gdk/gdkkeysyms.h)。
//   X11 の keysym から来てる? (/usr/X11R7/include/X11/keysymdef.h)。
//   ただし [A] と [SHIFT]+[A] が違うコードを返すというか、文字コードで
//   表現出来る時は文字コード、みたいな感じになっているので、ここでは
//   使いづらい。
//   左右の SHIFT は区別可能。
//   [^] と [SHIFT]+[^] もそもそも文字コードになるので '^' と '~' になる。
//
//  GetRawKeyFlags() は GDK の hardware keycode らしいけど、それが何か不明。
//   左右の SHIFT は違うコードを返す。
//   SHIFT を押しながらでもコードは変わらないっぽい。
//   [^] と [SHIFT]+[^] は同じコードを返す。
//
// wxOSX の場合
//  GetKeyCode()
//   これは左右の SHIFT が区別できないが、
//   それ以外はだいたい使えそうではある。
//
//  GetRawKeyCode()
//   [A] と [SHIFT]+[A] などは同じコードを返す。
//   左右の SHIFT は区別可能。
//
//  GetRawKeyFlags() は状態フラグか何かのようだ。
//

// キーを押した
void
WXMainView::OnKeyDown(wxKeyEvent& event)
{
	DKPRINTF("KeyDown ");
	int keycode = GetKeyCode(event);
	gKeyboard->MakeKey(keycode);

	// メニューのアクセラレータのために必要
	event.Skip();
}

// キーを離した
void
WXMainView::OnKeyUp(wxKeyEvent& event)
{
	DKPRINTF("KeyUp   ");
	int keycode = GetKeyCode(event);
	gKeyboard->BreakKey(keycode);
}

// 共通キーコードを取得する。
// 対応するキーがない場合は -1 を返す。
int
WXMainView::GetKeyCode(wxKeyEvent& event)
{
	uint32 wxcode   __unused = event.GetKeyCode();
	uint32 rawcode  __unused = event.GetRawKeyCode();
	uint32 rawflags __unused = event.GetRawKeyFlags();
	uint32 inp;
	int keycode;

#if defined(__WXGTK__)
	// wxGTK では RawKeyFlags を使う
	DKPRINTF("key=0x%03x raw=0x%04x <flags=0x%02x> ",
		wxcode, rawcode, rawflags);
	inp = rawflags;
#elif defined(__WXOSX__)
	// wxOSX では RawKeyCode を使う
	DKPRINTF("key=0x%03x <raw=0x%02x> flags=0x%06x ",
		wxcode, rawcode, rawflags);
	inp = rawcode;
#endif

	// どちらにしてもテーブルを引く
	if (inp < countof(keycode_table)) {
		keycode = keycode_table[inp];
	} else {
		keycode = KC_none;
	}
	DKPRINTF("keycode=%d %s\n", keycode, Keyboard::keyname[keycode]);
	return keycode;
}

#if defined(__WXGTK__)
// wxGTK の RawKeyFlags を共通キーコードに変換するテーブル。
/*static*/ const int
WXMainView::keycode_table[0xd4] = {
	KC_none,		// 0x00
	KC_none,		// 0x01
	KC_none,		// 0x02
	KC_none,		// 0x03
	KC_none,		// 0x04
	KC_none,		// 0x05
	KC_none,		// 0x06
	KC_none,		// 0x07
	KC_none,		// 0x08
	KC_none,		// 0x09
	KC_1,			// 0x0a
	KC_2,			// 0x0b
	KC_3,			// 0x0c
	KC_4,			// 0x0d
	KC_5,			// 0x0e
	KC_6,			// 0x0f

	KC_7,			// 0x10
	KC_8,			// 0x11
	KC_9,			// 0x12
	KC_0,			// 0x13
	KC_minus,		// 0x14
	KC_circum,		// 0x15
	KC_BS,			// 0x16
	KC_TAB,			// 0x17
	KC_Q,			// 0x18
	KC_W,			// 0x19
	KC_E,			// 0x1a
	KC_R,			// 0x1b
	KC_T,			// 0x1c
	KC_Y,			// 0x1d
	KC_U,			// 0x1e
	KC_I,			// 0x1f

	KC_O,			// 0x20
	KC_P,			// 0x21
	KC_at,			// 0x22
	KC_bracketleft,	// 0x23
	KC_enter,		// 0x24
	KC_CTRL,		// 0x25 (CTRL_L)
	KC_A,			// 0x26
	KC_S,			// 0x27
	KC_D,			// 0x28
	KC_F,			// 0x29
	KC_G,			// 0x2a
	KC_H,			// 0x2b
	KC_J,			// 0x2c
	KC_K,			// 0x2d
	KC_L,			// 0x2e
	KC_semicolon,	// 0x2f

	KC_colon,		// 0x30
	KC_ESC,			// 0x31
	KC_SHIFT_L,		// 0x32 (SHIFT_L)
	KC_bracketright,// 0x33
	KC_Z,			// 0x34
	KC_X,			// 0x35
	KC_C,			// 0x36
	KC_V,			// 0x37
	KC_B,			// 0x38
	KC_N,			// 0x39
	KC_M,			// 0x3a
	KC_comma,		// 0x3b
	KC_period,		// 0x3c
	KC_slash,		// 0x3d
	KC_SHIFT_R,		// 0x3e (SHIFT_R)
	KC_PAD_multiply,// 0x3f

	KC_none,		// 0x40
	KC_space,		// 0x41
	KC_none,		// 0x42
	KC_F1,			// 0x43
	KC_F2,			// 0x44
	KC_F3,			// 0x45
	KC_F4,			// 0x46
	KC_F5,			// 0x47
	KC_F6,			// 0x48
	KC_F7,			// 0x49
	KC_F8,			// 0x4a
	KC_F9,			// 0x4b
	KC_F10,			// 0x4c
	KC_none,		// 0x4d
	KC_none,		// 0x4e
	KC_PAD_7,		// 0x4f

	KC_PAD_8,		// 0x50
	KC_PAD_9,		// 0x51
	KC_PAD_minus,	// 0x52
	KC_PAD_4,		// 0x53
	KC_PAD_5,		// 0x54
	KC_PAD_6,		// 0x55
	KC_PAD_plus,	// 0x56
	KC_PAD_1,		// 0x57
	KC_PAD_2,		// 0x58
	KC_PAD_3,		// 0x59
	KC_PAD_0,		// 0x5a
	KC_PAD_decimal,	// 0x5b
	KC_none,		// 0x5c
	KC_none,		// 0x5d
	KC_none,		// 0x5e
	KC_none,		// 0x5f

	KC_none,		// 0x60
	KC_none,		// 0x61
	KC_up,			// 0x62
	KC_none,		// 0x63
	KC_left,		// 0x64
	KC_none,		// 0x65
	KC_right,		// 0x66
	KC_none,		// 0x67
	KC_down,		// 0x68
	KC_none,		// 0x69
	KC_INS,			// 0x6a
	KC_DEL,			// 0x6b
	KC_PAD_enter,	// 0x6c
	KC_CTRL,		// 0x6d (CTRL_R)
	KC_none,		// 0x6e
	KC_none,		// 0x6f

	KC_PAD_divide,	// 0x70
	KC_none,		// 0x71
	KC_none,		// 0x72
	KC_none,		// 0x73
	KC_none,		// 0x74
	KC_none,		// 0x75
	KC_none,		// 0x76
	KC_none,		// 0x77
	KC_none,		// 0x78
	KC_none,		// 0x79
	KC_none,		// 0x7a
	KC_none,		// 0x7b
	KC_none,		// 0x7c
	KC_none,		// 0x7d
	KC_none,		// 0x7e
	KC_none,		// 0x7f

	KC_none,		// 0x80
	KC_none,		// 0x81
	KC_none,		// 0x82
	KC_none,		// 0x83
	KC_none,		// 0x84
	KC_backslash,	// 0x85
	KC_none,		// 0x86
	KC_none,		// 0x87
	KC_none,		// 0x88
	KC_none,		// 0x89
	KC_none,		// 0x8a
	KC_none,		// 0x8b
	KC_none,		// 0x8c
	KC_none,		// 0x8d
	KC_none,		// 0x8e
	KC_none,		// 0x8f

	KC_none, KC_none, KC_none, KC_none, // 0x90
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,

	KC_none, KC_none, KC_none, KC_none, // 0xa0
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,

	KC_none, KC_none, KC_none, KC_none, // 0xb0
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,

	KC_none, KC_none, KC_none, KC_none, // 0xc0
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,
	KC_none, KC_none, KC_none, KC_none,

	KC_none,		// 0xd0
	KC_none,		// 0xd1
	KC_none,		// 0xd2
	KC_underscore,	// 0xd3
};
#endif // __WXGTK__

#if defined(__WXOSX__)
// wxOSX の RawKeyCode を共通キーコードに変換するテーブル。
// Mac の CAPS は押す離すで一回分の KeyDown しか来ないのでちょっと放置。
// [英数] と [かな] もちょっと放置。
/*static*/ const int
WXMainView::keycode_table[0x80] = {
	KC_A,			// 0x00
	KC_S,			// 0x01
	KC_D,			// 0x02
	KC_F,			// 0x03
	KC_H,			// 0x04
	KC_G,			// 0x05
	KC_Z,			// 0x06
	KC_X,			// 0x07
	KC_C,			// 0x08
	KC_V,			// 0x09
	KC_none,		// 0x0a
	KC_B,			// 0x0b
	KC_Q,			// 0x0c
	KC_W,			// 0x0d
	KC_E,			// 0x0e
	KC_R,			// 0x0f

	KC_Y,			// 0x10
	KC_T,			// 0x11
	KC_1,			// 0x12
	KC_2,			// 0x13
	KC_3,			// 0x14
	KC_4,			// 0x15
	KC_6,			// 0x16
	KC_5,			// 0x17
	KC_circum,		// 0x18
	KC_9,			// 0x19
	KC_7,			// 0x1a
	KC_minus,		// 0x1b
	KC_8,			// 0x1c
	KC_0,			// 0x1d
	KC_bracketleft,	// 0x1e
	KC_O,			// 0x1f

	KC_U,			// 0x20
	KC_at,			// 0x21
	KC_I,			// 0x22
	KC_P,			// 0x23
	KC_enter,		// 0x24
	KC_L,			// 0x25
	KC_J,			// 0x26
	KC_colon,		// 0x27
	KC_K,			// 0x28
	KC_semicolon,	// 0x29
	KC_bracketright,// 0x2a
	KC_comma,		// 0x2b
	KC_slash,		// 0x2c
	KC_N,			// 0x2d
	KC_M,			// 0x2e
	KC_period,		// 0x2f

	KC_TAB,			// 0x30
	KC_space,		// 0x31
	KC_none,		// 0x32
	KC_BS,			// 0x33
	KC_none,		// 0x34
	KC_ESC,			// 0x35
	KC_none,		// 0x36
	KC_none,		// 0x37
	KC_SHIFT_L,		// 0x38
	KC_none,		// 0x39
	KC_none,		// 0x3a
	KC_CTRL,		// 0x3b
	KC_SHIFT_R,		// 0x3c
	KC_none,		// 0x3d
	KC_none,		// 0x3e
	KC_none,		// 0x3f

	KC_none,		// 0x40
	KC_none,		// 0x41
	KC_none,		// 0x42
	KC_none,		// 0x43
	KC_none,		// 0x44
	KC_none,		// 0x45
	KC_none,		// 0x46
	KC_none,		// 0x47
	KC_none,		// 0x48
	KC_none,		// 0x49
	KC_none,		// 0x4a
	KC_none,		// 0x4b
	KC_none,		// 0x4c
	KC_none,		// 0x4d
	KC_none,		// 0x4e
	KC_none,		// 0x4f

	KC_none,		// 0x50
	KC_none,		// 0x51
	KC_none,		// 0x52
	KC_none,		// 0x53
	KC_none,		// 0x54
	KC_none,		// 0x55
	KC_none,		// 0x56
	KC_none,		// 0x57
	KC_none,		// 0x58
	KC_none,		// 0x59
	KC_none,		// 0x5a
	KC_none,		// 0x5b
	KC_none,		// 0x5c
	KC_backslash,	// 0x5d
	KC_underscore,	// 0x5e
	KC_none,		// 0x5f

	KC_F5,			// 0x60
	KC_F6,			// 0x61
	KC_F7,			// 0x62
	KC_F3,			// 0x63
	KC_F8,			// 0x64
	KC_F9,			// 0x65
	KC_none,		// 0x66
	KC_none,		// 0x67	F11
	KC_none,		// 0x68
	KC_none,		// 0x69
	KC_none,		// 0x6a
	KC_none,		// 0x6b
	KC_none,		// 0x6c
	KC_F10,			// 0x6d
	KC_none,		// 0x6e
	KC_none,		// 0x6f	F12

	KC_none,		// 0x70
	KC_none,		// 0x71
	KC_none,		// 0x72
	KC_none,		// 0x73
	KC_none,		// 0x74
	KC_none,		// 0x75
	KC_F4,			// 0x76
	KC_none,		// 0x77
	KC_F2,			// 0x78
	KC_none,		// 0x79
	KC_F1,			// 0x7a
	KC_left,		// 0x7b
	KC_right,		// 0x7c
	KC_down,		// 0x7d
	KC_up,			// 0x7e
	KC_none,		// 0x7f
};
#endif // __WXOSX__
