//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxsoftkey.h"
#include "wxcolor.h"
#include "wxtextscreen.h"
#include "sram.h"

//
// ソフトウェアキーボード (パネル) 共通部分
//

// エンター記号。
// 描画の際にキー刻印文字列の半角文字数から描画幅を決定しており、
// エンター記号は半角2文字幅で描いているので実際の文字列も2文字にしておく。
// そうすると "\\n" が意味的にもちょうどいい。
// これは幅を算出するのに使うだけで表示はされない。
#define EnterMark	"\\n"

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXSoftKeyPanel, inherited)
	EVT_LEFT_DOWN(WXSoftKeyPanel::OnLeftDown)
	EVT_LEFT_UP(WXSoftKeyPanel::OnLeftUp)
	EVT_LEFT_DCLICK(WXSoftKeyPanel::OnLeftDClick)
	EVT_RIGHT_DOWN(WXSoftKeyPanel::OnRightDown)
	EVT_PAINT(WXSoftKeyPanel::OnPaint)
	EVT_SIZE(WXSoftKeyPanel::OnSize)
wxEND_EVENT_TABLE()

// コンストラクタ
WXSoftKeyPanel::WXSoftKeyPanel(wxWindow *parent,
	int u_width_, int u_height_)
	: inherited(parent)
{
	u_width  = u_width_;
	u_height = u_height_;
}

// デストラクタ
WXSoftKeyPanel::~WXSoftKeyPanel()
{
	bitmap.reset();
	image.reset();
	bmpbuf.reset();
}

// フォントサイズ変更 (外部インタフェース)
void
WXSoftKeyPanel::SetFontSize(FontId new_fontid)
{
	// 同じなら何もしない
	if (fontid == new_fontid) {
		return;
	}

	// フォントを更新
	UpdateFont(new_fontid);

	// サイズ変更処理
	DoSize();
}

// サイズ変更処理
// サイズを変更したい時に呼ぶ (サイズが変更されたらではない)。
// ここではサイズは fontsize にのみ依存しているので、フォントサイズが
// 変化/確定した時に呼び出すこと。
void
WXSoftKeyPanel::DoSize()
{
	assert(fontid != FontId::None);

	// このパネル内の描画の基本単位
	unit = font_height / 4;

	// エンターキーだけ矩形でないのでここでリージョンを計算。
	// unit が確定したところで一回行う。
	//
	// |   ^   |   \   |   BS  |
	// |       |       |       |
	// +---+---+--(A)--+------(B)
	//     |       |           |
	//     |   [   |           |
	//     |       |    Enter  |
	// +---+---+--(F)-(E)      |
	// |       |       |       |
	// |   :   |   ]   |       |
	// |       |       |       |
	// +---+---+------(D)-----(C)
	//     |                   |
	//     |      SHIFT        |
	const keydata_t& keybs = FindKeydata(KC_BS);
	const keydata_t& keyleft = FindKeydata(KC_bracketleft);
	const keydata_t& keyright = FindKeydata(KC_bracketright);
	const keydata_t& keyshift = FindKeydata(KC_SHIFT_R);
	// 各点のピクセル座標
	enter_points = {
		{ keyleft.right() * unit, keyleft.y * unit },	// A
		{ keybs.right() * unit, keyleft.y * unit },		// B
		{ keyshift.right() * unit, keyshift.y * unit },	// C
		{ keyright.right() * unit, keyshift.y * unit },	// D
		{ keyright.right() * unit, keyright.y * unit },	// E
		{ keyleft.right() * unit, keyright.y * unit },	// F
	};
	// 当たり判定用のリージョン
	enter_region = wxRegion(enter_points.size(), enter_points.data());

	// コントロールの大きさを変更
	wxSize size(u_width * unit, u_height * unit);
	SetClientSize(size);
	SetMinSize(size);
}

// サイズ変更イベント
void
WXSoftKeyPanel::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある
	// GTK では 1 で来ることがある
	const wxSize& size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	view_width = size.x;
	view_height = size.y;

	// ビットマップを持っていれば一旦解放
	bitmap.reset();
	image.reset();
	bmpbuf.reset();

	// ビットマップ作成
	// 正しく描かれなかった時に分かりやすいよう初期値は濃いグレーにしておく
	bmpbuf.reset(new uint8 [view_width * view_height * 3]);
	memset(&bmpbuf[0], 0x70, view_width * view_height * 3);
	image.reset(new wxImage(view_width, view_height, bmpbuf.get(), true));
	bitmap.reset(new wxBitmap(*image));

	// 再描画を指示
	Refresh();
}

// 描画イベント (UI スレッドから呼ばれる)
void
WXSoftKeyPanel::OnPaint(wxPaintEvent& event)
{
	// メモリ DC に描画
	wxMemoryDC memDC;
	memDC.SelectObject(*bitmap);
	Draw(memDC);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	dstDC.Blit(0, 0, view_width, view_height, &memDC, 0, 0);
}

// 描画本体
void
WXSoftKeyPanel::Draw(wxDC& dc)
{
	dc.SetPen(wxPen(color.Background));
	dc.SetBrush(wxBrush(color.Background));
	dc.DrawRectangle(wxPoint(0, 0), GetSize());

	dc.SetPen(color.Text);
	SetTextForeColor(color.Text);
	for (const auto& key : GetKeydata()) {
		if (hit[key.keycode]) {
			// 押されている
			SetTextBackColor(color.KeyPressed);
			dc.SetBrush(color.KeyPressed);
		} else {
			// 押されていない
			SetTextBackColor(color.Keytop[key.coltype]);
			dc.SetBrush(wxBrush(color.Keytop[key.coltype]));
		}

		// エンターキーだけ枠の描き方が違うので別処理
		if (__predict_false(key.x < 0)) {
			DrawEnterKey(dc);
		} else {
			DrawKey(dc, key);
		}
	}

#if 0
	// X68k キーボードだとどこに表示するか
	SetTextBackColor(color.Background);
	wxString utfstr(wxT("右クリックで長押し"));
	const char *sjis = (const char *)utfstr.mb_str(conv);
	DrawStringSJIS(dc, 190 * unit, 2 * unit, sjis, TA::Disable);
#endif
}

// 指定のキーを一つ (枠から刻印まで) 描画する。
void
WXSoftKeyPanel::DrawKey(wxDC& dc, const keydata_t& key)
{
	// 枠
	wxPoint pt(key.x * unit, key.y * unit);
	wxSize sz(key.w * unit + 1, key.h * unit + 1);
	dc.DrawRectangle(pt, sz);

	// 刻印を選択
	const char *disp = key.disp[modifier];
	if (disp == NULL) {
		disp = key.disp[0];
	}

	// Shift_JIS に変換
	wxString utfstr(disp, wxConvUTF8);
	const char *sjis = (const char *)utfstr.mb_str(conv);
	// 改行があれば2行に分割
	char sjis1[8] {};
	char sjis2[8] {};
	int sjislen;
	int row;
	const char *lf = strchr(sjis, '\n');
	if (lf == NULL) {
		// 1行なら、Shift_JIS での文字数は半角換算の文字数と一致する。
		sjislen = strlen(sjis);
		row = 1;
	} else {
		// 2行に分割。センタリングは1行目の長さだけで調べる。
		strncpy(sjis1, sjis, lf - sjis);
		strcpy(sjis2, lf + 1);
		sjislen = strlen(sjis1);
		row = 2;
	}

	// センタリング
	int x = key.x + (key.w - sjislen * 2) / 2;
	int y = key.y + (key.h - row * 4) / 2;

	// LED を描画 (機種によって色々違うので個別クラス側で処理)
	if (key.led >= 0) {
		const wxPen oldpen = dc.GetPen();
		const wxBrush oldbrush = dc.GetBrush();

		y += DrawLED(dc, key);

		dc.SetPen(oldpen);
		dc.SetBrush(oldbrush);
	}

	// 文字を描画
	pt = wxPoint(x * unit + 1, y * unit + 1);
	if (strcmp(disp, EnterMark) == 0) {
		DrawEnterMark(dc, pt);
	} else {
		if (lf == NULL) {
			// 1行の場合
			DrawStringSJIS(dc, pt.x, pt.y, sjis);
		} else {
			// 2行の場合
			DrawStringSJIS(dc, pt.x, pt.y, sjis1);
			pt.y += 4 * unit;
			DrawStringSJIS(dc, pt.x, pt.y, sjis2);
		}
	}
}

// エンターキー
void
WXSoftKeyPanel::DrawEnterKey(wxDC& dc)
{
	// エンターキーを描画
	dc.DrawPolygon(enter_points.size(), enter_points.data());

	// エンター記号を描画
	DrawEnterMark(dc, wxPoint(
		(enter_points[0].x + enter_points[1].x) / 2 - (4 / 2),
		(enter_points[1].y + enter_points[2].y) / 2 - (4 / 2)));
}

// 指定の位置にエンター記号を描画する (エンターキー全体の描画ではない)。
// メインキーとテンキーの両方で同じ図柄を使う。
// 大きさは (3unit, 4unit) なので実質全角1文字分くらい。
void
WXSoftKeyPanel::DrawEnterMark(wxDC& dc, const wxPoint& basep)
{
	// エンター記号は幅 3 unit、高さ 4 unit。
	//    +0 +1 +2 +3
	// +0 .. .. .. 1.
	//    .. .. .. 1.
	// +1 .. .. .. 1.
	//    .. .. .. 1.
	// +2 .. 1. .. 1.
	//    .1 .. .. 1.
	// +3 11 11 11 1.
	//    .1 .. .. ..
	// +4 .. 1. .. ..
	//    .. .. .. ..
	static const std::vector<std::pair<int,int>> entermark {
		{  3, 0 },
		{  3, 3 },	// │
		{  0, 3 },	// ─
		{  1, 2 },	// ／
		{  0, 3 },	// (一筆書きなので戻る)
		{  1, 4 },	// ＼
	};
	for (int i = 1; i < entermark.size(); i++) {
		auto start = entermark[i - 1];
		auto end   = entermark[i];
		wxPoint startp(start.first * unit, start.second * unit);
		wxPoint endp(end.first * unit, end.second * unit);
		dc.DrawLine(basep + startp, basep + endp);
	}
}

// keycode に対応する keydata を返す
const keydata_t&
WXSoftKeyPanel::FindKeydata(uint keycode) const
{
	for (const auto& key : GetKeydata()) {
		if (key.keycode == keycode) {
			return key;
		}
	}
	// 見付からないはずはない
	assertmsg(0, "keycode %d not found", keycode);
}

// マウス左クリック (Down) イベント
void
WXSoftKeyPanel::OnLeftDown(wxMouseEvent& event)
{
	const wxPoint& pt = event.GetPosition();
	const keydata_t *key = GetKeydataFromPoint(pt);

	DoLeftDown(key);
}

// マウス左クリック (Down) 処理
void
WXSoftKeyPanel::DoLeftDown(const keydata_t *key)
{
	if (key != NULL) {
		uint code = key->keycode;
		if (hit[code]) {
			// 押下(長押し)中のキーならここで開放
			KeyUp(*key);
		} else {
			// 押下して、LeftUp のために覚えておく
			KeyDown(*key);
			leftdown_key = key;
		}
	}
}

// マウス左クリック (Up) イベント
void
WXSoftKeyPanel::OnLeftUp(wxMouseEvent& event)
{
	// Up 時は、これが起きた地点のキーコードではなく
	// 直近に LeftDown したキーコードをターゲットにする。
	const keydata_t *key = leftdown_key;

	DoLeftUp(key);
}

// マウス左クリック (Up) 処理
void
WXSoftKeyPanel::DoLeftUp(const keydata_t *key)
{
	if (key != NULL) {
		// 押されていたら開放
		uint code = key->keycode;
		if (hit[code]) {
			KeyUp(*key);
		}
	}
	leftdown_key = NULL;
}

// マウス左ダブルクリックイベント
// ダブルクリックという操作に特別何かを割り当ててはいないが、
// キーを連打するとその間隔によってはダブルクリックと扱われるため、
// 受け取ってここで処理する。
//
// DClick イベントは LeftDown、LeftUp、LeftDClick、LeftUp の順で
// 飛んでくるので、ここでは2回目の押下だけ処理すればよい。というか
// 処理自体は LeftDown とまったく同じになるので、イベントテーブルで
// 直接 EVT_LEFT_DCLICK => OnLeftDown に割り付けることも出来そうだが
// デバッグが紛らわしくなりそうなので独立したイベントとしておく。
void
WXSoftKeyPanel::OnLeftDClick(wxMouseEvent& event)
{
	const wxPoint& pt = event.GetPosition();
	const keydata_t *key = GetKeydataFromPoint(pt);

	DoLeftDown(key);
}

// マウス右クリック (Down) イベント
// 長押し開始/解除。
// 右クリックのほうは Down だけで Up は使わない。
void
WXSoftKeyPanel::OnRightDown(wxMouseEvent& event)
{
	const wxPoint& pt = event.GetPosition();
	const keydata_t *key = GetKeydataFromPoint(pt);

	if (key != NULL) {
		uint code = key->keycode;
		if (hit[code]) {
			// 押されていたら開放
			KeyUp(*key);
		} else {
			// 押されてなければ長押し開始
			KeyDown(*key);
		}
	}
}

// 座標からキーデータを返す。
// キーに当たっていなければ NULL を返す。
const keydata_t *
WXSoftKeyPanel::GetKeydataFromPoint(const wxPoint& pt) const
{
	// エンターキーだけ先に別判定
	if (enter_region.Contains(pt) == wxInRegion) {
		return &FindKeydata(KC_enter);
	}

	for (const auto& key : GetKeydata()) {
		if (key.x < 0) {
			continue;
		}
		// 始点と幅の +1/-1 は罫線の分
		wxRect rect(
			(key.x * unit) + 1,
			(key.y * unit) + 1,
			(key.w * unit) - 1,
			(key.h * unit) - 1);
		if (rect.Contains(pt)) {
			return &key;
		}
	}
	return NULL;
}

//
// LUNA ソフトウェアキーボード (パネル)
//

// コンストラクタ
WXLunaSoftKeyPanel::WXLunaSoftKeyPanel(wxWindow *parent)
	: inherited(parent, 226, 54)
{
	// LUNA では
	// BS キーの左隣 (0x5c) は円記号ではなくバックスラッシュ。
	// それを SHIFT したやつ(0x7c) は破線でないパイプ。
	// そのもう一つ左隣 (0x7e) はオーバーラインではなくチルダ。
	SetGlyph5C(true);
	SetGlyph7E(true);

	// キートップの濃淡はユニバーサルデザインの推奨色ではないけど
	// ここは見分けを付けるためではなくただのデザインなので。
	if (1) {
		// 白モデル
		color = {
			.Background	= UD_LIGHT_GREY,	// 現物新品は何色だったんだろう
			.Text		= *wxBLACK,
			.Keytop		= { *wxWHITE, wxColour(200, 200, 203) },
			// 白黒共通
			.KeyPressed	= UD_BLUE,
			.Border		= *wxBLACK,
			.LED		= { UD_RED },
		};
	} else {
		// 黒モデル
		color = {
			.Background	= *wxBLACK,
			.Text		= *wxWHITE,
			.Keytop		= { wxColour(50, 50, 53), *wxBLACK },
			// 白黒共通
			.KeyPressed	= UD_BLUE,
			.Border		= *wxBLACK,
			.LED		= { UD_RED },
		};
	}

	DoSize();
}

// デストラクタ
WXLunaSoftKeyPanel::~WXLunaSoftKeyPanel()
{
}

// LED を描画して刻印描画のための y offset を返す。
int
WXLunaSoftKeyPanel::DrawLED(wxDC& dc, const keydata_t& key)
{
	const std::vector<bool>& led = gKeyboard->GetLED();

	if (led[key.led]) {
		// 点灯時は赤、枠は黒
		dc.SetPen(color.Border);
		dc.SetBrush(color.LED[0]);
	} else {
		// 消灯時は元のキートップの色で、枠は薄めにしたい
		dc.SetPen(GetTextDisableColor());
		dc.SetBrush(color.Keytop[key.coltype]);
	}

	// LED を左肩に描く
	wxPoint pt = wxPoint(key.x * unit + 2, key.y * unit + 2);
	wxSize sz = wxSize(4 * unit - 2, 3 * unit - 2);
	dc.DrawRectangle(pt, sz);

	// そのため文字は少し下げる
	return +1;
}

// このキーが押された処理
void
WXLunaSoftKeyPanel::KeyDown(const keydata_t& key)
{
	uint code = key.keycode;

	// 押された
	hit[code] = true;

	// VM に通知 (Keyboard がそれをホストに通知するかはまた別)
	gKeyboard->MakeKey(code);

	// 修飾キー
	if (code == KC_SHIFT_L || code == KC_SHIFT_R) {
		modifier = ModShift;
	}

	// キー状態が変わった(かも知れない)ので再描画
	Refresh();
}

// このキーが放された処理
void
WXLunaSoftKeyPanel::KeyUp(const keydata_t& key)
{
	uint code = key.keycode;

	// 放された
	hit[code] = false;

	// VM に通知 (Keyboard がそれをホストに通知するかはまた別)
	gKeyboard->BreakKey(code);

	// 修飾キー
	if (code == KC_SHIFT_L || code == KC_SHIFT_R) {
		modifier = ModNormal;
	}

	// キー状態が変わった(かも知れない)ので再描画
	Refresh();
}

// キーデータ
#define H	 (8)	// キーの高さ
#define L0	 (2)	// ファンクションキーの行
#define L1	(12)	// ESC の行
#define L2	(20)	// TAB の行
#define L3	(28)	// CTRL の行
#define L4	(36)	// SHIFT の行
#define L5	(44)	// space の行
/*static*/ std::vector<keydata_t>
WXLunaSoftKeyPanel::keydata {
	//                +---- キートップの色(1:濃)
	//                |  +- 0 以上なら LED 番号
	//  X   Y   W  H  C  L	KeyCode				通常	SHIFT
	{   2, L0, 15, H, 1,-1,	KC_F1,				{ "PF1" } },
	{  17, L0, 15, H, 1,-1,	KC_F2,				{ "PF2" } },
	{  32, L0, 15, H, 1,-1,	KC_F3,				{ "PF3" } },
	{  47, L0, 15, H, 1,-1,	KC_F4,				{ "PF4" } },
	{  62, L0, 15, H, 1,-1,	KC_F5,				{ "PF5" } },
	{  81, L0, 15, H, 1,-1,	KC_F6,				{ "PF6" } },
	{  96, L0, 15, H, 1,-1,	KC_F7,				{ "PF7" } },
	{ 111, L0, 15, H, 1,-1,	KC_F8,				{ "PF8" } },
	{ 126, L0, 15, H, 1,-1,	KC_F9,				{ "PF9" } },
	{ 141, L0, 15, H, 1,-1,	KC_F10,				{ "PF10" } },

	{   2, L1, 12, H, 1,-1,	KC_ESC,				{ "ESC" } },
	{  14, L1, 10, H, 0,-1,	KC_1,				{ "1",	"!" } },
	{  24, L1, 10, H, 0,-1,	KC_2,				{ "2",	"\"" } },
	{  34, L1, 10, H, 0,-1,	KC_3,				{ "3",	"#" } },
	{  44, L1, 10, H, 0,-1,	KC_4,				{ "4",	"$" } },
	{  54, L1, 10, H, 0,-1,	KC_5,				{ "5",	"%" } },
	{  64, L1, 10, H, 0,-1,	KC_6,				{ "6",	"&" } },
	{  74, L1, 10, H, 0,-1,	KC_7,				{ "7",	"'" } },
	{  84, L1, 10, H, 0,-1,	KC_8,				{ "8",	"(" } },
	{  94, L1, 10, H, 0,-1,	KC_9,				{ "9",	")" } },
	{ 104, L1, 10, H, 0,-1,	KC_0,				{ "0",	" " } },
	{ 114, L1, 10, H, 0,-1,	KC_minus,			{ "-",	"=" } },
	{ 124, L1, 10, H, 0,-1,	KC_circum,			{ "^",	"~" } },
	{ 134, L1, 10, H, 0,-1,	KC_backslash,		{ "\\",	"|" } },
	{ 144, L1, 12, H, 1,-1,	KC_BS,				{ "BS" } },

	{   2, L2, 17, H, 1,-1,	KC_TAB,				{ "TAB" } },
	{  19, L2, 10, H, 0,-1,	KC_Q,				{ "q",	"Q" } },
	{  29, L2, 10, H, 0,-1,	KC_W,				{ "w",	"W" } },
	{  39, L2, 10, H, 0,-1,	KC_E,				{ "e",	"E" } },
	{  49, L2, 10, H, 0,-1,	KC_R,				{ "r",	"R" } },
	{  59, L2, 10, H, 0,-1,	KC_T,				{ "t",	"T" } },
	{  69, L2, 10, H, 0,-1,	KC_Y,				{ "y",	"Y" } },
	{  79, L2, 10, H, 0,-1,	KC_U,				{ "u",	"U" } },
	{  89, L2, 10, H, 0,-1,	KC_I,				{ "i",	"I" } },
	{  99, L2, 10, H, 0,-1,	KC_O,				{ "o",	"O" } },
	{ 109, L2, 10, H, 0,-1,	KC_P,				{ "p",	"P" } },
	{ 119, L2, 10, H, 0,-1,	KC_at,				{ "@",	"`" } },
	{ 129, L2, 10, H, 0,-1,	KC_bracketleft,		{ "[",	"{" } },

	{   2, L3, 20, H, 1,-1,	KC_CTRL,			{ "CTRL" } },
	{  22, L3, 10, H, 0,-1,	KC_A,				{ "a",	"A" } },
	{  32, L3, 10, H, 0,-1,	KC_S,				{ "s",	"S" } },
	{  42, L3, 10, H, 0,-1,	KC_D,				{ "d",	"D" } },
	{  52, L3, 10, H, 0,-1,	KC_F,				{ "f",	"F" } },
	{  62, L3, 10, H, 0,-1,	KC_G,				{ "g",	"G" } },
	{  72, L3, 10, H, 0,-1,	KC_H,				{ "h",	"H" } },
	{  82, L3, 10, H, 0,-1,	KC_J,				{ "j",	"J" } },
	{  92, L3, 10, H, 0,-1,	KC_K,				{ "k",	"K" } },
	{ 102, L3, 10, H, 0,-1,	KC_L,				{ "l",	"L" } },
	{ 112, L3, 10, H, 0,-1,	KC_semicolon,		{ ";",	"+" } },
	{ 122, L3, 10, H, 0,-1,	KC_colon,			{ ":",	"*" } },
	{ 132, L3, 10, H, 0,-1,	KC_bracketright,	{ "]",	"}" } },

	{   2, L4, 25, H, 1,-1,	KC_SHIFT_L,			{ "SHIFT" } },
	{  27, L4, 10, H, 0,-1,	KC_Z,				{ "z",	"Z" } },
	{  37, L4, 10, H, 0,-1,	KC_X,				{ "x",	"X" } },
	{  47, L4, 10, H, 0,-1,	KC_C,				{ "c",	"C" } },
	{  57, L4, 10, H, 0,-1,	KC_V,				{ "v",	"V" } },
	{  67, L4, 10, H, 0,-1,	KC_B,				{ "b",	"B" } },
	{  77, L4, 10, H, 0,-1,	KC_N,				{ "n",	"N" } },
	{  87, L4, 10, H, 0,-1,	KC_M,				{ "m",	"M" } },
	{  97, L4, 10, H, 0,-1,	KC_comma,			{ ",",	"<" } },
	{ 107, L4, 10, H, 0,-1,	KC_period,			{ ".",	">" } },
	{ 117, L4, 10, H, 0,-1,	KC_slash,			{ "/",	"?" } },
	{ 127, L4, 10, H, 0,-1,	KC_underscore,		{ " ",	"_" } },
	{ 137, L4, 19, H, 1,-1,	KC_SHIFT_R,			{ "SHIFT" } },

	{   2, L5, 20, H, 1,-1,	KC_VALID,			{ "確定" } },
	{  22, L5, 10, H, 1,-1,	KC_SF,				{ "前面" } },
	{  32, L5, 10, H, 1, 1,	KC_CAPS,			{ "CAP" } },
	{  42, L5, 80, H, 0,-1,	KC_space,			{ " " } },
	{ 122, L5, 10, H, 1, 0,	KC_kana,			{ "かな" } },
	{ 132, L5, 24, H, 1,-1,	KC_XFER,			{ "変換" } },

	// カーソルキー島
	{ 160, L1, 10, H, 1,-1,	KC_PF11,			{ "消去" } },
	{ 170, L1, 10, H, 1,-1,	KC_PF12,			{ "呼出" } },
	{ 160, L2, 10, H, 1,-1,	KC_PF13,			{ "文節\n←" } },
	{ 170, L2, 10, H, 1,-1,	KC_PF14,			{ "文節\n→" } },
	{ 160, L3, 20, H, 1,-1,	KC_up,				{ "↑" } },
	{ 160, L4, 10, H, 1,-1,	KC_left,			{ "←" } },
	{ 170, L4, 10, H, 1,-1,	KC_right,			{ "→" } },
	{ 160, L5, 20, H, 1,-1,	KC_down,			{ "↓" } },

	// テンキー島
	{ 184, L1, 10, H, 1,-1,	KC_DEL,				{ "DEL" } },
	{ 194, L1, 10, H, 1,-1,	KC_PAD_plus,		{ "+" } },
	{ 204, L1, 10, H, 1,-1,	KC_PAD_minus,		{ "-" } },
	{ 214, L1, 10, H, 1,-1,	KC_PAD_multiply,	{ "*" } },

	{ 184, L2, 10, H, 0,-1,	KC_PAD_7,			{ "7" } },
	{ 194, L2, 10, H, 0,-1,	KC_PAD_8,			{ "8" } },
	{ 204, L2, 10, H, 0,-1,	KC_PAD_9,			{ "9" } },
	{ 214, L2, 10, H, 1,-1,	KC_PAD_divide,		{ "/" } },

	{ 184, L3, 10, H, 0,-1,	KC_PAD_4,			{ "4" } },
	{ 194, L3, 10, H, 0,-1,	KC_PAD_5,			{ "5" } },
	{ 204, L3, 10, H, 0,-1,	KC_PAD_6,			{ "6" } },
	{ 214, L3, 10, H, 1,-1,	KC_PAD_equal,		{ "=" } },

	{ 184, L4, 10, H, 0,-1,	KC_PAD_1,			{ "1" } },
	{ 194, L4, 10, H, 0,-1,	KC_PAD_2,			{ "2" } },
	{ 204, L4, 10, H, 0,-1,	KC_PAD_3,			{ "3" } },
	{ 214, L4, 10, H, 1,-1,	KC_PAD_comma,		{ "," } },

	{ 184, L5, 10, H, 0,-1,	KC_PAD_0,			{ "0" } },
	{ 194, L5, 10, H, 0,-1,	KC_PAD_decimal,		{ "." } },
	{ 204, L5, 20, H, 1,-1,	KC_PAD_enter,		{ EnterMark } },

	// Enter だけ矩形でないので別処理するが
	// この構造体には全部のキー情報が統一的に並んでいるほうがよかろう。
	{ -1, -1, -1, -1, 1,-1,	KC_enter,			{ EnterMark } },
};


//
// X68k ソフトウェアキーボード (パネル)
//

// コンストラクタ
WXX68kSoftKeyPanel::WXX68kSoftKeyPanel(wxWindow *parent)
	: inherited(parent, 235, 54)
{
	// X68k では SRAM 設定に同期させる。
	uint8 sw = gSRAM->Peek8(0xed0059);
	SetGlyph5C((sw & 0x01));
	SetGlyph7E((sw & 0x02));
	SetGlyph7C((sw & 0x04));

	// とりあえず黒のみ
	color = {
		.Background	= wxColour(50, 50, 53),
		.Text		= *wxWHITE,
		.Keytop		= { *wxBLACK },
		.KeyPressed	= UD_BLUE,
		.Border		= *wxBLACK,
		.LED		= { UD_RED, UD_YELLOW_GREEN },
	};

	DoSize();
}

// デストラクタ
WXX68kSoftKeyPanel::~WXX68kSoftKeyPanel()
{
}

// LED を描画して刻印描画のための y offset を返す。
int
WXX68kSoftKeyPanel::DrawLED(wxDC& dc, const keydata_t& key)
{
	const std::vector<bool>& led = gKeyboard->GetLED();

	// 色を選択
	// idx	刻印	色
	// ----------------
	//  0	かな	赤
	//  1	ローマ	赤
	//  2	コード	赤
	//  3	CAPS	赤
	//  4	INS		赤
	//  5	ひら	緑
	//  6	全角	緑
	if (led[key.led]) {
		int n = (key.led < 5) ? 0 : 1;
		dc.SetPen(color.LED[n]);
		dc.SetBrush(color.LED[n]);
	} else {
		// 消灯時は元のキートップの色で、枠は薄めにしたい
		dc.SetPen(GetTextDisableColor());
		dc.SetBrush(color.Keytop[key.coltype]);
	}

	// LED は中央前方(画面でいうと下)
	wxPoint pt = wxPoint((key.x + 3) * unit, (key.y + key.h - 2) * unit);
	wxSize sz = wxSize(4 * unit, 2 * unit);
	dc.DrawRectangle(pt, sz);

	// そのため文字は少し上げる
	return -1;
}

// このキーが押された処理
void
WXX68kSoftKeyPanel::KeyDown(const keydata_t& key)
{
	uint code = key.keycode;

	// 押された
	hit[code] = true;

	// VM に通知 (Keyboard がそれをホストに通知するかはまた別)
	gKeyboard->MakeKey(code);

	// 修飾キー
	if (code == KC_SHIFT_L || code == KC_SHIFT_R) {
		modifier = ModShift;
	}

	// キー状態が変わった(かも知れない)ので再描画
	Refresh();
}

// このキーが放された処理
void
WXX68kSoftKeyPanel::KeyUp(const keydata_t& key)
{
	uint code = key.keycode;

	hit[code] = false;

	// VM に通知 (Keyboard がそれをホストに通知するかはまた別)
	gKeyboard->BreakKey(code);

	// 修飾キー
	if (code == KC_SHIFT_L || code == KC_SHIFT_R) {
		modifier = ModNormal;
	}

	// キー状態が変わった(かも知れない)ので再描画
	Refresh();
}

#define LF (L0 + 2)		// ファンクションキーは少し低い(横長い)
#define HT (H * 2)		// 縦2倍角のキー
/*static*/ std::vector<keydata_t>
WXX68kSoftKeyPanel::keydata {
	//                   +- 0以上なら LED 番号
	//  X   Y   W  H  C  L	KeyCode				通常	SHIFT
	{   2, L0, 10, H, 0,-1,	KC_BREAK,			{ "BRK" } },
	{  15, L0, 10, H, 0,-1,	KC_COPY,			{ "COPY" } },
	{  29, LF, 12, 6, 0,-1,	KC_F1,				{ "F1" } },
	{  41, LF, 12, 6, 0,-1,	KC_F2,				{ "F2" } },
	{  53, LF, 12, 6, 0,-1,	KC_F3,				{ "F3" } },
	{  65, LF, 12, 6, 0,-1,	KC_F4,				{ "F4" } },
	{  77, LF, 13, 6, 0,-1,	KC_F5,				{ "F5" } },
	{  94, LF, 13, 6, 0,-1,	KC_F6,				{ "F6" } },
	{ 106, LF, 12, 6, 0,-1,	KC_F7,				{ "F7" } },
	{ 118, LF, 12, 6, 0,-1,	KC_F8,				{ "F8" } },
	{ 130, LF, 12, 6, 0,-1,	KC_F9,				{ "F9" } },
	{ 142, LF, 13, 6, 0,-1,	KC_F10,				{ "F10" } },

	{   2, L1, 10, H, 0,-1,	KC_ESC,				{ "ESC" } },
	{  12, L1, 10, H, 0,-1,	KC_1,				{ "1",	"!" } },
	{  22, L1, 10, H, 0,-1,	KC_2,				{ "2",	"\"" } },
	{  32, L1, 10, H, 0,-1,	KC_3,				{ "3",	"#" } },
	{  42, L1, 10, H, 0,-1,	KC_4,				{ "4",	"$" } },
	{  52, L1, 10, H, 0,-1,	KC_5,				{ "5",	"%" } },
	{  62, L1, 10, H, 0,-1,	KC_6,				{ "6",	"&" } },
	{  72, L1, 10, H, 0,-1,	KC_7,				{ "7",	"'" } },
	{  82, L1, 10, H, 0,-1,	KC_8,				{ "8",	"(" } },
	{  92, L1, 10, H, 0,-1,	KC_9,				{ "9",	")" } },
	{ 102, L1, 10, H, 0,-1,	KC_0,				{ "0",	" " } },
	{ 112, L1, 10, H, 0,-1,	KC_minus,			{ "-",	"=" } },
	{ 122, L1, 10, H, 0,-1,	KC_circum,			{ "^",	"~" } },
	{ 132, L1, 10, H, 0,-1,	KC_backslash,		{ "\\",	"|" } },
	{ 142, L1, 13, H, 0,-1,	KC_BS,				{ "BS" } },

	{   2, L2, 16, H, 0,-1,	KC_TAB,				{ "TAB" } },
	{  18, L2, 10, H, 0,-1,	KC_Q,				{ "q",	"Q" } },
	{  28, L2, 10, H, 0,-1,	KC_W,				{ "w",	"W" } },
	{  38, L2, 10, H, 0,-1,	KC_E,				{ "e",	"E" } },
	{  48, L2, 10, H, 0,-1,	KC_R,				{ "r",	"R" } },
	{  58, L2, 10, H, 0,-1,	KC_T,				{ "t",	"T" } },
	{  68, L2, 10, H, 0,-1,	KC_Y,				{ "y",	"Y" } },
	{  78, L2, 10, H, 0,-1,	KC_U,				{ "u",	"U" } },
	{  88, L2, 10, H, 0,-1,	KC_I,				{ "i",	"I" } },
	{  98, L2, 10, H, 0,-1,	KC_O,				{ "o",	"O" } },
	{ 108, L2, 10, H, 0,-1,	KC_P,				{ "p",	"P" } },
	{ 118, L2, 10, H, 0,-1,	KC_at,				{ "@",	"`" } },
	{ 128, L2, 10, H, 0,-1,	KC_bracketleft,		{ "[",	"{" } },

	{   2, L3, 19, H, 0,-1,	KC_CTRL,			{ "CTRL" } },
	{  21, L3, 10, H, 0,-1,	KC_A,				{ "a",	"A" } },
	{  31, L3, 10, H, 0,-1,	KC_S,				{ "s",	"S" } },
	{  41, L3, 10, H, 0,-1,	KC_D,				{ "d",	"D" } },
	{  51, L3, 10, H, 0,-1,	KC_F,				{ "f",	"F" } },
	{  61, L3, 10, H, 0,-1,	KC_G,				{ "g",	"G" } },
	{  71, L3, 10, H, 0,-1,	KC_H,				{ "h",	"H" } },
	{  81, L3, 10, H, 0,-1,	KC_J,				{ "j",	"J" } },
	{  91, L3, 10, H, 0,-1,	KC_K,				{ "k",	"K" } },
	{ 101, L3, 10, H, 0,-1,	KC_L,				{ "l",	"L" } },
	{ 111, L3, 10, H, 0,-1,	KC_semicolon,		{ ";",	"+" } },
	{ 121, L3, 10, H, 0,-1,	KC_colon,			{ ":",	"*" } },
	{ 131, L3, 10, H, 0,-1,	KC_bracketright,	{ "]",	"}" } },

	{   2, L4, 24, H, 0,-1,	KC_SHIFT_L,			{ "SHIFT" } },
	{  26, L4, 10, H, 0,-1,	KC_Z,				{ "z",	"Z" } },
	{  36, L4, 10, H, 0,-1,	KC_X,				{ "x",	"X" } },
	{  46, L4, 10, H, 0,-1,	KC_C,				{ "c",	"C" } },
	{  56, L4, 10, H, 0,-1,	KC_V,				{ "v",	"V" } },
	{  66, L4, 10, H, 0,-1,	KC_B,				{ "b",	"B" } },
	{  76, L4, 10, H, 0,-1,	KC_N,				{ "n",	"N" } },
	{  86, L4, 10, H, 0,-1,	KC_M,				{ "m",	"M" } },
	{  96, L4, 10, H, 0,-1,	KC_comma,			{ ",",	"<" } },
	{ 106, L4, 10, H, 0,-1,	KC_period,			{ ".",	">" } },
	{ 116, L4, 10, H, 0,-1,	KC_slash,			{ "/",	"?" } },
	{ 126, L4, 10, H, 0,-1,	KC_underscore,		{ " ",	"_" } },
	{ 136, L4, 19, H, 0,-1,	KC_SHIFT_R,			{ "SHIFT" } },

	{  21, L5, 10, H, 0, 5,	KC_hira,			{ "ひら" } },
	{  31, L5, 12, H, 0,-1,	KC_XF1,				{ "XF1" } },
	{  43, L5, 13, H, 0,-1,	KC_XF2,				{ "XF2" } },
	{  56, L5, 36, H, 0,-1,	KC_space,			{ " " } },
	{  92, L5, 12, H, 0,-1,	KC_XF3,				{ "XF3" } },
	{ 104, L5, 12, H, 0,-1,	KC_XF4,				{ "XF4" } },
	{ 116, L5, 12, H, 0,-1,	KC_XF5,				{ "XF5" } },
	{ 128, L5, 10, H, 0, 6,	KC_zenkaku,			{ "全角" } },

	// カーソルキー島
	{ 159, L0, 10, H, 0, 0,	KC_kana,			{ "かな" } },
	{ 169, L0, 10, H, 0, 1,	KC_romaji,			{ "ﾛｰﾏ" } },
	{ 179, L0, 10, H, 0, 2,	KC_code,			{ "ｺｰﾄﾞ" } },
	{ 159, L1, 10, H, 0,-1,	KC_HOME,			{ "HOME" } },
	{ 169, L1, 10, H, 0, 4,	KC_INS,				{ "INS" } },
	{ 179, L1, 10, H, 0,-1,	KC_DEL,				{ "DEL" } },
	{ 159, L2, 10, H, 0,-1,	KC_ROLLUP,			{ "ROLL\n UP" } },
	{ 169, L2, 10, H, 0,-1,	KC_ROLLDOWN,		{ "ROLL\nDOWN" } },
	{ 179, L2, 10, H, 0,-1,	KC_UNDO,			{ "UNDO" } },
	{ 169, L3, 10, H, 0,-1,	KC_up,				{ "↑" } },
	{ 159, L3, 10, HT,0,-1,	KC_left,			{ "←" } },
	{ 179, L3, 10, HT,0,-1,	KC_right,			{ "→" } },
	{ 169, L4, 10, H, 0,-1,	KC_down,			{ "↓" } },
	{ 159, L5, 15, H, 0,-1,	KC_OPT1,			{ "OPT.1" } },
	{ 174, L5, 15, H, 0,-1,	KC_OPT2,			{ "OPT.2" } },

	// テンキー島
	{ 193, L0, 10, H, 0, 3,	KC_CAPS,			{ "CAPS" } },
	{ 203, L0, 10, H, 0,-1,	KC_kigou,			{ "記号" } },
	{ 213, L0, 10, H, 0,-1,	KC_touroku,			{ "登録" } },
	{ 223, L0, 10, H, 0,-1,	KC_HELP,			{ "HELP" } },
	{ 193, L1, 10, H, 0,-1,	KC_PAD_CLR,			{ "CLR" } },
	{ 203, L1, 10, H, 0,-1,	KC_PAD_divide,		{ "/" } },
	{ 213, L1, 10, H, 0,-1,	KC_PAD_multiply,	{ "*" } },
	{ 223, L1, 10, H, 0,-1,	KC_PAD_minus,		{ "-" } },
	{ 193, L2, 10, H, 0,-1,	KC_PAD_7,			{ "7" } },
	{ 203, L2, 10, H, 0,-1,	KC_PAD_8,			{ "8" } },
	{ 213, L2, 10, H, 0,-1,	KC_PAD_9,			{ "9" } },
	{ 223, L2, 10, H, 0,-1,	KC_PAD_plus,		{ "+" } },
	{ 193, L3, 10, H, 0,-1,	KC_PAD_4,			{ "4" } },
	{ 203, L3, 10, H, 0,-1,	KC_PAD_5,			{ "5" } },
	{ 213, L3, 10, H, 0,-1,	KC_PAD_6,			{ "6" } },
	{ 223, L3, 10, H, 0,-1,	KC_PAD_equal,		{ "=" } },
	{ 193, L4, 10, H, 0,-1,	KC_PAD_1,			{ "1" } },
	{ 203, L4, 10, H, 0,-1,	KC_PAD_2,			{ "2" } },
	{ 213, L4, 10, H, 0,-1,	KC_PAD_3,			{ "3" } },
	{ 223, L4, 10, HT,0,-1,	KC_PAD_enter,		{ "ENT" } },
	{ 193, L5, 10, H, 0,-1,	KC_PAD_0,			{ "0" } },
	{ 203, L5, 10, H, 0,-1,	KC_PAD_comma,		{ "," } },
	{ 213, L5, 10, H, 0,-1,	KC_PAD_decimal,		{ "." } },

	// Enter だけ矩形でないので別処理するが
	// この構造体には全部のキー情報が統一的に並んでいるほうがよかろう。
	{ -1, -1, -1, -1, 0,-1,	KC_enter,			{ EnterMark } },
};

//
// LUNA ソフトウェアキーボード (ウィンドウ)
//

// コンストラクタ
WXLunaSoftKeyWindow::WXLunaSoftKeyWindow(wxWindow *parent, wxWindowID id)
	: inherited(parent, id, _("Software Keyboard"))
{
	panel = new WXLunaSoftKeyPanel(this);

	// 大きさをそれに固定
	Layout();
	Fit();
}

// デストラクタ
WXLunaSoftKeyWindow::~WXLunaSoftKeyWindow()
{
}

//
// X680x0 ソフトウェアキーボード (ウィンドウ)
//

// コンストラクタ
WXX68kSoftKeyWindow::WXX68kSoftKeyWindow(wxWindow *parent, wxWindowID id)
	: inherited(parent, id, _("Software Keyboard"))
{
	panel = new WXX68kSoftKeyPanel(this);

	// 大きさをそれに固定
	Layout();
	Fit();
}

// デストラクタ
WXX68kSoftKeyWindow::~WXX68kSoftKeyWindow()
{
}
