//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#pragma once

#include "wxsubwindow.h"
#include "wxtextpanel.h"
#include "keyboard.h"
#include <array>
#include <vector>

// キー描画用のデータ構造体
struct keydata_t
{
	// フォントサイズの切り替えに対応するため、座標系の単位はピクセルでは
	// ない点に注意。フォントサイズ(高さ)の 1/4 を 1 と表現する。つまり
	// どのフォントサイズでも半角1文字は 2 x 4。
	int x;	// このキー枠の左上点
	int y;	// このキー枠の左上点
	int w;	// このキー枠の幅
	int h;	// このキー枠の高さ

	// 色種別
	// LUNA の場合はキートップ色の区別に使う。1=濃い, 0=通常。
	// X68k では使用しない。
	int coltype;

	int led;			// -1ならLEDなし、0以上なら LED 番号
	uint keycode;		// キーコード (keyboard.h の KC_*)

	// 刻印
	// [0]: 通常
	// [1]: SHIFT時 (NULL なら通常のまま)
	const char *disp[2];

	int right() const { return x + w; }		// このキー枠の右
	int bottom() const { return y + h; }	// このキー枠の下
};

// ソフトウェアキーボード (パネル) 共通部分
class WXSoftKeyPanel : public WXTextPanel
{
	using inherited = WXTextPanel;
 protected:
	// 色の集合(パレット)
	struct ColorPalette {
		wxColour Background;	// 背景というかガワの部分の色
		wxColour Text;			// 文字色
		wxColour Keytop[2];		// キートップ色。0:通常色、1:(あれば)濃色
		wxColour KeyPressed;	// キーが押された時のキートップ色
		wxColour Border;		// LED 枠の薄い輪郭色
		wxColour LED[2];		// LED。色は各機種定義 (今のところ0:赤 1:緑)
	};

 public:
	WXSoftKeyPanel(wxWindow *parent, int u_width_, int u_height_);
	virtual ~WXSoftKeyPanel() override;

	// フォントサイズ変更
	void SetFontSize(FontId fontid) override;

 protected:
	void InitLED();
	void DoSize();
	void OnSize(wxSizeEvent& event);
	void OnPaint(wxPaintEvent& event);

	void Draw(wxDC& dc);
	void DrawKey(wxDC& dc, const keydata_t& key);
	void DrawEnterKey(wxDC& dc);
	void DrawEnterMark(wxDC& dc, const wxPoint& basep);
	virtual int DrawLED(wxDC& dc, const keydata_t& key) = 0;
	const keydata_t& FindKeydata(uint) const;

	void OnLeftDown(wxMouseEvent& event);
	void DoLeftDown(const keydata_t *key);
	void OnLeftUp(wxMouseEvent& event);
	void DoLeftUp(const keydata_t *key);
	void OnLeftDClick(wxMouseEvent& event);
	void OnRightDown(wxMouseEvent& event);
	virtual void KeyDown(const keydata_t& key) = 0;
	virtual void KeyUp(const keydata_t& key) = 0;

	const keydata_t *GetKeydataFromPoint(const wxPoint& pt) const;

	// 押されていれば true (左クリック、右クリック区別なく両方)。
	// 共通キーコードによる配列。
	std::array<bool, KC_max> hit {};

	// 左クリック押下時のキー
	const keydata_t *leftdown_key = NULL;

	// 修飾キーの状態 (刻印を変更するため)
	enum {
		ModNormal = 0,
		ModShift,
	} modifier {};

	// フォントサイズに相対的な描画単位
	int unit = 0;

	// コントロールのサイズ [基準単位]
	int u_width = 0;
	int u_height = 0;

	// コントロールのサイズ [px]
	int view_width = 0;
	int view_height = 0;

	std::unique_ptr<uint8[]> bmpbuf {};
	std::unique_ptr<wxImage> image {};
	std::unique_ptr<wxBitmap> bitmap {};

	// エンター記号の周囲の点集合
	std::vector<wxPoint> enter_points;
	wxRegion enter_region;

	// 色
	ColorPalette color {};

	// UTF-8 -> CP932 変換
	// Unicode の "\\" は Shift_JIS に対応する文字がないため、これを
	// Unicode から Shift_JIS に変換すると "" になり、これではキートップが
	// 刻印できなくて困る。
	// CP932 なら "\\" を変換しても "\\" ('\x5c') になってくれるので、
	// これを使う。この \x5c をバックスラッシュだと思うか円記号だと思うかは
	// こちらで決める。
	wxCSConv conv { "CP932" };

	// キー描画データ
	virtual const std::vector<keydata_t>& GetKeydata() const = 0;

	// イベントテーブル
	wxDECLARE_EVENT_TABLE();
};

// LUNA ソフトウェアキーボード (パネル)
class WXLunaSoftKeyPanel : public WXSoftKeyPanel
{
	using inherited = WXSoftKeyPanel;
 public:
	WXLunaSoftKeyPanel(wxWindow *parent);
	virtual ~WXLunaSoftKeyPanel();

 private:
	virtual int DrawLED(wxDC& dc, const keydata_t& key);
	virtual void KeyDown(const keydata_t& key);
	virtual void KeyUp(const keydata_t& key);

	// キー描画データ
	virtual const std::vector<keydata_t>& GetKeydata() const { return keydata; }
	static std::vector<keydata_t> keydata;
};

// X68k ソフトウェアキーボード (パネル)
class WXX68kSoftKeyPanel : public WXSoftKeyPanel
{
	using inherited = WXSoftKeyPanel;
 public:
	WXX68kSoftKeyPanel(wxWindow *parent);
	virtual ~WXX68kSoftKeyPanel();

 private:
	virtual int DrawLED(wxDC& dc, const keydata_t& key);
	virtual void KeyDown(const keydata_t& key);
	virtual void KeyUp(const keydata_t& key);

	// キー描画データ
	virtual const std::vector<keydata_t>& GetKeydata() const { return keydata; }
	static std::vector<keydata_t> keydata;
};

// LUNA ソフトウェアキーボード (ウィンドウ)
class WXLunaSoftKeyWindow : public WXSubWindow
{
	using inherited = WXSubWindow;
 public:
	WXLunaSoftKeyWindow(wxWindow *parent, wxWindowID id);
	virtual ~WXLunaSoftKeyWindow();

 private:
	// キーボードパネル
	WXLunaSoftKeyPanel *panel = NULL;
};

// X68k ソフトウェアキーボード (ウィンドウ)
class WXX68kSoftKeyWindow : public WXSubWindow
{
	using inherited = WXSubWindow;
 public:
	WXX68kSoftKeyWindow(wxWindow *parent, wxWindowID id);
	virtual ~WXX68kSoftKeyWindow();

 private:
	// キーボードパネル
	WXX68kSoftKeyPanel *panel = NULL;
};
