//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxtextpanel.h"
#include "wxcolor.h"
#include "bitrev.h"
#include "cgrom.h"
#include "kaname12.h"

// wxOSX(-3.0.2) の wxLIGHT_GREY は値が(色も透明度も)正しくない。
// wxGTK の wxLIGHT_GREY は 0xffd3d3d3 だけどこれを Mac で見ると
// (Mac の統一感としては) 暗く見えるので色コード自体を変える。
#if defined(__WXOSX__)
#define LIGHT_GREY	wxColour(0xffe0e0e0)
#else
#define LIGHT_GREY	*wxLIGHT_GREY
#endif

//
// テキストフォントを管理するクラス
//

static const wxSize fontinfo[] = {
	// 順序はヘッダの FontId_* 識別子の順と揃えること
	FONT_6x12,
	FONT_8x16,
};
static_assert(countof(fontinfo) == FontId::Max, "");

// コンストラクタ
WXTextPanel::WXTextPanel(wxWindow *parent)
	: inherited(parent, wxID_ANY)
{
	UpdateFont(global_fontid);

	// デフォルト色
	SetTextColor(*wxBLACK, LIGHT_GREY, wxColour(0x50, 0x50, 0x50));
}

// デストラクタ
WXTextPanel::~WXTextPanel()
{
	FreeFont(fontid);
}

// フォント関連の内部情報をよしなに更新する
void
WXTextPanel::UpdateFont(FontId new_fontid)
{
	assert(fontid != new_fontid);

	// フォントデータを更新
	FreeFont(fontid);
	AllocFont(new_fontid);

	// 内部変数を更新
	fontid = new_fontid;
	fontsize = fontinfo[fontid];
	ascii_glyph = ascii_glyph_table[fontid];
	bold_glyph = bold_glyph_table[fontid];
}

// 必要ならフォントを用意する
/*static*/ void
WXTextPanel::AllocFont(FontId fontid)
{
	const uint8 *base;

	std::unique_lock<std::mutex> lock(global_font_mutex);
	if (font_refcnt[fontid]++ != 0) {
		return;
	}

	// 6x12, 8x16 半角は X68030 CGROM にあるのでそれを使う
	switch (fontid) {
	 case FontId::_6x12:
		base = BuiltinCGROM::Get6x12();
		break;
	 case FontId::_8x16:
		base = BuiltinCGROM::Get8x16();
		break;
	 default:
		assertmsg(0, "fontid=%d", fontid);
	}

	// フォントの行ストライド[byte] を計算。
	// ここは static 関数内なので WXTextPanel インスタンス内の変数
	// font_width, font_height は見えない。(shadow にもならない)
	int font_width = fontinfo[fontid].GetWidth();
	int font_height = fontinfo[fontid].GetHeight();
	int font_stride = (font_width + 7) / 8;
	int bytes = font_stride * font_height;
	std::vector<char> buf(bytes);

	wxBitmap **ascii_glyph = ascii_glyph_table[fontid];
	wxBitmap **bold_glyph = bold_glyph_table[fontid];
	// ASCII というか半角というかは通常 0x7f までだが
	// X680x0 の CGROM の 0x80, 0x81, 0x82 には '\\', '~', '|' が
	// 収納されているのでここまで含めてデータを作っておく。
	for (int ch = 0; ch < countof(ascii_glyph_table[0]); ch++) {
		const uint8 *ptr = base + ch * bytes;

		// wxBitmap は XBM 形式で 1bpp ビットマップを受け取る。
		// XBM は MSB が右なので、ビット左右を反転する
		for (int i = 0; i < bytes; i++) {
			uint8 src = ptr[i];
			buf[i] = bitrev(src);
		}
		// その XBM データから1文字分のビットマップオブジェクトを作成
		ascii_glyph[ch] = new wxBitmap(&buf[0], font_width, font_height, 1);

		// 続いてボールドに加工
		for (int i = 0; i < bytes; i++) {
			// XBM 形式は MSB が右なので左シフトすると一般的なボールド。
			buf[i] |= ((uint8)buf[i] << 1);
		}
		bold_glyph[ch] = new wxBitmap(&buf[0], font_width, font_height, 1);
	}
}

// 最後の一人ならフォントを解放する
/*static*/ void
WXTextPanel::FreeFont(FontId fontid)
{
	std::unique_lock<std::mutex> lock(global_font_mutex);

	if (fontid < 0 || fontid >= FontId::Max) {
		return;
	}
	if (--font_refcnt[fontid] != 0) {
		return;
	}

	wxBitmap **ascii_glyph = ascii_glyph_table[fontid];
	wxBitmap **bold_glyph = bold_glyph_table[fontid];
	for (int i = 0; i < countof(ascii_glyph_table[0]); i++) {
		if (ascii_glyph[i]) {
			delete ascii_glyph[i];
			ascii_glyph[i] = NULL;
		}
		if (bold_glyph[i]) {
			delete bold_glyph[i];
			bold_glyph[i] = NULL;
		}
	}
}

// テキスト色を指定する
void
WXTextPanel::SetTextColor(const wxColour& fg, const wxColour& bg,
	const wxColour& dc)
{
	text_fgcolor = fg;
	text_bgcolor = bg;
	text_disable_color = dc;
}

// (px, py) を左上とする座標から Shift_JIS 文字列を属性 attr で描画する。
// 文字列中で属性を変えることはできない。
// キャンバスをはみ出すかどうかについてはこちらでは関知しない。
// フォントサイズは現行のフォント。
void
WXTextPanel::DrawStringSJIS(wxDC& dc, int px, int py, const char *sjis,
	uint attr)
{
	const uint8 *s;

	for (s = (const uint8 *)sjis; *s; ) {
		if (__predict_true(*s < 0x80)) {
			// ASCII
			DrawChar1(dc, px, py, *s++, attr);
			px += font_width;
		} else if (0xa0 <= *s && *s < 0xe0) {
			// 半角カナ
			DrawCharKana(dc, px, py, *s++, attr);
			px += font_width;
		} else {
			// 全角
			int code;
			code = (*s++ << 8);
			code |= *s++;
			DrawChar2(dc, px, py, code, attr);
			px += font_width * 2;
		}
	}
}

// (px, py) を左上とする座標からASCII文字 code を属性 attr で描画する。
// code は 0x80 未満。
// フォントサイズは現行のフォント。
void
WXTextPanel::DrawChar1(wxDC& dc, int px, int py, uint code, uint attr)
{
	wxBitmap **glyph;

	assertmsg(code < 0x80, "code=0x%x", code);

	// ボールドかどうかでテーブルが別
	if ((attr & TA::Em)) {
		glyph = bold_glyph;
	} else {
		glyph = ascii_glyph;
	}

	// 変更可能なグリフ
	if (__predict_false(code == 0x5c)) {
		if (glyph_5c) {
			code = 0x80;
		}
	} else if (__predict_false(code == 0x7e)) {
		if (glyph_7e) {
			code = 0x81;
		}
	} else if (__predict_false(code == 0x7c)) {
		if (glyph_7c) {
			code = 0x82;
		}
	}
	DrawBitmap(dc, px, py, *glyph[code], attr);
}

// (px, py) を左上とする座標から半角カナ1文字を出力する。
// フォントサイズは現行のフォント。
void
WXTextPanel::DrawCharKana(wxDC& dc, int px, int py, uint sjiscode, uint attr)
{
	std::unique_ptr<wxBitmap> kana;
	kana.reset(KanaBitmap(sjiscode, attr));
	DrawBitmap(dc, px, py, *kana, attr);
}

// (px, py) を左上とする座標から漢字1文字を出力する。
// sjiscode は Shift_JIS。
// フォントサイズは現行のフォント。
void
WXTextPanel::DrawChar2(wxDC& dc, int px, int py, uint sjiscode, uint attr)
{
	std::unique_ptr<wxBitmap> kanji;
	kanji.reset(KanjiBitmap(sjiscode, attr));
	DrawBitmap(dc, px, py, *kanji, attr);
}

// (px, py) を左上とする座標に bitmap を描画する。
// この DrawBitmap() は dc のペン色を設定して dc.DrawBitmap() を呼ぶ。
// ちょっと紛らわしいので注意。
void
WXTextPanel::DrawBitmap(wxDC& dc, int px, int py, const wxBitmap& bitmap,
	uint attr)
{
	// 文字の背景は透過ではなく背景色で塗りつぶす
	// XXX もっと初期化時に近いところへ移してよさそうだが
	dc.SetBackgroundMode(wxSOLID);

	// 属性を設定
	switch (attr) {
	 case TA::On:
		dc.SetTextForeground(text_bgcolor);
		dc.SetTextBackground(text_fgcolor);
		break;
	 case TA::Disable:
		dc.SetTextForeground(text_disable_color);
		dc.SetTextBackground(text_bgcolor);
		break;
	 default:
		dc.SetTextForeground(text_fgcolor);
		dc.SetTextBackground(text_bgcolor);
		break;
	}

	// 1文字出力
	dc.DrawBitmap(bitmap, px, py);

	// 1文字ずつ属性を元に戻す
	dc.SetTextForeground(text_fgcolor);
	dc.SetTextBackground(text_bgcolor);
}

// Shift_JIS の半角カナ1文字分のビットマップを作成する。
wxBitmap *
WXTextPanel::KanaBitmap(uint16 sjiscode, uint attr)
{
	const uint8 *base;
	// X680x0 CGROM
	switch (fontid) {
	 case FontId::_6x12:
		base = BuiltinCGROM::Get6x12();
		break;
	 case FontId::_8x16:
		base = BuiltinCGROM::Get8x16();
		break;
	 default:
		__unreachable();
	}

	int bytes = ((font_width + 7) / 8) * font_height;
	std::vector<char> buf(bytes);

	const uint8 *ptr;
	ptr = base + sjiscode * bytes;

	for (int i = 0; i < bytes; i++) {
		buf[i] = bitrev(ptr[i]);
	}
	if (attr == TA::Em) {
		for (int i = 0; i < bytes; i++) {
			buf[i] |= buf[i] >> 1;
		}
	}

	// その XBM データから1文字分のビットマップオブジェクトを作成
	return new wxBitmap(&buf[0], font_width, font_height, 1);
}

// Shift_JIS の漢字1文字のビットマップを作成する。
wxBitmap *
WXTextPanel::KanjiBitmap(uint16 sjiscode, uint attr)
{
	// Shift_JIS を JIS というかコードポイントに変換
	uint sh = (sjiscode >> 8);
	uint sl = sjiscode & 0xff;
	uint jh;
	uint jl;
	if (sh < 0xa0) {
		jh = sh - 0x71;
	} else {
		jh = sh - 0xb1;
	}
	jh = jh * 2 + 1;
	jl = sl;
	if (sl > 0x7f) {
		jl--;
	}
	if (jl < 0x9e) {
		jl -= 0x1f;
	} else {
		jl -= 0x7d;
		jh++;
	}
	// ここで JIS。これを CGROM 格納コードポイントに変換。
	jh -= 0x21;
	jl -= 0x21;
	if (jh >= 15) {
		jh -= 7;
	}
	int code = jh * 94 + jl;

	// XXX 見付からなかったらゲタ文字にしたい
	if (code <0) code=0x6b;

	// フォント領域
	const uint8 *base;
	switch (fontid) {
	 case FontId::_6x12:
		// 12x12 は内蔵フォントデータ
		base = builtin_kanji12;
		break;
	 case FontId::_8x16:
		// 16x16 は X680x0 CGROM のを使う。
		base = BuiltinCGROM::Get16x16();
		break;
	 default:
		// ここより先に必ず半角のほうでエラーが出るはず
		__unreachable();
	}

	int bytes = (((font_width * 2) + 7) / 8) * font_height;
	std::vector<char> buf(bytes);

	const uint8 *ptr;
	ptr = base + code * bytes;

	for (int i = 0; i < bytes; i++) {
		buf[i] = bitrev(ptr[i]);
	}
	if (attr == TA::Em) {
		for (int i = 0; i < bytes; i++) {
			buf[i] |= buf[i] >> 1;
		}
	}

	// その XBM データから1文字分のビットマップオブジェクトを作成
	return new wxBitmap(&buf[0], font_width * 2, font_height, 1);
}

// スタティック変数
FontId WXTextPanel::global_fontid;
std::mutex WXTextPanel::global_font_mutex;
u_int WXTextPanel::font_refcnt[FontId::Max];
wxBitmap *WXTextPanel::ascii_glyph_table[FontId::Max][0x83];
wxBitmap *WXTextPanel::bold_glyph_table[FontId::Max][0x83];
