//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxtextscreen.h"
#include "wxmainframe.h"
#include "bitrev.h"
#include "cgrom.h"

//
// テキストスクリーンを表示するパネル
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXTextScreen, inherited)
	EVT_PAINT(WXTextScreen::OnPaint)
	EVT_SIZE(WXTextScreen::OnSize)
	EVT_TIMER(wxID_ANY, WXTextScreen::OnTimer)
wxEND_EVENT_TABLE()

// コンストラクタ
WXTextScreen::WXTextScreen(wxWindow *parent, IMonitor& monitor_)
	: inherited(parent)
	, monitor(monitor_)
{
	// モニタサイズから初期サイズを決定
	nnSize ms = monitor.GetMonitorSize();
	wxSize size = wxSize(ms.width * font_width, ms.height * font_height);
	SetClientSize(size);
	SetMinSize(size);

	timer.SetOwner(this);

	// 現在の設定値を適用
	SetRate(gMainFrame->GetMonitorRate());

	// 最初に一回描画を起こす
	wxTimerEvent dummy;
	AddPendingEvent(dummy);
}

// デストラクタ
WXTextScreen::~WXTextScreen()
{
	bitmap.reset();
	image.reset();
	bmpbuf.reset();
	prevbuf.reset();
}

// 画面更新頻度を Hz で設定する。
void
WXTextScreen::SetRate(int hz)
{
	timer.Start(1000 / hz);
}

// フォントサイズ変更 (外部インタフェース)
void
WXTextScreen::SetFontSize(FontId new_fontid)
{
	// 同じなら何もしない
	if (fontid == new_fontid) {
		return;
	}

	// フォントを更新
	UpdateFont(new_fontid);

	// 再描画を強制する。
	// テキストの桁数行数は変わってないので再確保はしなくていい。
	int col = ts.GetCol();
	int row = ts.GetRow();
	assert((bool)prevbuf);
	memset(&prevbuf[0], 0xff, col * row * sizeof(prevbuf[0]));

	// コントロールの大きさを変更
	wxSize size = wxSize(col * font_width, row * font_height);
	SetClientSize(size);
	SetMinSize(size);
}

// サイズ変更イベント
void
WXTextScreen::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある
	// GTK では 1 で来ることがある
	const wxSize& size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	// テキストサイズが変わった時だけテキストバッファを再構成。
	// (フォントサイズを変えると桁数×行数が変わらず画面サイズが変わる
	// ということは起きることに注意)
	int col = size.x / font_width;
	int row = size.y / font_height;
	if (col != ts.GetCol() || row != ts.GetRow()) {
		ts.Init(col, row);
		prevbuf.reset(new uint16 [col * row]);
		memset(&prevbuf[0], 0xff, col * row * sizeof(prevbuf[0]));
	}

	// ビットマップサイズが変わった時だけビットマップバッファを再構成。
	if (col * font_width != view_width || row * font_height != view_height) {
		// ビットマップを持っていれば (この順序で) 一旦解放
		bitmap.reset();
		image.reset();
		bmpbuf.reset();

		// ビットマップ作成
		view_width  = col * font_width;
		view_height = row * font_height;
		bmpbuf.reset(new uint8 [view_width * view_height * 3]);
		// 正しく描かれなかった時に分かりやすいよう初期値は濃いグレーにしておく
		memset(&bmpbuf[0], 0x70, view_width * view_height * 3);
		image.reset(new wxImage(view_width, view_height, bmpbuf.get(), true));
		bitmap.reset(new wxBitmap(*image));
	}
}

// タイマーイベント
void
WXTextScreen::OnTimer(wxTimerEvent& event)
{
	// モニタースクリーンを更新して再描画。
	monitor.MonitorUpdate(ts);
	RefreshIfUpdated();
}

// 必要なら画面を更新して再描画する
void
WXTextScreen::RefreshIfUpdated()
{
	int len = ts.GetCol() * ts.GetRow() * sizeof(prevbuf[0]);
	if (memcmp(ts.GetBuf(), &prevbuf[0], len) != 0) {
		Refresh();
	}
}

// 描画イベント (UI スレッドから呼ばれる)
void
WXTextScreen::OnPaint(wxPaintEvent& event)
{
	// メモリ DC に描画
	wxMemoryDC memDC;
	memDC.SelectObject(*bitmap);
	// 基本的には RefreshIfUpdated() により更新があった時だけ呼ばれる
	Draw(memDC);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	dstDC.Blit(0, 0, view_width, view_height, &memDC, 0, 0);
}

// テキストバッファの内容をパネルのバックグラウンドバッファに描画する。
void
WXTextScreen::Draw(wxDC& dc)
{
	const uint16 *text = ts.GetBuf();
	uint16 *prev = &prevbuf[0];
	int col = ts.GetCol();
	int row = ts.GetRow();
	int px = 0;
	int py = 0;
	for (int y = 0; y < row; y++, py += font_height) {
		px = 0;
		for (int x = 0; x < col; ) {
			uint16 data = *text;
			uint attr = (data & 0xff00);
			uint chr  = (data & 0xff);
			if (__predict_true(chr < 128)) {
				// ASCII
				if (*prev != data) {
					DrawChar1(dc, px, py, chr, attr);
					*prev = data;
				}
				prev++;
				text++;
				x++;
				px += font_width;
			} else {
				// マルチバイト文字は Shift_JIS として処理する
				uint code = (chr << 8) | (text[1] & 0xff);
				if (prev[0] != data || prev[1] != text[1]) {
					DrawChar2(dc, px, py, code, attr);
				}
				*prev++ = *text++;
				*prev++ = *text++;
				x += 2;
				px += font_width * 2;
			}
		}
	}
}
