//
// nono
// Copyright (C) 2020 nono project
// Licensed under nono-license.txt
//

#include "wxvideomonitor.h"
#include "bitmap.h"

//
// LUNA ビットマップパネル
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXLunaBitmapPanel, inherited)
	EVT_PAINT(WXLunaBitmapPanel::OnPaint)
	EVT_SIZE(WXLunaBitmapPanel::OnSize)
	EVT_SCROLLWIN(WXLunaBitmapPanel::OnScrollWin)
wxEND_EVENT_TABLE()

// コンストラクタ
WXLunaBitmapPanel::WXLunaBitmapPanel(wxWindow *parent)
	: inherited(parent, wxID_ANY, wxDefaultPosition,
		wxSize(512, 384))	// 初期値は適当
{
	src = gBitmap->GetBuf();

	// 仮想キャンバスサイズを設定
	SetVirtualSize(virtual_width, virtual_height);
	// スクロールバーを設定
	SetScrollRate(width_per_unit, height_per_unit);
	SetScrollbars(width_per_unit, height_per_unit,
		virtual_width / width_per_unit,
		virtual_height / height_per_unit);
}

// デストラクタ
WXLunaBitmapPanel::~WXLunaBitmapPanel()
{
}

// サイズ変更イベント
void
WXLunaBitmapPanel::OnSize(wxSizeEvent& event)
{
	// Mac ではウィンドウ作成時に 0 以下の値が来ることがある
	// GTK では 1 で来ることがある
	const wxSize& size = event.GetSize();
	if (size.x <= 1 || size.y <= 1) {
		return;
	}

	view_width = size.x;
	view_height = size.y;

	// BMP バッファを持っていれば一旦解放
	bmpbuf.reset();

	// BMP バッファ作成
	// 正しく描かれなかった時に分かりやすいよう初期値は濃いグレーにしておく
	bmpbuf.reset(new uint8 [view_width * view_height * 3]);
	memset(&bmpbuf[0], 0x70, view_width * view_height * 3);
}

// スクロールイベント
void
WXLunaBitmapPanel::OnScrollWin(wxScrollWinEvent& event)
{
	int orient = event.GetOrientation();
	int pos = event.GetPosition();

	if (orient == wxHORIZONTAL) {
		view_x = pos * width_per_unit;
	} else {
		view_y = pos * height_per_unit;
	}
}

// XXX 実際にはこの辺かどこかで更新チェックをしないといけない。
// 今は毎回全描画しているので画面サイズが大きくなると重くなる。

// 描画イベント (UI スレッドから呼ばれる)
void
WXLunaBitmapPanel::OnPaint(wxPaintEvent& event)
{
	// 更新があれば、ここで裏バッファに描画
	Draw();
	// バッファから wxBitmap を作成
	wxImage image(view_width, view_height, bmpbuf.get(), true);
	wxBitmap bitmap(image);

	// 実画面 DC にコピー
	wxPaintDC dstDC(this);
	dstDC.DrawBitmap(bitmap, 0, 0);
}

// ビットマッププレーンの内容をパネルのバックグラウンドバッファに描画する
void
WXLunaBitmapPanel::Draw()
{
	int width = std::min(2048, view_width);
	int height = std::min(1024, view_height);
	// X 方向は32bit(ロングワード)単位で処理
	int x_offset = view_x / width_per_unit;

	for (int y = 0; y < height; y++) {
		const uint32 *s = (const uint32 *)&src[(view_y + y) * 256];
		uint8 *d = &bmpbuf[y * (width * 3)];

		// dst を1ラインクリアしてから..
		memset(d, 0, width * 3);
		// ON になってるドットだけ打っていく
		for (int x = 0; x < width / 32; x++) {
			uint32 srcdata = s[x_offset + x];
			for (int i = 0; srcdata != 0; srcdata <<= 1, i++) {
				uint8 dst;
				if ((int32)srcdata < 0) {
					dst = 0xff;
				} else {
					dst = 0;
				}
				d[(x * 32 + i) * 3 + 0] = dst;
				d[(x * 32 + i) * 3 + 1] = dst;
				d[(x * 32 + i) * 3 + 2] = dst;
			}
		}
	}
}

//
// LUNA ビットマップウィンドウ
//

// イベントテーブル
wxBEGIN_EVENT_TABLE(WXLunaBitmapWindow, inherited)
	EVT_TIMER(wxID_ANY, WXLunaBitmapWindow::OnTimer)
wxEND_EVENT_TABLE()

// コンストラクタ
WXLunaBitmapWindow::WXLunaBitmapWindow(wxWindow *parent, wxWindowID id,
	const wxString& name)
	: inherited(parent, id, name,
		inherited::DEFAULT_STYLE | wxRESIZE_BORDER)
{
	panel = new WXLunaBitmapPanel(this);

	// 大きさをそれに固定
	Layout();
	Fit();

	// 更新タイマーはモニターパネル同様の 30Hz にしておく。
	timer.SetOwner(this);
	timer.Start(33);
}

// デストラクタ
WXLunaBitmapWindow::~WXLunaBitmapWindow()
{
}

// タイマーイベント
void
WXLunaBitmapWindow::OnTimer(wxTimerEvent& event)
{
	// モニタースクリーンを更新して、更新があれば画面を再描画
	panel->Refresh();
}
